package skuad.agentu;

import skuad.util.base.XmlManager;
import org.w3c.dom.*;

import skuad.util.value.Value;
import skuad.uda.UDA;
/**
 * Filtre possible :
 *
 *	- Node : ALL / NAME / IP
 *	- spot : ALL / NAME / ID
 *  - DEVICE : ALL / NAME / TYPE / ID (id global)
 * 
 */
import skuad.util.databytes.*;

public class PhysicalPlugRule  {
	public final static byte UP		=1;
	public final static byte DOWN	=2;

	//constante des type de clé
	public final static byte ALL	=0;	//pour node, spot et device
	public final static byte NAME	=1;	//pour node, spot et device
	public final static byte IP		=2;	//pour node
	public final static byte PORT	=3; //pour spot
	public final static byte TYPE	=4; //pour device
	public final static byte ID		=5; //pour device et spot (id global)
	
	int node_key, spot_key, device_key;
	String st_node; 	//pour les valeurs des clés IP et NAME de node
 	String st_spot;		//pour la valeur de clé NAME ou ID (id global) de spot
 	int int_spot;		//pour la valeur de clé PORT de spot
 	String st_device;	//pour la valeur NAME de device
 	int int_device;		//pour les valeurs des clés TYPE et ID device
 	
 	private PhysicalPlugRule(){} //on doit utiliser la frabrique PlugRule.parse(...)
 	
 	//lecture des champs
 	public int nodeKey(){ return(node_key); }
 	public int spotKey(){ return(spot_key); }
 	public int deviceKey(){ return(device_key); }
 	public String stringNodeFilter(){ return(st_node); }
 	public String stringSpotFilter(){ return(st_spot); }
 	public String stringDeviceFilter(){ return(st_device); }
 	public int intHostFilter(){ return(int_spot); }
 	public int intDeviceFilter(){ return(int_device); }
 	
 	public static PhysicalPlugRule parse(DataReader rdesc){ 
 		//rdesc = 	[node_key]{si node_key!=ALL: (string: node_value) }
 		//			[spot_key]{si spot_key==NAME: (string: spot_value) | si  spot_key==PORT: {short: port_value}}
 		//			[device_key]{si device_key==NAME: (string: device_value) | si  device_key==TYPE ou ID: (int: device_value)}
 		PhysicalPlugRule prule=new PhysicalPlugRule();
 		prule.node_key = rdesc.readByte();
 		if(prule.node_key != ALL) prule.st_node = rdesc.readString();
 		prule.spot_key = rdesc.readByte();
 		switch(prule.spot_key){
 		case NAME : case ID : prule.st_spot = rdesc.readString(); break;
 		case PORT : prule.int_spot = rdesc.readShort(); break;
 		}
 		prule.device_key = rdesc.readByte();
 		switch(prule.device_key){
 		case NAME : prule.st_device = rdesc.readString(); break;
 		case ID : case TYPE :  prule.int_device = rdesc.readInt(); break;
 		} 		
 		return(prule);
 	}
 	
 	public static PhysicalPlugRule ipRule(String ip) {
 		PhysicalPlugRule prule=new PhysicalPlugRule();
 		prule.node_key = IP; prule.st_node = ip;
 		prule.spot_key = ALL; prule.device_key=ALL;
 		return(prule);
 	}
 	
 	public static PhysicalPlugRule parse(Value rdesc){ //rdesc doit être un hash dont les champs suivants sont optionnels (vaudront ALL par défaut)
 		//retourne null et sans effet si rdesc n'est pas un hash !
 		// node : ALL ou [ NAME / IP , node_value ] 
 		// spot : ALL ou [ NAME / PORT / ID , spot_value]
 		// device : ALL ou [ NAME / ID , device_value] ou [ TYPE , 	LIGHT / BUTTON / LOGGER / TEMPERATURE / ROVER ]
 		PhysicalPlugRule prule=null;
 		if(rdesc!=null && rdesc.isHash()) {
 			prule=new PhysicalPlugRule();
 	 		prule.node_key = ALL;
 	 		prule.spot_key = ALL;
 	 		prule.device_key = ALL;
 			Value nk = rdesc.get("node");
 			if(nk.isArray()) {
 				String nk1 = nk.getString(0);
 				prule.st_node = nk.getString(1);
 				if("IP".equals(nk1)) prule.node_key = IP;
 				else if("NAME".equals(nk1)) prule.node_key = NAME;
 			}
 			
 			Value sk = rdesc.get("spot");
 			if(sk.isArray()) {
 				String sk1 = sk.getString(0);
 				if("PORT".equals(sk1)) {
 					prule.spot_key = PORT;
 					prule.int_spot = sk.getInt(1);
 				} else {
 					prule.st_spot = sk.getString(1);
 					if("NAME".equals(sk1)) prule.spot_key = NAME;
 					else if("ID".equals(sk1)) prule.spot_key = ID;
 					
 				}
 			} 			
 			
 			Value dk = rdesc.get("device");
 			if(dk.isArray()) {
 				String dk1 = dk.getString(0);
 				if("NAME".equals(dk1)) {
 					prule.device_key = NAME;
 					prule.st_device = dk.getString(1);
 				} else {
 					prule.int_device = dk.getInt(1);
 					if("ID".equals(dk1)) prule.device_key = ID;
 					else if("TYPE".equals(dk1)) {
 						prule.device_key = TYPE;
 						String dk2 = dk.getString(1);
 						switch(dk2) {
 						case "LIGHT" : prule.int_device = UDA.LIGHT; 	break;
 						case "BUTTON" : prule.int_device = UDA.BUTTON; 	break;
 						case "GPS" : prule.int_device = UDA.GPS; 	break;
 						case "LOGGER" : prule.int_device = UDA.LOGGER; 	break;
 						case "TEMPERATURE" : prule.int_device = UDA.TEMPERATURE; break;
 						case "ROVER" : prule.int_device = UDA.ROVER; 	break; 						
 						default : prule.device_key = ALL;
 						}
 					}
 					
 				}
 			} 
 		}
 		
 		return(prule);
 	}
 	
 	public static PhysicalPlugRule parse(Element rule_node){ 
//			<rule id='xxx' 
//	 				node_key='' node_value='' 
//	 				spot_key='' spot_value='' 
//	 				device_key='' device_value=''  
//	 			/> 
 		PhysicalPlugRule prule=new PhysicalPlugRule();
 			
 		prule.node_key = (byte) XmlManager.getIntAtt(rule_node, "node_key", 0);
 		if(prule.node_key != ALL) prule.st_node = rule_node.getAttribute("node_value");
 		prule.spot_key = (byte) XmlManager.getIntAtt(rule_node, "spot_key", 0);
 		switch(prule.spot_key){
 		case NAME :  case ID : prule.st_spot = rule_node.getAttribute("spot_value"); break;
 		case PORT : prule.int_spot = XmlManager.getIntAtt(rule_node, "spot_value", 0); break;
 		}
 		prule.device_key = (byte) XmlManager.getIntAtt(rule_node, "device_key", 0);
 		switch(prule.device_key){
 		case NAME : prule.st_device = rule_node.getAttribute("device_value"); break;
 		case ID : case TYPE :  prule.int_device = XmlManager.getIntAtt(rule_node, "device_value", 0); break;
 		} 		
 			
 		return(prule);
 	}
 	
 	public void write(Element rule_node, Document xml_doc){ 
//		<rule 
//			node_key='' node_value='' 
//			spot_key='' spot_value='' 
//			device_key='' device_value=''  
//		/> 		
 		XmlManager.setIntAtt(rule_node, "node_key", node_key);
 		if(node_key!=ALL) rule_node.setAttribute("node_value", st_node);

 		XmlManager.setIntAtt(rule_node, "spot_key", spot_key);
 		switch(spot_key){
 		case NAME :   case ID : rule_node.setAttribute("spot_value", st_spot); break;
 		case PORT :XmlManager.setIntAtt(rule_node, "spot_value", int_spot); break;
 		}
 		
 		XmlManager.setIntAtt(rule_node, "device_key", device_key);
 		switch(device_key){
 		case NAME :  rule_node.setAttribute("device_value", st_device); break;
 		case ID : case TYPE : XmlManager.setIntAtt(rule_node, "device_value", int_device); break;
 		}
	}
 	
 	public void write(DataWriter wdesc){ //[node_key]{si node_key!=ALL: (string: node_value) }[spot_key]{si spot_key==NAME: (string: spot_value) | si  spot_key==PORT: {short: port_value}}[device_key]{si device_key==NAME: (string: device_value) | si  device_key==TYPE ou ID: (int: device_value)}
 		wdesc.writeByte(node_key);
 		if(node_key != ALL) wdesc.writeString(st_node);
 		
 		wdesc.writeByte(spot_key);
 		switch(spot_key){
 		case NAME :  case ID : wdesc.writeString(st_spot); break;
 		case PORT : wdesc.writeShort(int_spot); break;
 		}
 		
 		wdesc.writeByte(device_key);
 		switch(device_key){
 		case NAME :  wdesc.writeString(st_device); break; 
 		case ID :  case TYPE : wdesc.writeInt(int_device); break;  
 		} 				 		
 	}
}