//Support d'exécution d'agent en temps réel
/** 
	NB: les serveurs persistants seront enregistré dans un fichier nom_du_serveur.xml dans le répertoire d'exécution. 
	Pour tout serveur persistant de nom NAME d'un spot, un idContext sera fixé pour la chaine contexte : "agentu_"+NAME
	
	API publique ServerAU :
	
		(static) ServerAU newServer()	--> nouveau Serveur agentU sans persistance
			==> plusieurs instances diffétentes peuvent cohabiter au sein d'un même contexte d'exécution, seul les codes disposants de la référence du seveur pourront agir sur lui localement
				(le serveur ne sera pas non plus exposé via le protocole réseau AgentU)
			
		(static) ServerAU get(String name)	--> Serveur agentU avec chargement/enregistrement des agents dans le fichier name.xml
			==> instance unique sur un contexte d'exécution donnée (pour éviter les accés concurents au fichier de sauvegarde)

		//Gestion des types d'agent
		int addTypeAgent(Class<? extends AgentU> classe)	 //ajoute le type d'agent indiqué (s'il n'est pas déjà enregistré) et retourne son code type, 0 en cas d'erreur, le type_name sera le nom de la classe
		int addTypeAgent(String type_name,  Class<? extends AgentU> classe)	 //ajoute le type d'agent indiqué (s'il n'est pas déjà enregistré) et retourne son code type, 0 en cas d'erreur
		Iterable<String> namesTypeAgent()	//retourne la liste des types agents existant
		int codeTypeAgent(String name_type_agent)
		
		//Création/suppresion des agents : --> retourne l'id de l'agent créé
 		int createAgent(Class<? extends AgentU> classe, String agent_name)
		int createAgent(int code_type, String agent_name)
		int createAgent(String type_name, String agent_name)
		boolean removeAgent(int agent_id)
		
		//Modification des agents
		boolean renameAgent(int agent_id, String name)
		boolean retypeAgent(int agent_id, int code_type)
		boolean retypeAgent(int agent_id, String type_name)
		
		//Consultation des agents
		Iterable<Integer> allAgentsId()
		Iterable<Integer> agentsId(String agent_name)
		boolean existAgent(int agentId)
		boolean existAgentWithName(String agent_name)
		int agentType(int agentId)
		String agentTypeName(int agentId)
		String agenteName(int agentId)
		byte agentState(int agentId)
		byte agentRunMode(int agentId)
		
		//Pilotage des agents
		boolean runAgent(int agentId)
		boolean stopAgent(int agentId)
		boolean pauseAgent(int agentId)
		boolean changeRunModeAgent(int agentId, byte run_mode)
	
		//Gestion des écouteurs d'agents
		void addAllAgentObserver(AgentUObserver obs)
		boolean removeAllAgentObserver(AgentUObserver obs)
		boolean addAgentObserver(int agentId, AgentUObserver obs)
		boolean removeAgentObserver(int agentId, AgentUObserver obs)

		//Gestion des slots sociaux
		ConstantIterable<SocialSpecif> socialsSpecif(int code_type_agent)
		ConstantIterable<SocialSpecif> socialsSpecif(String name_type_agent) 
		byte setSocialPlug(int agentId, String name_slot, String name_space, String password, String local_id)
		boolean existSocialPlug(int agentId, String name_slot)
		String getSocialPlugNameSpace(int agentId, String name_slot);
		boolean isSocialPlugPrivateSpace(int agentId, String name_slot);
		String getSocialSlotConfigLocalId(int agentId, String name_slot);
		String getSocialPlugCurrentLocalId(int agentId, String name_slot);

		//Gestion des slots physiques
		ConstantIterable<PhysicalSpecif> physicalsSpecif(int code_type_agent)
		ConstantIterable<PhysicalSpecif> physicalsSpecif(String name_type_agent)
		Iterable<Integer> physicalPlugRules(int agentId, String name_slot) //les ids des règles classé dans l'ordre
		int addPhysicalPlugRule(int agentId, String name_slot, DataReader rule)		// on va appliquer PhysicalPlugRule.parse(rule) pour traduire la règle, puis rule.dispose()
		int addPhysicalPlugRule(int agentId, String name_slot, PhysicalPlugRule rule) //retourne l'id, la règle est placé en dernière position sur le slot
		byte movePhysicalPlugRule(int agentId, String name_slot, byte rule_id, byte direction); //direction = PlugRule.UP ou plugRule.DOWN
		byte removePhysicalPlugRule(int agentId, String name_slot, byte rule_id)	
		boolean physicalSlotPlugged(int agentId, String name_slot)
		Host getPhysicalPlugHost(int agentId, String name_slot);	//null si non pluggé
		int getPhysicalPlugDeviceId(int agentId, String name_slot);
*/
package skuad.agentu;

import java.util.*;
import skuad.util.base.*;
import skuad.hubudp.*;
import skuad.util.value.Value;
import skuad.agentu.implementation.*;
import skuad.util.databytes.*;

public class ServerAU {
	
	private static Map<String, ServerAU> persitentServer = new  HashMap<String, ServerAU>(); //table interne des serveurs persistant, name->instance ServerAU
	RTAgentManager dbagent;
	
	//[Constructeurs
	private ServerAU(Hub hub, String name){	
		dbagent = new RTAgentManager(hub, this, name);
	}
	
	public static ServerAU newServer(Hub hub){	//--> Crée un nouveau Serveur agentU sans persistance des agents
		if(hub==null) hub=HubUdp.joinLocalHub();
		return(new ServerAU(hub, null));
	}
	
	public static synchronized ServerAU get(Hub hub, String name){	//--> Serveur agentU avec chargement/enregistrement des agents dans le fichier name.xml (instance unique)
		//NB: retourne null si le fichier de persistance est déjà ouvert sur un autre hub !
		if(hub==null) hub=HubUdp.joinLocalHub();
		ServerAU serv = persitentServer.get(name);
		if(serv==null){
			serv = new ServerAU(hub, name);
			persitentServer.put(name, serv);
		} else {
			if(serv.dbagent.kernel.hub != hub) return(null); //fichier de persistance déjà ouvert sur un autre hub !
		}
		return(serv);
	}
	//]

	//[Gestion des types d'agent
	public int addTypeAgent(Class<? extends AgentU> classe){	 //ajoute le type d'agent indiqué (s'il n'est pas déjà enregistré) et retourne son code type, 0 en cas d'erreur
		return(dbagent.addTypeAgent(classe.getName(), classe));
	}
	public int addTypeAgent(String type_name,  Class<? extends AgentU> classe){	 //ajoute le type d'agent indiqué (s'il n'est pas déjà enregistré) et retourne son code type, 0 en cas d'erreur
		return(dbagent.addTypeAgent(type_name, classe));
	}
	public Iterable<String> namesTypeAgent(){	//retourne la liste des types agents existant
		return(dbagent.namesTypeAgent());
	}
	public int codeTypeAgent(String name_type_agent){
		return(dbagent.codeTypeAgent(name_type_agent));
	}
	//]
	
	//[Création/suppresion des agents
	public int createAgent(Class<? extends AgentU> classe, String agent_name){
		int code_type = addTypeAgent(classe);
		return(dbagent.newAgent(code_type, agent_name, null));
	}
	public int createAgent(int code_type, String agent_name){
		return(dbagent.newAgent(code_type, agent_name, null));
	}
	public int createAgent(String type_name, String agent_name){
		return(dbagent.newAgent(type_name, agent_name, null));
	}
	public int createAgent(Class<? extends AgentU> classe, String agent_name, Value config){
		int code_type = addTypeAgent(classe);
		return(dbagent.newAgent(code_type, agent_name, config));
	}
	public int createAgent(int code_type, String agent_name, Value config){
		return(dbagent.newAgent(code_type, agent_name, config));
	}
	public int createAgent(String type_name, String agent_name, Value config){
		return(dbagent.newAgent(type_name, agent_name, config));
	}
	public boolean removeAgent(int agent_id){
		return(dbagent.removeAgent(agent_id));
	}
	//]
	
	//[Modification des agents
	boolean renameAgent(int agent_id, String name){
		return(dbagent.renameAgent(agent_id, name));
	}
	boolean retypeAgent(int agent_id, int code_type){
		return(dbagent.retypeAgent(agent_id, code_type));
	}
	boolean retypeAgent(int agent_id, String type_name){
		return(dbagent.retypeAgent(agent_id, type_name));
	}
	//]

	//Consultation des agents
	public Iterable<Integer> allAgentsId(){ return(dbagent.allAgentsId()); }
	public Iterable<Integer> agentsId(String agent_name){ return(dbagent.agentsId(agent_name)); }
	public boolean existAgent(int agentId){ return(dbagent.existAgent(agentId)); }
	public boolean existAgentWithName(String agent_name){ return(dbagent.existAgentWithName(agent_name)); }
	
	public int agentType(int agentId){ return(dbagent.agentType(agentId)); }
	public String agentTypeName(int agentId){ return(dbagent.agentTypeName(agentId)); }
	public String agentName(int agentId){ return(dbagent.agentName(agentId)); }
	public byte agentState(int agentId){ return(dbagent.agentState(agentId)); }
	public byte agentRunMode(int agentId){ return(dbagent.agentRunMode(agentId)); }
	
	//Pilotage des agents
	public boolean runAgent(int agentId){ return(dbagent.runAgent(agentId)); }
	public boolean stopAgent(int agentId){ return(dbagent.stopAgent(agentId)); }
	public boolean pauseAgent(int agentId){ return(dbagent.pauseAgent(agentId)); }
	public boolean changeRunModeAgent(int agentId, byte run_mode){ return(dbagent.changeRunModeAgent(agentId,run_mode)); }
	
	//[Gestion des écouteurs d'agents
	public void addAllAgentObserver(AgentUObserver obs){ dbagent.addAllAgentObserver(obs, false); }
	public boolean removeAllAgentObserver(AgentUObserver obs){ return(dbagent.removeAllAgentObserver(obs, false)); } 
	public boolean addAgentObserver(int agentId, AgentUObserver obs){ return(dbagent.addAgentObserver(agentId, obs, false)); }
	public boolean removeAgentObserver(int agentId, AgentUObserver obs){ return(dbagent.removeAgentObserver(agentId, obs, false)); }
	//]
	
	//Gestion des slots sociaux
	public ConstantIterable<SocialSpecif> socialsSpecif(int code_type_agent){ return(dbagent.socialsSpecif(code_type_agent)); }
	public ConstantIterable<SocialSpecif> socialsSpecif(String name_type_agent){ return(dbagent.socialsSpecif(name_type_agent)); }
	public byte setSocialPlug(int agentId, String name_slot, String name_space, String password, String local_id){ return(dbagent.setSocialPlug(agentId, name_slot, name_space, password, local_id)); }
	public boolean existSocialPlug(int agentId, String name_slot){ return(dbagent.existSocialPlug(agentId, name_slot)); }
	public String getSocialPlugNameSpace(int agentId, String name_slot){ return(dbagent.getSocialPlugNameSpace(agentId, name_slot)); }
	public boolean isSocialPlugPrivateSpace(int agentId, String name_slot){ return(dbagent.isSocialPlugPrivateSpace(agentId, name_slot)); }
	public String getSocialSlotConfigLocalId(int agentId, String name_slot){ return(dbagent.getSocialSlotConfigLocalId(agentId, name_slot)); }
	public String getSocialPlugCurrentLocalId(int agentId, String name_slot){ return(dbagent.getSocialPlugCurrentLocalId(agentId, name_slot)); }
	
	//Gestion des slots physiques
	public ConstantIterable<PhysicalSpecif> physicalsSpecif(int code_type_agent){ return(dbagent.physicalsSpecif(code_type_agent)); }
	public ConstantIterable<PhysicalSpecif> physicalsSpecif(String name_type_agent){ return(dbagent.physicalsSpecif(name_type_agent)); }
	public Iterable<Integer> physicalPlugRules(int agentId, String name_slot){ return(dbagent.physicalPlugRules(agentId, name_slot)); }
	public int addPhysicalPlugRule(int agentId, String name_slot, DataReader rule){
		PhysicalPlugRule pprule = PhysicalPlugRule.parse(rule);
		rule.dispose();
		return(dbagent.addPhysicalPlugRule(agentId, name_slot, pprule)); 
	}
	public int addPhysicalPlugRule(int agentId, String name_slot, String rule){
		return(addPhysicalPlugRule(agentId, name_slot, Value.parse(rule)));
	}
	public int addPhysicalPlugRule(int agentId, String name_slot, Value rule){
		PhysicalPlugRule pprule = PhysicalPlugRule.parse(rule);
		return(dbagent.addPhysicalPlugRule(agentId, name_slot, pprule)); 
	}
	public int addPhysicalPlugRule(int agentId, String name_slot, PhysicalPlugRule rule){ return(dbagent.addPhysicalPlugRule(agentId, name_slot, rule)); }
	public byte movePhysicalPlugRule(int agentId, String name_slot, byte rule_id, byte direction){ return(dbagent.movePhysicalPlugRule(agentId, name_slot, rule_id, direction)); } //direction = PlugRule.UP ou plugRule.DOWN
	public int removePhysicalPlugRule(int agentId, String name_slot, byte rule_id){ return(dbagent.removePhysicalPlugRule(agentId, name_slot, rule_id)); }	
	public boolean physicalSlotPlugged(int agentId, String name_slot){ return(dbagent.physicalSlotPlugged(agentId, name_slot)); }
	public Spot getPhysicalPlugSpot(int agentId, String name_slot){ return(dbagent.getPhysicalPlugSpot(agentId, name_slot)); }
	public int getPhysicalPlugDeviceId(int agentId, String name_slot){ return(dbagent.getPhysicalPlugDeviceId(agentId, name_slot)); }

}