/**
	Classe de gestion de la liste des agents et des types d'agents existants 
	
	 Syntaxe du fichier d'enregisrement pour le mode persistant :
	 
	 <dbagent next_agent_id='xxx'  next_code_type='xxx' >
	 
	 	//pour chaque type d'agent existant
	 	<type id = 'xxx'  name = 'xxx' class_name='xxxx' />
	 
	 	//pour chaque agent existant
	 	<agent id='xxx' name='xxx' type='xxx' run_mode='xxx'>
	 		
	 		//Configuration de l'agent
	 		<config>
	 		</config>
	 		
	 		//slot social de l'agent 
	 		<socialSlot name='xxx' space="name_space" password="...." local_id="..." />   //password et local_id optionnel

	 		...
	 	
	 		//slot physique ouvert de l'agent (ceux disposant d'au moins une règle)
	 		<physicalSlot name='xxx' >
	 			//pour chaque règle plug
	 			<rule id='xxx' 
	 				node_key='' node_value='' 
	 				spot_key='' spot_value='' 
	 				device_key='' device_value=''  
	 			/>
	 			...
	 		</physicalSlot>
	 		
	 		...
	 		
	 	</agent>
	 
	 	...
	 	
	 </dbagent>
 */
package skuad.agentu.implementation;

import java.util.*;
import skuad.agentu.*;
import skuad.util.base.*;
import skuad.util.databytes.*;
import skuad.hubudp.*;
import skuad.util.thread.*;
import skuad.util.value.Value;

import java.io.*;
import org.w3c.dom.*;
import skuad.SKUAD;
import skuad.agentu.implementation.amanager.*;

public class RTAgentManager extends AgentManager implements Runnable, SpotObserver {

	//Attributs généraux
	int dbam_id;
	boolean query_record;	//is !=false une demande d'enregistrement est en cours
	File db_file;			//fichier d'enregistrement de la liste d'agent, ==null si le fonctionnement et non persistant
	String server_name;
	
	public KernelSAU kernel;  //fixé via KernelSAU.register(...)
	Vector<Spot> spotAllAgentObservers;
	
	@SuppressWarnings("unchecked")
	public RTAgentManager(Hub hub, ServerAU vserver, String sname){
		server_name = sname;
		String file_name = "-none-";
		if(server_name!=null) {
			file_name=server_name+".xml";
			db_file = new File(file_name); 
			//on fixe la valeur d'id context agentu_[serv_name]
			hub.whenReady((LocalSpot localspot) -> localspot.fixIdContext("agentu_"+server_name));
		} else {
			server_name = "-anonyme-";
			db_file=null;
		}

		query_record = false;
				
		spotAllAgentObservers = new Vector<Spot>();
		
		if(db_file!=null && db_file.exists()){ //Chargement du fichier de persistance
			System.out.println("[DBAgentManager] - info - : loading persistant file agent \""+file_name+"\" ...");
			Element racine = XmlManager.readFile(db_file);
			if(racine!=null){
				setNextAgentId(XmlManager.getIntAtt(racine, "next_agent_id", 1));
				setNextCodeType(XmlManager.getIntAtt(racine, "next_code_type", 1));
				
				//analyse des noeuds <type> et <agent>
				NodeList node_list = racine.getChildNodes();
				for(int i=0; i<node_list.getLength(); i++){ Node cur_node=node_list.item(i); if(cur_node.getNodeType()==Node.ELEMENT_NODE){
					Element cur = (Element) cur_node;
					String node_name =  cur.getNodeName();
					
					if(node_name.equals("type")){ //noeud de difinition d'un type d'agent
						int type_id = XmlManager.getIntAtt(cur, "id", -1);
						String name = cur.getAttribute("name");
						String class_name = cur.getAttribute("class_name");
						if(type_id!=-1 && name!=null && class_name != null){
							TypeAgent tad = getTypeAgent(name); 	//On regarde si c'est un type déjà enregistré
							if(tad!=null){ //le type est déjà enregistré, on ignore cette définition
								 System.out.println("[DBAgentManager] * warning * : ... allready existe type name \""+name+"\" for type description number "+i+" ...");
							} else {
								try{								
									tad = new TypeAgent(type_id, name, (Class<? extends AgentU>)  Class.forName(class_name));
									internalAddTypeAgent(tad);
								}catch(Exception e){ System.out.println("[DBAgentManager] * warning * : ... unable to add type name \""+name+"\" for type description number "+i+" ..."); }
							}
						} else System.out.println("[DBAgentManager] * warning * : ... error on type description number "+i+" ...");
						
					} else if(node_name.equals("agent")){ //noeud de difinition d'un d'agent
						int agent_id = XmlManager.getIntAtt(cur, "id", -1);
						String name = cur.getAttribute("name");
						int code_type_agent = XmlManager.getIntAtt(cur, "type", -1);
						int code_retype_agent = XmlManager.getIntAtt(cur, "retype", -1);
						byte run_mode = (byte) XmlManager.getIntAtt(cur, "run_mode", 20);
						if(agent_id!=-1 && code_type_agent!=-1){
							TypeAgent type_agent = (code_retype_agent>0 ? getTypeAgent(code_retype_agent) : getTypeAgent(code_type_agent));
							if(type_agent==null){
								System.out.println("[DBAgentManager] * warning * : ... unknow code type agent \""+code_type_agent+"\" for agent description number "+i+" ...");
							} else {
								RTExistingAgentU eau = new RTExistingAgentU(this, type_agent, agent_id, name, null, false); //nb: les slots sociaux vont être chargé ci-aprés //OXO: on doit pouvoir rendre persistant la configuration de l'agent
								internalAddAgent(eau);

								//Analyse de la configuration des slots
								NodeList slot_node_list = cur.getChildNodes();
								for(int j=0; j<slot_node_list.getLength(); j++){ Node cur_slot_node=slot_node_list.item(j); if(cur_slot_node.getNodeType()==Node.ELEMENT_NODE){
									String name_node = cur_slot_node.getNodeName();

									if(name_node.equals("physicalSlot")){ //slot physique
										Element cur_slot = (Element) cur_slot_node;
										String slot_name =  cur_slot.getAttribute("name");
										PhysicalSlotManager smanager = eau.getPhysicalSlotManager(slot_name);
										if(smanager==null){
											if(code_retype_agent<=0) System.out.println("[DBAgentManager] * warning * : ... unknow slot name \""+slot_name+"\" for agent description number "+i+" ...");
										} else {
											//Analyse des règles pour ce slot
											NodeList rule_node_list = cur_slot.getChildNodes();
											for(int k=0; k<rule_node_list.getLength(); k++){ Node cur_rule_node=rule_node_list.item(k); if(cur_rule_node.getNodeType()==Node.ELEMENT_NODE && cur_rule_node.getNodeName().equals("rule")){
												Element cur_rule = (Element) cur_rule_node;
												PhysicalPlugRule prule = smanager.addRule(cur_rule);
												if(prule==null) System.out.println("[DBAgentManager] * warning * : ... error to perform rule number \""+k+"\" for agent description number "+i+" ...");
											}}
										}
									} else if(name_node.equals("socialSlot")){ //slot social : <socialSlot name='xxx' space="name_space" password="...." local_id="..." />   //password et local_id optionnel
										Element cur_slot = (Element) cur_slot_node;
										String slot_name =  cur_slot.getAttribute("name");
										String space =  cur_slot.getAttribute("space");
										String password =  XmlManager.getAtt(cur_slot, "password", null);
										String local_id =  XmlManager.getAtt(cur_slot, "local_id", null);
										
										eau.setSocialPlug(slot_name, space, password, local_id);
									}

								}}
								//on fixe le mode d'exécution de l'agent
								eau.checkSocialSlot();//on s'assure que les slots sociaux sont bien tous chargé
								eau.changeRunMode(run_mode, AgentU.INIT_SERVER);
								if(code_retype_agent>0) notifyChange(); //maj du fichier d'enregistrement dans 500ms maxi
							}
						} else System.out.println("[DBAgentManager] * warning * : ... error on agent description number "+i+" ...");
					}	  
				}}
					
				System.out.println("[DBAgentManager] - info - : ... loading file finished : "+nbAgent()+" agent(s) found.");
			} else System.out.println("[DBAgentManager] - info - : ... error when loading file.");
		} else if(db_file!=null){
			System.out.println("[DBAgentManager] * warning * : persistant file agent \""+file_name+"\" doesn't not exist.");
		}
//System.out.println("---------------------------------");		
		dbam_id = KernelSAU.register(hub, this); //on enregistre ce gestionnaire sur le noyau SAU
		
		//On engregistre le processus de détection de fin d'exécution
 		Runtime.getRuntime().addShutdownHook(new Thread(){
		    public void run(){
		    	stopAllAgent(AgentU.SHUTDOWN); //on va stopper l'exécution de tous les agents
		    }
		});
	}
	
	protected RunAgentContext getRunContext() { return(kernel); }
	
	protected Pulsor newPulsor(ExistingAgentU eau, String label, long delay, long periode, boolean fireUptodate_or_resetOnResume) {
		return(new RTPulsor(eau, label, delay, periode, fireUptodate_or_resetOnResume));
	}
	
	public String getServerName() { return(server_name); }
	
	protected void addAgentObserver(Spot sender, int agent_id){
		if(agent_id==0){ //écoute sur tous les agents
			synchronized(spotAllAgentObservers){if(!spotAllAgentObservers.contains(sender)){
				spotAllAgentObservers.add(sender);
				sender.addSpotObserver(this);
			}}
		} else { //écoute sur un agent en particulier
			RTExistingAgentU eau = (RTExistingAgentU) getAgent(agent_id);
			if(eau!=null) eau.addObserver(sender);	//l'agent exsite bien
		}			
	}
	protected void removeAgentObserver(Spot sender, int agent_id){
		if(agent_id==0){ //écoute sur tous les agents
			synchronized(spotAllAgentObservers){ 
				spotAllAgentObservers.remove(sender);
				sender.removeSpotObserver(this);
			}
		} else { //écoute sur un agent en particulier
			RTExistingAgentU eau = (RTExistingAgentU) getAgent(agent_id);
			if(eau!=null) eau.removeObserver(sender);	//l'agent exsite bien
		}			
	}
	//[Interface SpotObserver
	public void up(Spot dspot){} 
	public void updateSMD(Spot dspot, boolean first_definition){}
	public void down(Spot dspot){ spotAllAgentObservers.remove(dspot); }
	public void idContext(Spot dspot, String ch_context, String id_context){}
	//]
		
	public boolean isPersistant(){ return(db_file!=null); }
	
	//[Méthodes AgentManager internes surchargées  
	protected ExistingAgentU internalCreateExistingAgentU(TypeAgent type_agent, int agent_id, String name, Value config, boolean load_social_slot) {
		return(new RTExistingAgentU(this, type_agent, agent_id, name, config, load_social_slot));
	}	
	protected void notifyChange(){ //on doit réaliser une maj du fichier d'enregistrement dans au maximum 500ms
		if(!query_record && db_file!=null){
			query_record=true;
			ThreadStack.perform(this, 500);
		}
	}
	protected void notifyNewAgent(int agent_id, int code_type, String name) {
		if(!spotAllAgentObservers.isEmpty()){ synchronized(spotAllAgentObservers){
			//[AGENT_CREATE](int : request_id)(int: code_serveur) [bool: success][agent_id1][code_type_agent1](string: agent_name)
			DataReader mess = DataBytes.writer().b(KernelSAU.AGENT_CREATE).i(0).i(dbam_id).b(1).i(agent_id).b(code_type).s(name).flushReader();
			for(Spot target : spotAllAgentObservers)  target.send(SKUAD.AGENTU, mess, true, false);
			mess.dispose();
		}}		
	}
	protected void notifyNewTypeAgent(int code_type, String type_name) {
		if(!spotAllAgentObservers.isEmpty()){ synchronized(spotAllAgentObservers){
			//[TYPE_AGENT_ADDED](int : request_id)(int: code_serveur)[bool: success][code_type](string: name)[editable][error]{si error==true: (s: message_erreur)}
			DataReader mess = DataBytes.writer().b(KernelSAU.TYPE_AGENT_ADDED).i(0).i(dbam_id).b(1).b(code_type).s(type_name).b(false).b(false).flushReader();

			//notification des observeurs distants
			for(Spot target : spotAllAgentObservers)  target.send(SKUAD.AGENTU, mess, true, false);
			mess.dispose();
		}}		
	}
	//]
	//[Méthodes complémentaire à AgentManager 
	protected void notifyAgentRename(ExistingAgentU eau, int agent_id, String name){
		super.notifyAgentRename(eau, agent_id, name);
		Vector<Spot> spot_observers = ((RTExistingAgentU) eau).spot_observers;
		if(!spotAllAgentObservers.isEmpty() || (spot_observers!=null && !spot_observers.isEmpty()) ){
			//[AGENT_RENAMED](int : request_id)(int: code_serveur)[bool: success](int: agent_id)(s: name)
			DataReader mess = DataBytes.writer().b(KernelSAU.AGENT_RENAMED).i(0).i(dbam_id).b(1).i(agent_id).s(name).flushReader();

			//notification des observeurs distants
			if(!spotAllAgentObservers.isEmpty()){ synchronized(spotAllAgentObservers){
				for(Spot target : spotAllAgentObservers)  target.send(SKUAD.AGENTU, mess, true, false);
			}}
			
			if(spot_observers!=null && !spot_observers.isEmpty()){synchronized(spot_observers){
				for(Spot target : spot_observers)  target.send(SKUAD.AGENTU, mess, true, false);
			}}
			mess.dispose();
		}
	}	
	protected void notifyAgentRetype(ExistingAgentU eau, int agent_id, int code_type){
		super.notifyAgentRetype(eau, agent_id, code_type);
		Vector<Spot> spot_observers = ((RTExistingAgentU) eau).spot_observers;
		if(!spotAllAgentObservers.isEmpty() || (spot_observers!=null && !spot_observers.isEmpty()) ){
			//[AGENT_RETYPED](int : request_id)(int: code_serveur)[bool: success](int: agent_id)[code_type]
			DataReader mess = DataBytes.writer().b(KernelSAU.AGENT_RETYPED).i(0).i(dbam_id).b(1).i(agent_id).b(code_type).flushReader();

			//notification des observeurs distants
			if(!spotAllAgentObservers.isEmpty()){ synchronized(spotAllAgentObservers){
				for(Spot target : spotAllAgentObservers)  target.send(SKUAD.AGENTU, mess, true, false);
			}}
			
			if(spot_observers!=null && !spot_observers.isEmpty()){synchronized(spot_observers){
				for(Spot target : spot_observers) target.send(SKUAD.AGENTU, mess, true, false);
			}}
			mess.dispose();
		}
	}	
	protected void notifyAgentWaitRetype(ExistingAgentU eau, int agent_id, int code_type){
		super.notifyAgentWaitRetype(eau, agent_id, code_type);
		Vector<Spot> spot_observers = ((RTExistingAgentU) eau).spot_observers;
		if(!spotAllAgentObservers.isEmpty() || (spot_observers!=null && !spot_observers.isEmpty()) ){
			//[AGENT_WAIT_RETYPE](int : request_id)(int: code_serveur)[bool: success](int: agent_id)[code_type]
			DataReader mess = DataBytes.writer().b(KernelSAU.AGENT_WAIT_RETYPE).i(0).i(dbam_id).b(1).i(agent_id).b(code_type).flushReader();

			//notification des observeurs distants
			if(!spotAllAgentObservers.isEmpty()){ synchronized(spotAllAgentObservers){
				for(Spot target : spotAllAgentObservers) target.send(SKUAD.AGENTU, mess, true, false);
			}}
			
			if(spot_observers!=null && !spot_observers.isEmpty()){synchronized(spot_observers){
				for(Spot target : spot_observers)  target.send(SKUAD.AGENTU, mess, true, false);
			}}
			mess.dispose();
		}
	}
	protected void notifyChangeRunMode(ExistingAgentU eau, int agent_id, byte old_run_mode, byte new_run_mode){
		super.notifyChangeRunMode(eau, agent_id, old_run_mode, new_run_mode);
		Vector<Spot> spot_observers = ((RTExistingAgentU) eau).spot_observers;
		if(!spotAllAgentObservers.isEmpty() || (spot_observers!=null && !spot_observers.isEmpty()) ){
			//[AGENT_RUN_MODE](int : request_id)(int: code_serveur)[bool: success](int: agent_id)[run_mode]
			DataReader mess = DataBytes.writer().b(KernelSAU.AGENT_RUN_MODE).i(0).i(dbam_id).b(1).i(agent_id).b(new_run_mode).flushReader();

			//notification des observeurs distants
			if(!spotAllAgentObservers.isEmpty()){ synchronized(spotAllAgentObservers){
				for(Spot target : spotAllAgentObservers)  target.send(SKUAD.AGENTU, mess, true, false);
			}}
			
			if(spot_observers!=null && !spot_observers.isEmpty()){synchronized(spot_observers){
				for(Spot target : spot_observers)  target.send(SKUAD.AGENTU, mess, true, false);
			}}
			mess.dispose();
		}
	}
	protected void notifyLog(ExistingAgentU eau, int agent_id, String lmess, long time){
		super.notifyLog(eau, agent_id, lmess, time);
		Vector<Spot> spot_observers = ((RTExistingAgentU) eau).spot_observers;
		if(!spotAllAgentObservers.isEmpty() || (spot_observers!=null && !spot_observers.isEmpty()) ){
			//[AGENT_LOG](int : request_id)(int: code_serveur)[bool: success](int: agent_id)(s: log)(doubleWord: time)
			DataReader mess = DataBytes.writer().b(KernelSAU.AGENT_LOG).i(0).i(dbam_id).b(1).i(agent_id).s(lmess).d(time).flushReader();

			//notification des observeurs distants
			if(!spotAllAgentObservers.isEmpty()){ synchronized(spotAllAgentObservers){
				for(Spot target : spotAllAgentObservers)  target.send(SKUAD.AGENTU, mess, true, false);
			}}
			
			if(spot_observers!=null && !spot_observers.isEmpty()){synchronized(spot_observers){
				for(Spot target : spot_observers)  target.send(SKUAD.AGENTU, mess, true, false);
			}}
			mess.dispose();
		}		
	}
	protected void notifySocialChange(ExistingAgentU eau, int agent_id, String slot_name,  String name_space, String md5_space_password, String local_id ){
		super.notifySocialChange(eau, agent_id, slot_name, name_space, md5_space_password, local_id);
		Vector<Spot> spot_observers = ((RTExistingAgentU) eau).spot_observers;
		if(!spotAllAgentObservers.isEmpty() || (spot_observers!=null && !spot_observers.isEmpty()) ){
			//[SOCIAL_PLUG_CHANGE](int : request_id)(int: code_serveur)[bool: success](int: agent_id1)(string: slot_name)(string: space_name)[bool: is_private](string: local_id)
			DataReader mess = DataBytes.writer().b(KernelSAU.SOCIAL_PLUG_CHANGE).i(0).i(dbam_id).b(1).i(agent_id).s(slot_name).s(name_space).b(md5_space_password!=null).s(local_id).flushReader();

			//notification des observeurs distants
			if(!spotAllAgentObservers.isEmpty()){ synchronized(spotAllAgentObservers){
				for(Spot target : spotAllAgentObservers)  target.send(SKUAD.AGENTU, mess, true, false);
			}}
			
			if(spot_observers!=null && !spot_observers.isEmpty()){synchronized(spot_observers){
				for(Spot target : spot_observers)  target.send(SKUAD.AGENTU, mess, true, false);
			}}
			mess.dispose();
		}	
	}
	protected void notifySocialPlug(ExistingAgentU eau, int agent_id, String slot_name, String current_id, int ctx){
		super.notifySocialPlug(eau, agent_id, slot_name,current_id, ctx );
		Vector<Spot> spot_observers = ((RTExistingAgentU) eau).spot_observers;
		if(!spotAllAgentObservers.isEmpty() || (spot_observers!=null && !spot_observers.isEmpty()) ){
			//[SOCIAL_PLUG](int : request_id)(int: code_serveur)[bool: success](int: agent_id1)(string: slot_name)(string: current_id)[ctx]
			DataReader mess = DataBytes.writer().b(KernelSAU.SOCIAL_PLUG).i(0).i(dbam_id).b(1).i(agent_id).s(slot_name).s(current_id).b(ctx).flushReader();

			//notification des observeurs distants
			if(!spotAllAgentObservers.isEmpty()){ synchronized(spotAllAgentObservers){
				for(Spot target : spotAllAgentObservers)  target.send(SKUAD.AGENTU, mess, true, false);
			}}
			
			if(spot_observers!=null && !spot_observers.isEmpty()){synchronized(spot_observers){
				for(Spot target : spot_observers)  target.send(SKUAD.AGENTU, mess, true, false);
			}}
			mess.dispose();
		}		
	}
	protected void notifySocialUnplug(ExistingAgentU eau, int agent_id, String slot_name, int ctx){
		super.notifySocialUnplug(eau, agent_id, slot_name, ctx);
		Vector<Spot> spot_observers = ((RTExistingAgentU) eau).spot_observers;
		if(!spotAllAgentObservers.isEmpty() || (spot_observers!=null && !spot_observers.isEmpty()) ){
			//[SOCIAL_UNPLUG](int : request_id)(int: code_serveur)[bool: success](int: agent_id1)(string: slot_name)[ctx]
			DataReader mess = DataBytes.writer().b(KernelSAU.SOCIAL_UNPLUG).i(0).i(dbam_id).b(1).i(agent_id).s(slot_name).b(ctx).flushReader();

			//notification des observeurs distants
			if(!spotAllAgentObservers.isEmpty()){ synchronized(spotAllAgentObservers){
				for(Spot target : spotAllAgentObservers)  target.send(SKUAD.AGENTU, mess, true, false);
			}}
			
			if(spot_observers!=null && !spot_observers.isEmpty()){synchronized(spot_observers){
				for(Spot target : spot_observers)  target.send(SKUAD.AGENTU, mess, true, false);
			}}
			mess.dispose();
		}		
	}
	protected void notifyAddRule(ExistingAgentU eau, int agent_id, String name_slot, IdentifiedPlugRule ipr){
		super.notifyAddRule(eau, agent_id, name_slot, ipr);
		Vector<Spot> spot_observers = ((RTExistingAgentU) eau).spot_observers;
		if(!spotAllAgentObservers.isEmpty() || (spot_observers!=null && !spot_observers.isEmpty()) ){
			//[PLUG_RULE_ADDED](int : request_id)(int: code_serveur)[bool: success](int: agent_id)(string: slot_plug)[rule_id][rule_order][node_key]{si node_key!=ALL: (string: node_value) }[spot_key]{si spot_key==NAME: (string: spot_value) | si  spot_key==PORT: {short: port_value}}[device_key]{si device_key==NAME: (string: device_value) | si  device_key==TYPE ou ID: (int: device_value)}
			DataWriter wcmd = DataBytes.writer().b(KernelSAU.PLUG_RULE_ADDED).i(0).i(dbam_id).b(1).i(agent_id).s(name_slot);
			wcmd.b(ipr.id).b(ipr.order);
			ipr.rule.write(wcmd);
			DataReader mess = wcmd.flushReader();
			
			//notification des observeurs distants
			if(!spotAllAgentObservers.isEmpty()){ synchronized(spotAllAgentObservers){
//System.out.println("notifyAddRule 2 !!!");						
				for(Spot target : spotAllAgentObservers) target.send(SKUAD.AGENTU, mess, true, false);
			}}
			
			if(spot_observers!=null && !spot_observers.isEmpty()){synchronized(spot_observers){
				for(Spot target : spot_observers) target.send(SKUAD.AGENTU, mess, true, false);
			}}
			mess.dispose();
		}	
	}
	protected void notifyRemoveRule(ExistingAgentU eau, int agent_id, String name_slot, int rule_id){
		super.notifyRemoveRule(eau, agent_id, name_slot, rule_id);
		Vector<Spot> spot_observers = ((RTExistingAgentU) eau).spot_observers;
		if(!spotAllAgentObservers.isEmpty() || (spot_observers!=null && !spot_observers.isEmpty()) ){
			//[PLUG_RULE_REMOVED](int : request_id)(int: code_serveur)[bool: success](int: agent_id)(string: slot_plug)[rule_id]
			DataReader mess = DataBytes.writer().b(KernelSAU.PLUG_RULE_REMOVED).i(0).i(dbam_id).b(1).i(agent_id).s(name_slot).b(rule_id).flushReader();
			
			//notification des observeurs distants
			if(!spotAllAgentObservers.isEmpty()){ synchronized(spotAllAgentObservers){
				for(Spot target : spotAllAgentObservers)  target.send(SKUAD.AGENTU, mess, true, false);
			}}
			
			if(spot_observers!=null && !spot_observers.isEmpty()){synchronized(spot_observers){
				for(Spot target : spot_observers)  target.send(SKUAD.AGENTU, mess, true, false);
			}}
			mess.dispose();
		}
	}	
	protected void notifyPhysicalPlugRuleMove(ExistingAgentU eau, int agent_id, String name_slot, int rule_id, byte direction){
		super.notifyPhysicalPlugRuleMove(eau, agent_id, name_slot, rule_id,  direction);
		Vector<Spot> spot_observers = ((RTExistingAgentU) eau).spot_observers;
		if(!spotAllAgentObservers.isEmpty() || (spot_observers!=null && !spot_observers.isEmpty()) ){
			//[PLUG_RULE_MOVED](int : request_id)(int: code_serveur)(int: agent_id)(string: slot_plug)[rule_id][direction]
			DataReader mess = DataBytes.writer().b(KernelSAU.PLUG_RULE_MOVED).i(0).i(dbam_id).b(1).i(agent_id).s(name_slot).b(rule_id).b(direction).flushReader();

			//notification des observeurs distants
			if(!spotAllAgentObservers.isEmpty()){ synchronized(spotAllAgentObservers){
				for(Spot target : spotAllAgentObservers)  target.send(SKUAD.AGENTU, mess, true, false);
			}}
			
			if(spot_observers!=null && !spot_observers.isEmpty()){synchronized(spot_observers){
				for(Spot target : spot_observers) target.send(SKUAD.AGENTU, mess, true, false);
			}}
			mess.dispose();
		}	
	}
	protected void notifyPhysicalChange(ExistingAgentU eau, int agent_id, String name_slot, boolean optional_plug, Spot spoter, int device_id){
		super.notifyPhysicalChange(eau, agent_id, name_slot, optional_plug, spoter, device_id);
		Vector<Spot> spot_observers = ((RTExistingAgentU) eau).spot_observers;
		if(!spotAllAgentObservers.isEmpty() || (spot_observers!=null && !spot_observers.isEmpty()) ){
			//[PHYSICAL_PLUG_CHANGE](int : request_id)(int: code_serveur)[bool: success](int: agent_id1)(string: slot_name)[bool: optional](string: spot_owner_globalId)(int: device_id)
			DataWriter wcmd = DataBytes.writer().b(KernelSAU.PHYSICAL_PLUG_CHANGE).i(0).i(dbam_id).b(1).i(agent_id).s(name_slot).b(optional_plug).s(spoter.getIdGlobal());
			DataReader mess = wcmd.i(device_id).flushReader();

			//notification des observeurs distants
			if(!spotAllAgentObservers.isEmpty()){ synchronized(spotAllAgentObservers){
				for(Spot target : spotAllAgentObservers)  target.send(SKUAD.AGENTU, mess, true, false);
			}}
			
			if(spot_observers!=null && !spot_observers.isEmpty()){synchronized(spot_observers){
				for(Spot target : spot_observers)  target.send(SKUAD.AGENTU, mess, true, false);
			}}
			mess.dispose();
		}	
	}
	protected void notifyPhysicalPlug(ExistingAgentU eau, int agent_id, String name_slot, boolean optional_plug, Spot spoter, int device_id){
		super.notifyPhysicalPlug(eau, agent_id, name_slot, optional_plug, spoter, device_id);
		Vector<Spot> spot_observers = ((RTExistingAgentU) eau).spot_observers;
		if(!spotAllAgentObservers.isEmpty() || (spot_observers!=null && !spot_observers.isEmpty()) ){
			//[PHYSICAL_PLUG](int : request_id)(int: code_serveur)[bool: success](int: agent_id1)(string: slot_name)[bool: optional](string: owner_globalId)(int: device_id)
			DataWriter wcmd = DataBytes.writer().b(KernelSAU.PHYSICAL_PLUG).i(0).i(dbam_id).b(1).i(agent_id).s(name_slot).b(optional_plug).s(spoter.getIdGlobal());
			DataReader mess = wcmd.i(device_id).flushReader();

			//notification des observeurs distants
			if(!spotAllAgentObservers.isEmpty()){ synchronized(spotAllAgentObservers){
				for(Spot target : spotAllAgentObservers)  target.send(SKUAD.AGENTU, mess, true, false);
			}}
			
			if(spot_observers!=null && !spot_observers.isEmpty()){synchronized(spot_observers){
				for(Spot target : spot_observers)  target.send(SKUAD.AGENTU, mess, true, false);
			}}
			mess.dispose();
		}
	}	
	protected void notifyPhysicalUnplug(ExistingAgentU eau, int agent_id, String name_slot, boolean optional_plug){
		super.notifyPhysicalUnplug(eau, agent_id, name_slot, optional_plug);
		Vector<Spot> spot_observers = ((RTExistingAgentU) eau).spot_observers;
		if(!spotAllAgentObservers.isEmpty() || (spot_observers!=null && !spot_observers.isEmpty()) ){
			//[PHYSICAL_UNPLUG](int : request_id)(int: code_serveur)[bool: success](int: agent_id1)(string: slot_name)[bool: optional]
			DataReader mess = DataBytes.writer().b(KernelSAU.PHYSICAL_UNPLUG).i(0).i(dbam_id).b(1).i(agent_id).s(name_slot).b(optional_plug).flushReader();

			//notification des observeurs distants
			if(!spotAllAgentObservers.isEmpty()){ synchronized(spotAllAgentObservers){
				for(Spot target : spotAllAgentObservers)  target.send(SKUAD.AGENTU, mess, true, false);
			}}
			
			if(spot_observers!=null && !spot_observers.isEmpty()){synchronized(spot_observers){
				for(Spot target : spot_observers)  target.send(SKUAD.AGENTU, mess, true, false);
			}}
			mess.dispose();
		}
	}
	protected void notifyAgentStateChange(ExistingAgentU eau, int agent_id, byte old_state, byte new_state, byte ctx){
		super.notifyAgentStateChange(eau, agent_id, old_state, new_state, ctx);
		Vector<Spot> spot_observers = ((RTExistingAgentU) eau).spot_observers;
		//observeur distant (sauf si cela fait suite à un SHUTDOWN)
		if(ctx!=AgentU.SHUTDOWN && (!spotAllAgentObservers.isEmpty() || (spot_observers!=null && !spot_observers.isEmpty())) ){
			//[AGENT_STATE](int : request_id)(int: code_serveur)[bool: success](int: agent_id)[state]
			DataReader mess = DataBytes.writer().b(KernelSAU.AGENT_STATE).i(0).i(dbam_id).b(1).i(agent_id).b(new_state).flushReader();

			//notification des observeurs distants
			if(!spotAllAgentObservers.isEmpty()){ synchronized(spotAllAgentObservers){
				for(Spot target : spotAllAgentObservers)  target.send(SKUAD.AGENTU, mess, true, false);
			}}
			
			if(spot_observers!=null && !spot_observers.isEmpty()){synchronized(spot_observers){
				for(Spot target : spot_observers)  target.send(SKUAD.AGENTU, mess, true, false);
			}}
			mess.dispose();
		}
	}
	protected void notifyAgentRemove(ExistingAgentU eau, int agent_id){
		super.notifyAgentRemove(eau, agent_id);
		Vector<Spot> spot_observers = ((RTExistingAgentU) eau).spot_observers;
		if(!spotAllAgentObservers.isEmpty() || (spot_observers!=null && !spot_observers.isEmpty()) ){
			//[AGENT_REMOVED](int : request_id)(int: code_serveur)[bool: success](int: agent_id1)
			DataReader mess = DataBytes.writer().b(KernelSAU.AGENT_REMOVED).i(0).i(dbam_id).b(1).i(agent_id).flushReader();

			//notification des observeurs distants
			if(!spotAllAgentObservers.isEmpty()){ synchronized(spotAllAgentObservers){
				for(Spot target : spotAllAgentObservers)  target.send(SKUAD.AGENTU, mess, true, false);
			}}
			
			if(spot_observers!=null && !spot_observers.isEmpty()){synchronized(spot_observers){
				for(Spot target : spot_observers)  target.send(SKUAD.AGENTU, mess, true, false);
			}}
			mess.dispose();
		}
	}
	//]
	
	//[Tache de mise à jour du fichier de persistance
	public void run(){
		if(db_file==null) return; //normalement cette situation est impossible.

		query_record=false;
		System.out.println("[DBAgentManager] - info - : Enregistrement en cours...");
		Document xml_doc = XmlManager.newXMLDocument("dbagent");
		Element racine = xml_doc.getDocumentElement();
		if(writeToXml(xml_doc, racine)) {
			if(XmlManager.writeFile(racine, db_file)) System.out.println("[DBAgentManager] - info - : ... fin de l'enregistrement.");
			else System.out.println("[DBAgentManager] - info - : ... erreur de création du fichier d'enregistrement.");
		} else System.out.println("[DBAgentManager] * error * : ... impossible de construire la structure XML d'enregistrement.");
	}
	//]
}