package skuad.agentu.implementation;

import java.util.TimerTask;
import skuad.agentu.AgentU;
import skuad.util.thread.ThreadStack;
import skuad.agentu.implementation.amanager.*;

public class RTPulsor extends Pulsor implements Runnable {

	TimerTask task;
	long next_pulse; //timecode du prochain pulse (pour la gestion des reprises aprés pause) 
	
	public RTPulsor(ExistingAgentU viau, String vlabel, long vdelay,  long vperiode, boolean fireUptodate_or_resetOnResume){ //pulse cyclique
		//si reset_on_resume==true : le cycle de pulse est réinitialisé à la reprise de l'agent (le prochain pulse surviendra aprés 
		//l'écoulement du delay, puis les suivants a chaque période), avec reset_on_resume==false à la reprise aprés pause
		//les cycles de déclenchement reprenne comme si l'agent n'avait jamais été en pause.
		super(vlabel, vdelay, vperiode, fireUptodate_or_resetOnResume);
		task=null;
		next_pulse = System.currentTimeMillis()+vdelay;
		lingToExistingAgentU(viau);
	}
	
	public void run(){
//System.out.println("--- run du pulsor : "+label);				
		if(active && eau.getState()==AgentU.RUN) {
//System.out.println("   ---> run confirmé");			
			long tc= System.currentTimeMillis();
			//on met à jour son prochain timecode du prochain déclenchement
			if(periode>0) next_pulse= tc+periode; 
			else next_pulse=0;
			pulseBehavior(tc);
		}
		if(active && periode==0) cancel(true); //pulse ponctuel -> on le supprime
	}
	protected boolean remove(){
		if(super.remove()) {
			if(task!=null){ task.cancel(); task=null; }
			next_pulse=0;
			return(true);
		}
		return(false);
	}
	
	protected void pause(){
		if(task!=null){ task.cancel(); task=null; }
	}
	protected boolean startOrResume(boolean call_remove_pulsor_if_cancel){ //retourne false si le pulsor doit être supprimé
//System.out.println("--- demarrage du pulsor : "+label);		
		if(!active) return(false);
		if(task!=null) return(true); //le pulsor est déjà actif et pas en pause
		long cur_time = System.currentTimeMillis();
		long ddelay;
		if(periode==0){ //pulse ponctuel
//System.out.println("    ---> pulsor ponctuel");				
			if(!resume_reset && cur_time>next_pulse){  //on supprime le pulse
				cancel(call_remove_pulsor_if_cancel);
				return(false);
			}
			ddelay = next_pulse-cur_time;
			if(ddelay<=0) ddelay=0;
			task = ThreadStack.perform(this, ddelay);
		} else { //pulse cyclique
//System.out.println("    ---> pulsor cyclique");	
			if(resume_reset) ddelay=delay;
			else { ddelay = next_pulse-cur_time; if(ddelay<0) ddelay=0; }
			task = ThreadStack.perform(this, ddelay, periode);
		}
		return(true);
	}	
}