/**
	Classe de gestion de la liste des agents et des types d'agents existants 
 */
package skuad.agentu.implementation.amanager;

import java.util.*;
import skuad.agentu.*;
import skuad.util.base.*;
import skuad.util.databytes.*;
import skuad.hubudp.*;
import skuad.util.thread.*;
import skuad.util.value.Value;

import org.w3c.dom.*;

public abstract class AgentManager {

	//Les types d'agents
	int next_code_type;				//prochain code à utiliser pour les types
	Map<String, TypeAgent> nameTA;	//table des types d'agents connu, key : nom du type
	Map<Integer, TypeAgent> typeAgents;		//table des types d'agents connu, key : code_type
	
	//Les agents
	int next_agent_id;						//prochain id agent à utiliser	
	Map<Integer, ExistingAgentU> agents; 	//table des agents existants

	Vector<AgentUObserver> async_aagent_observers, sync_aagent_observers;
	
	public AgentManager(){
		async_aagent_observers = new Vector<AgentUObserver>();
		sync_aagent_observers = new Vector<AgentUObserver>();
		
		next_code_type=1;		
		nameTA = new HashMap<String, TypeAgent>();	
		typeAgents = new HashMap<Integer, TypeAgent>();		
		
		next_agent_id=1;
		agents = new HashMap<Integer, ExistingAgentU>();
	}
	
	public void reset() {
		async_aagent_observers.clear();
		sync_aagent_observers.clear();
		next_code_type=1;		
		nameTA.clear();	
		typeAgents.clear();	
		next_agent_id=1;
		agents.clear();
	}
	
	//[Doit être surchargé dans les classes filles
	protected abstract RunAgentContext getRunContext();
	protected abstract Pulsor newPulsor(ExistingAgentU eau, String label, long delay, long periode, boolean fireUptodate_or_resetOnResume);
	//]

	public long curTime() {		//lecture du temps courant
		return(getRunContext().getCurrentTime());
	}
	
	//[Pour une utilisation dans les classes filles
	protected void setNextCodeType(int nct) { next_code_type=nct; }
	protected void setNextAgentId(int nai) { next_agent_id=nai; }
	protected TypeAgent getTypeAgent(String name) { return(nameTA.get(name)); }
	protected TypeAgent getTypeAgent(int tid) { return(typeAgents.get(tid)); }
	protected ExistingAgentU getAgent(int aid) { return(agents.get(aid)); }
	protected void internalAddTypeAgent(TypeAgent ta) {
		nameTA.put(ta.name, ta);
		typeAgents.put(ta.code, ta);
	}
	protected void internalAddAgent(ExistingAgentU eau) {
		agents.put(eau.agent_id, eau);
	}
	
	protected void startAllAgent(byte ctx) { //on demarre les agents qui ont besoin de l'être
		synchronized(agents){ //on va stopper l'exécution de tous les agents
			for(ExistingAgentU a : agents.values()) a.refreshState(ctx);
		}
	}
	protected void stopAllAgent(byte ctx) {
		synchronized(agents){ //on va stopper l'exécution de tous les agents
			for(ExistingAgentU a : agents.values()) a.stop(ctx);
		}
	}
	/*
	 *  Syntaxe du fichier d'enregisrement pour le mode persistant :
	 
	 <dbagent next_agent_id='xxx'  next_code_type='xxx' >
	 
	 	//pour chaque type d'agent existant
	 	<type id = 'xxx'  name = 'xxx' class_name='xxxx' />
	 
	 	//pour chaque agent existant
	 	<agent id='xxx' name='xxx' type='xxx' run_mode='xxx'>
	 		
	 		//slot social de l'agent 
	 		<socialSlot name='xxx' space="name_space" password="...." local_id="..." />   //password et local_id optionnel

	 		...
	 	
	 		//slot physique ouvert de l'agent (ceux disposant d'au moins une règle)
	 		<physicalSlot name='xxx' >
	 			//pour chaque règle plug
	 			<rule id='xxx' 
	 				node_key='' node_value='' 
	 				spot_key='' spot_value='' 
	 				device_key='' device_value=''  
	 			/>
	 			...
	 		</physicalSlot>
	 		
	 		...
	 		
	 	</agent>
	 
	 	...
	 	
	 </dbagent>
	 */
	protected boolean writeToXml(Document xml_doc, Element racine) {
		if(racine==null) return(false); 
		synchronized(nameTA){ synchronized(agents){
			XmlManager.setIntAtt(racine, "next_agent_id", next_agent_id);
			XmlManager.setIntAtt(racine, "next_code_type", next_code_type);

			//on fixe les noeuds <type>
			for(TypeAgent ta : nameTA.values()){
				Element ta_node = xml_doc.createElement("type");
				XmlManager.setIntAtt(ta_node, "id", ta.code);
				ta_node.setAttribute("name", ta.name);
				ta_node.setAttribute("class_name", ta.classe.getName());

				racine.appendChild(ta_node);
			}

			for(ExistingAgentU eau : agents.values()){
				Element agent_node = xml_doc.createElement("agent");
				XmlManager.setIntAtt(agent_node, "id", eau.agent_id);
				agent_node.setAttribute("name", eau.name);
				XmlManager.setIntAtt(agent_node, "type", eau.type.code);
				if(eau.wait_retype!=null) XmlManager.setIntAtt(agent_node, "retype", eau.wait_retype.code);
				XmlManager.setIntAtt(agent_node, "run_mode", eau.run_mode);

				//On fixe la configuration des slots sociaux
				synchronized(eau.sslots){ for(SocialSlotManager mslot : eau.sslots.values()){
					if(mslot.specialConfig()){
						Element slot_node = xml_doc.createElement("socialSlot");
						slot_node.setAttribute("name", mslot.slot_specif.nameSlot());
						slot_node.setAttribute("space", mslot.space_name);
						XmlManager.setAttIfNotNull(slot_node, "password", mslot.space_password);
						XmlManager.setAttIfNotNull(slot_node, "local_id", mslot.local_id);

						agent_node.appendChild(slot_node);
					}
				}}

				//On fixe la configuration des slots physiques
				synchronized(eau.pslots){ for(PhysicalSlotManager mslot : eau.pslots.values()){ if(mslot.irules.size()>0){
					Element slot_node = xml_doc.createElement("physicalSlot");
					slot_node.setAttribute("name", mslot.slot_specif.nameSlot());

					synchronized(mslot.irules){ for(IdentifiedPlugRule ipr : mslot.irules){
						Element rule_node = xml_doc.createElement("rule");
						XmlManager.setIntAtt(rule_node, "id", ipr.id);
						ipr.rule.write(rule_node, xml_doc);
						slot_node.appendChild(rule_node);
					}}

					agent_node.appendChild(slot_node);
				}}}

				racine.appendChild(agent_node);
			}}
		}
		return(true);
	}
	//]
	
	//[A surcharger éventuellement dans les classe fille
	protected ExistingAgentU internalCreateExistingAgentU(TypeAgent type_agent, int agent_id, String name, Value config, boolean load_social_slot) {
		return(new ExistingAgentU(this, type_agent, agent_id, name, config, load_social_slot));
	}
	protected void notifyChange(){} //à surcharger pour réagir un une maj des données du manager (sur les agents ou les types)
	protected void notifyNewAgent(int agent_id, int code_type, String name) {}
	protected void notifyNewTypeAgent(int code_type, String type_name) {}
	//]
	
	public ExistingAgentU get(int aid){ return(agents.get(aid)); }
	public int codeTypeAgent(String name_type_agent){
		TypeAgent ta = nameTA.get(name_type_agent);
		if(ta!=null) return(ta.code);
		return(-1);
	}
	public int newAgent(int code_type, String name, Value config){ //retourne l'id de l'agent crée
		TypeAgent type_agent = typeAgents.get(code_type);
		if(type_agent!=null){ 
			int agent_id=0;
			synchronized(agents){
				do agent_id=next_agent_id++; while(agents.containsKey(agent_id));
				agents.put(agent_id, internalCreateExistingAgentU(type_agent, agent_id, name, config, true));
			}
			notifyChange(); //maj du fichier d'enregistrement dans 500ms maxi
	
			//notification des observeurs locaux
			int agentId = agent_id; //nécessaire pour valider l'écriture de la fonction lambda
			synchronized(async_aagent_observers) { for(AgentUObserver obs : async_aagent_observers) ThreadStack.perform(() -> obs.agentCreated(code_type, agentId)); }
			synchronized(sync_aagent_observers) { for(AgentUObserver obs : sync_aagent_observers) obs.agentCreated(code_type, agentId); }
			
			//notification des classes filles
			notifyNewAgent(agent_id, code_type, name); 

			return(agent_id);
		}
		return(-1);
	}
	public int newAgent(String name_type, String name, Value config){ //retourne l'id de l'agent crée
		TypeAgent type_agent = nameTA.get(name_type);
		if(type_agent!=null){ 
			int agent_id=0;
			synchronized(agents){
				agent_id=next_agent_id++;
				agents.put(agent_id, internalCreateExistingAgentU(type_agent, agent_id, name, config, true));
			}
			notifyChange(); //maj du fichier d'enregistrement dans 500ms maxi
			//notification des observeurs locaux
			int agentId = agent_id; //nécessaire pour valider l'écriture de la fonction lambda
			synchronized(async_aagent_observers) { for(AgentUObserver obs : async_aagent_observers) ThreadStack.perform( () -> obs.agentCreated(type_agent.code, agentId)); }
			synchronized(sync_aagent_observers) { for(AgentUObserver obs : sync_aagent_observers) obs.agentCreated(type_agent.code, agentId); }
			
			//notification des classes filles
			notifyNewAgent(agent_id, type_agent.code, name); 
			
			return(agent_id);
		}
		return(-1);
	}	
	
	public boolean removeAgent(int aid){ 
		ExistingAgentU eau = null;
		synchronized(agents){ eau = agents.remove(aid); }
		if(eau!=null){
			eau.remove(AgentU.LOCAL_ORDER); //<- va notifier les observers de la suppression
			notifyChange(); //maj du fichier d'enregistrement dans 500ms maxi		

			return(true);
		}
		return(false);
	}
		
	//[Modification des agents
	public boolean renameAgent(int agent_id, String name){
		ExistingAgentU eau = agents.get(agent_id);
		return(eau!=null && eau.rename(name));
	}
	public boolean retypeAgent(int agent_id, int code_type){
		TypeAgent type_agent = typeAgents.get(code_type);
		ExistingAgentU eau = agents.get(agent_id);
		return(eau!=null && type_agent!=null && eau.retype(type_agent));
	}
	public boolean retypeAgent(int agent_id, String type_name){
		TypeAgent type_agent = nameTA.get(type_name);		
		ExistingAgentU eau = agents.get(agent_id);
		return(eau!=null && type_agent!=null && eau.retype(type_agent));
	}
	//]
	
	//[Gestion des écouteurs d'agents
	public void addAllAgentObserver(AgentUObserver obs, boolean sync){ 
		if(sync) sync_aagent_observers.add(obs);
		else async_aagent_observers.add(obs);
	}
	public boolean removeAllAgentObserver(AgentUObserver obs, boolean sync){ 
		if(sync) return(sync_aagent_observers.remove(obs));	
		else  return(async_aagent_observers.remove(obs));
	}
	public boolean addAgentObserver(int agentId, AgentUObserver obs, boolean sync){
		ExistingAgentU eau = agents.get(agentId);
		if(eau!=null) return(eau.addAgentObserver(obs, sync));
		return(false); //l'agent demandé n'existe pas
	}
	public boolean removeAgentObserver(int agentId, AgentUObserver obs, boolean sync){
		ExistingAgentU eau = agents.get(agentId);
		if(eau!=null) return(eau.removeAgentObserver(obs, sync));
		return(false); //l'agent demandé n'existe pas		
	}
	//]
		
	public Iterable<String> namesTypeAgent(){
		synchronized(typeAgents){ 
			return(Util.copyStringIterable(nameTA.keySet()));
		}
	}
	public int nbAgentType(){
		return(nameTA.size());
	}
	public void writeAllTypeAgent(DataWriter dw){ //[nb_type]...|pour chaque type : [code_type1](string: name1)[editable][error]{si error==true: (s: message_erreur)}|...
		synchronized(nameTA){ 
			dw.b(nameTA.size());
			for(TypeAgent ta : nameTA.values()) {
				dw.b(ta.code).s(ta.name).b(ta.editable);
				//aucun erreur de typepossible pour la plateforme JAVA
				dw.b(false);
			}
		}
	}
	public void writeAllPhysisicalSlotSpecif(int code_type, DataWriter dw){ //[bool: exist](int: nb_slot)...|pour chaque slot: (string: name)(int: device_type)[bool: optional]|...
		TypeAgent type_agent = typeAgents.get(code_type);
		if(type_agent==null) dw.b(0);
		else {
			 dw.b(1);
			 dw.i(type_agent.physical_specif.size());
			 for(PhysicalSpecif spec : type_agent.physical_specif) dw.s(spec.nameSlot()).i(spec.deviceType()).b(spec.isOptional() ? 1 : 0);
		}
	}
	
	public void writeAllSocialSlotSpecif(int code_type, DataWriter dw){ //[bool: exist](int: nb_slot)...|pour chaque slot:(string: name)(string: space_name)(string: local_id)|...	
		TypeAgent type_agent = typeAgents.get(code_type);
		if(type_agent==null) dw.b(0);
		else {
			 dw.b(1);
			 dw.i(type_agent.social_specif.size());
			 for(SocialSpecif spec : type_agent.social_specif) dw.s(spec.nameSlot()).s(spec.nameSpace()).s(spec.localId());
		}
	}
	
	public void writeAllAgent(DataWriter dw){ //(int: nb_agent)...|pour chaque agent: (int: agent_id)[code_type_agent][code_retype](string: agent_name)[state][run_mode]|
		dw.i(agents.size());
		synchronized(agents){
			for(ExistingAgentU a : agents.values()) a.write(dw);
		}
	}
	
	public int addTypeAgent(String type_name,  Class<? extends AgentU> classe){ //ajoute le type d'agent indiqué (s'il n'est pas déjà enregistré) et retourne son code type, -1 en cas d'erreur
		if(type_name==null || classe==null) return(-1); //argument non valide
		int code_type;
		synchronized(typeAgents){ 
			//On regarde si c'est un type déjà enregistré
			TypeAgent tad = nameTA.get(type_name);
			if(tad!=null){ //le type est déjà enregistré
				if(tad.classe!=classe) return(-1); //si la classe n'est pas la même on signale l'erreur
				return(tad.code);
			}

			//C'est un nouveau type
			if(typeAgents.size()>250) return(-1);	//nombre de type maximum atteind

			//recherche du prochain code_id
			while(typeAgents.containsKey(next_code_type)) { ++next_code_type; if(next_code_type>255) next_code_type=1; }
			code_type = next_code_type;
			++next_code_type; if(next_code_type>255) next_code_type=1;

			tad = new TypeAgent(code_type, type_name, classe);
			nameTA.put(type_name, tad);
			typeAgents.put(code_type, tad);
		}
		
		//observeur locaux
		int codeType = code_type; //nécessaire pour valider l'écriture de la lambda fonction
		synchronized(async_aagent_observers) { for(AgentUObserver auo : async_aagent_observers) ThreadStack.perform( () -> auo.agentTypeAdded(type_name, codeType)); } //on notifie les observers
		synchronized(sync_aagent_observers) { for(AgentUObserver auo : sync_aagent_observers) auo.agentTypeAdded(type_name, codeType); } //on notifie les observers

		//notification des classes filles
		notifyNewTypeAgent(code_type, type_name);
		
		notifyChange(); //maj du fichier d'enregistrement dans 500ms maxi
		return(code_type);

	}

	public Iterable<Integer> allAgentsId(){
		synchronized(agents){ 
			return(Util.copyIntegerIterable(agents.keySet()));
		}
	}
	
	public int nbAgent(){
		return(agents.size());
	}
	
	public Iterable<Integer> agentsId(String name){
		if(name==null) return(null);
		synchronized(agents){ 
			Vector<Integer> rep = new Vector<Integer>();
			for(ExistingAgentU cur : agents.values()){
				if(name.equals(cur.name) && (cur.state >= AgentU.OFF)) rep.add(cur.agent_id);
			}
			return(rep);
		}
	}
	
	public boolean existAgent(int agent_id){
		ExistingAgentU eau = agents.get(agent_id);
		return(eau!=null && (eau.state >= AgentU.OFF));
	}
	
	public boolean existAgentWithName(String name){
		if(name==null) return(false);
		synchronized(agents){ 
			for(ExistingAgentU cur : agents.values()){
				if(name.equals(cur.name) && (cur.state >= AgentU.OFF)) return(true);
			}
			return(false);
		}
	}
	
	public int agentType(int agentId){
		ExistingAgentU eau = agents.get(agentId);
		if(eau!=null) return(eau.type.code);
		return(-1);
	}
	public String agentTypeName(int agentId){
		ExistingAgentU eau = agents.get(agentId);
		if(eau!=null) return(eau.type.name);
		return(null);		
	}
	public String agentName(int agentId){
		ExistingAgentU eau = agents.get(agentId);
		if(eau!=null) return(eau.name);
		return(null);		
	}
	public byte agentState(int agentId){
		ExistingAgentU eau = agents.get(agentId);
		if(eau!=null) return(eau.state);
		return(AgentU.NOT_EXIST);		
	}
	public byte agentRunMode(int agentId){
		ExistingAgentU eau = agents.get(agentId);
		if(eau!=null) return(eau.run_mode);
		return(AgentU.NOT_EXIST);			
	}
	
	//Pilotage des agents
	public boolean runAgent(int agentId){
		ExistingAgentU eau = agents.get(agentId);
		if(eau!=null) return(eau.run(AgentU.LOCAL_ORDER));
		return(false);		
	}
	public boolean stopAgent(int agentId){
		ExistingAgentU eau = agents.get(agentId);
		if(eau!=null) return(eau.stop(AgentU.LOCAL_ORDER));
		return(false);			
	}
	public boolean pauseAgent(int agentId){
		ExistingAgentU eau = agents.get(agentId);
		if(eau!=null) return(eau.pause(AgentU.LOCAL_ORDER));
		return(false);		
	}
	public boolean changeRunModeAgent(int agentId, byte run_mode){
		ExistingAgentU eau = agents.get(agentId);
		if(eau!=null) return(eau.changeRunMode(run_mode, AgentU.LOCAL_ORDER));
		return(false);		
	}
	
	public ConstantIterable<SocialSpecif> socialsSpecif(String name_type_agent){
		TypeAgent ta = nameTA.get(name_type_agent);
		if(ta!=null) return(ta.socialsSpecif());
		return(null);
	}
	public ConstantIterable<SocialSpecif> socialsSpecif(int code_type){
		TypeAgent ta = typeAgents.get(code_type);
		if(ta!=null) return(ta.socialsSpecif());
		return(null);
	}
	public byte setSocialPlug(int agentId, String name_slot, String name_space, String password, String local_id){
		ExistingAgentU eau = agents.get(agentId);
		if(eau!=null) return(eau.setSocialPlug(name_slot, name_space, password, local_id));
		return(AgentU.NOT_EXIST);			
	}
	public boolean existSocialPlug(int agentId, String name_slot){
		ExistingAgentU eau = agents.get(agentId);
		if(eau!=null) return(eau.existSocialPlug(name_slot));
		return(false);			
	}
	public String getSocialPlugNameSpace(int agentId, String name_slot){ 	//null si non pluggé
		ExistingAgentU eau = agents.get(agentId);
		if(eau!=null) return(eau.getSocialPlugNameSpace(name_slot));
		return(null);				
	}
	public boolean isSocialPlugPrivateSpace(int agentId, String name_slot){
		ExistingAgentU eau = agents.get(agentId);
		if(eau!=null) return(eau.isSocialPlugPrivateSpace(name_slot));
		return(false);			
	}
	public String getSocialSlotConfigLocalId(int agentId, String name_slot){ 	//null si non pluggé
		ExistingAgentU eau = agents.get(agentId);
		if(eau!=null) return(eau.getSocialSlotConfigLocalId(name_slot));
		return(null);				
	}	
	public String getSocialPlugCurrentLocalId(int agentId, String name_slot){ 	//null si non pluggé
		ExistingAgentU eau = agents.get(agentId);
		if(eau!=null) return(eau.getSocialPlugCurrentLocalId(name_slot));
		return(null);				
	}	
	
	//Gestion des slots physiques
	public ConstantIterable<PhysicalSpecif> physicalsSpecif(int code_type_agent){ 
		TypeAgent ta = typeAgents.get(code_type_agent);
		if(ta!=null) return(ta.physicalsSpecif());
		return(null);		
	}
	public ConstantIterable<PhysicalSpecif> physicalsSpecif(String name_type_agent){ 
		TypeAgent ta = nameTA.get(name_type_agent);
		if(ta!=null) return(ta.physicalsSpecif());
		return(null);		
	}
	public Iterable<Integer> physicalPlugRules(int agentId, String name_slot){ 
		ExistingAgentU eau = agents.get(agentId);
		if(eau!=null) return(eau.physicalPlugRules(name_slot));
		return(null);	 
	}
	public int addPhysicalPlugRule(int agentId, String name_slot, PhysicalPlugRule rule){
		ExistingAgentU eau = agents.get(agentId);
		if(eau!=null) return(eau.addPhysicalPlugRule(name_slot, rule));
		return(-1);	 
	}
	public byte movePhysicalPlugRule(int agentId, String name_slot, byte rule_id, byte direction){
		ExistingAgentU eau = agents.get(agentId);
		if(eau!=null) return(eau.movePhysicalPlugRule(name_slot, rule_id, direction));
		return(-1);	 
	}
	public int removePhysicalPlugRule(int agentId, String name_slot, int rule_id){
		ExistingAgentU eau = agents.get(agentId);
		if(eau!=null) return(eau.removePhysicalPlugRule(name_slot, rule_id));
		return(-1);	 
	}
	public boolean physicalSlotPlugged(int agentId, String name_slot){
		ExistingAgentU eau = agents.get(agentId);
		if(eau!=null) return(eau.physicalSlotPlugged(name_slot));
		return(false);	 
	}
	public Spot getPhysicalPlugSpot(int agentId, String name_slot){
		ExistingAgentU eau = agents.get(agentId);
		if(eau!=null) return(eau.getPhysicalPlugSpot(name_slot));
		return(null);	 
	}
	public int getPhysicalPlugDeviceId(int agentId, String name_slot){
		ExistingAgentU eau = agents.get(agentId);
		if(eau!=null) return(eau.getPhysicalPlugDeviceId(name_slot));
		return(-1);	 
	}		
	
	//[Pour ExistingAgentU	
	protected void notifyChangeRunMode(ExistingAgentU eau, int agent_id, byte old_run_mode, byte new_run_mode){
		//observeur locaux
		synchronized(async_aagent_observers) { for(AgentUObserver obs : async_aagent_observers) ThreadStack.perform(() -> obs.agentRunModeChange(agent_id, old_run_mode, new_run_mode)); }
		notifyChange(); //maj du fichier d'enregistrement dans 500ms maxi
		synchronized(sync_aagent_observers) { for(AgentUObserver obs : sync_aagent_observers) obs.agentRunModeChange(agent_id, old_run_mode, new_run_mode); }
	}
	protected void notifyLog(ExistingAgentU eau, int agent_id, String lmess, long time) {
		//observeur locaux
		synchronized(async_aagent_observers) { for(AgentUObserver obs : async_aagent_observers) ThreadStack.perform(() -> obs.agentLog(agent_id, lmess, time)); }
		synchronized(sync_aagent_observers) { for(AgentUObserver obs : sync_aagent_observers) obs.agentLog(agent_id, lmess, time); }
	}
	protected void notifyAgentRename(ExistingAgentU eau, int agent_id, String name){
		//observeur locaux
		synchronized(async_aagent_observers) { for(AgentUObserver obs : async_aagent_observers) ThreadStack.perform(() -> obs.agentRenamed(agent_id, name)); }
		notifyChange(); //maj du fichier d'enregistrement dans 500ms maxi
		synchronized(sync_aagent_observers) { for(AgentUObserver obs : sync_aagent_observers) obs.agentRenamed(agent_id, name); }
	}
	
	protected void notifyAgentRetype(ExistingAgentU eau, int agent_id, int code_type){
		//observeur locaux
		synchronized(async_aagent_observers) { for(AgentUObserver obs : async_aagent_observers) ThreadStack.perform(() -> obs.agentRetypep(agent_id, code_type)); }
		notifyChange(); //maj du fichier d'enregistrement dans 500ms maxi
		synchronized(sync_aagent_observers) { for(AgentUObserver obs : sync_aagent_observers) obs.agentRetypep(agent_id, code_type); }
	}
	
	protected void notifyAgentWaitRetype(ExistingAgentU eau, int agent_id, int code_type){
		//observeur locaux
		synchronized(async_aagent_observers) { for(AgentUObserver obs : async_aagent_observers) ThreadStack.perform(() -> obs.agentWaitRetype(agent_id, code_type)); }
		notifyChange(); //maj du fichier d'enregistrement dans 500ms maxi	
		synchronized(sync_aagent_observers) { for(AgentUObserver obs : sync_aagent_observers) obs.agentWaitRetype(agent_id, code_type); }
	}
	protected void notifyPhysicalPlugRuleMove(ExistingAgentU eau, int agent_id, String name_slot, int rule_id, byte direction){
		//observeur locaux
		synchronized(async_aagent_observers) { for(AgentUObserver obs : async_aagent_observers) ThreadStack.perform(() -> obs.physicalPlugRuleMoved(agent_id, name_slot, rule_id, direction)); }	
		notifyChange(); //maj du fichier d'enregistrement dans 500ms maxi
		synchronized(sync_aagent_observers) { for(AgentUObserver obs : sync_aagent_observers) obs.physicalPlugRuleMoved(agent_id, name_slot, rule_id, direction); }
	}
	
	protected void notifyAddRule(ExistingAgentU eau, int agent_id, String name_slot, IdentifiedPlugRule ipr){
		//observeur locaux
		synchronized(async_aagent_observers) { for(AgentUObserver obs : async_aagent_observers) ThreadStack.perform(() -> obs.physicalPlugRuleAdded(agent_id, name_slot, ipr.id, ipr.order, ipr.rule)); }
		notifyChange(); //maj du fichier d'enregistrement dans 500ms maxi
		synchronized(sync_aagent_observers) { for(AgentUObserver obs : sync_aagent_observers) obs.physicalPlugRuleAdded(agent_id, name_slot, ipr.id, ipr.order, ipr.rule); }
	}
	protected void notifyRemoveRule(ExistingAgentU eau, int agent_id, String name_slot, int rule_id){
		//observeur locaux
		synchronized(async_aagent_observers) { for(AgentUObserver obs : async_aagent_observers) ThreadStack.perform(() -> obs.physicalPlugRuleRemoved(agent_id, name_slot, rule_id)); }
		notifyChange(); //maj du fichier d'enregistrement dans 500ms maxi
		synchronized(sync_aagent_observers) { for(AgentUObserver obs : sync_aagent_observers) obs.physicalPlugRuleRemoved(agent_id, name_slot, rule_id); }
	}
	
	protected void notifySocialChange(ExistingAgentU eau, int agent_id, String slot_name,  String name_space, String md5_space_password, String local_id){
		//observeur locaux
		synchronized(async_aagent_observers) { for(AgentUObserver obs : async_aagent_observers) ThreadStack.perform(() -> obs.socialChange(agent_id, slot_name, name_space, md5_space_password, local_id)); }
		notifyChange(); //maj du fichier d'enregistrement dans 500ms maxi
		synchronized(sync_aagent_observers) { for(AgentUObserver obs : sync_aagent_observers) obs.socialChange(agent_id, slot_name, name_space, md5_space_password, local_id); }
	}
	
	protected void notifySocialPlug(ExistingAgentU eau, int agent_id, String slot_name, String current_id, int ctx){
		//observeur locaux
		synchronized(async_aagent_observers) { for(AgentUObserver obs : async_aagent_observers) ThreadStack.perform(() -> obs.socialPlug(agent_id, slot_name, current_id, ctx)); }
		synchronized(sync_aagent_observers) { for(AgentUObserver obs : sync_aagent_observers) obs.socialPlug(agent_id, slot_name, current_id, ctx); }
	}
	protected void notifySocialUnplug(ExistingAgentU eau, int agent_id, String slot_name, int ctx){
		//observeur locaux
		synchronized(async_aagent_observers) { for(AgentUObserver obs : async_aagent_observers) ThreadStack.perform(() -> obs.socialUnplug(agent_id, slot_name, ctx)); }
		synchronized(sync_aagent_observers) { for(AgentUObserver obs : sync_aagent_observers) obs.socialUnplug(agent_id, slot_name, ctx); }
	}
	
	protected void notifyPhysicalChange(ExistingAgentU eau, int agent_id, String name_slot, boolean optional_plug, Spot spoter, int device_id){
		//observeur locaux
		synchronized(async_aagent_observers) { for(AgentUObserver obs : async_aagent_observers) ThreadStack.perform(() -> obs.physicalChange(agent_id, name_slot, spoter, device_id)); }
		synchronized(sync_aagent_observers) { for(AgentUObserver obs : sync_aagent_observers) obs.physicalChange(agent_id, name_slot, spoter, device_id); }
	}
	
	protected void notifyPhysicalPlug(ExistingAgentU eau, int agent_id, String name_slot, boolean optional_plug, Spot spoter, int device_id){
		//observeur locaux
		synchronized(async_aagent_observers) { for(AgentUObserver obs : async_aagent_observers) ThreadStack.perform(() -> obs.physicalPlug(agent_id, name_slot, spoter, device_id)); }
		synchronized(sync_aagent_observers) { for(AgentUObserver obs : sync_aagent_observers) obs.physicalPlug(agent_id, name_slot, spoter, device_id); }
	}
	
	protected void notifyPhysicalUnplug(ExistingAgentU eau, int agent_id, String name_slot, boolean optional_plug){
		//observeur locaux
		synchronized(async_aagent_observers) { for(AgentUObserver obs : async_aagent_observers) ThreadStack.perform(() -> obs.physicalUnplug(agent_id, name_slot)); }
		synchronized(sync_aagent_observers) { for(AgentUObserver obs : sync_aagent_observers) obs.physicalUnplug(agent_id, name_slot); }
	}
	protected void notifyAgentStateChange(ExistingAgentU eau, int agent_id, byte old_state, byte new_state, byte ctx){
		//observeur locaux
		synchronized(async_aagent_observers) { for(AgentUObserver obs : async_aagent_observers) ThreadStack.perform(() -> obs.agentStateChange(agent_id, old_state, new_state, ctx)); }
		synchronized(sync_aagent_observers) { for(AgentUObserver obs : sync_aagent_observers) obs.agentStateChange(agent_id, old_state, new_state, ctx); }
	}
	protected void notifyAgentRemove(ExistingAgentU eau, int agent_id){
		//notification des observeurs locaux
		synchronized(async_aagent_observers) { for(AgentUObserver obs : async_aagent_observers) ThreadStack.perform(() -> obs.agentRemoved(agent_id)); }
		synchronized(sync_aagent_observers) { for(AgentUObserver obs : sync_aagent_observers) obs.agentRemoved(agent_id); }
		notifyChange(); //maj du fichier d'enregistrement dans 500ms maxi
	}
	//]
	
	//[
	public void performRuleToAnonymousDevice(HostedDevice hdevice){ //on doit déclencher l'analyse des règles sur ce device sans nom
		synchronized(agents){ for(ExistingAgentU agentu : agents.values()){
			agentu.performRuleToAnonymousDevice(hdevice);
		}}
	}
	public void performRuleToNamedDevice(HostedDevice hdevice){ //on doit déclencher l'analyse des règles sur ce device sans nom
		synchronized(agents){ for(ExistingAgentU agentu : agents.values()){
			agentu.performRuleToNamedDevice(hdevice);
		}}
	}
	//]
}