package skuad.agentu.implementation.amanager;

import java.util.Vector;
import skuad.hubudp.Spot;
import skuad.util.base.Network;
import skuad.agentu.PhysicalPlugRule;
import skuad.uda.*;

public class DeviceHoster {
	RunAgentContext racontext;
	int device_version;
	Spot spot;
	String spot_name, node_name;
	Vector<HostedDevice> devices; //liste des devices portés par ce noeud
	
	public DeviceHoster(RunAgentContext vracontext, Spot h, int dv){
		racontext=vracontext;
		spot = h;
		device_version = dv;
		devices = new Vector<HostedDevice>();
		spot_name=null; node_name=null;
	}
	
	public Spot getSpot() { return(spot); }
	
	public int getDeviceListVersion() {
		return(device_version);
	}
	
	public void setDeviceListVersion(int nv) {
		device_version=nv;
	}
	
	public HostedDevice addDevice(int did, int dt, String name) {
		HostedDevice hdevice = new HostedDevice(this, did, dt, name); 
		devices.add(hdevice);
		if(name!=null) performRuleToNamedDevice(hdevice);
		else performRuleToAnonymousDevice(hdevice);
		return(hdevice);
	}
	
	public boolean removeDevice(int did) {
		for(int i=0; i<devices.size(); i++) {
			HostedDevice hd = devices.get(i);
			if(hd.device_id==did) {
				hd.notifyDown();
				devices.remove(i);
				return(true);
			}
		}
		return(false);
	}
	public boolean removeDevice(HostedDevice hdevice) {
		if(devices.remove(hdevice)) {
			hdevice.notifyDown();
			return(true);
		}
		return(false);
	}
	
	public void setSpotName(String name) {
		spot_name=name;
		//OXO: pour le cas de changement de nom, cela devrait impacter l'analyse des règles :(
	}
	public void setNodeName(String name) {
		node_name=name;
		//OXO: pour le cas de changement de nom, cela devrait impacter l'analyse des règles :(
	}
	
	public int nbDevice() { return(devices.size()); }
	
	public Vector<Integer> getDeviceIds(){
		Vector<Integer> existing_devices = new Vector<Integer>();
		for(HostedDevice hdevice : devices) existing_devices.add(hdevice.device_id);
		return(existing_devices);
	}
	
	public void removeAllDevice(){ //tous les devices de cet spot doivent être supprimé
		for(HostedDevice hdevice : devices) hdevice.notifyDown();
		devices.removeAllElements();
	}
	
	public HostedDevice getDevice(int did) {
		for(HostedDevice hdevice : devices) if(hdevice.device_id==did) return(hdevice);
		return(null);
	}
/*	public HostedDevice getDevice(Device device) {
		for(HostedDevice hdevice : devices) if(hdevice.device==device) return(hdevice);
		return(null);
	}*/	
	public HostedDevice findDevice(PhysicalPlugRule prule) {
//System.out.println("find... node_name = "+node_name);		
		int node_key = prule.nodeKey();
		int spot_key = prule.spotKey();
//System.out.println("node_key = "+node_key+" // "+prule.stringNodeFilter());		
		String gid = spot.getIdGlobal();
		if( 	(	(node_key == PhysicalPlugRule.ALL) 
				|| 	(node_key == PhysicalPlugRule.NAME && node_name!=null && node_name.equals(prule.stringNodeFilter()))
				|| 	(node_key == PhysicalPlugRule.IP && Network.isSameIp(spot.getIp(), prule.stringNodeFilter()))
				)
				&& ( 	(spot_key == PhysicalPlugRule.ALL) 
						|| 	(spot_key == PhysicalPlugRule.NAME && spot_name!=null && spot_name.equals(prule.stringSpotFilter()))		
						||	 	(spot_key == PhysicalPlugRule.ID && gid!=null && gid.equals(prule.stringSpotFilter()))		
						)
				){ //on peut examiner les devices de cet hote
			int device_key = prule.deviceKey();
			for(HostedDevice device : devices){
				if(			(device_key == PhysicalPlugRule.ALL)
						|| 	(device_key == PhysicalPlugRule.NAME && device.device_name!=null && device.device_name.equals(prule.stringDeviceFilter()))
						|| 	(device_key == PhysicalPlugRule.TYPE && device.device_type==prule.intDeviceFilter())
						|| 	(device_key == PhysicalPlugRule.ID && device.device_id==prule.intDeviceFilter())
						) return(device);
			}
		}
		return(null);
	}
	
	protected void performRuleToAnonymousDevice(HostedDevice hdevice){ //on doit déclencher l'analyse des règles sur ce device sans nom
		racontext.performRuleToAnonymousDevice(hdevice);
	}

	protected void performRuleToNamedDevice(HostedDevice hdevice){ //on doit déclencher l'analyse des règles sur ce device sans nom
		racontext.performRuleToNamedDevice(hdevice);
	}
	protected Device linkDevice(HostedDevice hdevice) {
		 return(racontext.linkDevice(this, hdevice));
	}
	protected boolean unlinkDevice(HostedDevice hdevice, Device device) {
		 return(racontext.unlinkDevice(this, hdevice, device));
	}	
}