/**
 * Classe de gestion d'un agent existant
 * 
 */
package skuad.agentu.implementation.amanager;

import java.util.*;
import skuad.agentu.*;
import skuad.hubudp.*;
import skuad.ubiquity.space.Avatar;
import skuad.uda.*;
import skuad.util.base.*;
import skuad.util.thread.ThreadStack;
import skuad.util.value.Value;
import skuad.util.databytes.DataWriter;
import skuad.ubiquity.space.mailbox.*;

public class ExistingAgentU {
	
	public static int LOG_STACK_SIZE=20;
	
	//propriété statique
	int agent_id;
	String name;
	TypeAgent type;
	Value config;
	
	//propriétés dynamiques
	byte run_mode;								//mode d'exécution de l'agent
	TypeAgent wait_retype;						//si !=null signal un changement de type en attente
	Map<String, PhysicalSlotManager> pslots;	//slots physiques ouverts de l'agent (ceux disposant d'au moins une règle)
	Map<String, SocialSlotManager> sslots;		//slots sociaux ouverts de l'agent
	Vector<LogMessage> logs;					//pile des messages logs (taille maximum LOG_STACK_SIZE)
	
	byte state;									//état courant de l'agent
	AgentU agent;								//instance de l'agent si state>OFF (== RUN ou PAUSE)
	ImplAgentUDescriptor aud;
	int plugged_no_optional; 					//nbr de plugs non optionnel réalisés (social + physique)
	Vector<Pulsor> pulsors;						//table des pulsors actifs
	Vector<AgentUObserver> sync_agent_observers, async_agent_observers;
	AgentManager amanager;
	
	protected ExistingAgentU(AgentManager vamanager, TypeAgent ta, int aid, String vname, Value vconfig, boolean loadSocialSlot){
		amanager = vamanager;
		agent_id=aid;
		name = vname;
		config = (vconfig==null ? null : vconfig.toConstant());
		type = ta;
		agent = null;
		aud = new ImplAgentUDescriptor(this);
		pslots = new HashMap<String, PhysicalSlotManager>();
		sslots =  new HashMap<String, SocialSlotManager>();
		plugged_no_optional = 0;
		pulsors = new Vector<Pulsor>();	
		state = AgentU.OFF;
		run_mode = AgentU.MANUAL;
		sync_agent_observers = null; async_agent_observers=null;
		wait_retype = null; 
		logs = null;
		
		//on charge les slots physiques
		for(PhysicalSpecif devspec : ta.physical_specif){
			pslots.put(devspec.nameSlot(), new PhysicalSlotManager(this, devspec));
		}
		
		//on charge les slots sociaux par défaut
		if(loadSocialSlot){ 
			for(SocialSpecif slotspec : ta.social_specif){
				sslots.put(slotspec.nameSlot(), new SocialSlotManager(this, slotspec, true));
			}
		}
//System.out.println("agent_id = "+agent_id+"  loaded social slot = "+sslots.size()+" ta.social_specif.size() = "+ta.social_specif.size()+" loadSocialSlot = "+loadSocialSlot); //OXO	
	}
	
	public RunAgentContext getRunContext() { return(amanager.getRunContext()); }
	public long curTime(){ return(amanager.curTime()); }		//lecture du temps courant
	
	//[Pour ImplAgenUDescriptor
	protected Pulsor newPulsor(String label, long delay, long periode, boolean fireUptodate_or_resetOnResume) {
		return(amanager.newPulsor(this, label, delay, periode, fireUptodate_or_resetOnResume));
	}
	//]
	
	public int getState() { return(state); }
	
	public void checkSocialSlot(){
		for(SocialSpecif slotspec : type.social_specif){
			String name = slotspec.nameSlot();
			if(!sslots.containsKey(name)) sslots.put(name, new SocialSlotManager(this, slotspec, true));
		}		
	}
	
	public boolean rename(String vname){
		if(vname!=null){
			if(!name.equals(vname)){
				name = vname;
				
				if(async_agent_observers!=null) synchronized(async_agent_observers) { for(AgentUObserver obs : async_agent_observers) ThreadStack.perform( () -> obs.agentRenamed(agent_id, name)); }
				if(sync_agent_observers!=null) synchronized(sync_agent_observers) { for(AgentUObserver obs : sync_agent_observers) obs.agentRenamed(agent_id, name); }
				amanager.notifyAgentRename(this, agent_id, name);
			}
			return(true);
		} 
		return(false);
	}
	
	protected void applyRetype(TypeAgent type_agent){ //on applique le changement de type indiqué
		if(type!=type_agent){
			//ici on a forcément : state==AgentU.OFF
			//--> seules les configurations de slot de même nom que l'un de ceux du nouveau type seront conservé, les autres seront supprimé
			type = type_agent;

			//[Les slots physiques
			Vector<String> to_remove=new Vector<String>(); 
			for(PhysicalSlotManager psm : pslots.values()){
				String slot_name = psm.slot_specif.nameSlot();
				PhysicalSpecif ps = type_agent.getPhysicalSlotSpecif(slot_name);
				if(ps==null) {
					to_remove.add(slot_name); //on devra supprimer ce slot
					psm.unplug(true, false);
				} else psm.slot_specif = ps; //on met juste à jour la spécification du slot (en concervant sa configuration) 
			}
			//On supprime les slots obsolêtes
			for(String slot_name : to_remove) pslots.remove(slot_name);

			//on ajoute les nouveaux slots
			for(PhysicalSpecif devspec : type.physical_specif){
				String slot_name = devspec.nameSlot();
				if(!pslots.containsKey(slot_name)) pslots.put(slot_name, new PhysicalSlotManager(this, devspec));
			}
			//]


			//[Les slots sociaux
			to_remove.removeAllElements();
			for(SocialSlotManager ssm : sslots.values()){
				String slot_name = ssm.slot_specif.nameSlot();
				SocialSpecif ss = type_agent.getSocialSlotSpecif(slot_name);
				if(ss==null) to_remove.add(slot_name); //on devra supprimer ce slot, NB: l'avatar ne peut pas être instancié car on a state=OFF
				else {
					boolean load_def = !ssm.specialConfig();
					ssm.slot_specif = ss; //on met juste à jour la spécification du slot (en concervant sa configuration)
					if(load_def) ssm.loadDefautConfig();
				}
			}
			//On supprime les slots obsolêtes
			for(String slot_name : to_remove) sslots.remove(slot_name);

			//on ajoute les nouveaux slots
			for(SocialSpecif devspec : type.social_specif){
				String slot_name = devspec.nameSlot();
				if(!sslots.containsKey(slot_name)) sslots.put(slot_name, new SocialSlotManager(this, devspec, true));
			}
			//]

			if(async_agent_observers!=null) synchronized(async_agent_observers) { for(AgentUObserver obs : async_agent_observers) ThreadStack.perform( () -> obs.agentRetypep(agent_id, type_agent.code)); }
			if(sync_agent_observers!=null) synchronized(sync_agent_observers) { for(AgentUObserver obs : sync_agent_observers) obs.agentRetypep(agent_id, type_agent.code); }
			amanager.notifyAgentRetype(this, agent_id, type_agent.code);
		}
	}
	
	public boolean retype(TypeAgent type_agent){
		if(type_agent!=null && state>=AgentU.OFF){
			if(type_agent!=type){
				if(state==AgentU.OFF) applyRetype(type_agent); //on peut appliquer le changement de type maintenant
				else if(wait_retype!=type_agent){ //on doit mettre le changement de type en attente
					wait_retype = type_agent;

					if(async_agent_observers!=null) synchronized(async_agent_observers) { for(AgentUObserver obs : async_agent_observers) ThreadStack.perform( () -> obs.agentWaitRetype(agent_id, wait_retype.code)); }
					if(sync_agent_observers!=null) synchronized(sync_agent_observers) { for(AgentUObserver obs : sync_agent_observers) obs.agentWaitRetype(agent_id, wait_retype.code); }
					amanager.notifyAgentWaitRetype(this, agent_id, wait_retype.code);
				}
			}
			return(true);
		}
		return(false);
	}

	
	public boolean changeRunMode(int new_mode, byte ctx){
		if(run_mode!=new_mode && new_mode>=AgentU.MANUAL && new_mode<=AgentU.BREAKPLUG_PAUSE){
			byte old_run_mode = run_mode;
			run_mode = (byte) new_mode;
			
			//notification du changement de mode
			if(async_agent_observers!=null) synchronized(async_agent_observers) { for(AgentUObserver obs : async_agent_observers) ThreadStack.perform( () -> obs.agentRunModeChange(agent_id, old_run_mode, run_mode)); }
			if(sync_agent_observers!=null) synchronized(sync_agent_observers) { for(AgentUObserver obs : sync_agent_observers) obs.agentRunModeChange(agent_id, old_run_mode, run_mode); }
			amanager.notifyChangeRunMode(this, agent_id, old_run_mode, run_mode);

			//détermination de l'incidence sur l'état actuel de l'agent
			if(amanager.getRunContext().canRunAgent()) refreshState(ctx);
			return(true);
		}
		return(false);
	}
	
	public void refreshState(byte ctx) {	
		if(run_mode!=AgentU.MANUAL && state>=AgentU.OFF){
			if(state!=AgentU.RUN){ //doit-on démarrer l'agent ?
				if(run_mode==AgentU.ALLWAYS_RUN || (plugged_no_optional==type.nb_no_optional_plug)) run(ctx);
				else if(state==AgentU.PAUSE && run_mode==AgentU.BREAKPLUG_STOP) stop(ctx);
			} else { //ou l'arréter ?
				if(plugged_no_optional<type.nb_no_optional_plug){
					if(run_mode==AgentU.BREAKPLUG_PAUSE) pause(ctx);
					else if(run_mode==AgentU.BREAKPLUG_STOP) stop(ctx);
				}
			}
		}		
	}
	
	public synchronized void log(String mess){ 
		if(mess!=null){
			if(logs==null) logs = new Vector<LogMessage>();
			long time = amanager.getRunContext().getCurrentTime();
			LogMessage lm = new LogMessage(mess, time);
			logs.add(lm);
			if(logs.size()>LOG_STACK_SIZE) logs.removeElementAt(0); //si on dépasse la taille limite, on supprime le log le plus ancien
			
			//notification aux observeurs
			if(async_agent_observers!=null) synchronized(async_agent_observers) { for(AgentUObserver obs : async_agent_observers) ThreadStack.perform( () -> obs.agentLog(agent_id, mess, time)); }
			if(sync_agent_observers!=null) synchronized(sync_agent_observers) { for(AgentUObserver obs : sync_agent_observers) obs.agentLog(agent_id, mess, time); }
			
			amanager.notifyLog(this, agent_id, mess, time);	
		}
	}	
	
	//[Pour SocialSlotManager
	protected void fireSocialChange(String slot_name, String name_space, String space_password, String local_id){
		String md5_space_password=(space_password==null || space_password.length()==0 ? null : Util.cryptMD5(space_password) );
		if(async_agent_observers!=null) synchronized(async_agent_observers) { for(AgentUObserver obs : async_agent_observers) ThreadStack.perform( () -> obs.socialChange(agent_id, slot_name, name_space, md5_space_password, local_id)); }
		if(sync_agent_observers!=null) synchronized(sync_agent_observers) { for(AgentUObserver obs : sync_agent_observers) obs.socialChange(agent_id, slot_name, name_space, md5_space_password, local_id); }
		
		amanager.notifySocialChange(this, agent_id,slot_name, name_space, md5_space_password, local_id);
	}		
	protected void fireSocialPlug(String slot_name,  String current_id, int ctx){
		if(async_agent_observers!=null) synchronized(async_agent_observers) { for(AgentUObserver obs : async_agent_observers) ThreadStack.perform( () -> obs.socialPlug(agent_id, slot_name, current_id, ctx)); }
		if(sync_agent_observers!=null) synchronized(sync_agent_observers) { for(AgentUObserver obs : sync_agent_observers) obs.socialPlug(agent_id, slot_name, current_id, ctx); }

		amanager.notifySocialPlug(this, agent_id,slot_name, current_id, ctx);	
	}
	protected void fireSocialUnplug(String slot_name, int ctx){
		if(async_agent_observers!=null) synchronized(async_agent_observers) { for(AgentUObserver obs : async_agent_observers) ThreadStack.perform( () -> obs.socialUnplug(agent_id, slot_name, ctx)); }
		if(sync_agent_observers!=null) synchronized(sync_agent_observers) { for(AgentUObserver obs : sync_agent_observers) obs.socialUnplug(agent_id, slot_name, ctx); }
		
		amanager.notifySocialUnplug(this, agent_id,slot_name, ctx);			
	}
	protected void notifyReceiveMessage(Avatar avatar, MailBox box, Message mess){
		if(agent!=null && state==AgentU.RUN) agent.receive(aud, avatar, box, mess);
	}
	//]

	public void writeAllSocialPlug(DataWriter wcmd){ //(int: nb_slot)...|pour chaque slot: (string: name)(string: space_name)[bool: is_private](string: local_id)(string: current_id)|...
		synchronized(sslots){
//System.out.println("sslots.size() = "+sslots.size());
			wcmd.i(sslots.size());
			for(SocialSlotManager ssm : sslots.values()){
				wcmd.s(ssm.slot_specif.nameSlot()).s(ssm.space_name).b(ssm.space_password!=null).s(ssm.slot_specif.localId()).s(ssm.currentId());				
			}
		}
	}
	
	public void writeAllPhysicalPlug(DataWriter wcmd){ //(int: nb_slot)...|pour chaque slot: (string: name)(int: type)[bool: optional][bool: plugged]{si plugged1 : (int: true_type)(string: spot_owner_globalId)[device_id] }|...
		synchronized(pslots){
//System.out.println("Nombr de slot physique détecté : "+pslots.size());	
			wcmd.i(pslots.size());
			for(PhysicalSlotManager psm : pslots.values()){
				wcmd.s(psm.slot_specif.nameSlot()).i(psm.slot_specif.deviceType()).b(psm.slot_specif.isOptional());
				if(psm.plugged()){
					wcmd.b(true).i(psm.true_type).s(psm.plugSpot().getIdGlobal());
					wcmd.b(psm.plugDeviceId());
				} else wcmd.b(false);
			}
		}
	}
	
	public void writeAllPhysicalPlugRule(String name_slot, DataWriter wcmd){ 
		//[nb_rules]...|pour chaque règle: [rule_id][rule_order][node_key]{si node_key!=ALL: (string: node_value) }[spot_key]{si spot_key==NAME: (string: spot_value) | si  spot_key==PORT: {short: port_value}}[device_key]{si device_key==NAME: (string: device_value) | si  device_key==TYPE ou ID: (int: device_value)}|...
		PhysicalSlotManager psm = pslots.get(name_slot);
		if(psm!=null) psm.writeAllPhysicalPlugRule(wcmd);
		else wcmd.b(0);
	}

	public void writePhysicalPlugRule(String name_slot, int rule_id, DataWriter wcmd){ 
		//[rule_id][rule_order][node_key]{si node_key!=ALL: (string: node_value) }[spot_key]{si spot_key==NAME: (string: spot_value) | si  spot_key==PORT: {short: port_value}}[device_key]{si device_key==NAME: (string: device_value) | si  device_key==TYPE ou ID: (int: device_value)}
		PhysicalSlotManager psm = pslots.get(name_slot);
		if(psm!=null) psm.writePhysicalPlugRule(rule_id, wcmd);
	}
	
	public byte setSocialPlug(String name_slot, String name_space, String password, String local_id){ synchronized(sslots){
		if(name_space==null) return(AgentU.ERROR);
		SocialSlotManager ssm = sslots.get(name_slot);
		if(ssm==null){
			SocialSpecif ss = type.getSocialSlotSpecif(name_slot);
			if(ss!=null){
				ssm = new SocialSlotManager(this, ss, false);
				sslots.put(name_slot, ssm);
			} else return(AgentU.NOT_EXIST);
		}
		ssm.set(name_space, password, local_id);
		return(AgentU.OK);			
	}}
	
	public boolean existSocialPlug(String name_slot){
		return(sslots.containsKey(name_slot));							
	}
	
	public String getSocialPlugNameSpace(String name_slot){ 
		SocialSlotManager ssm = sslots.get(name_slot);
		if(ssm!=null) return(ssm.space_name);
		return(null);								
	}
	public boolean isSocialPlugPrivateSpace(String name_slot){
		SocialSlotManager ssm = sslots.get(name_slot);
		if(ssm!=null) return(ssm.space_password!=null);
		return(false);
	}	
	
	public String getSocialSlotConfigLocalId(String name_slot){ 	//null si non pluggé
		SocialSlotManager ssm = sslots.get(name_slot);
		if(ssm!=null) return(ssm.local_id);
		return(null);				
	}	
	public String getSocialPlugCurrentLocalId(String name_slot){ 	//null si non pluggé
		SocialSlotManager ssm = sslots.get(name_slot);
		if(ssm!=null && ssm.avatar!=null) return(ssm.avatar.participant().localId());
		return(null);				
	}	
	
	public Avatar getAvatar(String name_slot){ //null si l'avatar n'est pas instancié
		SocialSlotManager ssm = sslots.get(name_slot);
		if(ssm!=null) return(ssm.avatar);
		return(null);				
	}
	
	public Iterable<Integer> physicalPlugRules(String name_slot){ 
		PhysicalSlotManager psm = pslots.get(name_slot);
		if(psm!=null) return(psm.plugRulesIds());
		return(null);				
	}
	
	public int addPhysicalPlugRule(String name_slot, PhysicalPlugRule rule){
		PhysicalSlotManager psm = pslots.get(name_slot);
		if(psm!=null)  return(psm.addRule(rule));
		return(AgentU.NOT_EXIST);	 
	}
	public byte movePhysicalPlugRule(String name_slot,int rule_id, int direction){
		PhysicalSlotManager psm = pslots.get(name_slot);
		if(psm!=null){
			if(direction==PhysicalPlugRule.UP) return(psm.upRule(rule_id));
			else if(direction==PhysicalPlugRule.DOWN) return(psm.downRule(rule_id));
			else return(AgentU.ERROR);
		}
		return(AgentU.NOT_EXIST);	 
	}	
	public int removePhysicalPlugRule(String name_slot, int rule_id){
		PhysicalSlotManager psm = pslots.get(name_slot);
		if(psm!=null) return(psm.removeRule(rule_id));
		return(AgentU.NOT_EXIST);	 
	}
	public boolean physicalSlotPlugged(String name_slot){
		PhysicalSlotManager psm = pslots.get(name_slot);
		return(psm!=null && psm.plugged());
	}
	public Spot getPhysicalPlugSpot(String name_slot){
		PhysicalSlotManager psm = pslots.get(name_slot);
		if(psm!=null) return(psm.plugSpot());
		return(null);
	}
	public int getPhysicalPlugDeviceId(String name_slot){
		PhysicalSlotManager psm = pslots.get(name_slot);
		if(psm!=null) return(psm.plugDeviceId());
		return(-1);
	}
	
	public PhysicalSlotManager getPhysicalSlotManager(String slot_name){ return(pslots.get(slot_name)) ;}
	
	protected synchronized boolean addAgentObserver(AgentUObserver obs, boolean sync){
		if(state >= AgentU.OFF) {
			if(sync) { 
				if(sync_agent_observers==null) sync_agent_observers= new Vector<AgentUObserver>();
				synchronized(sync_agent_observers) { sync_agent_observers.add(obs); }
			} else { 
				if(async_agent_observers==null) async_agent_observers= new Vector<AgentUObserver>(); 
				synchronized(async_agent_observers) { async_agent_observers.add(obs); }
			}
			return(true);
		}
		return(false);
	}
	protected boolean removeAgentObserver(AgentUObserver obs, boolean sync){
		if(sync) {
			if(sync_agent_observers!=null) synchronized(sync_agent_observers) { return(sync_agent_observers.remove(obs)); }
		} else {
			if(async_agent_observers!=null) synchronized(async_agent_observers) { return(async_agent_observers.remove(obs)); }
		}
		return(false);
	}
	
	public void writeLogs(DataWriter wcmd){ //(int: nb_log)...|pour chaque log: (String: mess)(doubleWord: time)|...
		if(logs!=null){
			wcmd.i(logs.size());
			for(int i=0; i<logs.size(); i++){
				LogMessage cur = logs.get(i);
				wcmd.s(cur.getMess()).d(cur.getTime());
			}
		} else wcmd.i(0);
	}
		
	protected boolean addPulsor(Pulsor p){ synchronized(pulsors){
		if(state>=AgentU.OFF){
			pulsors.add(p);
			if(state==AgentU.RUN) p.startOrResume(true); //l'agent est en cours d'exécution, on met en route le pulse maintenant
			return(true);
		}
		return(false);
	}}
	protected boolean removePulsor(Pulsor p){ synchronized(pulsors){ return(pulsors.remove(p)); } }
	
	
	protected void write(DataWriter dw){//(int: agent_id)[code_type_agent](string: agent_name)[state][run_mode]|
		dw.i(agent_id).b(type.code).b(wait_retype!=null ? wait_retype.code : 0).s(name).b(state).b(run_mode);
	}
	
	protected void notifyAddRule(String name_slot, IdentifiedPlugRule ipr){
		//notification du changement de mode
		if(async_agent_observers!=null) synchronized(async_agent_observers) { for(AgentUObserver obs : async_agent_observers) ThreadStack.perform( () -> obs.physicalPlugRuleAdded(agent_id, name_slot, ipr.id, ipr.order, ipr.rule)); }
		if(sync_agent_observers!=null) synchronized(sync_agent_observers) { for(AgentUObserver obs : sync_agent_observers) obs.physicalPlugRuleAdded(agent_id, name_slot, ipr.id, ipr.order, ipr.rule); }
		amanager.notifyAddRule(this, agent_id, name_slot, ipr);
	}
	protected void notifyMoveRule(String name_slot, int rule_id, byte direction){
		//notification du changement de mode
		if(async_agent_observers!=null) synchronized(async_agent_observers) { for(AgentUObserver obs : async_agent_observers) ThreadStack.perform( () -> obs.physicalPlugRuleMoved(agent_id, name_slot, rule_id, direction)); }
		if(sync_agent_observers!=null) synchronized(sync_agent_observers) { for(AgentUObserver obs : sync_agent_observers) obs.physicalPlugRuleMoved(agent_id, name_slot, rule_id, direction); }
		amanager.notifyPhysicalPlugRuleMove(this, agent_id, name_slot, rule_id, direction);		
	}
	
	protected void notifyRemoveRule(String name_slot, int rule_id){
		//notification du changement de mode
		if(async_agent_observers!=null) synchronized(async_agent_observers) { for(AgentUObserver obs : async_agent_observers) ThreadStack.perform( () -> obs.physicalPlugRuleRemoved(agent_id, name_slot, rule_id)); }
		if(sync_agent_observers!=null) synchronized(sync_agent_observers) { for(AgentUObserver obs : sync_agent_observers) obs.physicalPlugRuleRemoved(agent_id, name_slot, rule_id); }
		amanager.notifyRemoveRule(this, agent_id, name_slot, rule_id);
	}
	
	protected void notifyPlug(PhysicalSlotManager mslot, boolean plug_replacement, HostedDevice hosted_device, Device device){ //pour PhysicalSlotManager
//System.out.println("pluggued on "+mslot.slot_specif.nameSlot());
		boolean plug_notified=false;
		String name_slot = mslot.slot_specif.nameSlot();
		if(!plug_replacement && !mslot.slot_specif.isOptional()) {
			plugged_no_optional++;
			if(plugged_no_optional==type.nb_no_optional_plug){
				if(state>AgentU.OFF){
					agent.notifyPlugDevice(aud, name_slot, device); plug_notified=true;
					agent.event(aud, AgentU.ALL_PLUGGED, AgentU.PLUG_NOTIFICATION);
				}
				if(run_mode==AgentU.BREAKPLUG_PAUSE || run_mode==AgentU.BREAKPLUG_STOP) { 
					plug_notified=true; //l'agent sera notifié des plugs à son démarrage 
					run(AgentU.RUN_MODE_PLUG); 
				}
			}
		}
		if(state>AgentU.OFF){
			if(plug_replacement) agent.notifyChangeDevice(aud, name_slot, device); //nb: ici on a forcément plug_notified==false
			else if(!plug_notified)  {
				agent.notifyPlugDevice(aud, name_slot, device);
			}
		}	
		if(plug_replacement){
			if(async_agent_observers!=null) synchronized(async_agent_observers) { for(AgentUObserver obs : async_agent_observers) ThreadStack.perform( () -> obs.physicalChange(agent_id, name_slot, hosted_device.hoster.spot, hosted_device.device_id)); }
			if(sync_agent_observers!=null) synchronized(sync_agent_observers) { for(AgentUObserver obs : sync_agent_observers) obs.physicalChange(agent_id, name_slot, hosted_device.hoster.spot, hosted_device.device_id); }
						
			amanager.notifyPhysicalChange(this, agent_id, name_slot, mslot.slot_specif.isOptional(), hosted_device.hoster.spot, hosted_device.device_id);
		} else{
			if(async_agent_observers!=null) synchronized(async_agent_observers) { for(AgentUObserver obs : async_agent_observers) ThreadStack.perform( () -> obs.physicalPlug(agent_id, name_slot, hosted_device.hoster.spot, hosted_device.device_id)); }
			if(sync_agent_observers!=null) synchronized(sync_agent_observers) { for(AgentUObserver obs : sync_agent_observers) obs.physicalPlug(agent_id, name_slot, hosted_device.hoster.spot, hosted_device.device_id); }
			amanager.notifyPhysicalPlug(this, agent_id, name_slot, mslot.slot_specif.isOptional(), hosted_device.hoster.spot, hosted_device.device_id);
		}
	}
	
	protected void notifyUnplug(PhysicalSlotManager mslot){ //pour SlotManager
		if(state>AgentU.OFF) agent.notifyUnplugDevice(aud, mslot.slot_specif.nameSlot());
		if(!mslot.slot_specif.isOptional()) {
			plugged_no_optional--;
			if(state>AgentU.OFF && plugged_no_optional==(type.nb_no_optional_plug-1)){
				agent.event(aud, AgentU.MISSING_PLUG, AgentU.UNPLUG_NOTIFICATION);
				if(run_mode==AgentU.BREAKPLUG_PAUSE) pause(AgentU.RUN_MODE_UNPLUG);
				else if(run_mode==AgentU.BREAKPLUG_STOP) stop(AgentU.RUN_MODE_UNPLUG);
			}
		}
		if(async_agent_observers!=null) synchronized(async_agent_observers) {
			for(AgentUObserver obs : async_agent_observers) ThreadStack.perform( () -> obs.physicalUnplug(agent_id, mslot.slot_specif.nameSlot()));
		}
		if(sync_agent_observers!=null) synchronized(sync_agent_observers) {
			for(AgentUObserver obs : sync_agent_observers) obs.physicalUnplug(agent_id, mslot.slot_specif.nameSlot());
		}
		amanager.notifyPhysicalUnplug(this, agent_id, mslot.slot_specif.nameSlot(),  mslot.slot_specif.isOptional());	
	}
	
	private void fireStateChange(byte old_state, byte ctx) {
		if(async_agent_observers!=null) synchronized(async_agent_observers) { for(AgentUObserver obs : async_agent_observers) ThreadStack.perform( () -> obs.agentStateChange(agent_id, old_state, state, ctx)); }
		if(sync_agent_observers!=null) synchronized(sync_agent_observers) { for(AgentUObserver obs : sync_agent_observers) obs.agentStateChange(agent_id, old_state, state, ctx); }
		amanager.notifyAgentStateChange(this, agent_id, old_state, state, ctx);
	}
	
	public boolean run(byte ctx){
		byte old_state = state;
		switch(state){
		case AgentU.NOT_EXIST: case AgentU.ERROR: return(false);
		case AgentU.OFF :	 //on doit instancier l'agent
			try{ agent = type.classe.newInstance(); } 
			catch(Exception e){ //impossible d'instancier l'agent
				e.printStackTrace(); //OXO				
				state = AgentU.ERROR;
				fireStateChange(old_state, ctx);
				return(false);
			}
			state = AgentU.RUN;
			agent.start(this.aud, ctx);
//System.out.println("agent_id = "+agent_id+" ready to plug "+sslots.size()+" slots");		//OXO	
			//[on plug tous les slots sociaux de l'agent
			synchronized(sslots){ for(SocialSlotManager mslot : sslots.values()){
				mslot.plug(AgentU.OP_START);
			}}
			//]
			//[on informe l'agent des plugs existants, OXO : voir aussi à ne pas linked réellement les devices avant ce traitement
			synchronized(pslots){ for(PhysicalSlotManager mslot : pslots.values()){
				if(mslot.device!=null){					
					agent.notifyPlugDevice(aud, mslot.slot_specif.nameSlot(), mslot.device);
				}
			}}
			//]
			if(plugged_no_optional == type.nb_no_optional_plug) agent.event(aud, AgentU.ALL_PLUGGED, ctx);			
			break;
		case AgentU.PAUSE:	//on fait un resume
			state = AgentU.RUN;
			agent.event(aud, AgentU.RESUMED, ctx);
			//[on plug tous les slots sociaux de l'agent
			synchronized(sslots){ for(SocialSlotManager mslot : sslots.values()){
				mslot.plug(AgentU.OP_RESUME);
			}}
			//]
			if(pulsors.size()>0){synchronized(pulsors){ //on remet en route tous les pulses
				Pulsor p;
				for(int i=0; i<pulsors.size(); i++){
					 p = pulsors.get(i);
					 if(!p.startOrResume(false)){ //on supprime ce pulsor
						 pulsors.remove(i);
						 i--;
					 }
				}
			}}
			break;
		case AgentU.RUN: break; //RAS
		}
		if(old_state!=state) { //l'agent à changé d'état
			fireStateChange(old_state, ctx); 
		}
		return(true);
	}
	public boolean stop(byte ctx){
		if(agent!=null){ //on a donc state==RUN ou PAUSE			
			synchronized(pulsors){
				for(Pulsor p : pulsors) p.remove(); //on détruit tous les pulses
				pulsors.removeAllElements();
			}
			//[on informe l'agent des déplug des plugs existants, OXO : voir aussi à ne pas délinker les devices suite à ce traitement
			synchronized(pslots){ for(PhysicalSlotManager mslot : pslots.values()){
				if(mslot.device!=null) agent.notifyUnplugDevice(aud, mslot.slot_specif.nameSlot());
			}}
			//]
			//[on deplug tous les slots sociaux de l'agent
			synchronized(sslots){ for(SocialSlotManager mslot : sslots.values()){
				mslot.unplug(AgentU.OP_STOP);
			}}
			//]			
			byte old_state = state;
			state=AgentU.OFF;
			agent.stop(aud, ctx);
			agent=null;
			fireStateChange(old_state, ctx);
			
			if(wait_retype!=null){ //on doit appliquer un changement de type
				this.applyRetype(wait_retype);
				wait_retype=null;
			}
			return(true);
		}
		return(false);
	}
	public boolean pause(byte ctx){
		if(state==AgentU.RUN){
			state=AgentU.PAUSE;
			synchronized(pulsors){ for(Pulsor p : pulsors) p.pause(); } //on met en pause tous les pulses
			//[on deplug tous les slots sociaux de l'agent
			synchronized(sslots){ for(SocialSlotManager mslot : sslots.values()){
				mslot.unplug(AgentU.OP_PAUSE);
			}}
			//]
			agent.event(aud, AgentU.PAUSED, ctx);
			fireStateChange(AgentU.RUN, ctx);
			return(true);
		}
		return(false);
	}	
	protected void remove(byte ctx){
		if(state!=AgentU.NOT_EXIST){
			stop(ctx); //on commence par stopper l'agent
			state = AgentU.NOT_EXIST;
			//On va déconnecter tous les pugs
			synchronized(pslots){ for(PhysicalSlotManager mslot : pslots.values()) mslot.unplug(true, false); }
			
			if(async_agent_observers!=null) synchronized(async_agent_observers) { for(AgentUObserver obs : async_agent_observers) ThreadStack.perform( () -> obs.agentRemoved(agent_id)); }
			if(sync_agent_observers!=null) synchronized(sync_agent_observers) { for(AgentUObserver obs : sync_agent_observers) obs.agentRemoved(agent_id); }
			amanager.notifyAgentRemove(this, agent_id);
		}
	}
	
	protected void performRuleToAnonymousDevice(HostedDevice hdevice){ //on doit déclencher l'analyse des règles sur ce device nommée 
		if(state>=AgentU.OFF){
			synchronized(pslots){ for(PhysicalSlotManager mslot : pslots.values()) mslot.performRuleToAnonymousDevice(hdevice); }
		}
	}

	protected void performRuleToNamedDevice(HostedDevice hdevice){ //on doit déclencher l'analyse des règles sur ce device nommée 
		if(state>=AgentU.OFF){
			synchronized(pslots){ for(PhysicalSlotManager mslot : pslots.values()) mslot.performRuleToNamedDevice(hdevice); }
		}
	}

}

