/**
 * Classe de gestion interne d'un device pluggé sur l'agent
 
 	irules = Tableau de PhysicalPlugRule classé par ordre de priorité
 		
 	NB: nombre maximal de règle : 100
 */
package skuad.agentu.implementation.amanager;

import java.util.*;
import skuad.agentu.*;
import org.w3c.dom.Element;
import skuad.uda.*;
import skuad.util.databytes.*;
import skuad.util.value.Value;
import skuad.util.base.Network;
import skuad.hubudp.*;

public class PhysicalSlotManager implements Monitor {
	
	ExistingAgentU iau;
	PhysicalSpecif slot_specif;
	Vector<IdentifiedPlugRule> irules;	 //règles de plug
	byte next_rule_id;
	
	//[si le slot est pluggué
	IdentifiedPlugRule active_rule; // != null si le slot est pluggé, et pointe sur la règle qui permet le plug courant
	HostedDevice hosted_device; //adresse du device actuellement pluggé (==null si aucun)
	boolean ready;
	int true_type;
	Device device; //= UdaUdp.linkDevice(DynIpPort target, int device_id)
	//]
	
	public PhysicalSlotManager(ExistingAgentU viau, PhysicalSpecif ss){
		iau=viau;
		slot_specif = ss;
		irules = new Vector<IdentifiedPlugRule>();
		next_rule_id = 1;
		
		//pas de plug en cours
		active_rule=null;
		hosted_device = null;
		ready=false;
		true_type=UDA.UNKNOW;
		device=null;
	}
	public boolean plugged(){ return(active_rule!=null); }
	public int plugDeviceId(){
		if(hosted_device!=null) return(hosted_device.device_id);
		return(-1);		
	}
	public Spot plugSpot(){ 
		if(hosted_device!=null) return(hosted_device.hoster.spot);
		return(null);
	}
	protected byte newRuleId(){synchronized(irules){ //génère un nouvel identifiant de règle
		if(irules.size()>=100) return(-1);
		byte rule_id = next_rule_id;
		for(int i=0; i<irules.size(); i++){
			if(irules.get(i).id==rule_id) {
				rule_id++; if(rule_id<0) rule_id=1;
				i=0;
			}
		}
		next_rule_id = (byte) (rule_id+1); if(next_rule_id<0) next_rule_id=1;
		return(rule_id);
	}}
	protected IdentifiedPlugRule getIRule(int rule_id){
		synchronized(irules){ for(IdentifiedPlugRule prule : irules) if(prule.id==rule_id) return(prule); }
		return(null);
	}

	protected PhysicalPlugRule getRule(int rule_id){
		synchronized(irules){ for(IdentifiedPlugRule prule : irules) if(prule.id==rule_id) return(prule.rule); }
		return(null);
	}

	public Iterable<Integer> plugRulesIds(){ 
		Vector<Integer> list = new Vector<Integer>();
		synchronized(irules){ for(IdentifiedPlugRule prule : irules) list.add((int)prule.id); }
		return(list);				
	}
	
	public synchronized int addRule(PhysicalPlugRule rule){synchronized(irules){
		int rule_id = newRuleId();
		if(rule_id!=-1){
			IdentifiedPlugRule ipr = new IdentifiedPlugRule(rule_id, irules.size(), rule);
			irules.add(ipr);
			
			iau.notifyAddRule(slot_specif.nameSlot(), ipr);
			
			if(active_rule==null && iau.state>=AgentU.OFF) { //Analyse des possibilité de plug sur la nouvelle règle ajouté
				HostedDevice hd = iau.getRunContext().findDevice(rule);
				if(hd != null) plug(hd, ipr); //un device a été trouvé pour cette règle			
			}
		}
		return(rule_id);
	}}
	
	public PhysicalPlugRule addRule(DataReader rdesc){ //rdesc = [node_key]{si node_key!=ALL: (string: node_value) }[spot_key]{si spot_key==NAME: (string: spot_value) | si  spot_key==PORT: {short: port_value}}[device_key]{si device_key==NAME: (string: device_value) | si  device_key==TYPE ou ID: (int: device_value)} 
		PhysicalPlugRule prule = PhysicalPlugRule.parse(rdesc);
		if(addRule(prule)!=-1) return(prule);
		return(null);
	}
	
	public PhysicalPlugRule addRule(Element cur_rule){
		PhysicalPlugRule prule = PhysicalPlugRule.parse(cur_rule);
		if(addRule(prule)!=-1) return(prule);
		return(null);
	}	
	
	protected void performRuleToAnonymousDevice(HostedDevice hdevice){ //on doit déclencher l'analyse des règles sur ce device non nommée
		DeviceHoster hoster = hdevice.hoster;
		String gid = hoster.spot.getIdGlobal();
		synchronized(irules){ for(IdentifiedPlugRule ipr : irules){
			PhysicalPlugRule rule = ipr.rule;
			
			if(active_rule!=null && active_rule.order<=ipr.order) return; //fin d'analye
			int device_key = rule.deviceKey();
			int node_key = rule.nodeKey();
			int spot_key = rule.spotKey();
			if(		  (		(device_key == PhysicalPlugRule.ALL)
						|| 	(device_key == PhysicalPlugRule.TYPE && hdevice.device_type==rule.intDeviceFilter())
						|| 	(device_key == PhysicalPlugRule.ID && hdevice.device_id==rule.intDeviceFilter())
					  )
					&&  ((node_key == PhysicalPlugRule.ALL) 
							  || 	(node_key == PhysicalPlugRule.NAME && hoster.node_name!=null && hoster.node_name.equals(rule.stringNodeFilter()))
							  || 	(node_key == PhysicalPlugRule.IP && Network.isSameIp(hoster.spot.getIp(), rule.stringNodeFilter()))
						)
					&& ( 	(spot_key == PhysicalPlugRule.ALL) 
						|| 	(spot_key == PhysicalPlugRule.NAME && hoster.spot_name!=null && hoster.spot_name.equals(rule.stringSpotFilter()))	
						|| 	(spot_key == PhysicalPlugRule.ID && gid!=null && gid.equals(rule.stringSpotFilter()))	
					)		
			 ){
				plug(hdevice, ipr);
				return;
			}
		}}			
	}

	protected void performRuleToNamedDevice(HostedDevice hdevice){ //on doit déclencher l'analyse des règles sur ce device nommée
		DeviceHoster hoster = hdevice.hoster;
		String gid = hoster.spot.getIdGlobal(); 
		synchronized(irules){ for(IdentifiedPlugRule ipr : irules){
			PhysicalPlugRule rule = ipr.rule;
			
			if(active_rule!=null && active_rule.order<=ipr.order) return; //fin d'analye
			int device_key = rule.deviceKey();
			int node_key = rule.nodeKey();
			int spot_key = rule.spotKey();
			if(			(device_key == PhysicalPlugRule.NAME && hdevice.device_name.equals(rule.stringDeviceFilter()))
					&&  ((node_key == PhysicalPlugRule.ALL) 
							  || 	(node_key == PhysicalPlugRule.NAME && hoster.node_name!=null && hoster.node_name.equals(rule.stringDeviceFilter()))
							  || 	(node_key == PhysicalPlugRule.IP && Network.isSameIp(hoster.spot.getIp(), rule.stringDeviceFilter()))
							  || 	(spot_key == PhysicalPlugRule.ID && gid!=null && gid.equals(rule.stringSpotFilter()))	
						)
					&& ( 	(spot_key == PhysicalPlugRule.ALL) 
						|| 	(spot_key == PhysicalPlugRule.NAME && hoster.spot_name!=null && hoster.spot_name.equals(rule.stringSpotFilter()))						
					)		
			 ){
				plug(hdevice, ipr);
				return;
			}
		}}			
	}	

	public int removeRule(int rule_id){synchronized(irules){
		for(int i=0; i<irules.size(); i++){		//recherche de la règle
			IdentifiedPlugRule cur = irules.get(i);
			if(cur.id == rule_id){
				irules.remove(i);
				for(int j=i; j<irules.size(); j++) irules.get(j).order=j; //on réordonne les règles suivantes
				
				iau.notifyRemoveRule(slot_specif.nameSlot(), rule_id);
				
				if(cur==active_rule){ //C'est la règle active qui vient d'être supprimé... on doit analyser les possibilités de cablage des autres règles moins prioritaire  
					//on a forcément : state>AgentU.OFF
					for(; i< irules.size(); i++){
						IdentifiedPlugRule orule = irules.get(i);
						HostedDevice hd = iau.getRunContext().findDevice(orule.rule);
						if(hd != null) {
							plug(hd, orule); //un device a été trouvé pour cette règle
							return(AgentU.OK);
						}
					}	
					unplug(true, false); //aucune autre règle ne peut être activé, on déplug le slot courant
				}				
				return(AgentU.OK);
			}
		}
		return(AgentU.NOT_EXIST);
	}}	
	
	public byte upRule(int rule_id){synchronized(irules){
		for(int i=0; i<irules.size(); i++){		//recherche de la règle
			IdentifiedPlugRule cur = irules.get(i);
			if(cur.id==rule_id){ //on a trouvé la règle
				if(i==0) return(AgentU.NOT_POSSIBLE); //impossible de remonter la priorité de cette règle
				//on change la position de la règle
				IdentifiedPlugRule rbefore = irules.get(i-1);
				cur.order--; rbefore.order++;
				irules.remove(i); irules.insertElementAt(cur, i-1);
				
				iau.notifyMoveRule(slot_specif.nameSlot(), rule_id, PhysicalPlugRule.UP);
				
				if(active_rule==rbefore){ //on a forcément : state>AgentU.OFF
					//on doit vérifier si cette promotion de règle ne le rend pas active par rapport à celle activée actuellement
					HostedDevice hd = iau.getRunContext().findDevice(cur.rule);
					if(hd != null) plug(hd, cur); //un device a été trouvé pour cette règle repriorisé				
				}	
				return(AgentU.OK);
			}
		}}
		return(AgentU.NOT_EXIST); //règle non trouvé
	}
	
	public byte downRule(int rule_id){synchronized(irules){
		for(int i=0; i<irules.size(); i++){		//recherche de la règle :
			IdentifiedPlugRule cur = irules.get(i);
			if(cur.id==rule_id){ //on a trouvé la règle
				if(i==(irules.size()-1)) return(AgentU.NOT_POSSIBLE); //impossible de descendre la priorité de cette règle
				//on change la position de la règle
				IdentifiedPlugRule rafter = irules.get(i+1);
				cur.order++; rafter.order--;
				irules.remove(i); irules.insertElementAt(cur, i+1);
				
				iau.notifyMoveRule(slot_specif.nameSlot(), rule_id, PhysicalPlugRule.DOWN);
				
				if(active_rule==cur){ //on a forcément : state>AgentU.OFF
					//on doit vérifier si cette régression de règle ne le rend pas active la règle situé aprés elle
					HostedDevice hd = iau.getRunContext().findDevice(rafter.rule);
					if(hd != null) plug(hd, rafter); //un device a été trouvé				
				}	
				return(AgentU.OK);
			}
		}}
		return(AgentU.NOT_EXIST); //règle non trouvé
	}

	//[Méthodes internes
	private void plug(HostedDevice hdevice, IdentifiedPlugRule irule){ 
//System.out.println("PSM : plug for slot = "+this.slot_specif.nameSlot());		
		boolean plug_replacement=false;
		if(hosted_device==null || !hosted_device.equals(hdevice)){
			if(hosted_device!=null){
				device.removeListener(this);
				plug_replacement=true;
				hosted_device.pluggedOff(this, true, device);
			}
			hosted_device = hdevice;
			device = hosted_device.pluggedOn(this);
			device.addListener(this);
			device.getState(this);
			
			iau.notifyPlug(this, plug_replacement, hosted_device, device);
		}
		active_rule = irule;
	}
	//]
	
	//[Pour HostedDevice
	protected void unplug(boolean remove_in_pluggedTo, boolean performRule){ 
		if(hosted_device!=null){
			device.removeListener(this);
			hosted_device.pluggedOff(this, remove_in_pluggedTo, device);
			device = null;
			hosted_device = null;			
			iau.notifyUnplug(this);
			int cur_order = (active_rule !=null ? active_rule.order : 0);
			active_rule = null;
			if(performRule){ //on doit réanalyser les règles à partir de celle courante
				synchronized(irules){ for(int i=cur_order; i< irules.size(); i++){
					IdentifiedPlugRule ipr = irules.get(i);
					HostedDevice hd = iau.getRunContext().findDevice(ipr.rule);
					if(hd != null) {
						plug(hd, ipr); //un device a été trouvé pour cette règle
						return;
					}
				}				
			}} 
		}
	}
	//]

	protected  void writeAllPhysicalPlugRule(DataWriter wcmd){ 
		//[nb_rules]...|pour chaque règle: [rule_id][rule_order][node_key]{si node_key!=ALL: (string: node_value) }[spot_key]{si spot_key==NAME: (string: spot_value) | si  spot_key==PORT: {short: port_value}}[device_key]{si device_key==NAME: (string: device_value) | si  device_key==TYPE ou ID: (int: device_value)}|...
		synchronized(irules){ 
			wcmd.b(irules.size());
			for(IdentifiedPlugRule cur : irules){
				wcmd.b(cur.id).b(cur.order);
				cur.rule.write(wcmd);
			}
		}
	}
	
	protected  void writePhysicalPlugRule(int rule_id, DataWriter wcmd){ 
		IdentifiedPlugRule cur = getIRule(rule_id);
		if(cur!=null){
			wcmd.b(cur.id).b(cur.order);
			cur.rule.write(wcmd);			
		}
	}
	
	//Interface Monitor
		//Réponse aux requêtes
	public void params(Device dev, String name, int type_device, int nb_sensor, int nb_actuator, boolean can_log){
		true_type=type_device;
	}
	public void moveCapacity(Device dev, int movement_type, int power_min, int power_moy, int power_max){}
	public void sensorParam(Device dev, int sensor_id, String name, int type_sensor, int type_value, boolean observable){} //nb : type_value sur 1 octet 
	public void actuatorParam(Device dev, int actuator_id, String name, int type_actuator, int type_value){}	//nb : type_value sur 1 octet
	//+ public void state(Device dev, boolean ready, boolean config_changed);
	//+ public void sensorValue(...);

		//Ecoute
	public void state(Device dev, boolean vready, boolean config_changed){
		ready=vready;
		if(iau.state>AgentU.OFF){
			if(ready && (true_type==UDA.UNKNOW || config_changed)) device.getParams(this);
			iau.agent.stateDevice(iau.aud, slot_specif.nameSlot(), device, vready, config_changed);
		}
	}
	public void sensorValue(Device dev, int sensor_id, int value){ 
		if(iau.state==AgentU.RUN) {
			Value v = Value.i(value);
			iau.agent.sensorChange(iau.aud, slot_specif.nameSlot(), sensor_id, v);
			iau.agent.sensorValue(iau.aud, slot_specif.nameSlot(), device, sensor_id, value); }
		}
	public void sensorValue(Device dev, int sensor_id, float value){  
		if(iau.state==AgentU.RUN) {
			Value v = Value.f(value);
			iau.agent.sensorChange(iau.aud, slot_specif.nameSlot(), sensor_id, v);
			iau.agent.sensorValue(iau.aud, slot_specif.nameSlot(), device, sensor_id, value); }
		}
	public void sensorValue(Device dev, int sensor_id, byte[] value){  
		if(iau.state==AgentU.RUN) {
			Value v = Value.bs(value);
			iau.agent.sensorChange(iau.aud, slot_specif.nameSlot(), sensor_id, v);
			iau.agent.sensorValue(iau.aud, slot_specif.nameSlot(), device, sensor_id, value); }
		}
	public void moved2D(Device dev, float dx, float dy, float da){}
	public void moved3D(Device dev, float dx, float dy, float dz, float ax, float ay, float az){}
	public void log(Device dev, String text){}
}