package skuad.agentu.implementation.amanager;

import skuad.util.time.Pulse;
import skuad.util.time.TimeFilter;

public abstract class Pulsor implements Pulse {
	protected String label;
	protected ExistingAgentU eau;
	protected boolean active, resume_reset;
	protected long delay, periode;
	int compteur;
	TimeFilter time_filter;
	
	public Pulsor(/*ExistingAgentU viau,*/ String vlabel, long vdelay,  long vperiode, boolean fireUptodate_or_resetOnResume){ //pulse cyclique
		//si reset_on_resume==true : le cycle de pulse est réinitialisé à la reprise de l'agent (le prochain pulse surviendra aprés 
		//l'écoulement du delay, puis les suivants a chaque période), avec reset_on_resume==false à la reprise aprés pause
		//les cycles de déclenchement reprenne comme si l'agent n'avait jamais été en pause.
		label = vlabel;
		active = true;
		delay = (vdelay>0 ? vdelay : 0); periode = (vperiode>0 ? vperiode : 0);
		resume_reset = fireUptodate_or_resetOnResume;
		compteur = 0;
		time_filter=null;
	}
	
	//[Interface Pulse
	public boolean is(String vlabel) {
		if(label==null) return(vlabel==null);
		else return(label.equals(vlabel));
	}
	public String label(){ return(label); }
	public boolean isLoop(){ return(periode>0); } 		//true ssi c'est une pulsation périodique 
	
	public boolean isActive(){ return(active); }
	public int count() { return(compteur); }
	public boolean cancel(){ return(cancel(true)); }
	public void setFilter(TimeFilter tf) {	//change le filtre d'activation de ce pulse, si tf=null cela revient à supprimer le filtre
		 time_filter=tf;
	}
	//]]
	
	//[Pour les classe fille --> Exécution du comportement via le pulse
	protected void lingToExistingAgentU(ExistingAgentU viau) { //pour une classe fille : cette méthode doit être invoqué à la fin de leur propre constructeur  (le dernier argument (auto_add) du constructeur super doit être fixé à false, et )
		eau = viau;
		eau.addPulsor(this);
	}
	protected void pulseBehavior(long tc) { 
		if(time_filter==null || time_filter.validTime(tc)) {
			compteur++; 
			eau.agent.behavior(eau.aud, this);
		}
	}
	//]

	public boolean cancel(boolean call_remove_pulsor_if_cancel){ 
		if(remove()) {
			if(call_remove_pulsor_if_cancel) eau.removePulsor(this);
			return(true);
		}
		return(false);
	}	
	protected boolean remove(){
		if(active){
			active=false;
			return(true);
		}
		return(false);
	}
	protected abstract void pause();
	protected abstract boolean startOrResume(boolean call_remove_pulsor_if_cancel); //false si le pulse est cancelé suite à cet appel

}