/**
 * 
 */
package skuad.agentu.implementation.amanager;

import skuad.agentu.*;
import skuad.ubiquity.space.*;
import skuad.ubiquity.space.mailbox.*;
import skuad.util.base.*;

public class SocialSlotManager implements ParticipantListener, MessageListener {
	
	ExistingAgentU iau;
	SocialSpecif slot_specif;
	
	//[si le slot est pluggué
	String space_name;				//nom du space pluggué
	String space_password;			//null ssi le space n'est pas protégé par un mot de passe
	String local_id;				//id local de l'agent spécifié dans la configuration du slot (si null, l'id réel de l'avatar est donnée par avatar.participant().localId()
	Avatar avatar; 					//avatar de l'agent dans le space, ==null si non instancié
	//]
	boolean auto_plug;
	
	public SocialSlotManager(ExistingAgentU viau, SocialSpecif ss, boolean load_defaut_config){
		iau=viau;
		slot_specif = ss;
		
		if(load_defaut_config) loadDefautConfig();
		else {
			space_name = null;
			space_password = null;
			local_id = null;
		}
		avatar=null;
		auto_plug=false;
	}
	
	public boolean specialConfig(){
		return( !space_name.equals(slot_specif.nameSpace()) ||  space_password!=null || !Util.same(local_id, slot_specif.localId()) );
	}

	public void loadDefautConfig(){
		space_name = slot_specif.nameSpace();
		space_password = null;
		local_id = slot_specif.localId();
	}

	public synchronized boolean set(String sname, String spass, String vlocal_id){ //false si sname==null ou aucun changement
		if(sname!=null && (!Util.same(space_name, sname) || !Util.same(space_password, spass) || !Util.same(local_id, vlocal_id))){
			unplug(AgentU.OP_SPACE_CHANGE);
			space_name = sname;
			space_password = spass;
			local_id = vlocal_id;
			
			//On notifify le changement de configuration
			iau.fireSocialChange(slot_specif.nameSlot(), space_name, space_password, local_id);
			
			if(iau.state==AgentU.RUN) plug(AgentU.OP_SPACE_CHANGE); //si l'agent est en cours d'éxécution, on instancie l'avatar
			return(true);
		}
		return(false); //aucun changement
	}
	
	public synchronized void remove(){
		unplug(AgentU.OP_REMOVE);
		space_name = null;
		space_password = null;
		local_id = null;
		//On notifify le changement de configuration
		iau.fireSocialChange(slot_specif.nameSlot(), null, null, null);	
	}
	
	protected boolean plug(int ctx){
		if(avatar==null && space_name!=null){
			Space space =  iau.getRunContext().getSpace(space_name, space_password);
//System.out.println("Plug to space "+space_name);			
			avatar =space.newAvatar(local_id);
			if(avatar!=null){ //on peut avoir avatar==null si local_id est un id déjà utilisé sur le spot local
				if(auto_plug){
					auto_plug=false;
					space.removeParticipantListener(this);
				}
				iau.agent.notifyPlugAvatar(iau.aud, slot_specif.nameSlot(), avatar, ctx);
			
				//System.out.println("fire social plug with avatar : "+avatar.participant().localId());			
				//On notifify le plug
				iau.fireSocialPlug(slot_specif.nameSlot(), avatar.participant().localId(), ctx);
				
				//on va écouter les messages recus sur la mailbox de cet avatar
				avatar.mailBox().addListener(this);
				return(true);
			} else { //on doit se mettre à l'écoute du space pour instancier le slot dés que possible
				if(!auto_plug){
					auto_plug=true;
					space.addParticipantListener(this);
				}
			}
		}
		return(false);
	}
	
	//[interface MessageListener
	public void receive(MailBox box, Message mess) {
		iau.notifyReceiveMessage(avatar, box, mess);
	}
	//]
	
	//[interface ParticipantListener {
	public void participantAdded(Participant p){}
	public void participantRemoved(Participant p){
		if(auto_plug && p.localId().equals(local_id)) plug(AgentU.OP_LOCAL_ID_RELEASE);
	}
	//]
	
	public String currentId(){
		if(avatar!=null) return(avatar.participant().localId());
		else return(null);
	}
	
	protected boolean unplug(int ctx){
		if(avatar!=null){
			if(auto_plug){
				auto_plug=false;
				iau.getRunContext().getSpace(space_name, space_password).removeParticipantListener(this);
			}
			iau.agent.beforeUnplugAvatar(iau.aud, slot_specif.nameSlot(), avatar, ctx);
			avatar.remove();
			avatar=null;
			iau.agent.notifyUnplugAvatar(iau.aud, slot_specif.nameSlot(), ctx);
			
			iau.fireSocialUnplug(slot_specif.nameSlot(), ctx);
			return(true);
		}		
		return(false);
	}
}