/**
 * Descripteur d'un type d'agent
 */
package skuad.agentu.implementation.amanager;

import skuad.agentu.*;
import skuad.uda.UDA;
import skuad.util.base.ConstantIterable;
import java.util.*;
import skuad.util.base.*;

import java.lang.reflect.*;

public class TypeAgent {
	int code;
	String name;
	boolean editable;
	Vector<PhysicalSpecif> physical_specif;
	Vector<SocialSpecif> social_specif;
	Class<? extends AgentU> classe;
	int nb_no_optional_plug;
	
	public TypeAgent(int vc, String type_name,  Class<? extends AgentU> vclasse){
		code = vc;
		name = type_name;
		classe = vclasse;
		editable = false; //tous les types implémenté en java ne sont pas éditable
		
		//Extraction de la spécification des slots
		try{	//les slots sociaux
			Field f = vclasse.getField("social_specif");
			social_specif = parseSocialSpecif((String) f.get(null));
//System.out.println("[TypeAgent code="+code+"] nb social_slot = "+ 	social_specif.size());		
		}catch(Exception e){
System.out.println("[TypeAgent] -warning- aucune spécification sociale détectée pour la classe : "+vclasse.getName());	
			social_specif = new Vector<SocialSpecif>();
//			e.printStackTrace();
		}

		try{	//les slots physiques
			Field f = vclasse.getField("physical_specif");
			physical_specif = parsePhysicalSpecif((String) f.get(null));
		}catch(Exception e){
System.out.println("[TypeAgent] -warning- aucune spécification physique détectée pour classe : "+vclasse.getName());	
			physical_specif = new Vector<PhysicalSpecif>();
//			e.printStackTrace();
		}
		
		//on va compter combien de slot sont marqués non optionnel
		nb_no_optional_plug = 0;
		for(PhysicalSpecif slotspec : physical_specif) if(!slotspec.isOptional()) nb_no_optional_plug++;
	}
	
	public SocialSpecif getSocialSlotSpecif(String slot_name){ // true ssi le slot est effectivement définie pour ce type d'agent
		for(SocialSpecif slotspec : social_specif){
			if(slotspec.nameSlot().equals(slot_name)) return(slotspec);
		}
		return(null);
	}
	
	public PhysicalSpecif getPhysicalSlotSpecif(String slot_name){ // true ssi le slot est effectivement définie pour ce type d'agent
		for(PhysicalSpecif slotspec : physical_specif){
			if(slotspec.nameSlot().equals(slot_name)) return(slotspec);
		}
		return(null);
	}
	
	public ConstantIterable<SocialSpecif> socialsSpecif(){
		return(new ConstantIterable<SocialSpecif>(social_specif));
	}
	
	public ConstantIterable<PhysicalSpecif> physicalsSpecif(){ 
		return(new ConstantIterable<PhysicalSpecif>(physical_specif));
	}
	
	//[Méthodes internes
	private static char[] SEP_SPECIF={',', ':'};
	public static int parseDeviceType(String value){
		int rep=UDA.UNKNOW;
		try{
			rep = Integer.parseInt(value);
		}catch(Exception e1){
			//On essais de voir si c'est une constante UDA
			try{
				Field f = UDA.class.getField(value);
				rep = f.getInt(null);
			}catch(Exception e2){}
		}
		return(rep);
	}
	public static Vector<PhysicalSpecif> parsePhysicalSpecif(String ch){ //"name_slot2 : type_device2 , optional ;"
		Vector<PhysicalSpecif> res = new Vector<PhysicalSpecif>();
		if(ch != null){
			String[] lines = ch.split(";");
			for(int l = 0; l<lines.length; l++){
				Vector<String> cur = Util.explode(lines[l], SEP_SPECIF);
				if(cur.size()>0){
					String name = cur.get(0);
					int type = UDA.UNKNOW;
					boolean optional=false;
					if(cur.size()>1) type = parseDeviceType(cur.get(1));
					if(cur.size()>2) optional = cur.get(2).toLowerCase().equals("true");
					res.add(new PhysicalSpecif(name, type, optional));
				}
			}
		}
		return(res);
	}
	public static Vector<SocialSpecif> parseSocialSpecif(String ch){ //"name_slot2 : space_name , local_id ;"
//System.out.println("parseSocialSpecif sur ["+ch+"] ...");		
		Vector<SocialSpecif> res = new Vector<SocialSpecif>();
		if(ch != null){
			String[] lines = ch.split(";");
			for(int l = 0; l<lines.length; l++){
				Vector<String> cur = Util.explode(lines[l], SEP_SPECIF);
				if(cur.size()>0){
					String name = cur.get(0);
					String space = "world";
					String local_id = null;
					if(cur.size()>1) space = cur.get(1);
					if(cur.size()>2) { local_id = cur.get(2); if(local_id.equals("*")) local_id=null; }
					//(String vname, String vspace_name, String vlid, boolean vopt)
					res.add(new SocialSpecif(name, space, local_id));
				}
			}
		}
//System.out.println("\t nbr de slot trouvé --> "+res.size());		
		return(res);
	}	
	//]
}