/**
 * 
 * Classe LocalHost :
 		Host +
 		 //écriture
	void setByte(int code, int value [,boolean mode_add = false] )		// -> fixe la valeur du code à value sur 1 octet
	void setString(int code, String value [,boolean mode_add = false])	// 	-> fixe la valeur du code à value au format [nb_lettre]{...lettres...}
	void setBuff(int code, byte[] value [,boolean mode_add = false])		//-> fixe la valeur du code à value au format tableau d'octet, value est un buffer ou un tableau
	void setShort(int code, int value [,boolean mode_add = false])		//	-> fixe la valeur du code à value au format entier sur 2 octet
		 //suppression
	void removeAll(int code);					//	-> supprime toute les valeurs pour le code indiqué
	void removeByte(int code, int value);		//	-> supprime le couple (code, value) s'il existe
	public void removeShort(int code, int value);	//	-> supprime le couple (code, value) s'il existe
 */
package skuad.hubudp;

import skuad.util.base.Network;
import skuad.util.value.Value;

/**
 * This class, extends the <em>Spot</em> class and describes the <em>LocalSpot</em> i.e. the unique representation of an execution context in a given <em>Hub</em>.	
 * @see Spot
 */
public abstract class LocalSpot extends Spot {
	
	protected LocalSpot(int vct) {
		super(vct);
		is_up = true; //les spots locaux sont toujours up
	}
	
	public String getIp() { return(Network.LOCALHOST); }
	
	/**
	 * To know if the spot is <em>local</em> or not. If it is, the <i>Spot</i> will return <b>true</b>.
	 * @return <b>boolean</b>
	 * 			Return <em>true</em> if the spot is local. Return <em>false</em> if not.
	 */
	public boolean isLocalSpot() { return(true); }	//true s'il s'agit de le spot local
	/**
	 * To know if the <i>host</i> of the <i>Spot</i> is <em>local</em> or not. In this case of a <i>
	 * Local Spot</i>, it will return <b>true</b>.
	 * @return <b>boolean</b>
	 * 			Return <em>true</em> if the host is local. Return <em>false</em> if not.
	 */
	public boolean isLocalHost(){ return(true); }
	
	//gestion des idContext
	/**
	 * Set the Context ID with the <i>String</i> parameter
	 * @param  ch_contexte
	 * 			The String which contains the Context ID you want to set
	 * @return <b>String</b>
	 * 			Return the Context ID that you just fixed.
	 * @see Spot
	 */
	public abstract String fixIdContext(String ch_contexte); //	<- fixe un idContext sur le spot local, retourne l'idContext attribué
	
	//écriture de la définition SMD de l'host local
		 //écriture
	/**
	 * Set the <u>code</u> with the <i>integer</i> value in the parameters. It's possible to set the <u>mode_add</u> to true,
	 * in this case, the value will be added to the old value.
	 * @param code The code of the SMD value that you want to set.
	 * @param value The value that you want to set.
	 * @param mode_add To enable the addition_mode. In this case, the old value is added to the new one.
	 */
	public abstract void setInt(int code, int value , boolean mode_add);			//	-> fixe la valeur du code à value sur 1 octet
	/**
	 * Set the <u>code</u> with the <i>integer</i> value in the parameters.
	 * @param code The code of the SMD value that you want to set.
	 * @param value The value that you want to set.
	 */
	public void setInt(int code, int value){ setInt(code, value, false); } 	
	/**
	 * Set the <u>code</u> with the <i>String</i> value in the parameters.
	 * @param code The code of the SMD value that you want to set.
	 * @param value The value that you want to set.
	 */
	public void setString(int code, String value){ setString(code, value, false); }
	/**
	 * Set the <u>code</u> with the <i>String</i> value in the parameters. It's possible to set the <u>mode_add</u> to true,
	 * in this case, the value will be added to the old value.
	 * @param code The code of the SMD value that you want to set.
	 * @param value The value that you want to set.
	 * @param mode_add To enable the addition_mode. In this case, the old value is added to the new one.
	 */
	public abstract void setString(int code, String value, boolean mode_add);
	/**
	 * Set the <u>code</u> with the <i>buffer</i> (i.e. byte table)value in the parameters.
	 * @param code The code of the SMD value that you want to set.
	 * @param value The value that you want to set.
	 */
	public void setBuff(int code, byte[] value){ setBuff(code, value, false); }
	/**
	 * Set the <u>code</u> with the <i>buffer</i> (i.e. byte table) value in the parameters. It's possible to set the <u>mode_add</u> to true,
	 * in this case, the value will be added to the old value.
	 * @param code The code of the SMD value that you want to set.
	 * @param value The value that you want to set.
	 * @param mode_add To enable the addition_mode. In this case, the old value is added to the new one.
	 */
	public abstract void setBuff(int code, byte[] value, boolean mode_add);	//-> fixe la valeur du code à value au format tableau d'octet, value est un buffer ou un tableau	
	/**
	 * Set the <u>code</u> with the <i>boolean</i> value in the parameters. 
	 * @param code The code of the SMD value that you want to set.
	 * @param value The value that you want to set.
	 */
	public void setBool(int code, boolean v){ setInt(code, (v ? 1 : 0), false); }
	
		 //suppression
	/**
	 * Remove all values assigned to the given code.
	 * @param code The code where all values will be deleted.
	 */
	public abstract void removeAll(int code);				//	-> supprime toute les valeurs pour le code indiqué
	/**
	 * Remove the given value assigned to the given code.
	 * @param code The code where the value will be deleted.
	 * @param value The value which will be deleted.
	 */
	public abstract boolean removeInt(int code, int value);	//	-> supprime le couple (code, value) s'il existe
	
	//gestion des Tag
	/**
	 * Add the given value to the tag with the given name.
	 * @param name The name of the tag which will receive the given value.
	 * @param value The value that you want to assign to the given tag.
	 */
	public abstract void addTag(String name, Value value);	 	//ajoute la valeur 'value' au tag de nom 'name'
	/**
	 * To verify if the tag with the given name and value exists.
	 * @param name The name of the tag that you want to verify.
	 * @param value The value that you want to verify in the given tag.
	 * @return <b>boolean</b> Return true if the tag with the given name and the given value exists.
	 */
	public abstract boolean haveTag(String name, Value value);	//teste si le tag name possède la valeur value
	/**
	 * Add the given value to the tag with the given name if the value isn't already assigned to this tag.
	 * Works like <i><b>addTag()</b></i> but check if the value is already assigned to the tag before doing it.
	 * @param name The name of the tag that you want to verify.
	 * @param value The value that you want to verify in the given tag.
	 * @return <b>boolean</b> Return true if the tag with the given value is already assigned to the tag.
	 */
	public abstract boolean addOnce(String name, Value value);	//ajoute la valeur value au tag name s'il ne la possède pas déjà, true si l'ajout est effectif
	/**
	 * Remove the given value in the tag with the given name.
	 * @param name The name of the tag that you want to remove.
	 * @param value The value that you want to remove from the given tag.
	 * @return <b>boolean</b> Return true if the remove operation happened correctly.
	 */
	public abstract boolean removeTag(String name, Value value);//supprime la valeur 'value' du tag de nom 'name'
	/**
	 * Remove all values in the tag with the given name.
	 * @param name The name of the tag that you want to remove.
	 * @return <b>boolean</b> Return true if the remove operation happened correctly.
	 */
	public abstract boolean clearTag(String name);				//efface toutes les valeurs du tag de nom 'name'
	//]
}
