package skuad.hubudp.implementation;

import java.util.*;
import skuad.hubudp.*;
import skuad.util.databytes.*;
import skuad.util.thread.ThreadStack;

public abstract class HubImpl extends Hub {

	//[Attribut
	protected HubManager hubManager;
	protected LocalSpotImpl localspot;
	private Vector<SpotObserver> observers;				//les observers de spot
	protected Map<Integer, Vector<MessageReceiverDescriptor>> mreceivers;  	//table des gestionnaires de réception : clé -> channel, value : tableau de MessageReceiver
	protected Map<String, EContextMember> econtextes;  				//table des contextes connectés à ce hub : clé -> 
	protected boolean ready;
//	byte state;
	private Vector<HubReadyListener> readyListeners;	
	//]
	
	protected HubImpl(HubManager vhubManager) { //l'instance fille doit impérativement invoquer  registerLocalSpot(...) dans son constructeur !
		state = Hub.UNKNOW;
		hubManager = vhubManager;
		ready=false;
		
		observers = new Vector<SpotObserver>();
		mreceivers = new HashMap<Integer, Vector<MessageReceiverDescriptor>>();
		econtextes = new HashMap<String, EContextMember>();
		
		localspot = null; 
		readyListeners = null;
	}
	
	public synchronized void whenReady(HubReadyListener hrl) {
		if(ready) ThreadStack.perform(()-> hrl.ready(localspot));
		else {
			if(readyListeners==null) readyListeners=new Vector<HubReadyListener>();
			readyListeners.add(hrl);
		}
	}
	
	protected void setState(byte value, String notify_mess){
		state = value;
		hubManager.fireHubEvent(this, state, notify_mess);
	}
	public void logError(String mess){
		hubManager.fireHubEvent(this, Hub.LOG_ERROR, mess);
	}
	public void logWarning(String mess){
		hubManager.fireHubEvent(this, Hub.LOG_WARNING, mess);
	}
	public void logInfo(String mess){
		hubManager.fireHubEvent(this, Hub.LOG_INFO, mess);
	}
	public void setError(String mess) {
		state = Hub.HUB_ERROR;
		hubManager.fireHubEvent(this, Hub.HUB_ERROR, mess);//On notifie les hubListeners
	}
		
	protected String getIdContext(String ch_context) {
		return(hubManager.getIdContext(ch_context));
	}
	
	protected void receiveLocalAppMessage(SpotSender sender, int channel, DataReader mess) {
		//Recherche des gestionnaires pour le traitement du message
		Vector<MessageReceiverDescriptor> listpm = mreceivers.get(channel);

		if(listpm!=null && listpm.size()>0) { //on a des channel à servir
			synchronized(listpm) { for(int i=0; i<listpm.size(); i++) new RunnerManager(listpm.get(i).receiver, sender.getSpot(), channel, mess); }
		}
		mess.dispose();
	}
	
	//[Pour le kernel
	public void performMessage(SpotSender sender, SpotReceiver receiver, int protocole, DataReader mess) { //on peut traiter maintenant ce message recu de sender
		//mess = { ...content... }	, NB: ici receiver!=null
		if(protocole==-1) performSpotMessage(sender, receiver, mess); //cas du protocole Spot
		else if(receiver!=null) receiver.receiveAppMessage(sender, protocole, mess); 
	}
	public synchronized void localSpotLinked(int ctarget, int p) { //le noyau informe que le spotlocal est maintenant connecté au hub sous l'id-local ctarget
		if(localspot!=null) hubManager.fireHubEvent(this, Hub.LOG_WARNING, "re-invocation of localSpotLinked()");
		else localspot = instanciateLocalSpot(ctarget, p);
	}
	public synchronized void setReady() {
		if(!ready) {
			ready=true;
			setState(Hub.HUB_READY,  getHubName());
			
			if(readyListeners!=null) {
				for(HubReadyListener hrl : readyListeners) ThreadStack.perform(()-> hrl.ready(localspot));
				readyListeners.removeAllElements();
				readyListeners=null;
			}
			
			//On doit déclencher tous les up/updateSMD/IdContext de tous les spots connus sur tous les observers
			synchronized(observers) { for(SpotObserver so : observers) notifyAllSpot(so); }
		}		
	}
	protected abstract LocalSpotImpl instanciateLocalSpot(int ctarget, int port);
	
	protected EContextMember newEContextMember(SpotImpl spot) {
		return(new EContextMember(spot));
	}
	
	public void enter(SpotImpl spot, int rinstance) { //ce spot vient de rejoindre le hub
		String econtext_id = spot.getIpPort()+"i"+rinstance; 
		synchronized(econtextes) {
			EContextMember cl = econtextes.get(econtext_id);
			if(cl==null) { //c'est un nouveau contexte
				cl = newEContextMember(spot);
//System.out.println("-- new econtext for : "+spot.getIdTarget()); int x;				
				econtextes.put(econtext_id, cl);
			} else cl.add(spot);
		}
		if(ready) synchronized(observers) { for(SpotObserver so : observers) ThreadStack.perform(() -> so.up(spot)); }
	}

	public void exit(SpotImpl spot, int rinstance) { //ce spot vient de quitter le hub
		String econtext_id = spot.getIpPort()+"i"+rinstance; 
		synchronized(econtextes) {
			EContextMember cl = econtextes.get(econtext_id);
			if(cl!=null) {
				cl.remove(spot);
				if(cl.count()==0) econtextes.remove(econtext_id); //plus de spot dans ce contexte
				spot.goDown(); //il s'occupera de la notification down(...) aux observeurs
			}
		}
	}
	protected void idContextChanged(String ch_context, String idc){ //le noyau informe que l'idContext de la chaine contextuelle ch_context à changé
		if(localspot!=null) {
			String current = localspot.getIdContext(ch_context); //-> si le spot local à fixé cet id, le changement doit être répercuté et notifié sur le hub
			if(current != null && !current.equals(idc)){ //on doit changer ce idContext
				localspot.setIdContext(ch_context, idc);
			}
		}
	}
	//]
	
	//[Méthodes internes
	protected abstract void performSpotMessage(SpotSender sender, SpotReceiver receiver, DataReader mess);
	
    protected void checkIdContextConflict(Spot howner, String chaine_context, String id_context){
    	if(localspot!=null) {
    		//On controle que cet id_context ne rentre pas en conflit avec un id_context du spot local :
    		String local_idc = localspot.getIdContext(chaine_context);
    		if(local_idc!=null && local_idc.equals(id_context) && howner.getIdTarget().compareTo(localspot.getIdTarget())<0){ //conflit détecté !
    			//on doit mettre en oeuvre le protocole de changement d'id_context
    			hubManager.idcManager.newIdContext(chaine_context, !howner.isLocalHost(), true);
    		}
    	}
    }
	//]
    
    protected abstract void notifyLocalSpotSetIdContext(String ch_context, String idc);
	protected abstract void notifyLocalSpotUpdateSMD();
	
	//[API des hubs
	//envoies de message en mode broadcast
	protected abstract void localBroadcastAppMessage(SpotSender spot_sender, int channel, DataReader mess, boolean serve_sender);
	public abstract boolean broadcastByte(int channel, byte content, int option);	//-> emission du message ne contenant qu'un byte sur l'ensemble du hub
	public abstract boolean broadcastBuff(int channel, byte[] buff , int offset, int size,  int option);
	public abstract boolean broadcast(int channel, DataReader content, int option);	//-> emission du message en mode broadcast sur l'ensemble du hub, dispose() sera appliqué à l'instance content que si auto_dispose==true
	
	//gestion des spots
	public LocalSpot localSpot() { return(localspot); }
	public void  addSpotObserver(SpotObserver obs) { synchronized(observers) { //dès son enregistrement l'observer recoit tous les évènements up pour chaque spot UP et les IdContext déjà produit sur ces spots, ainsi qu'un updateSMD(spot, true) pour les spots qui ont déjà inintialisés leur SMD 
		if(!observers.contains(obs)) {
			observers.add(obs);
			if(ready) notifyAllSpot(obs); //On doit notifier cet observer de tous les UP, updateSMD et idContext !
		}
	} }
	protected void notifyAllSpot(SpotObserver obs){ //On doit notifier cet observer de tous les UP, updateSMD et idContext !
		//NB: ici on est forcément dans l'éta ready
		ThreadStack.perform(() -> obs.up(localspot));
		ThreadStack.perform(() -> obs.updateSMD(localspot, true));
		ThreadStack.perform(() -> localspot.exploreAllIdContext(obs));
		synchronized(econtextes) {
			for(EContextMember cl : econtextes.values()) { 
				for(int i=0; i<cl.spots.size(); i++) {
					SpotImpl cur = cl.spots.get(i);
					ThreadStack.perform(() -> obs.up(cur));
					if(cur.readySMD()) ThreadStack.perform(() -> obs.updateSMD(cur, true));
					cur.exploreAllIdContext(obs);
				}
			}
		}
	}
	protected void downAllNoLocalspot(){ 
		synchronized(econtextes) {
			for(EContextMember cl : econtextes.values()) { 
				for(int i=0; i<cl.spots.size(); i++) {
					SpotImpl cur = cl.spots.get(i);
					fireDown(cur);
				}
				cl.spots.removeAllElements();
			}
			econtextes.clear();
		}
	}
	public boolean removeSpotObserver(SpotObserver obs){ synchronized(observers) { return(observers.remove(obs)); }	}
	public void explore(SpotExplorer cb) {
		if(ready) {
			cb.spot(localspot);
			synchronized(econtextes) {
				for(EContextMember cl : econtextes.values()) { 
					for(int i=0; i<cl.spots.size(); i++) cb.spot(cl.spots.get(i));
				}
			}
		}
	}
	public abstract Spot getSpotByIdTarget(String idTarget);
	public abstract Spot getSpotByIdLocal(int local_id);
	
	public Spot getSpotByIdGlobal(String global_id){
		if(global_id==null) return(null);
		if(localspot!=null && global_id.equals(localspot.getIdGlobal())) return(localspot);
		synchronized(econtextes) {
			for(EContextMember cl : econtextes.values()) {
				for(SpotImpl target : cl.spots) if(global_id.equals(target.getIdGlobal())) return(target);
			}
		}		
		return(null); //non trouvé
	}
	public Spot getSpotByIdContext(String chaine_contexte, String id_context){
		if(chaine_contexte==null || id_context==null) return(null);
		if(localspot!=null && id_context.equals(localspot.getIdContext(chaine_contexte))) return(localspot);
		synchronized(econtextes) {
			for(EContextMember cl : econtextes.values()) { 
				for(SpotImpl target : cl.spots) if(id_context.equals(target.getIdContext(chaine_contexte))) return(target);
			}
		}		
		return(null); //non trouvé
	}
	//administration du hub
	private static boolean containsReceiver(Vector<MessageReceiverDescriptor> list, MessageReceiver receiver, String name) {
		for(MessageReceiverDescriptor mrd : list) {
			if(mrd.receiver==receiver && name.equals(mrd.name)) return(true);
		}
		return(false);
	}
	public boolean addReceiver(int channel, String name, MessageReceiver receiver) {
		synchronized(mreceivers) { //le code protocole doit être positif !
			if(name==null) name="";
			Vector<MessageReceiverDescriptor> listpm = mreceivers.get(channel);
			if(listpm==null) {
				listpm = new Vector<MessageReceiverDescriptor>();
				mreceivers.put(channel, listpm);
			} 
			synchronized(listpm) { 
				if(containsReceiver(listpm, receiver, name)) return(true); //gestionnaire déjà enregistré
				listpm.add(new MessageReceiverDescriptor(receiver, name));
			}
			
			//on notifie cet ajout aux hublistener
			hubManager.fireHubEvent(this, Hub.MANAGER_ADDED, name);
		}
		return(true);
	}
	public boolean removeReceiver(int channel, String name, MessageReceiver receiver) {
		synchronized(mreceivers) {
			if(name==null) name="";
			Vector<MessageReceiverDescriptor> listpm = mreceivers.get(channel);
			if(listpm==null) return(false);
			synchronized(listpm) { 
				for(int i=0; i<listpm.size(); i++) {
					MessageReceiverDescriptor mrd = listpm.get(i);
					if(mrd.receiver==receiver && name.equals(mrd.name)) {
						listpm.removeElementAt(i);
						//on notifie les hublistener
						hubManager.fireHubEvent(this, Hub.MANAGER_REMOVED, name);	
						return(true);
					}
				}
			}
		}
		return(false);
	}

	public abstract boolean linkPortal(String ip_portal, int port_portal , String link_key);
//	public abstract  boolean localPortal(int port_portal);
//	public abstract boolean fusion(Hub hub_b);
	
	//Notification des SpotObserver
	protected void fireUp(Spot spot){
		if(ready) synchronized(observers) { for(SpotObserver so : observers) ThreadStack.perform(() -> so.up(spot)); } 
	}
	public void fireDown(Spot spot){
		if(ready) synchronized(observers) { for(SpotObserver so : observers) ThreadStack.perform(() -> so.down(spot)); } 
	}
	protected void fireIdContext(Spot spot, String ch_context, String id_context){
		if(ready) synchronized(observers){
			for(SpotObserver so : observers) ThreadStack.perform(() -> so.idContext(spot, ch_context, id_context));
			if(spot==localspot) notifyLocalSpotSetIdContext(ch_context, id_context);
		}
	}
	protected void fireUpdateSMD(Spot spot, boolean first_definition){ //Pour chaque méthode mmm de l'interface XXXListener
//System.out.println("..hub fireUpdateSMD pour "+spot.getIdTarget());
		if(ready) {
			if(spot==localspot) notifyLocalSpotUpdateSMD();
			synchronized(observers){
				for(SpotObserver so : observers) ThreadStack.perform(() -> so.updateSMD(spot, first_definition));
			}
		}		
	}
	//]

	//]	
}