package skuad.hubudp.implementation;
import java.util.*;
import skuad.hubudp.*;
import skuad.util.base.Util;
import skuad.util.thread.ThreadStack;
import skuad.hubudp.implementation.portal.*;
import skuad.hubudp.implementation.local.*;

public class HubManager {

	private Vector<HubListener> listeners;	//les écouteurs de hub
	boolean hub_event_logged;
	//[Attributs initialisé lors de la 1ère connexion à un hub
	Map<String, HubLocal> localHubs;			//table des hubs connecté au contexte d'exécution local, key : id_hub   -> instance Hub
	Map<String, Joiner> distantHubs;			//table des hubs connecté via un portail, key : id_portal  -> instance Hub
	IdContextManager idcManager;
	//]
	
	UdpKernel udpKernel;
	
	public HubManager() {
		listeners =  new Vector<HubListener>();
		hub_event_logged = false;
		localHubs = new HashMap<String, HubLocal>();
		distantHubs = new HashMap<String, Joiner>();
		idcManager =new IdContextManager(this);
		udpKernel = null;
	}
	
	public String getIdContext(String ch_context) {
		return(idcManager.getIdContext(ch_context));
	}
	
	//[Pour HubLocal
	public void registerLocalHub(String id_hub, HubLocal hl) {
		synchronized(localHubs) { localHubs.put(id_hub, hl); }
	}
	public boolean existLocalHub(String id_hub) { return(localHubs.containsKey(id_hub)); }
	public HubLocal getLocalHub(String id_hub) { return(localHubs.get(id_hub)); }
	public Map<String, HubLocal> getAllLocalHubs() { return(localHubs); }
	//]
		
	//[Pour HubUdp
	public synchronized Hub joinLocalHub(String name, String key) {
		if(key==null || key.length()==0) key=null;	//c'est un hub public
		if(name==null || name.length()==0) name=Hub.DEFAULT_HUB;
		else if(name.indexOf(':')!=-1) return(null);	//le symbole ':' est interdit dans le nom d'un hub

		String id_hub = null;
		if(key==null) id_hub=name;
		else id_hub=name+":"+hashKey(key);

		HubLocal hub = localHubs.get(id_hub);
		if(hub==null){ //on doit instancier ce hub
			if(udpKernel==null) udpKernel = new UdpKernel(this);
			return(udpKernel.joinLocalHub(id_hub, name, key));
		} else return(hub); //sinon le hub est déjà connecté, donc RAS
	}
	public void addListener(HubListener obs) { synchronized(listeners) { 
		if(!listeners.contains(obs)) {
			listeners.add(obs); 
			//On doit notifier HUB_READY pour tous les hubs locaux ready 			
			synchronized(localHubs) {
				for(HubImpl ihub : localHubs.values()) if(ihub.ready) ThreadStack.perform(() -> obs.hubEvent(ihub, Hub.HUB_READY, ihub.getHubName()));
			}
			//et les hub distants
			synchronized(distantHubs) {
				for(HubImpl ihub : distantHubs.values()) if(ihub.ready) ThreadStack.perform(() -> obs.hubEvent(ihub, Hub.HUB_READY, ihub.getHubName()));
			}
		} 
	}}
	public boolean removeListener(HubListener obs) {  synchronized(listeners) { return(listeners.remove(obs)); } }
	public synchronized void logHubEvent() { 
		if(!hub_event_logged) {
			hub_event_logged = true;
			addListener((Hub vhub, byte event, String args) -> {
				//OXO: afficher le nom de l'event plutôt que sa valeur INT !!!
				System.out.println("#HubEvent# hub = "+(vhub!=null ? vhub.getHubName() : "-all-")+" , event = "+Hub.eventName(event)+" : "+args);
			});
		}
	}
	public Hub joinDistantHub(String ip_portal, int port_portal , String key) { 
		if(key==null || key.length()==0) key=null;
		String id_portal = ip_portal+':'+port_portal;
		if(key!=null) id_portal+=":"+hashKey(key);

		Joiner joiner = distantHubs.get(id_portal);
		if(joiner==null) { //on doit instancier ce hub
			joiner = new Joiner(this, ip_portal, port_portal, key);
			distantHubs.put(id_portal, joiner);
		}
		return(joiner); 
	}
	public boolean newPortal(int port_portal, String link_key) {	//<- crée une instance de portail de jonction sur le port port_portal. Cette méthode ne crée pas de hub
		//false si le port demandé est déjà utilisé ou si port_portal<=0
		if(port_portal<=0) return(false);
		Portal portal = new Portal(this, port_portal, link_key); 
		return(portal.inRun());
	}
	public Vector<LocalSpot> haveIdContext(String ch_contex){ //retourne tous les spots locaux qui ont fixé l'idContext correspondant à la chaine contextuelle indiqué		
		Vector<LocalSpot> rep = new Vector<LocalSpot>();
		if(localHubs!=null) synchronized(localHubs) {
			for(HubImpl ihub : localHubs.values()) if(ihub.ready && ihub.localSpot().getIdContext(ch_contex)!=null) rep.add(ihub.localSpot());
		}
		return(rep); 
	}
	//]	
	
	//[Méthode pour HubImpl
	public void fireHubEvent(Hub hub, byte type, String mess) {
		 synchronized(listeners) { for(HubListener hl : listeners) ThreadStack.perform(() -> hl.hubEvent(hub, type, mess)); }
	}
	//]
	
	public void fireKernelLog(byte type, String mess) {
		 synchronized(listeners) { for(HubListener hl : listeners) ThreadStack.perform(() -> hl.hubEvent(null, type, mess)); }
	}
	public static int nextVersion(int nv) {
		nv++;
		if(nv>65500) nv=1;
		return(nv);
	}
	public static String hashKey(String key) { //retourne le hash du mot de passe fournit en argument
		String hk=Util.cryptMD5(key);
		return(hk.substring(0,10)); //on ne conserve que les 10 premières lettres 
	}
	//]
}
//]