package skuad.hubudp.implementation;

import java.io.File;
import org.w3c.dom.*;
import skuad.util.base.*;

public class IdContextManager {
	
	File idcontext_file;	//fichier de persistence des idContext
	HubManager hmanager;
	
	public IdContextManager(HubManager k) {
		idcontext_file = new File("idcontext.xml"); 
		hmanager = k;
	}
			
	protected synchronized String getIdContext(String ch_context){ //charge l'idContexte demandé, et génère sa valeur s'il n'existe pas encore
		if(idcontext_file.exists()){ 
			Element racine = XmlManager.readFile(idcontext_file);
			if(racine!=null){
				//analyse des noeuds <idc context="" id="" />
				NodeList node_list = racine.getChildNodes();
				for(int i=0; i<node_list.getLength(); i++){ 
					Node cur_node=node_list.item(i); 
					if(cur_node.getNodeType()==Node.ELEMENT_NODE){
						Element cur = (Element) cur_node;
						String node_name =  cur.getNodeName();
						if(node_name.equals("idc")){
							String cur_ch = XmlManager.getAtt(cur, "context", null);
							if(cur_ch!=null && cur_ch.equals(ch_context)) return(XmlManager.getAtt(cur, "id", null));
						}
					}
				}
			}
		}
		//on doit générer une nouvelle valeur d'idContext	
		return(newIdContext(ch_context, true, false)); //NB: pas besoin de diffuser les changements, car c'est une nouvelle clé
	}
	
	protected synchronized String newIdContext(String ch_context, boolean save, boolean notifie_change){ 
		String idc=null;
		do { idc = Util.randString(6); }
		while(existIdContext(ch_context, idc));
		
		if(save) saveIdContext(ch_context, idc); //on enregistre ce nouvel idc

		if(notifie_change){//notification de l'affectation de ce nouvel idContext
			synchronized(hmanager.localHubs) {
				for(HubImpl hub : hmanager.localHubs.values()) hub.idContextChanged(ch_context, idc);
			}
		}
		return(idc);
	}	
	
	//[Méthodes internes
	private void saveIdContext(String ch_context, String id_context){
		Element racine = null; Document xml_doc=null; Element idc_node=null;
		if(idcontext_file.exists()) racine = XmlManager.readFile(idcontext_file);
		if(racine==null) {
			xml_doc = XmlManager.newXMLDocument("idcontext");
			racine = xml_doc.getDocumentElement();
		} else {
			xml_doc = racine.getOwnerDocument();
			//on regarde si l'association existe déjà ou pas
			NodeList node_list = racine.getChildNodes();
			for(int i=0; i<node_list.getLength(); i++){ 
				Node cur_node=node_list.item(i); 
				if(cur_node.getNodeType()==Node.ELEMENT_NODE){
					Element cur = (Element) cur_node;
					String node_name =  cur.getNodeName();
					if(node_name.equals("idc")){ //noeud de difinition d'un type d'agent
						String cur_ch = XmlManager.getAtt(cur, "context", null);
						if(cur_ch!=null && cur_ch.equals(ch_context)) {
							idc_node = cur;
							break;
						}
					}
				}
			}
		}
		if(idc_node==null) {
			idc_node = xml_doc.createElement("idc");
			idc_node.setAttribute("context", ch_context);
			racine.appendChild(idc_node);
		}			
		idc_node.setAttribute("id", id_context);
		
		//renregistrement
		XmlManager.writeFile(racine, idcontext_file);
	}
	private boolean existIdContext(String ch_context, String id){
		synchronized(hmanager.localHubs) {
			for(HubImpl hub : hmanager.localHubs.values()){
				if(hub.getSpotByIdContext(ch_context, id)!=null) return(true);
			}
		}
		return(false);
	}
	//]
}