package skuad.hubudp.implementation;

import java.io.DataOutputStream;
import java.util.HashMap;
import java.util.Map;
import java.util.Vector;
import skuad.hubudp.*;
import skuad.util.value.Value;
import skuad.util.databytes.DataBytes;
import skuad.util.databytes.DataReader;
import skuad.util.databytes.DataWriter;
import skuad.util.base.Network;
import skuad.util.base.Util;

public abstract class LocalSpotImpl extends LocalSpot implements SpotSender, SpotReceiver {
	
	protected HubImpl hub;
	int version_smd;
	Vector<ValueSMD> local_dsmd; 	//définition SMD local brute
	DataReader compiled_dsmd;		//définition SMD compilé :		//[SMD][def_version](i: nb_value)|...pour chaque valeur: (i: code)[value_size]{...value...}...|
	int port;
	
	public LocalSpotImpl(HubImpl vhub, int vctarget, int p) {
		super(vctarget);
		hub = vhub;
		version_smd=-1;
		local_dsmd = new Vector<ValueSMD>();
		port = p;
		
		//On renseigne le code SMD GLOBAL_ID
		global_id = Util.randString(8);
		setString(Spot.GLOBAL_ID, global_id);
		
		//On renseigne le code SMD OS
		if(Util.isWindows()) setInt(Spot.OS, Spot.WINDOWS);
		else if(Util.isMac()) setInt(Spot.OS, Spot.MAC);
		else if(Util.isLinux()) setInt(Spot.OS, Spot.LINUX);
		setInt(Spot.SOFT, Spot.JAVA);
	}
	public int getPort() {return(port); }
	public Hub getHub() { return(hub); }
	public void receiveAppMessage(SpotSender sender, int channel, DataReader mess) {
		hub.receiveLocalAppMessage(sender, channel, mess);
	}
	
	public Spot getSpot() { return(this); }
	public void queryGet(SpotSender sender){ ///[GET] : on doit répondre ->  SMD et ALL_ID_CONTEXT
//System.out.println("QueryGet de "+sender.getSpot().getIdTarget());	int x;	
		//[ALL_ID_CONTEXT](int: nb_id_context)|...pour chaque id: (string: chaine_context)(string: id_context) ...|
		synchronized(idContexts){
			DataWriter dw = DataBytes.writer().b(ProtocoleSpot.ALL_ID_CONTEXT).i(idContexts.size());
			for(String ctx : idContexts.keySet()){
				dw.s(ctx).s(idContexts.get(ctx));
			}
			sender.isend(this, -1, dw.flushReader(), true, true); 
		}
		//[SMD][def_version](i: nb_value)|...pour chaque valeur: (i: code)[value_size]{...value...}...|
//System.out.println("Send SMD vers -> "+sender.getSpot().getIdTarget()); int w;		
		sender.isend(this, -1, getDefSMD(), true, false); //NB: pas de dispose sur la définition SMD
	}
	public void queryTag(SpotSender sender, String name) { //[GET_TAG](string: name) : on doit réponse -> [TAG_VALUE](string: name)(int: count)...|pour chaque valeur: (value: v)|...
		DataWriter dw = DataBytes.writer().b(ProtocoleSpot.TAG_VALUE).s(name);
		Vector<Value> tab_val=tags.get(name);
		if(tab_val==null || tab_val.size()==0) dw.i(0);
		else synchronized(tab_val){
			dw.i(tab_val.size());
			for(Value v : tab_val) dw.writeValue(v);
		}
	
		sender.isend(this, -1, dw.flushReader(), true, true);	
	}
	//]
	public void goDown() {
		//Cette méthode ne devrait jamais être déclenchée sur le spotlocal
		hub.logWarning("Error call goDown(...) applied on LocaSpot");
	}
	//[Interface SpotSender
	//public Spot getSpot() { return(this); } //déjà défini
	public void internalFixIdContext(String ch_context, String id_context){
		//Cette méthode ne devrait jamais être déclenchée sur le spotlocal
		hub.logWarning("Error call internalFixIdContext(...) applied on LocaSpot");		
	}
	public void setSMD(int vhi, DataReader newDefSMD){
		//Cette méthode ne devrait jamais être déclenchée sur le spotlocal
		hub.logWarning("Error call setSMD(...) applied on LocaSpot");		
	}
	public void receiveTag(DataReader mess) {
		//Cette méthode ne devrait jamais être déclenchée sur le spotlocal
		hub.logWarning("Error call receiveTag(...) applied on LocaSpot");				
	}
	//]
	
	//[Pour Kernel
	public void changeCTargetAndPort(int vctarget, int vp) {
		this.changeIdTarget(vctarget);
		port = vp;
	}
	public void notifyExistGlobalId(String egi){ //SpotInfoKernel informe local_spot de l'existence éventuel d'un nouveau global_id afin qu'il s'assure d'en avoir un différent.
		if(egi!=null && egi.equals(global_id)){ //on doit changer d'identifiant global !
			global_id=null; String new_global_id;
			do {
				new_global_id = Util.randString(8);
			}while(hub.getSpotByIdGlobal(new_global_id)!=null);
			global_id = new_global_id;
			setString(Spot.GLOBAL_ID, global_id); //NB: la notification du changement d'id global découlera de la notification d'update SMD
		}
	}
	public int smdVersion(){ return(version_smd); }
	//]
	
	//[ Méthode de la classe Spot
	public String getIpPort() {		//ip:port du contexte portant ce spot
		return(getIp()+':'+port);
	}
	
	 //[lecture des méta-donnée SMD (Small Méta-Data)
	public boolean readySMD() { return(true); }	
	public int count(int code) {						//	-> retourne le nombre d'occurence existante pour le code SMD indiqué
		int nb=0;
		synchronized(local_dsmd) { for(int i=0; i<local_dsmd.size(); i++) {
			if(local_dsmd.get(i).code==code) nb++;
		}}
		return(nb);
	}		
	public int length(int code, int rang) {				//	-> retourne le nombre d'octet de la valeur du code SMD indiqué
		ValueSMD vhia = findCode(code, rang);
		if(vhia!=null) {
			switch(vhia.type){
			case ValueSMD.NUMBER :
				if(vhia.num_value==0) return(0);
				else if(vhia.num_value<256)  return(1);
				else if(vhia.num_value<65536)  return(2);
				else return(4);
			case ValueSMD.STRING :
				if(vhia.st_value==null)  return(0);
				else return(vhia.st_value.length());
			case ValueSMD.BUFFER :
				if(vhia.buffer_value==null)  return(0);
				else return(vhia.buffer_value.length);
			}
		}
		return(0);
	}	
	public int getInt(int code, int rang) {				//	-> lecture de l'octet numéro 0 de la valeur du code SMD indiqué
		ValueSMD vhia = findCode(code, rang);
		if(vhia!=null) {
			if(vhia.type==ValueSMD.NUMBER) return(vhia.num_value);
			else return(-1); //mauvais type
		}
		return(0); //non trouvé
	}	
	public boolean haveInt(int code , int value) {			//	-> true ssi le code SMD indiqué possède la valeur d'octet value
		ValueSMD cur;
		synchronized(local_dsmd) { for(int i=0; i<local_dsmd.size(); i++) {
			cur = local_dsmd.get(i);
			if(cur.code==code) {
				if(cur.type==ValueSMD.NUMBER && cur.num_value==value) return(true);
			}
		}}
		return(false); //code non trouvé
	}	
	public String getString(int code , int rang) {		//	-> lecture de la valeur du code SMD indiqué sous la forme d'une chaine de caractère
		ValueSMD vhia = findCode(code, rang);
		if(vhia!=null) {
			if(vhia.type==ValueSMD.STRING) return(vhia.st_value);
			else return(null); //mauvais type
		}
		return(null); //non trouvé
	}	
	public byte[] getBuff(int code , int rang) {			//	-> lecture du buffer de la valeur du code SMD indiqué
		ValueSMD vhia = findCode(code, rang);
		if(vhia!=null) {
			if(vhia.type==ValueSMD.BUFFER && vhia.buffer_value!=null) {
				byte[] buff = new byte[vhia.buffer_value.length];
				for(int i=0; i<buff.length; i++) buff[i]=vhia.buffer_value[i];
				return(buff);
			} else return(null); //mauvais type
		}
		return(null); //non trouvé
	}	
	public int readBuff(int code, int pos , int rang) { 	//	-> lecture de l'octet pos de la valeur du code SMD indiqué
		ValueSMD vhia = findCode(code, rang);
		if(vhia!=null && vhia.type==ValueSMD.BUFFER && vhia.buffer_value!=null && pos<vhia.buffer_value.length) return(vhia.buffer_value[pos]);
		return(-1); //non trouvé
	}	
	//]
	
	//[gestion des Tags
	Map<String, Vector<Value>> tags = new HashMap<String, Vector<Value>>(); //table des valeurs des tags
	public static boolean containValue(Vector<Value> tab_value, Value v){
		if(tab_value!=null){ synchronized(tab_value){
			for(Value value : tab_value) if(value.equals(v)) return(true);
		}}
		return(false);
	}
	public void addTag(String name, Value value){	 		//ajoute la valeur 'value' au tag de nom 'name'
		synchronized(tags){ 
			Vector<Value> tab_value = tags.get(name);
			if(tab_value==null){
				tab_value = new Vector<Value>();
				tags.put(name, tab_value);
			}
			synchronized(tab_value){ tab_value.add(value.toConstant()); }
		}
	}
	public boolean haveTag(String name, Value value){	//teste si le tag name possède la valeur value
		return(containValue(tags.get(name), value));
	}
	public boolean addOnce(String name, Value value){	//ajoute la valeur value au tag name s'il ne la possède pas déjà, true si l'ajout est effectif
		if(!containValue(tags.get(name), value)){
			addTag(name, value);
			return(true);
		}
		return(false);
	}
	public boolean removeTag(String name, Value v){ //supprime la valeur 'value' du tag de nom 'name'
		Vector<Value> tab_value = tags.get(name);
		if(tab_value!=null){ synchronized(tab_value){
			for(int i=0; i<tab_value.size(); i++) if(tab_value.get(i).equals(v)){
				tab_value.removeElement(i);
				return(true);
			}
		}}
		return(false);		
	}
	public boolean clearTag(String name){				//efface toutes les valeurs du tag de nom 'name'
		Vector<Value> tab_value = tags.get(name);
		if(tab_value!=null){synchronized(tab_value){
			tab_value.removeAllElements();
			return(true);
		}}
		return(false);
	}
	public void getTag(String name, TagReader callback){ 	//--> déclenche callback(name, tab_value)
		if(callback!=null){
			Vector<Value> tab_value = tags.get(name);
			if(tab_value==null) callback.tagValues(this, name, new Value[0]);
			else callback.tagValues(this, name, (Value[]) tab_value.toArray()); 
		}
	}
	//]
	
 	//[envoie de message vers ce spot, c'est le spot local donc les messages seront traité directement sans passer par le réseau
	public boolean send(int channel, byte content, boolean need_ack) {	//-> emission du message ne contenant qu'un byte
		if(channel>=0) {
			DataWriter mess = DataBytes.writer().b(content);
			return(ireceive(this, channel, mess.flushReader(), true));
		}
		return(false);
	}	

	public boolean send(int channel, DataReader content, boolean need_ack, boolean auto_dispose) { 	//-> emission du message, dispose() sera appliquée à l'instance content que si auto_dispose==true
		return(channel>=0 && ireceive( this, channel,  content,  auto_dispose));
	}	
	public boolean send(int channel, byte[] buff , int offset, int size, boolean need_ack) {	//-> emission du buffer buff	
		if(channel>=0) {
			DataWriter mess = DataBytes.writer();
			mess.copyBytes(buff, offset, size);
			return(ireceive(this, channel, mess.flushReader(), true));
		}
		return(false);
	}	
	public boolean isend(SpotSender sender, int channel, DataReader content, boolean need_ack, boolean auto_dispose) {
		return(ireceive(sender, channel,content, auto_dispose));
	}
	public boolean ireceive(SpotSender sender, int channel, DataReader content, boolean auto_dispose) {
		if(isUp()) {
			DataWriter mess = DataBytes.writer();
			mess.writeData(content);
			hub.performMessage(sender, this, channel, mess.flushReader()); //mess = { ...content... }
			if(auto_dispose) content.dispose();
			return(true);
		}
		return(false);	
	}
	//]
	
	
	//[Méthode de la classe LocalSpot
	public String fixIdContext(String ch_context){ //	<- fixe un idContext sur le spot local, retourne l'idContext attribué
		if(ch_context==null || ch_context.length()==0) return(null); //la chaine de contexte ne peut pas être vide
		String idc = idContexts.get(ch_context);
		if(idc==null){ //on doit charger ce idContext			
			idc = hub.getIdContext(ch_context); 
			setIdContext(ch_context, idc);
		}
		return(idc);
	}
	protected void setIdContext(String ch_context, String idc){
		idContexts.put(ch_context, idc);
		
		if(isUp()) {
			fireIdContext(ch_context, idc);//On notifie cette affectation aux SpotObserver
		
			hub.fireIdContext(this, ch_context, idc);//on notifie cette affectation au hub
		}
	}
		//écriture de la définition SMD de l'host local
	public void setInt(int code, int value ,boolean mode_add) {	
		boolean need_update=false;
		ValueSMD vhia=null;
		synchronized(local_dsmd) {
			if(!mode_add)  vhia = findCode(code, 0);
			if(vhia==null) {
				vhia = new ValueSMD(code, value);
				local_dsmd.add(vhia);
				need_update = true;
			} else {
				if(vhia.type!=ValueSMD.NUMBER || vhia.num_value!=value) {
					vhia.type=ValueSMD.NUMBER;
					vhia.num_value = value;
					need_update = true;
				}
			}
			if(need_update) needUpdateSMD();
		}
	}	
	public void setString(int code, String value, boolean mode_add) { // si != null la chaine automatiquement tronquée à 250 caractère maximum
		boolean need_update=false;
		ValueSMD vhia=null;
		if(value!=null && value.length()>250) value=value.substring(0,250);
		synchronized(local_dsmd) {
			if(!mode_add)  vhia = findCode(code, 0);
			if(vhia==null) {
				vhia = new ValueSMD(code, value);
				local_dsmd.add(vhia);
				need_update = true;
			} else {
				if(vhia.type!=ValueSMD.STRING || (vhia.st_value!=null && value==null) ||  (vhia.st_value==null && value!=null) || (vhia.st_value!=null && !vhia.st_value.equals(value)) ) {
					vhia.type=ValueSMD.STRING;
					vhia.st_value = value;
					need_update = true;
				}
			}
			if(need_update) needUpdateSMD();
		}
	}	
	
	public void setBuff(int code, byte[] value, boolean mode_add) {	//-> fixe la valeur du code à value au format tableau d'octet, si != null la chaine automatiquement tronquée à 255 éléments maximum	
		boolean need_update=false;
		ValueSMD vhia=null;
		synchronized(local_dsmd) {
			if(!mode_add)  vhia = findCode(code, 0);
			if(vhia==null) {
				vhia = new ValueSMD(code, value);
				local_dsmd.add(vhia);
				need_update = true;
			} else {
				if(!vhia.isSameBuffer(value)) {
					vhia.type=ValueSMD.BUFFER;
					int length = Math.min(value.length, 250);
					vhia.buffer_value = new byte[length];
					for(int i=0; i<length; i++) vhia.buffer_value[i] = value[i];
					need_update = true;
				}
			}
			if(need_update) needUpdateSMD();
		}
	}	
		 //suppression
	public void removeAll(int code) {				//	-> supprime toute les valeurs pour le code indiqué
		boolean need_update=false;
		for(int i=0; i<local_dsmd.size(); i++) {
			if(local_dsmd.get(i).code==code){
				local_dsmd.removeElementAt(i); 
				i--;
				need_update=true;
			}
		}
		if(need_update) needUpdateSMD();
	}	

	public boolean removeInt(int code, int value) {	//	-> supprime le couple (code, value) s'il existe
		boolean need_update=false;
		for(int i=0; i<local_dsmd.size(); i++) {
			ValueSMD cur = local_dsmd.get(i); 
			if(cur.code==code && cur.type==ValueSMD.NUMBER && cur.num_value==value){
				local_dsmd.removeElementAt(i); 
				need_update=true;
				break; //on sort du for
			}
		}
		if(need_update) {
			needUpdateSMD();
			return(true);
		}
		return(false);
	}
	//]
	
	//[Méthode interne
	public DataReader getDefSMD(){ //retourne le buffer compilé de la définition SMD: //[SMD][def_version](i: nb_value)|...pour chaque valeur: (i: code)[value_size]{...value...}...|
		synchronized(local_dsmd) { if(compiled_dsmd==null){ //on doit recompiler la difinition SMD
			++version_smd;
			//[SMD][def_version](i: nb_value)|...pour chaque valeur: (i: code)[value_size]{...value...}...|
			DataWriter dw = DataBytes.writer().b(ProtocoleSpot.SMD).b(version_smd).i(local_dsmd.size());
			for(ValueSMD vhia : local_dsmd){
				dw.i(vhia.code);
				switch(vhia.type){
				case ValueSMD.NUMBER :
					if(vhia.num_value==0) dw.b(0);
					else if(vhia.num_value<256) dw.b(1).b(vhia.num_value);
					else if(vhia.num_value<65536) dw.b(2).sh(vhia.num_value);
					else dw.b(4).w(vhia.num_value);
					break;
				case ValueSMD.STRING :
					if(vhia.st_value==null) dw.b(0);
					else {
						dw.b(vhia.st_value.length());
						dw.writeBruteString(vhia.st_value);
					}
					break;
				case ValueSMD.BUFFER :
					if(vhia.buffer_value==null)  dw.b(0);
					else {
						dw.b(vhia.buffer_value.length);
						dw.copyBytes(vhia.buffer_value);
					}
					break;
				}
			}
			compiled_dsmd=dw.flushReader();
		}
		return(compiled_dsmd);
	}}
	public DataReader getRawSMD(){ //retourne la définition SMD: (i: nb_value)|...pour chaque valeur: (i: code)[value_size]{...value...}...|
		DataReader c_dsmd = getDefSMD();
		return(c_dsmd.splitClone(2));
	}
	protected  ValueSMD findCode(int code, int rang){
		int num = 0; ValueSMD cur;
		synchronized(local_dsmd) { for(int i=0; i<local_dsmd.size(); i++) {
			cur = local_dsmd.get(i);
			if(cur.code==code) {
				if(num==rang) return(cur);
				num++;
			}
		}}
		return(null); //code non trouvé
	}
	boolean first_definition=true;
	protected synchronized void needUpdateSMD(){
		if(compiled_dsmd!=null) compiled_dsmd.dispose();
		compiled_dsmd=null; //on va devoir recompiler la difinition SMD
		if(isUp()) {
			fireUpdateSMD(first_definition);
			hub.fireUpdateSMD(this, first_definition);
			first_definition=false;
		}
	}
	public boolean sendAllIdContext(DataOutputStream out) { //envoie toutes les définitions id_context dans le stream: (int: nb){... (string: ch_context)(string: id_context }
		if(!idContexts.isEmpty()) synchronized(idContexts) {
			try {
				out.writeInt(idContexts.size());
				for(String ch_ctx : idContexts.keySet()) {
					out.writeUTF(ch_ctx);
					out.writeUTF(idContexts.get(ch_ctx));
				}
			}catch(Exception e) { return(false); }
		}
		return(true);	
	}
	//]
}
