package skuad.hubudp.implementation;

import skuad.hubudp.*;
import skuad.util.databytes.*;
import skuad.util.value.Value;

import java.util.*;

public abstract class SpotImpl extends Spot implements SpotSender {
	
	DataReader raw_smd; //(i: nb_value)|...pour chaque valeur: (i: code)[value_size]{...value...}...|
	int version_smd;
	protected HubImpl hub;
	private int port;
	private String ipPort;
	
	public SpotImpl(HubImpl vhub, int ct, int p) {
		super(ct);
		hub = vhub;
		is_up = true;
		raw_smd=null;
		version_smd=-1;
		port = p;
		ipPort= null;
	}	
	public abstract boolean isLocalHost();
	public abstract String getIp();
	public abstract boolean isLocalEContext();
	public HubImpl getHub() { return(hub); }
	public int getPort() {	return(port); }
	public String getIpPort() {		//ip:port du contexte portant ce spot
		if(ipPort==null) ipPort = getIp() +':'+getPort();
		return(ipPort); 
	}
	protected void changePort(int p) {
		port = p;
		ipPort=null;
	}

	
	//[Interface SpotSender
	public Spot getSpot() { return(this); }
	public void internalFixIdContext(String ch_context, String id_context){ //	
		if(ch_context==null || ch_context.length()==0) return; //la chaine de contexte ne peut pas être vide
		String idc = idContexts.get(ch_context);
		if(idc==null || !idc.equals(id_context)){
			idContexts.put(ch_context, id_context);
			//notification de l'affectation de ce nouvelle idContext
			fireIdContext(ch_context, id_context); //au niveau des écoutes locales
			hub.fireIdContext(this, ch_context, id_context); //et global
		}
	}
	//]
	
	//[Méthodes pour le Kernel
	public synchronized void goDown() {
		if(is_up) {
			is_up = false;
			internalDown();
			
			fireDown();//on notifie les SpotObserver au niveau des écoutes locales
			hub.fireDown(this);//et global
		}
	}
	protected void internalDown() {
		if(raw_smd!=null) { synchronized(raw_smd) {
			raw_smd.dispose();
			raw_smd=null;
		}}
	}
	public DataReader getRawSMD(){ 	//(i: nb_value)|...pour chaque valeur: (i: code)[value_size]{...value...}...|
		if(raw_smd!=null) return(raw_smd.clone());
		return(DataBytes.writer().i(0).flushReader());
	}
	protected DataReader getDefSMD(){ //[SMD][def_version](i: nb_value)|...pour chaque valeur: (i: code)[value_size]{...value...}...|
		DataWriter dw = DataBytes.writer().b(ProtocoleSpot.SMD).b(version_smd);
		if(raw_smd!=null) { synchronized(raw_smd) {
			raw_smd.reset();
			dw.writeData(raw_smd);
		}} else dw.i(0);
		return(dw.flushReader());
	}
	//]
	
	 //[lecture des méta-donnée SMD (Small Méta-Data)	
	public boolean readySMD() {								//	-> retourne true ssi les SMD du spot sont disponible (première notification updateSMD(...) déjà réalisé)
		return(raw_smd!=null);
	}	
	public int count(int code) {							//	-> retourne le nombre d'occurence existante pour le code SMD indiqué
		int nb=0;
		if(raw_smd!=null){ synchronized(raw_smd) {
			raw_smd.reset();
			raw_smd.readInt(); //length
			while(raw_smd.canRead()){
				if(raw_smd.readInt()==code) nb++;
				raw_smd.skip(raw_smd.readByte());
			}
		}}
		return(nb);
	}		
	public int length(int code, int rang) {				//	-> retourne le nombre d'octet de la valeur du code SMD indiqué
		if(raw_smd!=null){ synchronized(raw_smd) { if(findCode(code, rang)) return(raw_smd.readByte()); } }
		return(0);
	}	
	public int getInt(int code, int rang) {				//	-> lecture de la valeur int du code SMD indiqué
		if(raw_smd!=null){ synchronized(raw_smd) { if(findCode(code, rang)) {
			switch(raw_smd.readByte()) {
			case 0 : return(0);
			case 1 : return(raw_smd.readByte());
			case 2 : return(raw_smd.readShort());
			case 4 : return(raw_smd.readWord());
			}
		}}}
		return(0); //code non trouvé
	}	
	public boolean haveInt(int code , int value) {			//	-> true ssi le code SMD indiqué possède la valeur d'octet value
		if(raw_smd!=null){synchronized(raw_smd) {
			raw_smd.reset();
			raw_smd.readInt(); //length
			while(raw_smd.canRead()){
				if(raw_smd.readInt()==code){
					int l = raw_smd.readByte();
					switch(l) {
					case 0 : if(value==0) return(true); break;
					case 1 : if(value==raw_smd.readByte()) return(true);  break;
					case 2 : if(value==raw_smd.readShort()) return(true);  break;
					case 4 : if(value==raw_smd.readWord()) return(true);  break;
					default : raw_smd.skip(l);
					}
				} else raw_smd.skip(raw_smd.readByte());
			}
		}}
		return(false);
	}	
	public String getString(int code , int rang) {		//	-> lecture de la valeur du code SMD indiqué sous la forme d'une chaine de caractère
		if(raw_smd!=null){ synchronized(raw_smd) { if(findCode(code, rang)) {
			int length = raw_smd.readByte();
			return(raw_smd.readString(length));
		}}}
		return(null); //code non trouvé
	}	
	public byte[] getBuff(int code , int rang) {			//	-> lecture du buffer de la valeur du code SMD indiqué
		if(raw_smd!=null){ synchronized(raw_smd) { if(findCode(code, rang)) {
			int length = raw_smd.readByte();
			return(raw_smd.readBytes(null, length));
		} } }
		return(null); //code non trouvé
	}	
	public int readBuff(int code, int pos , int rang) { 	//	-> lecture de l'octet pos de la valeur du code SMD indiqué
		if(raw_smd!=null){ synchronized(raw_smd) { if(findCode(code, rang)){ //on va réaliser la lecture
			int length = raw_smd.readByte();
			if(pos<length) {
				raw_smd.skip(pos);
				return(raw_smd.readByte());
			}
		} } }
		return(-1);
	}		
	//]
	
	//[lecture des tags
	Map<String, Vector<TagReader>> tags_waiters = new HashMap<String, Vector<TagReader>>(); //table des callback en attente de lecture des tags
	public void getTag(String name, TagReader callback){ 	//-->lecture asynchrone du tag name
		synchronized(tags_waiters) {
			Vector<TagReader> waiters = tags_waiters.get(name);
			if(waiters==null){ //on enregistre un nouveau tag_waiters
				waiters = new Vector<TagReader>();
				tags_waiters.put(name, waiters);
				//on lance la requête de lecture du tag
				//[GET_TAG](string: name)
				isend(hub.localspot, -1, DataBytes.writer().b(ProtocoleSpot.GET_TAG).s(name).flushReader(), true,  true);
			}
			waiters.add(callback);
		}
	}
	
	public void receiveTag(DataReader mess) { //mess = (string: name)(int: count)...|pour chaque valeur: (value: v)|...
		String name = mess.readString();
		synchronized(tags_waiters) {
			Vector<TagReader> waiters = tags_waiters.remove(name);
			if(waiters!=null && waiters.size()>0){
				Vector<Value> tab_val=new Vector<Value>();
				int nb = mess.readInt();
				for(int i=0; i<nb; i++) tab_val.add(mess.readValue());

				for(TagReader tr : waiters) tr.tagValues(this, name, (Value[]) tab_val.toArray());
			}
		}
	}
	//]
	
	//[Gestion des SMD
	public int smdVersion(){ return(version_smd); }
	public void setSMD(int vhi, DataReader newDefSMD){ //retourne true ssi c'est la 1er définition SMD
		version_smd = vhi; 
		boolean first_def=(raw_smd==null);
		if(!first_def) synchronized(raw_smd) { raw_smd.dispose(); }
		raw_smd = newDefSMD;

		//on met à jour la valeur global_id si définie
		global_id = getString(Spot.GLOBAL_ID, 0);

		//on informe les écouteurs
		fireUpdateSMD(first_def);
		hub.fireUpdateSMD(this, first_def);
	}
	protected boolean findCode(int code, int rang){ //positionne le pointeur de lecture juste aprés la position du code indiqué
		if(raw_smd!=null){
			raw_smd.reset();
			raw_smd.readInt(); //length
			int num=0;
			while(raw_smd.canRead()){
				if(raw_smd.readInt()==code){
					if(num == rang) return(true); //positionnement réalisé
					num++;
				}
				raw_smd.skip(raw_smd.readByte());
			}
		}
		return(false); //(code, rang) non trouvé
	}
	//]
	
 	//[envoie de message vers ce spot
	public boolean send(int channel, byte content, boolean need_ack) {	//-> emission du message ne contenant qu'un byte
		return(channel>=0 && isend(hub.localspot, channel,  content,  need_ack));
	}
	public abstract boolean isend(SpotSender sender, int channel, byte content, boolean need_ack);	//-> emission du message ne contenant qu'un byte
	public boolean send(int channel, DataReader content, boolean need_ack, boolean auto_dispose) { 	//-> emission du message, dispose() sera appliquée à l'instance content que si auto_dispose==true
		return(channel>=0 && isend(hub.localspot, channel, content, need_ack, auto_dispose) );
	}
	public abstract boolean isend(SpotSender sender, int channel, DataReader content, boolean need_ack, boolean auto_dispose);  //nb: sur cette méthode interne l'utilisation du canal -1 (PROTO_SPOT) est possible
	public boolean send(int channel, byte[] buff , int offset, int size, boolean need_ack) {	//-> emission du buffer buff	
		return(channel>=0 && isend(hub.localspot, channel, buff , offset, size, need_ack));
	}
	public abstract boolean isend(SpotSender sender, int channel, byte[] buff , int offset, int size, boolean need_ack); //nb: sur cette méthode interne l'utilisation du canal -1 (PROTO_SPOT) est possible	
	//]
}
