package skuad.hubudp.implementation.local;

import java.util.Vector;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;
import skuad.util.databytes.DataReader;
import skuad.hubudp.implementation.*;

public class EContextMemberUDP extends EContextMember { //instance de description des contextes exécutions connecté à ce hub (utile pour les envoies de message broadcast)

	HubLocal hub;
	
	public EContextMemberUDP(HubLocal vhub, SpotImpl spot) {
		super(spot);
		hub = vhub;
		if(spot instanceof SpotUdp) ((SpotUdp) spot).econtext=this;
	}

	public synchronized void add(SpotImpl spot) {
		super.add(spot);
		if(spot instanceof SpotUdp) ((SpotUdp) spot).econtext=this;
	}	
		
	//[gestion des acquittements et ré-ordonnancement des messages vers et depuis ce contexte d'exécution
	protected Lock ackLock = new ReentrantLock();
	private int next_receive_ack=1;											//prochain numéro à acquité pour les messages recus de ce spot
	private int next_send_ack=1;											//prochain numéro d'acquittement à utiliser pour les messages à envoyer vers ce spot
	private Vector<Integer> unacks= new Vector<Integer>();					//pile des numéros de message non encore acquité (ordonnés par ordre d'envoie)
	private Vector<WaitingMessage> waitmess=new Vector<WaitingMessage> ();	//pile des messages en attente de ré-ordonnancement
	public synchronized boolean isNextReceiveAck(int ack) { //true ssi ack est le prochain numéro d'acquittement, et dans ce cas incrémente la valeur next_ack de 1
		if(ack==next_receive_ack) {
			next_receive_ack++;
			if(next_receive_ack>65000) next_receive_ack=1;
			return(true);
		} else return(false);
	}
	public synchronized int nextSendAck() { //génére la prochaine valeur de numéro d'acquitement à utiliser pour l'envoie du message
		int nack = next_send_ack++; //ack capture la valeur avant l'incrément
		if(next_send_ack>65000) next_send_ack=1;
		unacks.add(nack);
		return(nack);
	}
	public synchronized int getLunack(int cur_ack) {
		synchronized(unacks) {
			if(unacks.isEmpty()) return(cur_ack);
			else return(unacks.get(0));
		}		
	}
	public synchronized void notifyAck(int ack) { //le kernel signal que le message de numéro ack vient d'être acquité
		for(int i=0; i<unacks.size(); i++) if(unacks.get(i)==ack) {
			unacks.removeElementAt(i);
			return;
		}
	}
	public synchronized void needWaiting(SpotSender sender, int ack, SpotReceiver receiver, int proto, DataReader mess) {
		if((ack<5000  && next_receive_ack>60000) || ack>next_receive_ack) {
			for(int i=0; i<waitmess.size(); i++) {
				WaitingMessage cur = waitmess.get(i);
				if(ack == cur.ack) return; //ce paquet est déjà en attente
			}
		 	waitmess.add(new WaitingMessage(sender, ack, receiver, proto, mess)); //on doit mettre le message en attente afin de le ré-ordonner avant son traitement
		}
	}
	public synchronized void checkWaiting() {
		for(int i=0; i<waitmess.size(); i++) {
			WaitingMessage cur = waitmess.get(i);
			if(isNextReceiveAck(cur.ack)) { //on peut traiter ce message
//System.out.println("-checkWaiting.performMessage ("+cur.sender.getSpot().getIdTarget()+")-> ack = "+cur.ack+" receiver = "+cur.receiver.getSpot().getIdTarget());				
				waitmess.removeElementAt(i); 
				i=-1; //on va recommencer l'analyse depuis le début de la liste d'attente (i++ => i=0 pour le prochain tour)
				hub.performMessage(cur.sender, cur.receiver, cur.protocole, cur.mess); //nb: on ignore les messages dont le destinataire n'est pas connu
			}
		}
	}
	//]
}

class WaitingMessage{ //message en attente de ré-ordonnancement
	SpotSender sender;
	int ack, protocole;
	SpotReceiver receiver;
	DataReader mess;
	
	public WaitingMessage(SpotSender vsender, int vack, SpotReceiver vreceiver, int vproto, DataReader vmess) {
		sender = vsender;
		ack = vack;
		receiver = vreceiver;
		protocole = vproto;
		mess = vmess;
	}
}