package skuad.hubudp.implementation.local;

import java.util.*;
import skuad.hubudp.*;
import skuad.util.databytes.*;
import skuad.hubudp.implementation.*;
import skuad.hubudp.implementation.portal.*;
import skuad.util.thread.ThreadStack;

public class HubLocal extends HubImpl implements HubLinker {

	String id_hub, name, pass;
	ManagedLocalSpot managed_localspot;
	UdpKernel uKernel;
	Map<String, Linker> linkers;
	
	public HubLocal(HubManager hmanager, UdpKernel uk, String idh, String vname, String vpass) {
		super(hmanager);
		id_hub = idh;
		name = vname;
		pass = vpass;
		uKernel = uk;
		managed_localspot = null;
		hmanager.registerLocalHub(id_hub, this);
		linkers = new HashMap<String, Linker>();
		setState(Hub.INITIALIZING, vname); //jusqu'à ce que localSpotLinked() soit délcenché par le kernel
	}
	protected LocalSpotImpl instanciateLocalSpot(int ctarget, int port) {
		managed_localspot = new ManagedLocalSpot(this, ctarget, port);
		uKernel.registerManagedLocalSpot(ctarget, managed_localspot);
		return(managed_localspot);
	}
	
	protected EContextMember newEContextMember(SpotImpl spot) {
		return(new EContextMemberUDP(this, spot));
	}
	
	public void enter(SpotImpl spot, int rinstance) { //ce spot viens de rejoindre le hub
		super.enter(spot, rinstance);
		if(ready && !spot.isLocalEContext()) { //On doit envoyer la commande GET vers ce nouveau Spot pour obtenir ses définitions méta-data : SMD et IdContext
//System.out.println("--> Get sended to : "+spot.getIdTarget()); int x;				
			spot.isend(localspot, -1, ProtocoleSpot.GET, true);
		}
	}

	//]	
	public String getHubName() { return(name); }
	public boolean isPrivate() { return(pass!=null && pass.length()>0); }
	public String getHubId() { return(id_hub); }
	public String getLocalHub() { return(id_hub); }
	public ManagedLocalSpot getManagedLocalSpot(){		//-> retourne le descripteur du spot local
		return(managed_localspot);
	}
	public int getLocalSpotIdLocal() { return(managed_localspot!=null ? managed_localspot.getIdLocal() : 0); }
	
//OXO	protected EContextMember getEContext(String econtext_id) { return(econtextes.get(econtext_id));}
	
	public synchronized void setReady() {
		if(!ready) {
			super.setReady();

			//On doit envoyer un GET à tous les spots non locaux connus
			synchronized(econtextes) {
				for(EContextMember cl : econtextes.values()) { if(!cl.isLocalEContext()) { 
					for(int i=0; i<cl.spots.size(); i++) {
						SpotImpl cur = cl.spots.get(i);
						cur.isend(localspot, -1, ProtocoleSpot.GET, true);
					}
				}}
			}
		}		
	}
	
	//[Méthodes internes
	protected void localBroadcastAppMessage(SpotSender spot_sender, int protocole, DataReader mess, boolean serve_sender) {
//System.out.println("localBroadcastAppMessage , serve_sender = "+serve_sender);		
		if(serve_sender || !spot_sender.getSpot().isLocalSpot()) receiveLocalAppMessage(spot_sender, protocole, mess.clone()); //-> on sert les manager du spot local 

		//-> ainsi que les spots proxy connecté via un portail:
		synchronized(linkers) {
			for(Linker lk : linkers.values()) lk.broadcast(spot_sender.getSpot(), protocole, mess.clone(), serve_sender);
		}
		mess.dispose();
	}
	public void performSpotMessage(SpotSender sender, SpotReceiver receiver, DataReader mess) { //mess = [cmd]{ ...content... }
		int cmd = mess.readByte();
//System.out.println("--traitement du message PROTO_SPOT recu de : "+sender.getSpot().getIpPort()+" , code commande = "+cmd);		
		switch(cmd) {
		case ProtocoleSpot.ACK : //[ACK](short: ack)(short: your_rinstance)[status] -> le sender ACK le message de numéro ack qu'il a reçu
			int ack= mess.readShort();
			int ri = mess.readShort();
			if(ri == uKernel.randInstance) {
				if(sender instanceof SpotUdp) uKernel.performAck((SpotUdp) sender, ack);
				else logWarning("ACK message receive from non SpotUdp instance in HubLocal.performSpotMessage(...)");
			}
			break;
		case ProtocoleSpot.PART : //[PART](int : num)(int : total)[level_protocole|channel]{...content_part...}
			if(sender instanceof SpotUdp) {
				((SpotUdp) sender).recomposeMessage(receiver, mess);
			} else logWarning("PART message receive from non SpotUdp instance in HubLocal.performSpotMessage(...)");	
			break;
		case ProtocoleSpot.BROADCAST : //[BROADCAST][X][Y]{...content_part...} , si X == PROTO_SPOT ->pas de Y, sinon si X = PROTO_APP -> Y = code protocole
			//on doit livrer tous les spots de ce hub gérés localement  
			int conf = mess.readByte();
			if(conf==ProtocoleGlobal.PROTO_SPOT) {
				performSpotMessage(sender, receiver, mess); //protocole spot (/!\ ca doit forcément être une commande SET_ID_CONTEXT ou SMD qui sont les seules autorisées en mode broadcast sur ce protocole)				
			} else if(conf==ProtocoleGlobal.PROTO_APP){ 
				int protocole = mess.readByte();
//System.out.println("===== BROADCAST APP sur protocole = "+protocole);			
				localBroadcastAppMessage(sender, protocole, mess.cloneRest(), true);
			} //else : message non conforme, il est ignoré
			break;
		case ProtocoleSpot.ALL_ID_CONTEXT: //[ALL_ID_CONTEXT](int: nb_id_context)|...pour chaque id: (string: chaine_context)(string: id_context) ...|
			int nb = mess.readInt();
			for(int i=0; i<nb; i++) {
				String ch_ctx = mess.readString();
				String idc = mess.readString();
				sender.internalFixIdContext(ch_ctx, idc);
				checkIdContextConflict(sender.getSpot(), ch_ctx, idc); //On controle que cet id_context ne rentre pas en conflit avec un id_context du spot local
			}
			break;
		case ProtocoleSpot.SET_ID_CONTEXT : //[SET_ID_CONTEXT](string: chaine_context)(string: id_context)	
			String ch_ctx = mess.readString();
			String idc = mess.readString();
			sender.internalFixIdContext(ch_ctx, idc);
			checkIdContextConflict(sender.getSpot(), ch_ctx, idc); //On controle que cet id_context ne rentre pas en conflit avec un id_context du spot local
			break;
		case ProtocoleSpot.GET : //[GET] : on doit répondre ->  SMD et ALL_ID_CONTEXT
			receiver.queryGet(sender);
			break;
		case ProtocoleSpot.SMD : //réception d'une définition SMD
//System.out.println("Réception SMD de "+sender.getSpot().getIdTarget());	int x;	
			//message = //[SMD][def_version](i: nb_value)|...pour chaque valeur: (i: code)[value_size]{...value...}...|
			int smdv = sender.smdVersion();
			int new_version = mess.readByte();
			if(new_version!=smdv){			
				sender.setSMD(new_version, mess.cloneRest());
				if(managed_localspot!=null) managed_localspot.notifyExistGlobalId(sender.getSpot().getIdGlobal()); //on informe local_spot de l'existence éventuel d'un nouveau global_id afin qu'il s'assure d'en avoir un différent.
			}
			break;	
		case ProtocoleSpot.GET_TAG : //[GET_TAG](string: name)
			receiver.queryTag(sender, mess.readString());
			break;
		case ProtocoleSpot.TAG_VALUE: //[TAG_VALUE](string: name)(int: count)...|pour chaque valeur: (value: v)|...
			sender.receiveTag( mess);
			break;
		default :
			logWarning("Unknow command spot receive for master mode ["+cmd+"] (sender: "+sender.getIdLocal()+")");		
		}
		mess.dispose();
	}
	//]
	
	protected void notifyLocalSpotSetIdContext(String ch_context, String idc) {
		if(ready) {
			//on doit notifier cette affectation via un message SET_ID_CONTEXT 
			//[SET_ID_CONTEXT](string: chaine_context)(string: id_context)
			ibroadcast(localspot, -1, DataBytes.writer().b(ProtocoleSpot.SET_ID_CONTEXT).s(ch_context).s(idc).flushReader(), Hub.NO_SERVE_LOCAL);
		}
	}
	
	boolean timer_update=false;
	Runnable smd_sender=null;
	protected synchronized void notifyLocalSpotUpdateSMD() {
		if(!timer_update){ //on va programmer l'envoie de l'update SMD
			timer_update=true;	
			if(smd_sender==null) smd_sender = new Runnable(){
				public void run(){
					timer_update = false;
					DataReader dhi = localspot.getDefSMD();
					ibroadcast(localspot, -1, dhi, Hub.NEED_ACK);
				}
			};
			ThreadStack.perform(smd_sender, 200);
		}
	}
	
	//[méthode de broadcast
	public boolean broadcastByte(int channel, byte content, int option) {	//-> emission du message ne contenant qu'un byte sur l'ensemble du hub
		if(ready && channel>=0) {
			ibroadcastByte(localspot, channel, content, option);
			return(true);
		}
		return(false);
	}
	public void ibroadcastByte(SpotSender sender, int channel, byte content, int option) {	//-> emission du message ne contenant qu'un byte sur l'ensemble du hub
		//on va envoyer le message dans le protocole spot avec la commande BROADCAST à un spot de chaque contexte connecté au hub  
		byte[] buff_broadcast = null; //[BROADCAST][channel]{...content_part...}
		boolean need_ack = (option & NEED_ACK)!=0;
		boolean serve_local = (option & SERVE_LOCAL)!=0;
		int length = (channel==-1 ? 3 : 4);

		//Le contexte local
		if(serve_local || (channel>=0 && !linkers.isEmpty())) { 
			DataReader mess = DataBytes.writer().b(content).flushReader();
			if(channel<0) performSpotMessage(sender, localspot, mess);
			else localBroadcastAppMessage(sender, channel, mess, serve_local);
		}
		
		//Les autres contexte
		synchronized(econtextes) {
			for(EContextMember cl : econtextes.values()) {
				if(!cl.isLocalEContext()) {
					if(buff_broadcast==null) {
						buff_broadcast = new byte[length];
						buff_broadcast[0] = ProtocoleSpot.BROADCAST;
						if(channel==-1) {
							buff_broadcast[1] = ProtocoleGlobal.PROTO_SPOT;			
							buff_broadcast[2] = content;
						} else {
							buff_broadcast[1] = ProtocoleGlobal.PROTO_APP;			
							buff_broadcast[2] = (byte) channel;
							buff_broadcast[3] = content;
						}
					}		
					SpotImpl target =  cl.one();
					target.isend(sender, -1, buff_broadcast, 0, length, need_ack); //c'est un message PROTO_SPOT
				}
			}
		}
	}	
	public boolean broadcastBuff(int channel, byte[] buff , int offset, int size,  int option) { //-> emission du buffer en mode broadcast sur l'ensemble du réseau
		if(ready && channel>=0) {
			ibroadcastBuff(localspot, channel, buff, offset, size, option);
			return(true);
		}
		return(false);
	}	
	public void ibroadcastBuff(SpotSender sender, int channel, byte[] buff , int offset, int size,  int option) {	//-> emission du message ne contenant qu'un byte sur l'ensemble du hub		
		//on va envoyer le message dans le protocole spot avec la commande BROADCAST à un spotde chaque contexte connecté au hub
		//[BROADCAST][channel]{...content_part...}  
		DataReader br_content = null; 
		boolean need_ack = (option & NEED_ACK)!=0;
		boolean serve_local = (option & SERVE_LOCAL)!=0;
	
		//Le contexte local
		if(serve_local || (channel>=0 && !linkers.isEmpty())) { 
			DataReader mess = DataBytes.reader(buff, offset, size);
			if(channel<0) performSpotMessage(sender, localspot, mess);
			else localBroadcastAppMessage(sender, channel, mess, serve_local);
		}
		
		//Les autres contexte
		synchronized(econtextes) {
			for(EContextMember cl : econtextes.values()) {
				if(!cl.isLocalEContext()) {	
					if(br_content==null) {
						DataWriter dw= DataBytes.writer().b(ProtocoleSpot.BROADCAST);
						if(channel==-1) dw.b(ProtocoleGlobal.PROTO_SPOT);
						else dw.b(ProtocoleGlobal.PROTO_APP).b(channel);
						dw.copyBytes(buff, offset, size);
						br_content = dw.flushReader();
					}		
					SpotImpl target =  cl.one();
					target.isend(sender, -1, br_content, need_ack, false); //c'est un message PROTO_SPOT
				}
			}
			if(br_content!=null) br_content.dispose();
		}
	}	
	public boolean broadcast(int channel, DataReader content, int option){	//-> emission du message en mode broadcast sur l'ensemble du hub, dispose() sera appliqué à l'instance content que si auto_dispose==true
		if(ready && channel>=0) {
			ibroadcast(localspot, channel, content, option);
			return(true);
		} else return(false);
	}
	public void ibroadcast(SpotSender sender, int channel, DataReader content, int option){	//-> emission du message en mode broadcast sur l'ensemble du hub, dispose() sera appliqué à l'instance content que si auto_dispose==true
		//on va envoyer le message dans le protocole spot avec la commande BROADCAST à un spot (peut importe qu'il soit local ou proxy) de chaque contexte connecté au hub
		//[BROADCAST][channel]{...content_part...}
		DataReader br_content = null; 
		boolean need_ack = (option & NEED_ACK)!=0;
		boolean serve_local = (option & SERVE_LOCAL)!=0;
				
		//Le contexte local
		if(serve_local || (channel>=0 && !linkers.isEmpty())) { 
			if(channel<0) performSpotMessage(sender, localspot, content.clone());
			else localBroadcastAppMessage(sender, channel, content.clone(), serve_local);
		}
		
		//Les autres contexte
		synchronized(econtextes) {
			for(EContextMember cl : econtextes.values()) {
				if(!cl.isLocalEContext()) {	
					if(br_content==null) {
						DataWriter dw = DataBytes.writer().b(ProtocoleSpot.BROADCAST);
						if(channel==-1) dw.b(ProtocoleGlobal.PROTO_SPOT);
						else dw.b(ProtocoleGlobal.PROTO_APP).b(channel);
						dw.writeData(content);
						br_content = dw.flushReader();
					}		
					SpotImpl target =  cl.one();
//System.out.println("ibroadcast send to "+target.getIdTarget());					
					target.isend(sender, -1, br_content, need_ack, false); //c'est un message PROTO_SPOT
				}
			}
			if(br_content!=null) br_content.dispose();
		}
		if((option & AUTO_DISPOSE)!=0) content.dispose();
	}
	//]
	
	//[Méthode de relégation (pour les spotproxy) d'envoie de message
	public void internal_send(SpotSender sender, String ip_target, int ctarget, int channel, DataReader content, boolean need_ack) {
		Spot target = uKernel.getSpot(ip_target, ctarget);
		if(target!=null) {
			if(target.isLocalSpot()) { //la cible est le spot local
				localspot.ireceive(sender, channel, content, true);
			} else { 
				((SpotImpl) target).isend(sender, channel, content, need_ack, true);
			}
		} else logWarning("internal_send to unknow target : "+ip_target+"#"+ctarget);
	}
	public void internal_sendByte(SpotSender sender, String ip_target, int ctarget, int channel, byte content, boolean need_ack) {
		Spot target = uKernel.getSpot(ip_target, ctarget);
		if(target!=null) {
			if(target.isLocalSpot()) { //la cible est le spot local
				localspot.ireceive(sender, channel, DataBytes.writer().b(content).flushReader(), true);
			} else { 
				((SpotImpl) target).isend(sender, channel, content, need_ack);
			}
		} else logWarning("internal_send to unknow target : "+ip_target+"#"+ctarget);
	}
	public void internal_sendBuff(SpotSender sender, String ip_target, int ctarget, int channel,  byte[] content, boolean need_ack) {
		Spot target = uKernel.getSpot(ip_target, ctarget);
		if(target!=null) {
			if(target.isLocalSpot()) { //la cible est le spot local
				localspot.ireceive(sender, channel, DataBytes.writer().b(content).flushReader(), true);
			} else { 
				((SpotImpl) target).isend(sender, channel, content, 0, content.length, need_ack);
			}
		} else logWarning("internal_send to unknow target : "+ip_target+"#"+ctarget);
	}
	//]
	
	//[Méthode pour les SpotImpl connecté à ce hub
	public void sendTo(SpotSender sender, SpotUdp spot_target, int ack, int lunack, int channel, byte content) { 	//-> emission du message vers spot_target
		//[X][Y](short: ctarget_sender)(short: ctarget_destination)(short: ack)(short: lunack){content} , Y = protocole app si X==PROTO_APP
		DataWriter dw = DataBytes.writer();
		if(channel==-1) dw.b(ProtocoleGlobal.PROTO_SPOT);
		else  dw.b(ProtocoleGlobal.PROTO_APP).b(channel);
		dw.sh(sender.getIdLocal()).sh(spot_target.getIdLocal()).sh(ack).sh(lunack).b(content);
//if(content==ProtocoleSpot.GET) System.out.println("****** GET sended with ack = "+ack); int x;
		uKernel.spotSend(spot_target, dw, ack);		
	}	
	public void sendTo(SpotSender sender, SpotUdp spot_target, int ack, int lunack, int channel, DataReader content) { 	//-> emission du message vers spot_target
		//[X][Y](short: ctarget_sender)(short: ctarget_destination)(short: ack)(short: lunack){content} , Y = protocole app
		//ici on a content.size()<=Protocole.CONTENT_SPOT_SIZE , sinon il faut utiliser sendChunckTo(...)
		DataWriter dw = DataBytes.writer();
		if(channel==-1) dw.b(ProtocoleGlobal.PROTO_SPOT);
		else  dw.b(ProtocoleGlobal.PROTO_APP).b(channel);
		dw.sh(sender.getIdLocal()).sh(spot_target.getIdLocal()).sh(ack).sh(lunack);
		dw.writeData(content);
		
		uKernel.spotSend(spot_target, dw, ack);
	}	
	public void sendChunckTo(SpotSender sender, SpotUdp spot_target, int ack, int lunack, int num, int total, int channel, DataReader content, int size) { 	//-> emission du morceau d'un message vers spot_target
		//[X][Y](short: ctarget_sender)(short: ctarget_destination)(short: ack)(short: lunack){content} , Y = protocole app
		//[PART](int : num)(int : total)[level_protocole|channel]{...content_part...}	<- paquet à recomposer (acquitement obligatoire)		
		DataWriter dw = DataBytes.writer().b(ProtocoleGlobal.PROTO_SPOT).sh(sender.getIdLocal()).sh(spot_target.getIdLocal()).sh(ack).sh(lunack).b(ProtocoleSpot.PART).i(num).i(total);
		if(channel==-1) dw.b(ProtocoleGlobal.PROTO_SPOT);
		else  dw.b(ProtocoleGlobal.PROTO_APP).b(channel);
		dw.writeData(content, size);
		
		uKernel.spotSend(spot_target, dw, ack);
	}
	public void sendTo(SpotSender sender, SpotUdp spot_target, int ack, int lunack, int channel, byte[] buff , int offset, int size) { 	//-> emission du message vers spot_target
		//[X][Y](short: ctarget_sender)(short: ctarget_destination)(short: ack)(short: lunack){content} , Y = protocole app
		//ici on a size<=Protocole.CONTENT_SPOT_SIZE , sinon il faut utiliser sendChunckTo(...)		
		DataWriter dw = DataBytes.writer();
		if(channel==-1) dw.b(ProtocoleGlobal.PROTO_SPOT);
		else  dw.b(ProtocoleGlobal.PROTO_APP).b(channel);
		dw.sh(sender.getIdLocal()).sh(spot_target.getIdLocal()).sh(ack).sh(lunack);
		dw.copyBytes(buff, offset, size);
		
		uKernel.spotSend(spot_target, dw, ack);
	}
	public void sendChunckTo(SpotSender sender, SpotUdp spot_target, int ack, int lunack, int num, int total, int channel, byte[] buff , int offset, int size) { 	//-> emission du morceau d'un message vers spot_target
		//[X][Y](short: ctarget_sender)(short: ctarget_destination)(short: ack)(short: lunack){content} , Y = protocole app
		//[PART](int : num)(int : total)[level_protocole|channel]{...content_part...}	<- paquet à recomposer (acquitement obligatoire)
		DataWriter dw = DataBytes.writer().b(ProtocoleGlobal.PROTO_SPOT).sh(sender.getIdLocal()).sh(spot_target.getIdLocal()).sh(ack).sh(lunack).b(ProtocoleSpot.PART).i(num).i(total);
		if(channel==-1) dw.b(ProtocoleGlobal.PROTO_SPOT);
		else  dw.b(ProtocoleGlobal.PROTO_APP).b(channel);
		dw.copyBytes(buff, offset, size);
		
		uKernel.spotSend(spot_target, dw, ack);
	}
	//]
	
	public Spot getSpotByIdTarget(String idTarget) {
		return(uKernel.getSpotByIdTarget(idTarget));
	}
	public Spot getSpotByIdLocal(int local_id){
		return(uKernel.getSpotByIdLocal(local_id));
	}

	public boolean linkPortal(String ip_portal, int port_portal , String link_key) {
		String linker_id=ip_portal+':'+port_portal+(link_key!=null && link_key.length()>0 ? "$"+link_key : "");
		synchronized(linkers) { if(!linkers.containsKey(linker_id)) linkers.put(link_key, new Linker(this, ip_portal, port_portal, link_key, pass)); }
		return(true);
	}
/*	public boolean localPortal(int port_portal) {
		return(false); //OWO : sera mis en place dans une future version
	}
	public boolean fusion(Hub hub_b) {
		return(false); //OWO : sera mis en place dans une future version
	}
*/	
	//[Interface HubLinker
	public void connectionError(Linker l, String err_mess) { //le linker signal une erreur au moment de la tentative de connection
		logError(err_mess);
	}
	public void linked(Linker l) { 	//le linker signal que sa liaision avec le portail est établie
		logInfo("link to portal ["+l.portal_id+"] enabled");
	}
	public void unlinked(Linker vl, String err_mess) { //le linker signal que sa liaision avec le portail est perdu
		logInfo("link to portal ["+vl.portal_id+"] disabled");
		uKernel.removeSpotProxy(this, vl, 0); //--> on down tous les spot proxy de ce linker
	}
	public void addSpotProxy(Linker vl, int vjid) { //enregistre un spot distant dans le hub --> l'appel linker.notifyProxyAdded(...) sera déclenché quand l'ajout sera effectif
		uKernel.addSpotProxy(this, vl, vjid);
	}
	public void removeSpotProxy(Linker vl, int vjid) { //désenregistre un spot distant dans le hub
		uKernel.removeSpotProxy(this, vl, vjid);
	}
	public ManagedSpotProxy getSpotProxy(Linker vl, int vjid) {
		return(uKernel.getSpotProxy(vl, vjid));
	}
	public DataReader getRawSMD(Spot spot) {
		DataReader smd = null;
		if(spot.isLocalSpot()) smd = managed_localspot.getRawSMD();
		else smd = ((SpotImpl) spot).getRawSMD();
		return(smd!=null ? smd.clone() : null);
	}
	public int getInternalVersionSMD(Spot spot) {
		if(spot.isLocalSpot()) return(managed_localspot.smdVersion());
		else return(((SpotImpl) spot).smdVersion());
	}
	//]
}