package skuad.hubudp.implementation.local;

import java.util.*;
import java.net.*;
import skuad.util.databytes.*;
import skuad.hubudp.implementation.*;

/**
 * Décrit la liste des spots (non instancié, sous forme de couple ctarget/port) d'un host (identifié par son ip)
 * 	=> ip , hubs={hash:  id_hub =>   [...{ctarget: port: }...]   }			NB: id_hub =  "nom:hash_key"
 * 
 *  == l'ensemble des hubs auquel l'host est connecté, et l'ensemble des spots targets (identifié par un code ctarget) qu'il possède dans chacun de ces hubs
 */
public class SpotMap {
	
	String ip;
	byte[] byte_ip;
	private InetAddress iadr;
	int rinstance;
	int version;
	int wd;			//wd du master de l'hote
	
	Map<String, Vector<Target>> hubs; //hubs={hash:  id_hub =>   [...{ctarget: port: }...]   }	
	int count_target;	//nombre de cible total dans cette map
	
	public SpotMap(String vip, InetAddress viadr) {
		ip = vip;
		iadr = viadr;
		byte_ip = null;
		hubs = new HashMap<String,  Vector<Target> >();
		count_target = 0;
		rinstance = 0;
		version = 0;
		wd=0;
	}
	
	public synchronized Vector<Target> targetOnPort(int port){ //liste des targets pour le port indiqué
		Vector<Target> rep = new Vector<Target>();
		for(Vector<Target> targets : hubs.values() ){
			for(int i=0; i<targets.size(); i++) {
				Target cur = targets.get(i);
				if(port==cur.port) rep.add(cur);
			}
		}
		return(rep);
	}
	
	public void setRInstance(int r, int v) { rinstance = r; version=v;}
	
	public void writeIp(DataWriter dw) {
		if(byte_ip==null) {
			if(iadr==null) try{ iadr = InetAddress.getByName(ip); } catch(Exception e){ e.printStackTrace(); }
			byte_ip = iadr.getAddress();
		}
		for(int i=0; i<4; i++) dw.b(byte_ip[i]);
	}
	
	public InetAddress getIadr() {
		if(iadr==null) try{ iadr = InetAddress.getByName(ip); } catch(Exception e) { e.printStackTrace(); }
		return(iadr);
	}
	
	public synchronized boolean targetExist(int ctarget, int port, int ri) { //true ssi la cible d'id ctarget existe sur cet host (tout hub et contexte d'exécution confondu)
		for(Vector<Target> targets : hubs.values() ){
			for(int i=0; i<targets.size(); i++) {
				Target cur = targets.get(i);
				if(cur.ctarget == ctarget && (port==0 || port==cur.port) && (ri==0 || ri==cur.rinstance)) return(true);
			}
		}
		return(false);
	}
	
	public int targetCount() {
		return(count_target);	//nombre de cible total dans cette map
	}
	
	public synchronized void addTarget(UdpKernel uk, String id_hub, int ctarget, int port, int ri, int new_version, boolean notify_change){ //new_version==0 pour auto-incrémentation de la version
//System.out.println("=== addTarget "+id_hub+"#"+ctarget);	
		//va déclencher k.targetInMap(...) aprés modification
		Vector<Target> targets = hubs.get(id_hub);
		if(targets==null){
			targets = new Vector<Target>();
			hubs.put(id_hub, targets);
		}
		Target t = new Target(id_hub, ctarget, port, ri);
		targets.add(t);
		count_target++;
		
		if(new_version==0) new_version = HubManager.nextVersion(version);
		version = new_version;
		uk.targetInMap(this, t, notify_change);
	}
	
	public synchronized void removeTarget(UdpKernel uk, String id_hub, int ctarget, int new_version, boolean notify_change){ //new_version==0 pour auto-incrémentation de la version
		Vector<Target> targets = hubs.get(id_hub);
		if(targets!=null){
			for(int i=0; i<targets.size(); i++) {
				Target cur = targets.get(i); 
				if(cur.ctarget==ctarget) {
					targets.removeElementAt(i);
					count_target--;
					if(new_version==0) new_version = HubManager.nextVersion(version);
					version = new_version;
					uk.targetOutMap(this, cur, notify_change);
					return;
				}
			}
		}		
	}
	
	public synchronized void searchTargetForHub(TargetFinder finder, String id_hub, boolean notify_change) { //va déclencher k.targetInMap(...) pour chaque cible connecté au hub d'id id_hub
//System.out.println("=== searchTargetForHub : "+id_hub);
		//NB: les spots gérés par le contexte courant seront jamais traité par cet appel grace à la valeur omit_port
		Vector<Target> targets = hubs.get(id_hub);
		if(targets!=null) for(int i=0; i<targets.size(); i++) finder.targetInMap(this, targets.get(i), notify_change);
	}

	public synchronized int buildSpotMap(int cmd, int randInstance, int version, Vector<DataWriter> stack, int page, DataWriter cdw) { //remplie le stack avec les paquets de description de la SpotMap 'map'
		//--> [85][cmd](short: rand_instance)(short: version)[page][nb_page] {...| (s: id_hub)(i:nb_target) $ {...|  (short: ctarget)(short: port)(short: rinstance) $ |...} |...}
		//depuis la page courante 'page' et le paquet courant 'cdw'
		//-> retourne le nouveau nombre de page
		for(String id_hub : hubs.keySet()) {
			Vector<Target> targets = hubs.get(id_hub);
			int old_pos = cdw.size(); 
			cdw.s(id_hub).i(targets.size());
			if(cdw.size()>ProtocoleLocal.MAX_UDP_BUFFER_SIZE) { //on doit opérer sur un nouveau paquet
				cdw.setSize(old_pos);
				page++;
				cdw = DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(cmd).sh(randInstance).sh(version).b(page);
				cdw.b(0); //[7] nb_page sera fixé par le code appelant
				stack.add(cdw);
				cdw.s(id_hub).i(targets.size());
			}

			for(Target target : targets) {
				if(cdw.size()+6>ProtocoleLocal.MAX_UDP_BUFFER_SIZE) { //on doit opérer sur un nouveau paquet
					page++;
					cdw = DataBytes.writer().b(ProtocoleGlobal.PROTO_KERNEL).b(cmd).sh(randInstance).sh(version).b(page);
					cdw.b(0); //[7] nb_page sera fixé en (1)
					stack.add(cdw);
				}
				cdw.sh(target.ctarget).sh(target.port).sh(target.rinstance);
			}
		}
		return(page);
	}
	
	public synchronized void removeTargetOnPort(UdpKernel uk, int port, int ri, int new_version, boolean notify_change) { //port==0 pour down tous les ports
		//NB: si notify_change===false il n'y aura pas de suppression dans les listes et pas de changement de version de la map (==> la map complête est sensée être supprimé aprés ce traitement)
		for(Vector<Target> targets : hubs.values() ){
			for(int i=0; i<targets.size(); i++) {
				Target cur = targets.get(i);
				if((port==0 || cur.port==port) && (ri==0 || cur.rinstance==ri)) {
					targets.removeElementAt(i);	i--; count_target--;
					if(new_version==0) new_version = HubManager.nextVersion(version);
					version = new_version;
					uk.targetOutMap(this, cur, notify_change);
				}
			}
		}		
	}
	
	public synchronized void removeHub(UdpKernel uk, String id_hub, int new_version, boolean notify_change) { //port==0 pour down tous les ports
		//NB: les spots gérés par le contexte courant seront jamais traité par cet appel
		//NB: si notify_change===false il n'y aura pas de suppression dans les listes et pas de changement de version de la map (==> la map complête est sensée être supprimé aprés ce traitement)
		Vector<Target> targets = hubs.remove(id_hub);
		if(targets!=null) {
			for(int i=0; i<targets.size(); i++) {
				Target cur = targets.get(i);
				count_target--;
				uk.targetOutMap(this, cur, notify_change);
			}
		}
		if(new_version==0) new_version = HubManager.nextVersion(version);
		version = new_version;
	}
}