package skuad.hubudp.implementation.local;

import skuad.util.databytes.*;
import skuad.util.base.Network;
import java.net.*;
import skuad.hubudp.implementation.*;

public class SpotUdp extends SpotImpl {
	
	InetAddress address;
	boolean is_local_host;
	String ip;
	HubLocal localhub;
	EContextMemberUDP econtext;
	int rinstance;
	
	public SpotUdp(HubLocal vhub, int ct, String vip, InetAddress iadr, int p, int ri) {
		super(vhub, ct, p);
		localhub = vhub;
		address = iadr;
		is_local_host = Network.isLocalAddress(address);
		ip = vip;
		econtext = null;
		rinstance = ri;
	}	
	public boolean isLocalHost(){ return(is_local_host); }
	public String getIp() { return(ip); }
	public boolean isLocalEContext() { return(false); }
	
	//[Interface SpotSender
	DataWriter recomposer=null;	//paquet en cours de recomposition, null si aucun
	int next_num_recomp=1;		//numéro du prochain paquet attendu
	public synchronized void recomposeMessage(SpotReceiver receiver, DataReader mess) { //mess =  (int : num)(int : total)[level_protocole]|[channel]|{...content_part...}
		if(!is_up) return;
		int num = mess.readInt();
		int total = mess.readInt();
		int protocole;
		if(mess.readByte()==ProtocoleGlobal.PROTO_APP) protocole = mess.readByte();
		else protocole=-1;

		if(num==1) { //c'est le 1er morceau
			if(recomposer!=null) {
				recomposer.dispose(); //normalement cette situation ne devrait pas se produire
				hub.logWarning("Error of recomposition message in internal protocole SPOT (current not null)");
			}
			recomposer = DataBytes.writer();
		} else if(next_num_recomp!=num) {
			hub.logWarning("Error of recomposition message in internal protocole SPOT (bad order)");
			return; //normalement cette situation ne devrait pas se produire
		}
		recomposer.writeData(mess);
		next_num_recomp=num+1;
		
		if(num==total) { //fin de recomposition du paquet
			DataReader mess_complet = recomposer.flushReader();
			recomposer=null;
			hub.performMessage(this, receiver, protocole, mess_complet); //le message est complet, on peut le traiter maintenant
		}
	}
	//]
	
	//[Méthodes pour le Kernel
	protected void internalDown() {
		//--> purger: waitmess et recomposer, ...
		if(recomposer!=null) {
			recomposer.dispose();
			recomposer=null;
		}
/*OXO: purge des WaitingMessage de cette source ?		
  synchronized(waitmess) {
			for(WaitingMessage wm : waitmess) {
				wm.mess.dispose();
			}
			waitmess.removeAllElements();
		}
*/		
		super.internalDown();
	}
	public void messageFrom(SpotReceiver receiver, int ack, int lunack, int protocole, DataReader mess) { //le message mess vient d'arriver en provenance du spot this à destination du spot receiver
//System.out.println("-messageFrom ("+this.getIdTarget()+")-> ack = "+ack+" receiver = "+receiver.getSpot().getIdTarget());
		//mess = { ...content... }
			//OXO: aucun message de code ack inférieur à lunack ne sera renvoyé par l'expéditeur (sauf cas particulier de rotation de numérotation) 
			//	-> on devrait ajuster next_receive_ack si nécessaire (et débloquer les attentes fantomes dans waitmess ?)
			
			if(ack==0 || econtext.isNextReceiveAck(ack)) { //on peut traiter ce message tout de suite
//System.out.println("-messageFrom.performMessage NOW ("+this.getIdTarget()+")-> ack = "+ack+" receiver = "+receiver.getSpot().getIdTarget());	
				hub.performMessage(this, receiver, protocole, mess);
				if(ack!=0) econtext.checkWaiting(); //on regarde si d'autres messages peuvent être sorti de la liste d'attente
			} else {
//System.out.println("-messageFrom.performMessage WAIT ("+this.getIdTarget()+")-> ack = "+ack+" receiver = "+receiver.getSpot().getIdTarget());				
				econtext.needWaiting(this, ack, receiver, protocole, mess); //si nécessaire le paquet sera mis en attente de réception des paquets intermédiaire
			}
	}
	//]
	
 	//[envoie de message vers ce spot
	public boolean isend(SpotSender sender, int channel, byte content, boolean need_ack) {	//-> emission du message ne contenant qu'un byte du sender vers le spot this
//System.out.println("--isend for "+getIdLocal());		
		if(isUp()) {
			int ack=0, lunack=0;
			if(need_ack) {
				ack = econtext.nextSendAck();
				lunack = econtext.getLunack(ack);		
			}
			localhub.sendTo(sender, this, ack, lunack, channel, content);
			return(true);
		}
		return(false);
	}	
	public boolean isend(SpotSender sender, int channel, DataReader content, boolean need_ack, boolean auto_dispose) { //nb: sur cette méthode interne l'utilisation du protocole -1 (PROTO_SPOT) est possible
		if(isUp()) {
			int ack=0, lunack=0;
			int size = content.avaible();
			if(size>(ProtocoleLocal.CONTENT_SPOT_SIZE)) { //on doit découper le message en plusieurs morceaux
				econtext.ackLock.lock(); 	//le numéro de paquet doivent se suivre
					int nb = (int) Math.ceil((double) size / (double) ProtocoleLocal.CONTENT_SPOT_SIZE);
					DataReader chuncked_content = content.clone();
					for(int i=1; i<=nb; i++) {
						ack = econtext.nextSendAck(); 	
						if(lunack==0) lunack = econtext.getLunack(ack);		
						int cur_size = Math.min(size, ProtocoleLocal.CONTENT_SPOT_SIZE);
						localhub.sendChunckTo(sender, this, ack, lunack, i, nb, channel, chuncked_content, cur_size);
						chuncked_content.skip(cur_size);
						size-=cur_size;
					}
					chuncked_content.dispose();
				econtext.ackLock.unlock();
			} else {
				if(need_ack) {
					ack = econtext.nextSendAck();
					lunack = econtext.getLunack(ack);			
				}
//System.out.println("-"+this.getIdTarget()+"- isend sur channel = "+channel+" ack = "+ack+" cmd = "+content.getByte(0));				
				localhub.sendTo(sender, this, ack, lunack, channel, content);
			}
			if(auto_dispose) content.dispose();
			return(true);
		}
		return(false);
	}	
	public boolean isend(SpotSender sender, int channel, byte[] buff , int offset, int size, boolean need_ack) {	//nb: sur cette méthode interne l'utilisation du protocole -1 (PROTO_SPOT) est possible	
		if(isUp()) {			
			int ack=0, lunack=0;
			if(size>(ProtocoleLocal.CONTENT_SPOT_SIZE)) { //on doit découper le message en plusieurs morceaux
				econtext.ackLock.lock(); 	//le numéro de paquet doivent se suivre
					int nb = (int) Math.ceil((double) size / (double) ProtocoleLocal.CONTENT_SPOT_SIZE);
					for(int i=1; i<=nb; i++) {
						ack = econtext.nextSendAck(); 	
						if(lunack==0) lunack = econtext.getLunack(ack);		
						int cur_size = Math.min(size, ProtocoleLocal.CONTENT_SPOT_SIZE);
						localhub.sendChunckTo(sender, this, ack, lunack, i, nb, channel, buff , offset, cur_size);
						offset+=cur_size;
						size-=cur_size;
					}
					econtext.ackLock.unlock();
			} else {
				if(need_ack) {
					ack = econtext.nextSendAck();
					lunack = econtext.getLunack(ack);		
				}
				localhub.sendTo(sender, this, ack, lunack, channel, buff , offset, size);
			}
			return(true);
		}
		return(false);
	}	
	//]
}
