/**
 * Classe d'implémentation du composant de liaison Hub -> portail de jonction.
 * 
 */
package skuad.hubudp.implementation.portal;

import java.net.*;
import java.util.*;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;
import java.io.*;
import skuad.util.base.Network;
import skuad.util.thread.*;
import skuad.hubudp.implementation.*;
import skuad.hubudp.*;
import skuad.util.databytes.DataBytes;
import skuad.util.databytes.DataReader;

public class Joiner extends HubImpl implements Runnable {
	
	public final static byte ERROR=0;
	public final static byte IN_RUN=1;
	
	Socket socket_client;
	DataInputStream from_server;
	DataOutputStream to_server;
	String portal_id;
	String ip_portal;
	int port_portal;
	String join_key;
	String hub_name;
	Lock sendlock;
	private Map<String, SpotPortal> otherSpots;		//table des spots connecté au hub autre que le spotlocal, cle = idTarget
		
	public Joiner(HubManager hm, String vip_portal, int vport_portal, String vkey) {
		super(hm);
		ip_portal = vip_portal;
		port_portal = vport_portal;
		portal_id = ip_portal+":"+port_portal;
		join_key = (vkey == null ? "" : vkey);
		hub_name=null;
		localSpotLinked(-1, -1); //va conduire à instancier le localspot du joiner via Joiner.instanciateLocalSpot()
			//NB: l'instance du localspot d'un joiner est toujours disponible quelque soit l'état du hub
		sendlock=new ReentrantLock();
		otherSpots = new HashMap<String, SpotPortal>();
		
		setState(Hub.TRY_JOIN_PORTAL, portal_id);
		ThreadStack.perform(this);
	}
	protected LocalSpotImpl instanciateLocalSpot(int ctarget, int p) {
		return(new LocalSpotPortal(this, ctarget, p));
	}
	
	//boucle d'écoute TCP
	public void run() {
		while(true){
			if(join()){ //on essais de se connecter au portail
				
				//On est connecté au Portail, on passe dans l'état INITIALIZING
				setState(Hub.INITIALIZING, portal_id);
				
				while(true) { //boucle d'écoute des messages du portail
					byte cmd=0;
					String idTarget;
					int ctarget, length, protocole, port;
					DataReader mess;
					SpotPortal spotp;
					try { 
						cmd = from_server.readByte(); 
						switch(cmd) {
						case ProtocolePortal.PORTAL_NO_ACTIVATED : //le portail n'est pas ou plus activé
							downHub();
							setState(Hub.PORTAL_DISABLED, portal_id); //on va attentre la lecture du status LINKED du portail						
							break; //on va attendre le prochain message
						case ProtocolePortal.PORTAL_ACTIVATED : //le portail vient d'être activé
							//OWO: tester si on est bien avec un spotlocal à down ?
							setState(Hub.PORTAL_ENABLED, portal_id);				
							break;
						case ProtocolePortal.BAD_JOIN_KEY : //le portail vient d'être activé mais la clé join_key n'est pas la bonne
							setState(Hub.BAD_KEY, "Bad join key to connect on portal ["+portal_id+"]");
							break;	
							
						case ProtocolePortal.LOCALSPOT_LINKED : //J<-P: [LOCALSPOT_LINKED](int: ctarget)(int: port)(s: hubname)
//System.out.println("-Joiner : LOCALSPOT_LINKED");		
							ctarget =  from_server.readInt(); 
							port =  from_server.readInt();
							hub_name = from_server.readUTF();
							localspot.changeCTargetAndPort(ctarget, port);
							setReady(); //on pass le hub à l'état READY
							
							notifyLocalSpotUpdateSMD(); //pour déclencher l'envoie des SMD du localspot vers le portail
							if(localspot.nbIdContext()>0) { //on envoie aussi tous les ID CONTEXT
								sendlock.lock(); try {  
									//[ALL_ID_CONTEXT](int: nb){... (string: ch_context)(string: id_context }
									to_server.writeByte(ProtocolePortal.ALL_ID_CONTEXT);
									if(!localspot.sendAllIdContext(to_server)) {
										logWarning("Error read/write with portal ["+portal_id+"] (sendAllIdContext)");
										close(); //va conduire à mettre fin à la boucle d'écoute du portail										
									}
								} catch(Exception e){
									logWarning("Error read/write with portal ["+portal_id+"] ("+e.getMessage()+")");
									close(); //va conduire à mettre fin à la boucle d'écoute du portail	
								} finally {
									sendlock.unlock();
								}
							}
							break;
						case ProtocolePortal.NOTIFY_UP: 	//[NOTIFY_UP](string: ip)(int : port)(int: ctarget)
							String ip = from_server.readUTF();
							port = from_server.readInt();
							ctarget = from_server.readInt();
							spotp = new SpotPortal(this, ip, port, ctarget);
							otherSpots.put(spotp.getIdTarget(), spotp);
							enter(spotp, 0);
							break;
						case ProtocolePortal.NOTIFY_DOWN: 	//[NOTIFY_DOWN](string: idTarget)
							idTarget = from_server.readUTF();
							spotp = otherSpots.remove(idTarget);
							if(spotp!=null) exit(spotp, 0);
							break;			
						case ProtocolePortal.NOTIFY_IDCONTEXT: 	//[NOTIFY_IDCONTEXT](string: idTarget)(st: ch_context)(st: id_context)
							idTarget = from_server.readUTF();
							spotp = otherSpots.get(idTarget);
							if(spotp!=null) spotp.internalFixIdContext(from_server.readUTF(), from_server.readUTF());
							break;			
						case ProtocolePortal.NOTIFY_UPDATE_SMD: 	//[NOTIFY_UPDATE_SMD](string: idtarget)(int: version)(int: length)(DataReader: smd)		
							idTarget = from_server.readUTF();
							spotp = otherSpots.get(idTarget);
							int version =  from_server.readInt();
							length = from_server.readInt();
							DataReader data= DataBytes.reader(from_server, length);
							if(spotp!=null) spotp.setSMD(version, data); 
							break;	
						case ProtocolePortal.DELIVER_BROADCAST: 	//[DELIVER_BROADCAST](string: idtarget_sender)(int: protocole)(int: length)(DataReader: mess)
							idTarget = from_server.readUTF();
							spotp = otherSpots.get(idTarget);
							protocole = from_server.readInt();
							length = from_server.readInt();
							mess= DataBytes.reader(from_server, length);
							if(spotp!=null) {
								//-> on va servir les manager du spot local 
								Vector<MessageReceiverDescriptor> listpm = mreceivers.get(protocole);
								if(listpm!=null && listpm.size()>0) { //on a des protocoles à servir
									synchronized(listpm) { for(int i=0; i<listpm.size(); i++) new RunnerManager(listpm.get(i).receiver, spotp, protocole, mess); }
								}
							}
							break;
						case ProtocolePortal.QUERY_TAG: //[QUERY_TAG](string: idtarget_requester)(int: jid)(string: tag_name)
							idTarget = from_server.readUTF();
							spotp = otherSpots.get(idTarget);
							from_server.readInt(); //jid, inutile ici
							String tag_name = from_server.readUTF();
							if(spotp!=null) localspot.queryTag(spotp, tag_name);
							break;							
						case ProtocolePortal.RECEIVE_APP_MESSAGE: //[RECEIVE_APP_MESSAGE](string: idtarget_sender)(int: jid)(int: protocole)(int: length)(DataReader: mess)
							idTarget = from_server.readUTF();
							spotp = otherSpots.get(idTarget);
							from_server.readInt(); //jid, inutile ici
							protocole = from_server.readInt();
							length = from_server.readInt();
							mess= DataBytes.reader(from_server, length);
							if(spotp!=null) localspot.receiveAppMessage(spotp, protocole, mess);
							break;
						}
					}
					catch(Exception e) {
						logError("Read error with portal ["+portal_id+"] ("+e.getMessage()+")");
						break; //on sort de la boucle d'écoute
					}
				}			
				//Ici on est déconnecté du portail -> libération des ressources
				close();
				downHub();
				setState(Hub.TRY_JOIN_PORTAL, portal_id);
			} //si on est pas arrivé à se connceter, ou en cas d'erreur on fera une nouvelle tentative dans 10 secondes
			try{ Thread.sleep(10000); } 
			catch(Exception e){ e.printStackTrace(); }		
		}
	}
	
	private void downHub() {
		if(ready){
			ready=false;
			downAllNoLocalspot();
			localspot.goDown();	
		}
	}
	
	private boolean join(){
		//création de la socket TCP
		socket_client=null;
		try { socket_client=new Socket(ip_portal, port_portal); } 
		catch(Exception e) {
			logWarning("Error on connect to portal ["+portal_id+"] ("+e.getMessage()+")");
			return(false);
		}
		
		//ouverture des flux de communication
		from_server = null; to_server=null;
		try {
			from_server=new DataInputStream(socket_client.getInputStream());
			to_server=new DataOutputStream(socket_client.getOutputStream());
		}catch(Exception e) {
			logWarning("Error when open stream with portal ["+portal_id+"] ("+e.getMessage()+")");
			close();
			return(false);
		}
		
		//Protocole de connexion
		try {
			to_server.writeByte(ProtocolePortal.CLIENT_PORTAL);
			int cmd =  from_server.readByte();
			if(cmd!=ProtocolePortal.SERVER_PORTAL) {
				logWarning("Bad protocole in negociation with portal ["+portal_id+"]");
				close();
				return(false);
			}
			
			to_server.writeByte(ProtocolePortal.JOIN);
			to_server.writeUTF(join_key);
			cmd =  from_server.readByte();
			if(cmd == ProtocolePortal.OK) return(true);				

			logWarning("Bad response when connect to portal ["+portal_id+"]");
			close();
			return(false);
		}catch(Exception e) {
			logWarning("Error read/write with portal ["+portal_id+"] ("+e.getMessage()+")");
			close();
			return(false);
		}
	}
	
	private void close() {
		try { to_server.close(); } catch(Exception e){}
		try { from_server.close(); } catch(Exception e){}
		try { socket_client.close(); } catch(Exception e){}
	}
	
	//[Méthode abstraite HubImpl
	public String getHubName() {
		return(hub_name);
	}
	public boolean isPrivate() { return(join_key!=null && join_key.length()>0); }
	protected void performSpotMessage(SpotSender sender, SpotReceiver receiver, DataReader mess) {
		//normalement un hub joiner ne peut pas recevoir de SpotMessage !
		logWarning("Error call performSpotMessage(...) applied on Joiner !");
	}
	protected void localBroadcastAppMessage(SpotSender spot_sender, int protocole, DataReader mess, boolean serve_sender) {
		if(serve_sender || !spot_sender.getSpot().isLocalSpot()) receiveLocalAppMessage(spot_sender, protocole, mess); //-> on sert les manager du spot local 
		else mess.dispose();
	}
	
	protected void notifyLocalSpotSetIdContext(String ch_context, String idc) {
		if(ready) {
			sendlock.lock(); try {
				//[SET_ID_CONTEXT](string: ch_context)(string: id_contex)
				to_server.writeByte(ProtocolePortal.SET_ID_CONTEXT);
				to_server.writeUTF(ch_context);
				to_server.writeUTF(idc);
			} catch(Exception e){
				logWarning("Error read/write with portal ["+portal_id+"] ("+e.getMessage()+")");
				close(); //va conduire à mettre fin à la boucle d'écoute du portail	
			} finally {
				sendlock.unlock();
			}
		}
	}
	protected void notifyLocalSpotUpdateSMD() {
		if(ready) {
			sendlock.lock(); try {
				//[UPDATE_SMD](int: version_smd)(int: length)(DataReader: raw_smd)
				to_server.writeByte(ProtocolePortal.UPDATE_SMD);
				to_server.writeInt(localspot.smdVersion());
				DataReader dsmd = localspot.getRawSMD();
				int length = (dsmd!=null ? dsmd.avaible() : 0);
				to_server.writeInt(length);
				if(length>0) dsmd.writeToStream(to_server, length);
			} catch(Exception e){
				logWarning("Error read/write with portal ["+portal_id+"] ("+e.getMessage()+")");
				close(); //va conduire à mettre fin à la boucle d'écoute du portail	
			} finally {
				sendlock.unlock();
			}			
		}		
	}
	//envoies de message en mode broadcast
	public boolean broadcastByte(int channel, byte content, int option) {	//-> emission du message ne contenant qu'un byte sur l'ensemble du hub
		if(ready && channel>=0) {
			sendlock.lock(); try {
				//[BROADCAST_BUFF](int: option)(int: protocole)(int: length)[content]
				to_server.writeByte(ProtocolePortal.BROADCAST_BUFF);
				to_server.writeInt(option);
				to_server.writeInt(channel);
				to_server.writeInt(1);
				to_server.writeByte(content);
			} catch(Exception e){
				logWarning("Error read/write with portal ["+portal_id+"] ("+e.getMessage()+")");
				close(); //va conduire à mettre fin à la boucle d'écoute du portail	
			} finally {
				sendlock.unlock();
			}				
			return(true);
		}
		return(false);
	}
	public boolean broadcastBuff(int channel, byte[] buff , int offset, int size,  int option) { //-> emission du buffer en mode broadcast sur l'ensemble du réseau
		if(ready && channel>=0) { 
			sendlock.lock(); try {
				//[BROADCAST_BUFF](int: option)(int: protocole)(int: length){content}
				to_server.writeByte(ProtocolePortal.BROADCAST_BUFF);
				to_server.writeInt(option);
				to_server.writeInt(channel);
				to_server.writeInt(size);
				to_server.write(buff, offset, size);
			} catch(Exception e){
				logWarning("Error read/write with portal ["+portal_id+"] ("+e.getMessage()+")");
				close(); //va conduire à mettre fin à la boucle d'écoute du portail	
			} finally {
				sendlock.unlock();
			}				
			return(true);
		}
		return(false);
	}
	public boolean broadcast(int channel, DataReader content, int option){	//-> emission du message en mode broadcast sur l'ensemble du hub, dispose() sera appliqué à l'instance content que si auto_dispose==true
		if(ready && channel>=0 ) {
			ibroadcast(channel, content, option);
			return(true);
		} else return(false);
	}
	public void ibroadcast(int channel, DataReader content, int option){	//-> emission du message en mode broadcast sur l'ensemble du hub, dispose() sera appliqué à l'instance content que si auto_dispose==true
		sendlock.lock(); try {
			//[BROADCAST](int: option)(int: protocole)(int: length)(DataReader: mess)
			to_server.writeByte(ProtocolePortal.BROADCAST);
			to_server.writeInt(option);
			to_server.writeInt(channel);
			int length = (content!=null ? content.avaible() : 0);
			to_server.writeInt(length);
			if(length>0) content.writeToStream(to_server, length);
		} catch(Exception e){
			logWarning("Error read/write with portal ["+portal_id+"] ("+e.getMessage()+")");
			close(); //va conduire à mettre fin à la boucle d'écoute du portail	
		} finally {
			sendlock.unlock();
		}		
		if((option & AUTO_DISPOSE)!=0) content.dispose();
		else content.reset();
	}
	
	//[Méthode pour les SpotPortal connecté à ce hub
	public void sendTo(SpotImpl spot_target, boolean need_ack, int channel, byte content) {
		sendlock.lock(); try {
			//J->P : [SEND_BUFF](string: ip)(int ctarget)[need_ack](int: protocole)(int: length)[content]
			to_server.writeByte(ProtocolePortal.SEND_BUFF);
			to_server.writeUTF(spot_target.getIp());
			to_server.writeInt(spot_target.getIdLocal());
			to_server.writeByte(need_ack ? 1 : 0);
			to_server.writeInt(channel);
			to_server.writeInt(1);
			to_server.writeByte(content);
		} catch(Exception e){
			logWarning("Error read/write with portal ["+portal_id+"] ("+e.getMessage()+")");
			close(); //va conduire à mettre fin à la boucle d'écoute du portail	
		} finally {
			sendlock.unlock();
		}		
	}
	public void sendTo(SpotImpl spot_target, boolean need_ack, int channel, DataReader content) {		
//System.out.println("--sendTo "+spot_target.getIdTarget()+" channel = "+channel);		
		sendlock.lock(); try {
			//J->P : [SEND](string: ip_target)(int: ctarget)[need_ack](int: protocole)(int: length)(DataReader: mess)
			to_server.writeByte(ProtocolePortal.SEND);
			to_server.writeUTF(spot_target.getIp());
			to_server.writeInt(spot_target.getIdLocal());
			to_server.writeByte(need_ack ? 1 : 0);
			to_server.writeInt(channel);
			int length = (content==null ? 0 : content.avaible());
			to_server.writeInt(length);
			if(length>0) content.writeToStream(to_server, length);
		} catch(Exception e){
			logWarning("Error read/write with portal ["+portal_id+"] ("+e.getMessage()+")");
			close(); //va conduire à mettre fin à la boucle d'écoute du portail	
		} finally {
			sendlock.unlock();
		}
	}
	public void sendTo(SpotImpl spot_target, boolean need_ack,  int channel, byte[] buff , int offset, int size) {
		sendlock.lock(); try {
			//J->P : [SEND_BUFF](string: ip_target)(int: ctarget)[need_ack](int: protocole)(int: length)[content]
			to_server.writeByte(ProtocolePortal.SEND_BUFF);
			to_server.writeUTF(spot_target.getIp());
			to_server.writeInt(spot_target.getIdLocal());
			to_server.writeByte(need_ack ? 1 : 0);
			to_server.writeInt(channel);
			to_server.writeInt(size);
			to_server.write(buff, offset, size);
		} catch(Exception e){
			logWarning("Error read/write with portal ["+portal_id+"] ("+e.getMessage()+")");
			close(); //va conduire à mettre fin à la boucle d'écoute du portail	
		} finally {
			sendlock.unlock();
		}
	}
	//]
	
	public Spot getSpotByIdTarget(String idTarget){ //idTarget == ip#ctarget
		Spot spot = otherSpots.get(idTarget);
		if(spot==null) {
			//il faut testé si spot n'est pas un spot géré localement
			int ddotp = idTarget.indexOf("#");
			if(ddotp>0) {
				String ip = idTarget.substring(0, ddotp);
				if(Network.isLocalAddress(ip)) {
					try {
						int ct = Integer.parseInt(idTarget.substring(ddotp+1));
						if(localspot.getIdLocal()==ct) spot = localspot;
					}catch(Exception e) {}
				}
			}
		}
		return(spot);		
	}
	public Spot getSpotByIdLocal(int local_id){
		Spot spot = null;
		//on regarde si la cible est dans le contexte d'exécution local
		if(localspot.getIdLocal()==local_id) spot=localspot;
		if(spot==null) spot = otherSpots.get(Network.LOCALHOST+'#'+local_id); //on regarde si la cible est dans un autre contexte de la machine hote
		return(spot);
	}
	
	//hinibition des méthodes relative au portail: 
	public boolean linkPortal(String ip_portal, int port_portal , String link_key) { return(false); } //opération non permise sur un hub distant
	public boolean localPortal(int port_portal){ return(false); } //opération non permise sur un hub distant
	public boolean fusion(Hub hub_b){ return(false); } //opération non permise sur un hub distant
	//]
}