/**
 * Classe d'implémentation du composant de liaison Hub -> portail de jonction.
 * 
 * Chaque client du portail donnera lieu à l'instanciation locale d'un spot proxy
 * connecté à ce linker.
 * Le linker va recevoir et emmettre les commandes des clients distant via le portail.
 * 
 */
package skuad.hubudp.implementation.portal;

import java.net.*;
import java.io.*;
import skuad.hubudp.*;
import skuad.util.databytes.DataBytes;
import skuad.util.databytes.DataReader;
import skuad.util.thread.*;
import java.util.concurrent.locks.*;
import skuad.hubudp.implementation.*;

public class Linker implements Runnable, SpotObserver {

	public final static byte ERROR=0;
	public final static byte IN_RUN=1;

	HubLinker hub;
	String ip_portal;
	int port_portal;
	public String portal_id;
	String link_key, join_key;
	Lock sendlock;
	String last_errormess;
	
	Socket socket_client=null;
	DataInputStream from_server=null;
	DataOutputStream to_server=null;
	
	public Linker(HubLinker vhub, String vip_portal, int vport_portal, String vlink_key, String vjoin_key) {
		hub = vhub;
		ip_portal = vip_portal;
		port_portal = vport_portal;
		portal_id = ip_portal+":"+port_portal;
		link_key = (vlink_key == null ? "" : vlink_key);
		join_key = (vjoin_key == null ? "" : vjoin_key);
		sendlock=new ReentrantLock();
		last_errormess=null;
		
		ThreadStack.perform(this); //boucle de connexion
	}
	
	//boucle d'écoute TCP
	public void run() {
		while(true){
			if(link()){ //on essais de se linker au portail
				hub.linked(this); 
				hub.addSpotObserver(this);
				byte cmd=0; 
				int option, length, protocole; boolean need_ack;
				String ip; int ctarget;
				ManagedSpotProxy mspot;
				DataReader mess;
				
				while(true) { //boucle d'écoute des messages du portail (B)
					try { 
						cmd = from_server.readByte(); 
//System.out.println("*** receive Portal command = "+cmd);							
						int jid;
						switch(cmd) {
						case ProtocolePortal.ADD_JOINER : //[ADD_JOINER](int: jid)				
							jid = from_server.readInt();
//System.out.println("* add joiner, jid = "+jid);							
							hub.addSpotProxy(this, jid);
							break;
						case ProtocolePortal.REMOVE_JOINER:  //L<-P: [REMOVE_JOINER](int: jid)
							jid = from_server.readInt();
							hub.removeSpotProxy(this, jid);  
							break;
						case ProtocolePortal.SEND_BROADCAST_BUFF : //L<-P: [SEND_BROADCAST_BUFF](int: jid)(int: option)(int: protocole)(int: length)[content]
							jid = from_server.readInt();
							option = from_server.readInt();
							protocole = from_server.readInt();
							length =  from_server.readInt();
							mspot = hub.getSpotProxy(this, jid);
							if(length==1) {
								byte content = from_server.readByte();
								if(mspot!=null) hub.ibroadcastByte(mspot, protocole, content, option);
							} else { //c'est un buffer
								byte[] buff = new byte[length];
								int off=0;
								while(length>0) {
									int s = from_server.read(buff, off, length);
									off+=s;
									length-=s;
								}
								if(mspot!=null) hub.ibroadcastBuff(mspot, protocole, buff, 0, length, option);								
							}
							break;
						case ProtocolePortal.SEND_BROADCAST : //L<-P: [SEND_BROADCAST](int: jid)(int: option)(int: protocole)(int: length)(DataReader: mess)
							jid = from_server.readInt();
							option = from_server.readInt();
							protocole = from_server.readInt();
							length =  from_server.readInt();
							mess= DataBytes.reader(from_server, length);
							mspot = hub.getSpotProxy(this, jid);
							if(mspot!=null) hub.ibroadcast(mspot, protocole, mess, option);	
							break;		
						case ProtocolePortal.SEND_UPDATE_SMD ://[SEND_UPDATE_SMD](int: jid)(int: version_smd)(int: length)(DataReader: raw_smd)
							jid = from_server.readInt();
							int version = from_server.readInt();
							length =  from_server.readInt();
							mess= DataBytes.reader(from_server, length);
							mspot = hub.getSpotProxy(this, jid);
							if(mspot!=null) mspot.setSMD(version, mess);
							break;
						case ProtocolePortal.SEND_ID_CONTEXT : //[SEND_ID_CONTEXT](int: jid)(string: ch_context)(string: id_contex)
							jid = from_server.readInt();
							String ch_context = from_server.readUTF();
							String id_context = from_server.readUTF();
							mspot = hub.getSpotProxy(this, jid);
							if(mspot!=null) mspot.internalFixIdContext(ch_context, id_context);						
							break;
						case ProtocolePortal.SEND_MESS : //L<-P:  [SEND_MESS](int: jid)(string: ip_target)(int: ctarget_target)[need_ack](int: protocole)(int: length)(DataReader: mess)
							jid = from_server.readInt();
							ip = from_server.readUTF();
							ctarget = from_server.readInt();
							need_ack = from_server.readByte()!=0;
							protocole = from_server.readInt();
							length =  from_server.readInt();
							mess= DataBytes.reader(from_server, length);
							mspot = hub.getSpotProxy(this, jid);
//System.out.println("--SEND_MESS-receive "+ip+":"+ctarget+" channel = "+protocole+" jid = "+jid);									
							if(mspot!=null) {
								hub.internal_send(mspot, ip, ctarget, protocole, mess, need_ack);	
							}
							break;		
						case ProtocolePortal.SEND_MESS_BUFF : //L<-P:  [SEND_MESS_BUFF](int: jid)(string: ip_target)(int: ctarget)[need_ack](int: protocole)(int: length){...[content] ...}
							jid = from_server.readInt();
							ip = from_server.readUTF();
							ctarget = from_server.readInt();
							need_ack = from_server.readByte()!=0;
							protocole = from_server.readInt();
							length =  from_server.readInt();
							mspot = hub.getSpotProxy(this, jid);
							if(length==1) {
								byte content = from_server.readByte();
								if(mspot!=null) hub.internal_sendByte(mspot, ip, ctarget, protocole, content, need_ack);	
							} else { //c'est un buffer
								byte[] buff = new byte[length];
								int off=0;
								while(length>0) {
									int s = from_server.read(buff, off, length);
									off+=s;
									length-=s;
								}
								if(mspot!=null) hub.internal_sendBuff(mspot,  ip, ctarget, protocole, buff, need_ack);								
							}
							break;												
						}
					} catch(Exception e) {
						if(last_errormess==null) last_errormess = "Read error with portal ["+portal_id+"] ("+e.getMessage()+")";
						break; //on sort de la boucle d'écoute
					}					
				}			
				//Ici on est déconnecté du portail -> libération des ressources
				hub.removeSpotObserver(this);
				close();
				hub.unlinked(this, last_errormess); last_errormess=null;
			}//si on est pas arrivé à se linker, ou en cas d'erreur on fera une nouvelle tentative dans 10 secondes
			try{ Thread.sleep(10000); } 
			catch(Exception e){ e.printStackTrace(); }
		}
	}

	private void close() {
		try { if(to_server!=null) { to_server.close(); to_server=null; } } catch(Exception e){}
		try { if(from_server!=null) { from_server.close(); from_server=null; }} catch(Exception e){}
		try { if(socket_client!=null) { socket_client.close(); socket_client=null; }} catch(Exception e){}
	}
	
	private boolean link(){
		//création de la socket TCP
		socket_client=null;
		try { socket_client=new Socket(ip_portal, port_portal); } 
		catch(Exception e) {
			hub.connectionError(this, "Enable connect to portal ["+portal_id+"] ("+e.getMessage()+")");
			return(false);
		}	

		//ouverture des flux de communication
		from_server = null; to_server=null;
		try {
			from_server=new DataInputStream(socket_client.getInputStream());
			to_server=new DataOutputStream(socket_client.getOutputStream());
		}catch(Exception e) {
			hub.connectionError(this, "Enable to open stream with portal ["+portal_id+"] ("+e.getMessage()+")");
			close();
			return(false);
		}

		//Protocole de connexion
		try {
			to_server.writeByte(ProtocolePortal.CLIENT_PORTAL);
			int cmd =  from_server.readByte();
			if(cmd!=ProtocolePortal.SERVER_PORTAL) {
				hub.connectionError(this, "Bad protocole in negociation with portal ["+portal_id+"]");
				close();
				return(false);
			}

			to_server.writeByte(ProtocolePortal.LINK);
			to_server.writeUTF(link_key);
			to_server.writeUTF(join_key);
			to_server.writeUTF(hub.getHubName());
			cmd =  from_server.readByte();
			switch(cmd) {
			case ProtocolePortal.BAD_LINK_KEY :
				hub.connectionError(this,"Bad key to link portal ["+portal_id+"]");
				break;
			case ProtocolePortal.ALLREADY_LINKED:
				hub.connectionError(this,"Portal ["+portal_id+"] allready linked");
				break;
			case ProtocolePortal.OK:				
				return(true); //link ok
			default:
				hub.connectionError(this, "Bad response when connect to portal ["+portal_id+"]");
			}
		}catch(Exception e) {
			hub.connectionError(this, "Error read/write with portal ["+portal_id+"] ("+e.getMessage()+")");
		}	
		close();
		return(false);
	}

	//[interface SpotObserver --> va propager l'info vers le portail distant
	public void up(Spot spot) {
//System.out.println("**up**");		
		if(socket_client==null) return;
		sendlock.lock(); try {
			//[NOTIFY_UP](string: ip)(int: port)(int: ctarget)(int: jid) //jid==0 si ce n'est pas un spot joiner sur ce portail
//System.out.println("*send newSpotUp : "+spot.getIp()+":"+spot.getPort()+" ctarget = "+spot.getIdLocal());			
			to_server.writeByte(ProtocolePortal.NOTIFY_UP);
			to_server.writeUTF(spot.getIp());
			to_server.writeInt(spot.getPort());
			to_server.writeInt(spot.getIdLocal());
			if(spot instanceof ManagedSpotProxy) { //c'est SpotProxy
				ManagedSpotProxy spot_proxy = (ManagedSpotProxy) spot;
				if(spot_proxy.isLinkedTo(this)) to_server.writeInt(spot_proxy.getJid());
				else to_server.writeInt(0);
			} else to_server.writeInt(0);
		} catch(Exception e){
			last_errormess="Write error with portal ["+portal_id+"] ("+e.getMessage()+")";
			close(); //va conduire à mettre fin à la boucle d'écoute (B)		
		} finally {
			sendlock.unlock();
		}
	}
	public void down(Spot spot) { //NB: jamais invoqué sur le spotLocal
		if(socket_client==null) return;
		sendlock.lock(); try {
			//[NOTIFY_DOWN](string: idTarget)
			to_server.writeByte(ProtocolePortal.NOTIFY_DOWN);
			to_server.writeUTF(spot.getIdTarget());
		} catch(Exception e){
			last_errormess="Write error with portal ["+portal_id+"] ("+e.getMessage()+")";
			close(); //va conduire à mettre fin à la boucle d'écoute (B)		
		} finally {
			sendlock.unlock();
		}
	}
	public void idContext(Spot spot, String ch_context, String id_context) {
		if(socket_client==null) return;
		sendlock.lock(); try {
			//[NOTIFY_IDCONTEXT](string: idTarget)(st: ch_context)(st: id_context)
			to_server.writeByte(ProtocolePortal.NOTIFY_IDCONTEXT);
			to_server.writeUTF(spot.getIdTarget());
			to_server.writeUTF(ch_context);
			to_server.writeUTF(id_context);
		} catch(Exception e){
			last_errormess="Write error with portal ["+portal_id+"] ("+e.getMessage()+")";
			close(); //va conduire à mettre fin à la boucle d'écoute (B)		
		} finally {
			sendlock.unlock();
		}
	}
	public void updateSMD(Spot spot, boolean first_definition) {
		if(socket_client==null) return;
		sendlock.lock(); try {
			//[NOTIFY_UPDATE_SMD](string: idtarget)(int: version)(int: length)(DataReader: smd) , smd = (i: nb_value)|...pour chaque valeur: (i: code)[value_size]{...value...}...|
			to_server.writeByte(ProtocolePortal.NOTIFY_UPDATE_SMD);
			to_server.writeUTF(spot.getIdTarget());
			int version =hub.getInternalVersionSMD(spot);
			to_server.writeInt(version);
			DataReader smd = hub.getRawSMD(spot);
			int length = (smd!=null ? smd.avaible() : 0);
			to_server.writeInt(length);
			if(length>0) smd.writeToStream(to_server, length);
			smd.dispose();
		} catch(Exception e){
			last_errormess="Write error with portal ["+portal_id+"] ("+e.getMessage()+")";
			close(); //va conduire à mettre fin à la boucle d'écoute (B)		
		} finally {
			sendlock.unlock();
		}
	}
	//]
	
	public void broadcast(Spot spot_sender, int protocole, DataReader mess, boolean serve_sender) {
		if(socket_client==null) return;
		sendlock.lock(); try {
			//[DELIVER_BROADCAST](string: idtarget_sender)[serve_sender](int: protocole)(int: length)(DataReader: mess)
			to_server.writeByte(ProtocolePortal.DELIVER_BROADCAST);
			to_server.writeUTF(spot_sender.getIdTarget());
			to_server.writeByte(serve_sender ? 1 : 0);
			to_server.writeInt(protocole);
			int length = (mess!=null ? mess.avaible() : 0);
			to_server.writeInt(length);
			if(length>0) mess.writeToStream(to_server, length);
			mess.dispose();
		} catch(Exception e){
			last_errormess="Write error with portal ["+portal_id+"] ("+e.getMessage()+")";
			close(); //va conduire à mettre fin à la boucle d'écoute (B)		
		} finally {
			sendlock.unlock();
		}
	}
	
	public void queryTag(int jid, Spot requester, String name) { //on doit remonter la demande jusqu'au portail, pour qu'il la relègue au joiner distant
		if(socket_client==null) return;
		sendlock.lock(); try {
			//[QUERY_TAG](string: idtarget_requester)(int: jid)(string: tag_name)
			to_server.writeByte(ProtocolePortal.QUERY_TAG);
			to_server.writeUTF(requester.getIdTarget());
			to_server.writeInt(jid);
			to_server.writeUTF(name);
		} catch(Exception e){
			last_errormess="Write error with portal ["+portal_id+"] ("+e.getMessage()+")";
			close(); //va conduire à mettre fin à la boucle d'écoute (B)		
		} finally {
			sendlock.unlock();
		}		
	}
	
	public void receiveAppMessage(int jid, Spot sender, int protocole, DataReader mess) {
//System.out.println("-receiveAppMessage -> jid = "+jid);		
		if(socket_client==null) return;
		sendlock.lock(); try {
			//[RECEIVE_APP_MESSAGE](string: idtarget_sender)(int: jid)(int: protocole)(int: length)(DataReader: mess)
			to_server.writeByte(ProtocolePortal.RECEIVE_APP_MESSAGE);
			to_server.writeUTF(sender.getIdTarget());
			to_server.writeInt(jid);
			to_server.writeInt(protocole);
			int length = (mess!=null ? mess.avaible() : 0);
			to_server.writeInt(length);
			if(length>0) mess.writeToStream(to_server, length);
			mess.dispose();
		} catch(Exception e){
			last_errormess="Write error with portal ["+portal_id+"] ("+e.getMessage()+")";
			close(); //va conduire à mettre fin à la boucle d'écoute (B)		
		} finally {
			sendlock.unlock();
		}					
	}
}