/**
 * Classe d'implémentation d'un portail de jonction.
 */
package skuad.hubudp.implementation.portal;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.net.*;
import skuad.util.thread.*;
import skuad.hubudp.*;
import skuad.util.databytes.DataBytes;
import skuad.util.databytes.DataReader;
import skuad.hubudp.implementation.*;
import java.util.*;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

public class Portal implements Runnable {
	
	public final static byte ERROR=0;
	public final static byte IN_RUN=1;
	
	int port;
	String link_key, join_key;
	byte status;
	ServerSocket socket_server;
	HubManager hubManager;
	ProcessClient linker;
	Vector<ProcessClient> joiners;
	Vector<SpotDescriptor> spots; //liste des spots up
	static int next_jid=1;
	String hubname;
	
	public Portal(HubManager hm, int vport, String vlink_key) {
		port = vport;
		link_key = (vlink_key == null ? "" : vlink_key);
		join_key = null;
		linker = null;
		joiners = new Vector<ProcessClient>();
		spots = new Vector<SpotDescriptor>();
		hubManager = hm;
		hubname =null;
		
		//création de la socket TCP
		socket_server=null;
		try { socket_server=new ServerSocket(port); } 
		catch(Exception e) {
			hubManager.fireKernelLog(Hub.LOG_ERROR, "Enable to open port ["+port+"] to launch portal ("+e.getMessage()+")");
			status = ERROR;
			return;
		}
		
		status = IN_RUN;
		ThreadStack.perform(this); //lancement du processus d'écoute
	}
	
	public boolean inRun() { return(status == IN_RUN); }
	
	protected int newJid(){
		if(next_jid==0) next_jid=1;
		while(getJoiner(next_jid)!=null) { next_jid++; if(next_jid==0) next_jid=1; }
		return(next_jid++);
	}
	
	protected ProcessClient getJoiner(int jid) {
		synchronized(joiners) { //on doit initialiser tous les joiners en attente ==> PORTAL_ACTIVATED
			for(ProcessClient pc : joiners){
				if(pc.jid==jid) return(pc);
			}
		}
		return(null);
	}
	
	//boucle d'accueil des clients TCP
	public void run() {
		
		Socket socket_client=null;
		while(true) {
			try { socket_client=socket_server.accept();}
			catch(Exception e) {
				hubManager.fireKernelLog(Hub.LOG_ERROR, "Error on accept connexion on portal port ["+port+"] ("+e.getMessage()+")");
				continue;
			}
//System.out.println("-client accept : "+socket_client.getPort());			
			ProcessClient pclient = new ProcessClient(this, socket_client);
			ThreadStack.perform(pclient); //lancement du processus de gestion de ce client
		}
	}
	
	//[Pour la classe ProcessClient
	protected boolean addJoiner(ProcessClient joiner) { //enregistre le client dans la liste, retourne true ssi le portail est activé
		synchronized(joiners) {  		
			if(linker!=null) {
				try{ joiner.to_client.writeByte(ProtocolePortal.PORTAL_ACTIVATED); }catch(Exception e) { return(false); } //nb: verron d'écriture non nécessaire car c'est un tout nouveau joiner
				joiners.add(joiner);				

				linker.sendlock.lock();
				try { //on enregistre ce joiner sur le linker
					linker.to_client.writeByte(ProtocolePortal.ADD_JOINER); 
					linker.to_client.writeInt(joiner.jid);
				}catch(Exception e) {
					linker.close();  //va conduire à l'appel unlink(...)
				} finally { linker.sendlock.unlock(); }		
			} else {	
				try{ joiner.to_client.writeByte(ProtocolePortal.PORTAL_NO_ACTIVATED); }catch(Exception e) { return(false); } //nb: verron d'écriture non nécessaire car c'est un tout nouveau joiner
				joiners.add(joiner);
			}
			return(true);
		}
	}
	protected void removeJoiner(ProcessClient joiner) {
		synchronized(joiners) {
			joiner.close();
			if(joiners.remove(joiner) && linker!=null){ 
				linker.sendlock.lock();
				try { //on desenregistre ce joiner sur le linker
					linker.to_client.writeByte(ProtocolePortal.REMOVE_JOINER); 
					linker.to_client.writeInt(joiner.jid);
				}catch(Exception e) {
					linker.close();  //va conduire à l'appel unlink(...)
				} finally { linker.sendlock.unlock(); }
			}
		}
	}
	protected synchronized boolean link(ProcessClient pc, String kp, String jk, String hn){ //nb: kp!=null
		if(linker==null){	
			if(!kp.equals(link_key)) try { pc.to_client.writeByte(ProtocolePortal.BAD_LINK_KEY); return(false); }catch(Exception e) { return(false); }
			try { pc.to_client.writeByte(ProtocolePortal.OK); }catch(Exception e) { return(false); }
			linker = pc;
			join_key = jk;
			hubname = hn;
			//Lancement du processus d'écoute du linker
			ThreadStack.perform( () -> pc.listenLinker() );
			synchronized(joiners) { //on doit initialiser tous les joiners en attente ==> PORTAL_ACTIVATED
				for(int i=0; i<joiners.size(); i++){
					ProcessClient joiner = joiners.get(i);
					joiner.sendlock.lock();
					try{ 
						if(joiner.join_key.equals(join_key)) {
							joiner.actif_join = true;
							joiner.to_client.writeByte(ProtocolePortal.PORTAL_ACTIVATED); 
						} else {
							joiner.actif_join = false;
							joiner.to_client.writeByte(ProtocolePortal.BAD_JOIN_KEY); 
							continue; //on passe au joiner suivant
						}
					} catch(Exception e) {
						joiners.removeElementAt(i); i--;
						joiner.close();
						continue; 
					} finally { joiner.sendlock.unlock(); }
					
					linker.sendlock.lock();
					try { //on enregistre ce joiner sur le linker
						linker.to_client.writeByte(ProtocolePortal.ADD_JOINER); 
						linker.to_client.writeInt(joiner.jid);
					}catch(Exception e) {
						return(false);  //va conduire à l'appel linker.close() pui unlink(...)
					} finally { linker.sendlock.unlock(); }		
				}
			}
			return(true);
		} else {
			try { pc.to_client.writeByte(ProtocolePortal.ALLREADY_LINKED); }catch(Exception e) {}
			return(false);
		}
	}
	protected synchronized void unlink(ProcessClient pc){
		if(linker==pc){
			linker=null;
			synchronized(spots) { spots.removeAllElements(); }
			synchronized(joiners) {
				for(int i=0; i<joiners.size(); i++){
					ProcessClient joiner = joiners.get(i);
					joiner.linked=null;
					joiner.sendlock.lock();
					try{ joiner.to_client.writeByte(ProtocolePortal.PORTAL_NO_ACTIVATED); }
					catch(Exception e) {
						joiners.removeElementAt(i); i--;
						joiner.close();
						continue; 
					} finally { joiner.sendlock.unlock(); }	
				}
			}
			hubname =null;
		}
	}
	protected void notifyAllSpot(ProcessClient joiner) { //on notify la présence et les données de tous les spots up à ce joiner (sauf ceux de son localspot !)
		synchronized(spots) { 
			joiner.sendlock.lock();
			try {
				for(SpotDescriptor spotd : spots) { if(spotd!=joiner.linked) {
					//UP
					//[NOTIFY_UP](string: ip)(int : port)(int: ctarget)
					joiner.to_client.writeByte(ProtocolePortal.NOTIFY_UP); 
					joiner.to_client.writeUTF(spotd.ip);
					joiner.to_client.writeInt(spotd.port);
					joiner.to_client.writeInt(spotd.ctarget);
					
					if(spotd.smd!=null) { //UPDATE_SMD
						joiner.to_client.writeByte(ProtocolePortal.NOTIFY_UPDATE_SMD); 
						joiner.to_client.writeUTF(spotd.id_target);
						joiner.to_client.writeInt(spotd.version_smd);
						int length = spotd.smd.avaible();
						joiner.to_client.writeInt(length);
						spotd.smd.writeToStream(joiner.to_client, length);
						spotd.smd.reset();
					}
					
					//IDCONTEXT
					synchronized(spotd.idContexts){
						for(int i=0; i<spotd.idContexts.size(); i++) {
							String[] entry = spotd.idContexts.get(i);
							joiner.to_client.writeByte(ProtocolePortal.NOTIFY_IDCONTEXT); 
							joiner.to_client.writeUTF(spotd.id_target);
							joiner.to_client.writeUTF(entry[0]);
							joiner.to_client.writeUTF(entry[1]);
						}
					}
				}}
			} catch(Exception e) { removeJoiner(joiner); }
			finally { joiner.sendlock.unlock(); }	
		}
	}
	protected void newSpotUp(String ip, int port, int ctarget, int jid) {
//System.out.println("*newSpotUp : "+ip+":"+port+" ctarget = "+ctarget);
		SpotDescriptor spot_descriptor = new SpotDescriptor(ip, port, ctarget);
		synchronized(spots) { spots.add(spot_descriptor); }
		synchronized(joiners) { //on notify le up à tous les joiners
			for(int i=0; i<joiners.size(); i++){
				ProcessClient joiner = joiners.get(i);
				joiner.sendlock.lock();
				try{ 
					if(joiner.jid == jid) {
						joiner.linked=spot_descriptor;
						joiner.to_client.writeByte(ProtocolePortal.LOCALSPOT_LINKED);
						joiner.to_client.writeInt(ctarget);
						joiner.to_client.writeInt(port);
						joiner.to_client.writeUTF(hubname);
						notifyAllSpot(joiner); //on notify la présence et les données de tous les spots up à ce joiner
					} else if(joiner.linked!=null) {
						joiner.to_client.writeByte(ProtocolePortal.NOTIFY_UP); 
						joiner.to_client.writeUTF(ip);
						joiner.to_client.writeInt(port);
						joiner.to_client.writeInt(ctarget);
					}
				} catch(Exception e) { removeJoiner(joiner); i--; }
				finally { joiner.sendlock.unlock(); }	
			}
		}	
	}
	protected void spotIdContext(String idtarget, String ch_context, String id_context){
		SpotDescriptor spotd = null;
		synchronized(spots) {
			for(int i=0; i<spots.size(); i++) {
				spotd = spots.get(i);
				if(spotd.isIt(idtarget)) {
					spotd.setIdContext(ch_context, id_context);
					break;
				}
			}
			if(spotd==null) return; //cette situation est normalement impossible
		}
		synchronized(joiners) { //on notify tous les joiners
			for(int i=0; i<joiners.size(); i++){
				ProcessClient joiner = joiners.get(i);
				if(joiner.linked!=null && joiner.linked!=spotd) {
					joiner.sendlock.lock();
					try{ 
						joiner.to_client.writeByte(ProtocolePortal.NOTIFY_IDCONTEXT); 
						joiner.to_client.writeUTF(idtarget);
						joiner.to_client.writeUTF(ch_context);
						joiner.to_client.writeUTF(id_context);
					} catch(Exception e) { removeJoiner(joiner); i--; }
					finally { joiner.sendlock.unlock(); }	
				}
			}
		}		
	}
	protected void spotSMD(String idtarget, int version, DataReader smd){
		SpotDescriptor spotd = null;
		synchronized(spots) {
			for(int i=0; i<spots.size(); i++) {
				spotd = spots.get(i);
				if(spotd.isIt(idtarget)) { 
					if(spotd.smd!=null) spotd.smd.dispose();
					spotd.smd = smd;
					spotd.version_smd = version;
					break;
				}
			}
			if(spotd==null) return; //cette situation est normalement impossible
		}
		int length = smd.avaible();
		synchronized(joiners) { //on notify tous les joiners, sauf le porteur de ces SMD
			for(int i=0; i<joiners.size(); i++){
				ProcessClient joiner = joiners.get(i);
				if(joiner.linked!=null && joiner.linked!=spotd) {
					joiner.sendlock.lock();
					try{ 
						joiner.to_client.writeByte(ProtocolePortal.NOTIFY_UPDATE_SMD); 
						joiner.to_client.writeUTF(idtarget);
						joiner.to_client.writeInt(version);
						joiner.to_client.writeInt(length);
						smd.writeToStream(joiner.to_client, length);
						smd.reset();
					} catch(Exception e) { removeJoiner(joiner); i--; }
					finally { joiner.sendlock.unlock(); }
				}
			}
		}		
	}	
	protected void spotDown(String idTarget) {
		SpotDescriptor spotd = null;
		synchronized(spots) {
			for(int i=0; i<spots.size(); i++) {
				spotd = spots.get(i);
				if(spotd.isIt(idTarget)) { 
					spots.removeElementAt(i);
					break;
				}
			}
			if(spotd==null) return; //cette situation est normalement impossible
		}
		synchronized(joiners) { //on notify le down à tous les joiners
			for(int i=0; i<joiners.size(); i++){
				ProcessClient joiner = joiners.get(i);
				if(joiner.linked!=null) {
					joiner.sendlock.lock();
					try{ 
						joiner.to_client.writeByte(ProtocolePortal.NOTIFY_DOWN); 
						joiner.to_client.writeUTF(idTarget);
					} catch(Exception e) { removeJoiner(joiner); i--; }
					finally { joiner.sendlock.unlock(); }
				}
			}
		}	
	}	
	
	protected void deliver_broadcast(String idtarget, int protocole, DataReader mess, boolean serve_sender) {
		synchronized(joiners) {
			int length = mess.avaible();
			for(int i=0; i<joiners.size(); i++){
				ProcessClient joiner = joiners.get(i);
				if(joiner.linked==null || (!serve_sender &&  joiner.linked.isIt(idtarget))) continue;
				joiner.sendlock.lock();
				try{ 
					joiner.to_client.writeByte(ProtocolePortal.DELIVER_BROADCAST); 
					joiner.to_client.writeUTF(idtarget);
					joiner.to_client.writeInt(protocole);
					joiner.to_client.writeInt(length);
					mess.writeToStream(joiner.to_client, length);
					mess.reset();
				} catch(Exception e) { removeJoiner(joiner); i--; }
				finally { joiner.sendlock.unlock(); }	
			}
			mess.dispose();
		}		
	}
	
	protected void queryTag(String idtarget, int jid, String tag_name) {
		synchronized(joiners) {
			for(ProcessClient cur : joiners){
				if(cur.jid == jid) {
					cur.sendlock.lock();
					try{
						cur.to_client.writeByte(ProtocolePortal.QUERY_TAG); 
						cur.to_client.writeUTF(idtarget);
						cur.to_client.writeInt(jid);
						cur.to_client.writeUTF(tag_name);
					} catch(Exception e) { removeJoiner(cur); }
					finally { cur.sendlock.unlock(); }	
					return;
				}
			}
		}
	}
	protected void receiveAppMessage(String idtarget, int jid, int protocole, DataReader mess) {
//System.out.println("-receiveAppMessage -> jid = "+jid);	
		synchronized(joiners) {
			for(ProcessClient cur : joiners){
				if(cur.jid == jid) {
					cur.sendlock.lock();
					try{
						//[RECEIVE_APP_MESSAGE](string: idtarget_sender)(int: jid)(int: protocole)(int: length)(DataReader: mess)
						cur.to_client.writeByte(ProtocolePortal.RECEIVE_APP_MESSAGE);
						cur.to_client.writeUTF(idtarget);
						cur.to_client.writeInt(jid);
						cur.to_client.writeInt(protocole);
						int length = (mess!=null ? mess.avaible() : 0);
						cur.to_client.writeInt(length);
						if(length>0) mess.writeToStream(cur.to_client, length);
						mess.dispose();
					} catch(Exception e) { removeJoiner(cur); }
					finally { cur.sendlock.unlock(); }	
					return;
				}
			}
		}
	}
	protected void broadcast_byte(int jid, int protocole, int content, int option){
		if(linker!=null) {
			linker.sendlock.lock();
			try { //[SEND_BROADCAST_BUFF](int: jid)(int: option)(int: protocole)(int : length)[content]
				linker.to_client.writeByte(ProtocolePortal.SEND_BROADCAST_BUFF); 
				linker.to_client.writeInt(jid);				
				linker.to_client.writeInt(option);
				linker.to_client.writeInt(protocole);
				linker.to_client.writeInt(1);
				linker.to_client.writeByte(content);
			}catch(Exception e) {
				linker.close();  //va conduire à l'appel unlink(...)
			} finally { linker.sendlock.unlock(); }		
		}
	}
	protected void broadcast_buff(int jid, int protocole, byte[] content, int option){
		if(linker!=null) {
			linker.sendlock.lock();
			try { //[SEND_BROADCAST_BUFF](int: jid)(int: option)(int: protocole)(int : length)[content]
				linker.to_client.writeByte(ProtocolePortal.SEND_BROADCAST_BUFF); 
				linker.to_client.writeInt(jid);				
				linker.to_client.writeInt(option);
				linker.to_client.writeInt(protocole);
				linker.to_client.writeInt(content.length);
				linker.to_client.write(content);
			}catch(Exception e) {
				linker.close();  //va conduire à l'appel unlink(...)
			} finally { linker.sendlock.unlock(); }		
		}
	}	
	protected void broadcast(int jid, int protocole, DataReader content, int option){
		if(linker!=null) {
			linker.sendlock.lock();
			try { //[SEND_BROADCAST](int: jid)(int: option)(int: protocole)(int: length)(DataReader: mess)
				linker.to_client.writeByte(ProtocolePortal.SEND_BROADCAST); 
				linker.to_client.writeInt(jid);				
				linker.to_client.writeInt(option);
				linker.to_client.writeInt(protocole);
				int length = (content!=null ? content.avaible() : 0);
				linker.to_client.writeInt(length);
				if(length>0) content.writeToStream(linker.to_client, length);
			}catch(Exception e) {
				linker.close();  //va conduire à l'appel unlink(...)
			} finally { linker.sendlock.unlock(); }		
		}
	}	
	protected void updateSMD(int jid, int version, int length, DataReader raw_smd){
		if(linker!=null) {
			linker.sendlock.lock();
			try { //[SEND_UPDATE_SMD](int: jid)(int: version_smd)(int: length)(DataReader: raw_smd)
				linker.to_client.writeByte(ProtocolePortal.SEND_UPDATE_SMD); 
				linker.to_client.writeInt(jid);
				linker.to_client.writeInt(version);	
				linker.to_client.writeInt(length);
				if(length>0) raw_smd.writeToStream(linker.to_client, length);
			}catch(Exception e) {
				linker.close();  //va conduire à l'appel unlink(...)
			} finally { linker.sendlock.unlock(); }		
		}
	}
	protected void setIdContext(int jid, String ch_context, String id_contex) {
		if(linker!=null) {
			linker.sendlock.lock();
			try { //[SEND_ID_CONTEXT](int: jid)(string: ch_context)(string: id_contex)
				linker.to_client.writeByte(ProtocolePortal.SEND_ID_CONTEXT); 
				linker.to_client.writeInt(jid);
				linker.to_client.writeUTF(ch_context);	
				linker.to_client.writeUTF(id_contex);
			}catch(Exception e) {
				linker.close();  //va conduire à l'appel unlink(...)
			} finally { linker.sendlock.unlock(); }		
		}		
	}
	protected void send_byte(int jid, String ip, int ctarget, int protocole, int content, boolean need_ack){
//System.out.println("-----sendbyte : proto ="+protocole+" , content = "+content);		
		if(linker!=null) {
			linker.sendlock.lock();
			try { //[SEND_MESS_BUFF](int: jid)(string: ip_target)(int: ctarget_target)[need_ack](int: protocole)(int: length){...[content] ...}
				linker.to_client.writeByte(ProtocolePortal.SEND_MESS_BUFF); 
				linker.to_client.writeInt(jid);
				linker.to_client.writeUTF(ip);
				linker.to_client.writeInt(ctarget);
				linker.to_client.writeByte(need_ack ? 1 : 0);
				linker.to_client.writeInt(protocole);
				linker.to_client.writeInt(1);
				linker.to_client.writeByte(content);	
			}catch(Exception e) {
				linker.close();  //va conduire à l'appel unlink(...)
			} finally { linker.sendlock.unlock(); }		
		}			
	}
	protected void send_buff(int jid, String ip, int ctarget, int protocole, byte[] content, boolean need_ack){
		if(linker!=null) {
			linker.sendlock.lock();
			try { //[SEND_MESS_BUFF](int: jid)(string: ip_target)(int: ctarget_target)[need_ack](int: protocole)(int: length){...[content] ...}
				linker.to_client.writeByte(ProtocolePortal.SEND_MESS_BUFF); 
				linker.to_client.writeInt(jid);
				linker.to_client.writeUTF(ip);
				linker.to_client.writeInt(ctarget);
				linker.to_client.writeByte(need_ack ? 1 : 0);
				linker.to_client.writeInt(protocole);
				linker.to_client.writeInt(content.length);
				linker.to_client.write(content);	
			}catch(Exception e) {
				linker.close();  //va conduire à l'appel unlink(...)
			} finally { linker.sendlock.unlock(); }		
		}			
	}	
	protected void send(int jid, String ip, int ctarget, int protocole, int length, DataReader mess, boolean need_ack) {
//System.out.println("--SEND_MESS-(send to linker) "+ip+":"+ctarget+" channel = "+protocole+" jid = "+jid);			
		if(linker!=null) {
			linker.sendlock.lock();
			try { //[SEND_MESS](int: jid)(string: ip_target)(int: ctarget_target)[need_ack](int: protocole)(int: length)(DataReader: mess)
				linker.to_client.writeByte(ProtocolePortal.SEND_MESS); 
				linker.to_client.writeInt(jid);
				linker.to_client.writeUTF(ip);
				linker.to_client.writeInt(ctarget);
				linker.to_client.writeByte(need_ack ? 1 : 0);
				linker.to_client.writeInt(protocole);
				linker.to_client.writeInt(length);
				mess.writeToStream(linker.to_client, length);
			}catch(Exception e) {
				linker.close();  //va conduire à l'appel unlink(...)
			} finally { linker.sendlock.unlock(); }		
		}		
	}
	//] 
}

class ProcessClient implements Runnable {
	Socket socket_client;
	DataInputStream from_client; 
	DataOutputStream to_client;
	Portal portal;
	int jid; String join_key;
	Lock sendlock;
	boolean actif_join; 
	SpotDescriptor linked;
		
	public ProcessClient(Portal p, Socket vsocket_client) {
		portal = p;
		socket_client = vsocket_client;
		jid = 0;
		actif_join = false;
		linked = null;
		join_key = null;
		sendlock=new ReentrantLock();
	}
	
	public void run() {
		//ouverture des flux de communication
		from_client = null; 
		to_client = null;
		try {
			from_client=new DataInputStream(socket_client.getInputStream());
			to_client=new DataOutputStream(socket_client.getOutputStream());
		}catch(Exception e) { //impossible d'ouvrire les flux de communication avec le client
			 close();
			return;
		}
//System.out.println("-ProcessClient : "+socket_client.getPort()+" : begin");		
		//Procole de connexion
		try {
			byte cmd = from_client.readByte();
			if(cmd!=ProtocolePortal.CLIENT_PORTAL) {
				to_client.writeByte(ProtocolePortal.BAD_PROTOCOL);
				close();
				return;
			}
			to_client.writeByte(ProtocolePortal.SERVER_PORTAL);
//System.out.println("-ProcessClient : "+socket_client.getPort()+" : CLIENT_PORTAL ok");			
			
			//identification du mode link ou join
			cmd = from_client.readByte();
			switch(cmd){
			case ProtocolePortal.JOIN:
//System.out.println("-ProcessClient : "+socket_client.getPort()+" : JOIN");				
				jid = portal.newJid();
				join_key = from_client.readUTF();
				to_client.writeByte(ProtocolePortal.OK);
				if(portal.addJoiner(this)) listenJoiner();
				else close();
				break;
			case ProtocolePortal.LINK:				
				String lk = from_client.readUTF();
				String jk = from_client.readUTF();
				String hn = from_client.readUTF();

//System.out.println("-ProcessClient : "+socket_client.getPort()+" : LINK -> lk = "+lk+" // jk = "+jk+" // hn = "+hn);				
				if(lk!=null){
					if(!portal.link(this, lk, jk, hn)) close(); //fin du dialogue
				} else {
//System.out.println("-ProcessClient : "+socket_client.getPort()+" : ---> LINK BAD_PROTOCOL");					
					to_client.writeByte(ProtocolePortal.BAD_PROTOCOL);
					close(); //fin du dialogue	
				}
				break;
			}
		}catch(Exception e) { //error critique dans le protocole de connexion du client
//e.printStackTrace();		
			close(); //fin du dialogue	
		} 
	}
	
	public void close() {
/*int pclient=0;		
try { pclient = socket_client.getPort(); } catch(Exception e){}		
System.out.println("-ProcessClient : "+pclient+" : ---> close()");
*/		
		try { to_client.close(); } catch(Exception e){}
		try { from_client.close(); } catch(Exception e){}
		try { socket_client.close(); } catch(Exception e){}
	}
	
	protected void listenJoiner() { //Boucle d'écoute du joiner, le portail est activé (A)
		try {
			while(true) {
				byte cmd = from_client.readByte();
				int protocole, option, length;
				String ip; int ctarget;
				boolean need_ack;
				DataReader mess;
				String ch_context, id_contex;
				
				switch(cmd) {
				case ProtocolePortal.BROADCAST_BUFF: //[BROADCAST_BUFF](int: option)(int: protocole)(int : length)[content]
					option =  from_client.readInt();
					protocole =  from_client.readInt();
					length = from_client.readInt();
					if(length==1) {
						int content = from_client.readByte();
						portal.broadcast_byte(jid, protocole, content, option);
					} else { //c'est un buffer
						byte[] buff = new byte[length];
						int off=0;
						while(length>0) {
							int s = from_client.read(buff, off, length);
							off+=s;
							length-=s;
						}
						portal.broadcast_buff(jid, protocole, buff, option);
					}
					break;
				case ProtocolePortal.BROADCAST: //[BROADCAST](int: option)(int: protocole)(int: length)(DataReader: mess)
					option =  from_client.readInt();
					protocole =  from_client.readInt();
					length = from_client.readInt();
					mess= DataBytes.reader(from_client, length);
					portal.broadcast(jid, protocole, mess, option);
					break;
				case ProtocolePortal.UPDATE_SMD: //[UPDATE_SMD](int: version_smd)(int: length)(DataReader: raw_smd)
					int version =  from_client.readInt();
					length = from_client.readInt();
					mess= DataBytes.reader(from_client, length);
					portal.updateSMD(jid, version, length, mess);
					break;	
				case ProtocolePortal.SET_ID_CONTEXT: //[SET_ID_CONTEXT](string: ch_context)(string: id_contex)
					ch_context = from_client.readUTF();
					id_contex = from_client.readUTF();
					portal.setIdContext(jid, ch_context, id_contex);
					break;
				case ProtocolePortal.ALL_ID_CONTEXT: //[ALL_ID_CONTEXT](int: nb){... (string: ch_context)(string: id_context) ... }
					int nb = from_client.readInt();
					for(int i=0; i<nb; i++) {
						ch_context = from_client.readUTF();
						id_contex = from_client.readUTF();
						portal.setIdContext(jid, ch_context, id_contex);						
					}
					break;
				case ProtocolePortal.SEND_BUFF: //[SEND_BUFF](string: ip_target)(int: ctarget)[need_ack](int: protocole)(int: length)[content]
					ip =  from_client.readUTF();
					ctarget =  from_client.readInt();
					need_ack = from_client.readByte() != 0;
					protocole =  from_client.readInt();
					length = from_client.readInt();
					if(length==1) {
						int content = from_client.readByte();
						portal.send_byte(jid, ip, ctarget, protocole, content, need_ack);
					} else { //c'est un buffer
						byte[] buff = new byte[length];
						int off=0;
						while(length>0) {
							int s = from_client.read(buff, off, length);
							off+=s;
							length-=s;
						}
						portal.send_buff(jid, ip, ctarget, protocole, buff, need_ack);
					}
					break;	
				case ProtocolePortal.SEND: //[SEND](string: ip_target)(int: ctarget)[need_ack](int: protocole)(int: length)(DataReader: mess)
//System.out.println("--SEND--");							
					ip =  from_client.readUTF();
					ctarget =  from_client.readInt();
					need_ack = from_client.readByte() != 0;
					protocole =  from_client.readInt();
					length = from_client.readInt();
					mess= DataBytes.reader(from_client, length);
					portal.send(jid, ip, ctarget, protocole, length, mess, need_ack);
					break;					
				}
			}
		} catch(Exception e) { //error critique dans le protocole de connexion du client
			portal.removeJoiner(this);
		} 
	}
	
	protected void listenLinker() { //Boucle d'écoute du linker
		try {
			while(true) {
				byte cmd = from_client.readByte();
				int jid, ctarget, length, protocole;
				DataReader mess;
				String idTarget;
				
				switch(cmd) {				
				case ProtocolePortal.NOTIFY_UP: 	//[NOTIFY_UP](string: ip)(int: port)(int: ctarget)(int: jid) //jid==0 si ce n'est pas un spot joiner sur ce portail
					String ip = from_client.readUTF();
					int port = from_client.readInt();
					ctarget = from_client.readInt();
					jid = from_client.readInt();
					portal.newSpotUp(ip, port, ctarget, jid);
					break;
				case ProtocolePortal.NOTIFY_DOWN: 	//[NOTIFY_DOWN](string: idTarget)
					idTarget = from_client.readUTF();
					portal.spotDown(idTarget);
					break;			
				case ProtocolePortal.NOTIFY_IDCONTEXT: 	//[NOTIFY_IDCONTEXT](string: idTarget)(st: ch_context)(st: id_context)
					idTarget = from_client.readUTF();
					portal.spotIdContext(idTarget, from_client.readUTF(), from_client.readUTF());
					break;			
				case ProtocolePortal.NOTIFY_UPDATE_SMD: 	//[NOTIFY_UPDATE_SMD](string: idtarget)(int: version)(int: length)(DataReader: smd)
					idTarget = from_client.readUTF();
					int version = from_client.readInt();
					length = from_client.readInt();
					DataReader data= DataBytes.reader(from_client, length);
					portal.spotSMD(idTarget, version, data);
					break;
				case ProtocolePortal.DELIVER_BROADCAST: //[DELIVER_BROADCAST](string: idtarget_sender)[serve_sender](int: protocole)(int: length)(DataReader: mess)
					idTarget = from_client.readUTF();
					boolean serve_sender = from_client.readByte() != 0;
					protocole =  from_client.readInt();
					length = from_client.readInt();
					mess= DataBytes.reader(from_client, length);
					portal.deliver_broadcast(idTarget, protocole, mess, serve_sender);					
					break;
				case ProtocolePortal.QUERY_TAG: //[QUERY_TAG](string: idtarget_requester)(int: jid)(string: tag_name)
					idTarget = from_client.readUTF();
					jid =  from_client.readInt();
					String tag_name = from_client.readUTF();
					portal.queryTag(idTarget, jid, tag_name);		
					break;		
				case ProtocolePortal.RECEIVE_APP_MESSAGE: //[RECEIVE_APP_MESSAGE](string: idtarget_sender)(int: jid)(int: protocole)(int: length)(DataReader: mess)
					idTarget = from_client.readUTF();
					jid =  from_client.readInt();
					protocole = from_client.readInt();
					length = from_client.readInt();
					mess= DataBytes.reader(from_client, length);
					portal.receiveAppMessage(idTarget, jid, protocole, mess);		
					break;
				}
			}
		} catch(Exception e) { //error critique dans le protocole de connexion du client
//e.printStackTrace();			
			close(); //fin du dialogue
			portal.unlink(this);
		} 
	}	
}

class SpotDescriptor {
	String ip;
	int port;
	int ctarget;
	DataReader smd;
	int version_smd;
	String id_target;
	Vector<String[]> idContexts;
	
	public SpotDescriptor(String vip, int vport, int vct) {
		ip = vip;
		port = vport;
		ctarget = vct;
		id_target = ip+'#'+ctarget;
		idContexts = new Vector<String[]>();
		smd=null; version_smd=0;
	}
	public void setIdContext(String ch_context, String id_context) {
		synchronized(idContexts){
			for(int i=0; i<idContexts.size(); i++) {
				if(idContexts.get(i)[0].equals(ch_context)) {
					idContexts.get(i)[1] = id_context;
					return;
				}
			}
			String[] entry = new String[2];
			entry[0]=ch_context; entry[1]=id_context;
			idContexts.add(entry);
		}
	}

	public boolean isIt(String idTarget) { 
		return(id_target.equals(idTarget));
	}
}
