/**
 *  Procole Portal :
 * 	
 * - Procole de connexion 
 * 		client [ProtocolePortal.CLIENT_PORTAL] -> serveur	
 * 					<- 	serveur	: 	- [ProtocolePortal.SERVER_PORTAL] si tout est ok
 * 									- [ProtocolePortal.BAD_PROTOCOL] en cas d'erreur
 * 
 * 			- si le client veut établir un link				
 * 		client [ProcolePortal.LINK](utf : link_key)(utf : join_key)(utf: hubname) -> serveur
 * 				<- serveur :	- [ProtocolePortal.BAD_LINK_KEY]	si la link_key n'est pas la bonne
 * 								- [ProtocolePortal.ALLREADY_LINKED] si le portail est déjà associé à un hub
 * 								- [ProtocolePortal.BAD_PROTOCOL] en cas d'erreur
 * 								- [ProtocolePortal.OK] si le link est établie, on passe alors dans le protocole link
 * 
 * 			- si le client veut établir un join				
 * 		client [ProcolePortal.JOIN](utf : join_key) -> serveur		
 * 				<- serveur :	- [ProtocolePortal.OK] si la connexion est acceptée, on passe alors dans le protocole join
 *  
 */
package skuad.hubudp.implementation.portal;

public class ProtocolePortal {
	
	//[Constantes générales
	//]

	//[-----------> Procole de connexion :
	public final static byte CLIENT_PORTAL  = 80;	
	public final static byte SERVER_PORTAL 	= 90;	

	public final static byte BAD_PROTOCOL 	= 1;
	public final static byte LINK  			= 2;	
		//à partir de l'ensemble des spot du hub linker sont observés et les events updateSMD et idContext de tous ces spots seront remontés vers le portail qui les transmettra vers tous les joiners
		//... et ce jusqu'à la ruprture de la liaison avec le portail
	public final static byte JOIN  			= 3;	
		
	public final static byte BAD_LINK_KEY	= 4;
	public final static byte ALLREADY_LINKED = 5;
	public final static byte OK = 6;
	
	//[-----------> Procole link
	public final static byte ADD_JOINER = 10;			//L<-P: [ADD_JOINER](int: jid) , enregistre un joiner distant sur le linker
	public final static byte REMOVE_JOINER = 11;		//L<-P: [REMOVE_JOINER](int: jid) , désenregistre un joiner distant sur le linker
		
	public final static byte NOTIFY_UP = 13;			//L->P : [NOTIFY_UP](string: ip)(int : port)(int: ctarget)(int: jid) 	//jid==0 si ce n'est pas un spot joiner sur ce portail
														//P->J : [NOTIFY_UP](string: ip)(int : port)(int: ctarget)	

	public final static byte NOTIFY_DOWN = 14;			//L->P // P-> J : [NOTIFY_DOWN](string: idTarget)
	public final static byte NOTIFY_IDCONTEXT = 15;		//L->P // P-> J : [NOTIFY_IDCONTEXT](string: idTarget)(st: ch_context)(st: id_context)
	public final static byte NOTIFY_UPDATE_SMD = 16;	//L->P // P-> J : [NOTIFY_UPDATE_SMD](string: idtarget)(int: version)(int: length)(DataReader: smd)
	
	public final static byte DELIVER_BROADCAST = 17;	//L->P // P-> J : [DELIVER_BROADCAST](string: idtarget)[serve_sender](int: protocole)(int: length)(DataReader: mess)
	public final static byte QUERY_TAG = 18;			//L->P // P-> J : [QUERY_TAG](string: idtarget_requester)(int: jid)(string: tag_name)
	public final static byte RECEIVE_APP_MESSAGE = 19;	//L->P // P-> J : //[RECEIVE_APP_MESSAGE](string: idtarget_sender)(int: jid)(int: protocole)(int: length)(DataReader: mess)
	
	public final static byte SEND_BROADCAST_BUFF = 20; 	//L<-P : [SEND_BROADCAST_BUFF](int: jid)(int: option)(int: protocole)(int : length){...[content]...}
	public final static byte SEND_BROADCAST = 21; 		//L<-P : [SEND_BROADCAST](int: jid)(int: option)(int: protocole)(int: length)(DataReader: mess)
	public final static byte SEND_UPDATE_SMD = 22; 		//L<-P : [SEND_UPDATE_SMD](int: jid)(int: version_smd)(int: length)(DataReader: raw_smd)
	public final static byte SEND_ID_CONTEXT = 23; 		//L<-P : [SEND_ID_CONTEXT](int: jid)(string: ch_context)(string: id_contex)
	
	public final static byte SEND_MESS_BUFF = 24; 		//L<-P : [SEND_MESS_BUFF](int: jid)(string: ip_target)(int: ctarget)[need_ack](int: protocole)(int: length){...[content] ...}
	public final static byte SEND_MESS = 25; 			//L<-P : [SEND_MESS](int: jid)(string: ip_target)(int: ctarget)[need_ack](int: protocole)(int: length)(DataReader: mess)
	//]
	
	//[-----------> Procole join
	public final static byte PORTAL_NO_ACTIVATED = 30; 	//J<-P: [PORTAL_NO_ACTIVATED] , le portail n'est pas activé ou il vient de se désactiver 
	public final static byte PORTAL_ACTIVATED = 31;		//J<-P: [PORTAL_ACTIVATED], si le portail est activé, ou il vient de s'activer, et la clé join_key est bonne
	public final static byte BAD_JOIN_KEY = 32;			//J<-P: [PORTAL_ACTIVATED], si le portail est activé, ou il vient de s'activer, mais la clé join_key n'est pas la bonne
	
	public final static byte LOCALSPOT_LINKED = 33;		//J<-P: [LOCALSPOT_LINKED](int: ctarget)(int: port)(s: hubname), signal au joiner que son spot local est connecté et lui indique son ctarget
		//à partir de la le spotlocal du joiner est observé et les events updateSMD et idContext sont remontés vers le portail qui les transmettra vers le linker
		//... et ce jusqu'au message : PORTAL_NO_ACTIVATED ou la ruprture de la liaison avec le portail
	
	public final static byte BROADCAST_BUFF = 34; 	//J->P : [BROADCAST_BUFF](int: option)(int: protocole)(int: length)[content]
	public final static byte BROADCAST = 35; 		//J->P : [BROADCAST](int: option)(int: protocole)(int: length)(DataReader: mess)
	public final static byte ALL_ID_CONTEXT = 36; 	//J->P : [ALL_ID_CONTEXT](int: nb){... (string: ch_context)(string: id_context) }
	public final static byte SET_ID_CONTEXT = 37; 	//J->P : [SET_ID_CONTEXT](string: ch_context)(string: id_contex)
	public final static byte UPDATE_SMD = 38; 		//J->P : [UPDATE_SMD](int: version_smd)(int: length)(DataReader: raw_smd)
	public final static byte SEND_BUFF = 39; 		//J->P : [SEND_BUFF](string: ip_target)(int: ctarget)[need_ack](int: protocole)(int: length){...[content] ...}
	public final static byte SEND = 40; 			//J->P : [SEND](string: ip_target)(int: ctarget)[need_ack](int: protocole)(int: length)(DataReader: mess)
	//]
}