package skuad.sim;

import skuad.uda.*;
import skuad.hubudp.*;
import skuad.util.databytes.DataReader;
import skuad.util.value.Value;

public abstract class VirtualSpot extends Spot {
	
	protected String ipPort;
	
	protected VirtualSpot(int vct) {
		super(vct);
		ipPort=null;
	}
	
	public abstract void goUp();
	public abstract void goDown();
	public abstract boolean addDevice(Device device, int type);
	public abstract boolean removeDevice(Device device);

	//[Modification des méta-données de ce spot virtuel
	public void setInt(int code, int value){ setInt(code, value, false); } 
	public void setString(int code, String value){ setString(code, value, false); }
	public void setBuff(int code, byte[] value){ setBuff(code, value, false); }
	public void setBool(int code, boolean v){ setInt(code, (v ? 1 : 0), false); }

	public abstract void setInt(int code, int value, boolean mode_add);
	public abstract void setString(int code, String value, boolean mode_add);
	public abstract void setBuff(int code, byte[] value, boolean mode_add);
	public abstract void removeAll(int code);
	public abstract boolean removeInt(int code, int value);
	
	public abstract void addTag(String name, Value value);	 	//ajoute la valeur 'value' au tag de nom 'name'
	public abstract boolean haveTag(String name, Value value);	//teste si le tag name possède la valeur value
	public abstract boolean addOnce(String name, Value value);	//ajoute la valeur value au tag name s'il ne la possède pas déjà, true si l'ajout est effectif
	public abstract boolean removeTag(String name, Value value);//supprime la valeur 'value' du tag de nom 'name'
	public abstract boolean clearTag(String name);				//efface toutes les valeurs du tag de nom 'name'
	//]
	
	//[Méthode de la classe Spot
	public boolean isLocalHost() { return(false); }
	public Hub getHub() { return(null); }
	public String getIpPort() {		//ip:port du contexte portant ce spot
		if(ipPort==null) ipPort = getIp() +':'+getPort();
		return(ipPort); 
	}

	public abstract String getIp();
	public int getPort() { return( this.getCTarget()); } //les VirtualSpot ont toujours un port == à leur ctarget
	public abstract boolean readySMD();
	public abstract int count(int code);
	public abstract int length(int code, int rang);
	public abstract int getInt(int code, int rang);
	public abstract boolean haveInt(int code, int value);
	public abstract String getString(int code, int rang);
	public abstract byte[] getBuff(int code, int rang);
	public abstract int readBuff(int code, int pos, int rang);

	public abstract void getTag(String name, TagReader callback);

	//Aucun envoie de message possible
	public boolean send(int channel, DataReader content, boolean need_ack, boolean auto_dispose) { return(false); }
	public boolean send(int channel, byte content, boolean need_ack) { return(false); }
	public boolean send(int channel, byte[] buff, int offset, int size, boolean need_ack) { return(false); }
	//]
}