package skuad.sim.implementation.simamanager;

import java.util.*;
import skuad.util.base.Util;
import skuad.sim.implementation.*;
import skuad.sim.simop.*;
import skuad.agentu.PhysicalPlugRule;
import skuad.agentu.implementation.amanager.*;
import skuad.ubiquity.implementation.ImplSpace;
import skuad.ubiquity.space.Space;
import skuad.uda.Device;

public class AgentUSimContext extends SimContext implements RunAgentContext {

	SimAgentManager sagent_manager;
	ImplStartSimOperation sim_op;
	Map<String, ImplSpace> allSpaces; //tables des spaces virtuels existant, key = idspace_local
	
	public AgentUSimContext() {
		sagent_manager = new SimAgentManager(this);	
		sim_op = new ImplStartSimOperation(this, sagent_manager);
		allSpaces = new HashMap<String, ImplSpace>();
	}
		
	//[Les simop
	public ReadSimOperation readSimOperation() { return(sim_op); }
	public StartSimOperation startSimOperation() { return(sim_op); }
	public RunSimOperation runSimOperation() { return(sagent_manager); }
	//]
	
	protected synchronized void reset() { //on annule toutes les sims opérations !
		super.reset();
		for(ImplSpace ispace : allSpaces.values()) ispace.destroy();
		allSpaces.clear();
		sagent_manager.reset();
		sim_op.reset();
	}
	
	//[interface RunAgentContext
	//OXO: pour l'instant les spaces sont à réactivité immédiate, il faudra changer ça pour introduire un delta temps sur les opérations 
	//d'écriture dans le space
	public Space getSpace(String space_name, String space_password) {
		if(space_password!=null && space_password.length()==0) space_password=null;
		if(space_name==null || space_name.length()==0) return(new ImplSpace(sim_op.localspot(), null, space_password)); //il s'agit d'un space anonyme
		if(space_name.equals("#new")){ 
			//génération d'un nouveau space avec nom aléatoire
			String pass_part = (space_password==null ? null : "#"+space_password);
			do space_name=Util.randString(6); while(allSpaces.containsKey(pass_part==null?space_name:space_name+pass_part));
		} else if(space_name.contains("#")) return(null); //l'usage du symbole # est interdit dans le nom d'un space
		
		String idSpace=(space_password==null?space_name:space_name+"#"+space_password);
		ImplSpace space=allSpaces.get(idSpace);
		if(space==null) { //création de l'objet space
			space=new ImplSpace(sim_op.localspot(), space_name, space_password);
			allSpaces.put(idSpace, space);
		}
		return(space);	
	}
	
	//[Gestion des devices
	public HostedDevice findDevice(PhysicalPlugRule prule) {
		return(sim_op.findDevice(prule));
	}
	public Device linkDevice(DeviceHoster hoster, HostedDevice hdevice) {
		return(sim_op.findDevice(hoster, hdevice));
	}
	public boolean unlinkDevice(DeviceHoster hoster, HostedDevice hdevice, Device device) {
		return(true); //RAS
	}
	
	public void performRuleToAnonymousDevice(HostedDevice hdevice){ //on doit déclencher l'analyse des règles sur ce device sans nom
		sagent_manager.performRuleToAnonymousDevice(hdevice);
	}
	public void performRuleToNamedDevice(HostedDevice hdevice){ //on doit déclencher l'analyse des règles sur ce device sans nom
		sagent_manager.performRuleToNamedDevice(hdevice);
	}
	//]

	public boolean canRunAgent() { //on peut lancer l'exécution d'un agent qui si la simulation est démarrée
		return(getState()==SimPilot.IN_RUN); 
	}
	
	public long getCurrentTime() { 
		return(curTime()); 
	}
	
	public void startAllAgent() { //on demarre les agents qui ont besoin de l'être
		sagent_manager.startAllAgent();
	}
	public void stopAllAgent() {//on demarre les agents qui ont besoin de l'être
		sagent_manager.stopAllAgent();
	}
	//]
}