package skuad.sim.implementation.simamanager;

import skuad.agentu.*;
import skuad.agentu.implementation.amanager.DeviceHoster;
import skuad.agentu.implementation.amanager.HostedDevice;
import skuad.util.base.ConstantIterable;
import java.util.Vector;
import skuad.uda.*;
import skuad.hubudp.Spot;
import skuad.hubudp.SpotExplorer;
import skuad.hubudp.SpotObserver;
import skuad.sim.VirtualSpot;
import skuad.sim.simop.*;

public class ImplReadSimOperation implements ReadSimOperation { //Opération de lecture possibles à tout moment de la simulation
	
	protected AgentUSimContext au_sim_context;
	SimAgentManager samanager;
	Vector<SpotObserver> spot_observers;	//les observeurs de spot
	Vector<ImplVirtualSpot> spots;
	
	public ImplReadSimOperation(AgentUSimContext vau_sim_context, SimAgentManager vsamanager) {
		au_sim_context = vau_sim_context;
		samanager = vsamanager;
		spots = new Vector<ImplVirtualSpot>();
		spot_observers = new Vector<SpotObserver>();
	}

	protected void reset() { //on annule toutes les sims opérations !
		spot_observers.clear();
		spots.clear();
	}
	
	//[Gestion des spots
	public void  addSpotObserver(SpotObserver obs) { synchronized(spot_observers) { 
		if(!spot_observers.contains(obs)) {
			spot_observers.add(obs);
			notifyAllSpot(obs); //On doit notifier cet observer de tous les UP et updateSMD
		}
	} }
	public boolean removeSpotObserver(SpotObserver obs){ synchronized(spot_observers) { return(spot_observers.remove(obs)); } }
	public void explore(SpotExplorer cb, boolean include_down_spot) {
		synchronized(spots) {
			for(VirtualSpot spot : spots) {
				if(include_down_spot || spot.isUp()) cb.spot(spot);
			}
		}
	}	
	private void notifyAllSpot(SpotObserver obs){ //On doit notifier cet observer de tous les UP, updateSMD 
		synchronized(spots) {
			for(VirtualSpot spot : spots) { 
				if(spot.isUp()) obs.up(spot);
				if(spot.readySMD()) obs.updateSMD(spot, true);
			}
		}
	}
	//[Pour les instances VirtualSpot
	protected void fireDown(VirtualSpot spot){
		synchronized(spot_observers) { for(SpotObserver so : spot_observers) so.down(spot); } 
	}
	protected void fireUp(Spot spot){
		synchronized(spot_observers) { for(SpotObserver so : spot_observers) so.up(spot); } 
	}
	protected void fireUpdateSMD(Spot spot, boolean first_definition){ //Pour chaque méthode mmm de l'interface XXXListener
		synchronized(spot_observers){
			for(SpotObserver so : spot_observers) so.updateSMD(spot, first_definition);
		}
	}
	//]
	//[Pour AgentUSimContext
	protected HostedDevice findDevice(PhysicalPlugRule prule){
		HostedDevice hdevice=null;
		for(ImplVirtualSpot vspot : spots) {
			if(vspot.isUp()) {		
				DeviceHoster dh = vspot.dhoster;
				hdevice = dh.findDevice(prule);
				if(hdevice!=null) break;	
			}
		}
		return(hdevice);
	}	
	protected Device findDevice(DeviceHoster hoster, HostedDevice hdevice){
		for(ImplVirtualSpot vspot : spots) {
			if(vspot.dhoster==hoster && vspot.isUp()) return(vspot.getDevice(hdevice));
		}
		return(null);
	}
	//]
	//]
	
	//[Configuration des générales
	public int getState() { return(au_sim_context.getState()); }
	public long curTime() { return(au_sim_context.curTime()); }
	
	public long getStartTime() { return(au_sim_context.getStartTime()); }
	public long getPingTime() { return(au_sim_context.getPingTime()); }
	public long getEndTime() { return(au_sim_context.getEndTime()); }
	//]
		
  //[[AGENTU
		
	//[Gestion des types d'agent
	public Iterable<String> namesTypeAgent(){ return(samanager.namesTypeAgent()); }	//retourne la liste des types agents existant
	public int codeTypeAgent(String name_type_agent){ return(samanager.codeTypeAgent(name_type_agent)); }
	//]		
	
	//[Consultation des agents
	public void addAllAgentObserver(AgentUObserver obs){ samanager.addAllAgentObserver(obs, true); }
	public boolean removeAllAgentObserver(AgentUObserver obs){ return(samanager.removeAllAgentObserver(obs, true)); }
	
	public boolean addAgentObserver(int agentId, AgentUObserver obs){ return(samanager.addAgentObserver(agentId, obs, true)); }
	public boolean removeAgentObserver(int agentId, AgentUObserver obs){ return(samanager.removeAgentObserver(agentId, obs, true)); }

	public Iterable<Integer> allAgentsId(){ return(samanager.allAgentsId()); }	
	public Iterable<Integer> agentsId(String agent_name){ return(samanager.agentsId(agent_name)); }	
	public boolean existAgent(int agentId){ return(samanager.existAgent(agentId)); }	
	public boolean existAgentWithName(String agent_name){ return(samanager.existAgentWithName(agent_name)); }	
	
	public int agentType(int agentId){ return(samanager.agentType(agentId)); }
	public String agentTypeName(int agentId){ return(samanager.agentTypeName(agentId)); }
	public String agentName(int agentId){ return(samanager.agentName(agentId)); }
	public byte agentState(int agentId){ return(samanager.agentState(agentId)); }
	public byte agentRunMode(int agentId){ return(samanager.agentRunMode(agentId)); }
	//]
	
	//[Configuration des agents
	public ConstantIterable<SocialSpecif> socialsSpecif(int code_type_agent){ return(samanager.socialsSpecif(code_type_agent)); }
	public ConstantIterable<SocialSpecif> socialsSpecif(String name_type_agent){ return(samanager.socialsSpecif(name_type_agent)); }
	public String getSocialPlugNameSpace(int agentId, String name_slot){ return(samanager.getSocialPlugNameSpace(agentId, name_slot)); }
	public boolean isSocialPlugPrivateSpace(int agentId, String name_slot){ return(samanager.isSocialPlugPrivateSpace(agentId, name_slot)); }
	public String getSocialSlotConfigLocalId(int agentId, String name_slot){ return(samanager.getSocialSlotConfigLocalId(agentId, name_slot)); }

	public boolean existSocialPlug(int agentId, String name_slot){ return(samanager.existSocialPlug(agentId, name_slot)); }
	public String getSocialPlugCurrentLocalId(int agentId, String name_slot){ return(samanager.getSocialPlugCurrentLocalId(agentId, name_slot)); }
	
	public ConstantIterable<PhysicalSpecif> physicalsSpecif(int code_type_agent){ return(samanager.physicalsSpecif(code_type_agent)); }
	public ConstantIterable<PhysicalSpecif> physicalsSpecif(String name_type_agent){ return(samanager.physicalsSpecif(name_type_agent)); }
	public Iterable<Integer> physicalPlugRules(int agentId, String name_slot){ return(samanager.physicalPlugRules(agentId, name_slot)); }
	
	public boolean physicalSlotPlugged(int agentId, String name_slot){ return(samanager.physicalSlotPlugged(agentId, name_slot)); }
	public int getPhysicalPlugDeviceId(int agentId, String name_slot){ return(samanager.getPhysicalPlugDeviceId(agentId, name_slot)); }
	
	public VirtualSpot getPhysicalPlugSpot(int agentId, String name_slot){ 
		Spot spot = samanager.getPhysicalPlugSpot(agentId, name_slot); 
		if(spot==null) return(null);
		return((VirtualSpot) spot);
	}
	//]	
  //AGENTU]]		
}