package skuad.sim.implementation.simamanager;

import java.util.HashMap;

import skuad.hubudp.Spot;
import skuad.util.databytes.DataReader;
import skuad.util.value.Value;
import skuad.sim.*;
import skuad.agentu.AgentU;
import skuad.agentu.PhysicalPlugRule;
import skuad.sim.simop.*;
import skuad.uda.Device;

public class ImplStartSimOperation extends ImplReadSimOperation implements StartSimOperation  { //Opération possibles au moment du lancement de la simulation
	
	HashMap<String, SmdValue[]> host_root_smd;
	VirtualSpot localspot;
	int[] auto_ip; //pour l'ip des spots créés automatiquement
	
	public ImplStartSimOperation(AgentUSimContext vau_sim_context, SimAgentManager vsamanager) {
		super(vau_sim_context, vsamanager);
		host_root_smd = new HashMap<String, SmdValue[]>();	
		localspot=null;
		auto_ip = new int[4];
		auto_ip[0]=200; auto_ip[1]=0; auto_ip[2]=0; auto_ip[3]=0;
	}
	
	protected void reset() { //on annule toutes les sims opérations !
		super.reset();
		host_root_smd.clear();
		localspot=null;
	}
	
//Ecriture
	//[Configuration des générales
	public boolean loadXeloc(String filename) { //seuls les opérations permise en fonction de l'état de la simulation seront traité
		//OXO: --> prévoir des blocs conditionnés sur l'état preSim/inSim
		return(false);
	}
	//]

	public void log(String mess) {
		au_sim_context.log(mess);
	}
	
	//[Gestion des environnements virtuels
	public void addVirtualEnv(VirtualEnv venv) { au_sim_context.addVirtualEnv(venv); }
	public boolean removeVirtualEnv(VirtualEnv venv){ return(au_sim_context.removeVirtualEnv(venv)); }
	//]
	
	//[Gestion des pulsables
	public void addPulsable(Pulsable pulsable) { au_sim_context.addPulsable(pulsable); }
	public boolean removePulsable(Pulsable pulsable){ return(au_sim_context.removePulsable(pulsable)); }
	//]

  //[[AGENTU		
	//[Configuration des spots
	public void setHostRootSMD(String host_ip, String host_name) {
		SmdValue[] smd = new SmdValue[1];
		smd[0] = new SmdValue(Spot.HOST_NAME, host_name);
		host_root_smd.put(host_ip, smd);
	}
	public void setHostRootSMD(String host_ip, SmdValue[] smd) {
		host_root_smd.put(host_ip, smd);
	}
	public VirtualSpot newSpot(String host_ip) {	
		ImplVirtualSpot spot = new ImplVirtualSpot(au_sim_context, this,  host_ip);
		synchronized(spots) { spots.add(spot); }
		
		//On regarde si l'hote possède des smd
		SmdValue[] smd = host_root_smd.get(host_ip);
		if(smd!=null) {
			for(int i=0; i<smd.length; i++) {
				SmdValue cur = smd[i];
				
				if(cur!=null) cur.fixe(spot);
			}
		}
		
		return(spot);
	}
	private String nextAutoIp() {
		auto_ip[3]++;
		if(auto_ip[3]>255) {
			auto_ip[3]=0; auto_ip[2]++;
			if(auto_ip[2]>255) {
				auto_ip[2]=0; auto_ip[1]++;
				if(auto_ip[1]>255) { auto_ip[1]=0; auto_ip[0]++; }
			}
		}
		return(auto_ip[0]+"."+auto_ip[1]+"."+auto_ip[2]+"."+auto_ip[3]);
	}
	public PhysicalPlugRule addDevice(Device d) { //enregistre le device sur un nouveau spot et retourne une règle de plug permettant de le cibler
		String ip=nextAutoIp();
		VirtualSpot spot = newSpot(ip);
		spot.addDevice(d, 0);
		spot.goUp();
		return(PhysicalPlugRule.ipRule(ip));
	}
	
	public boolean removeDevice(Device d) {
		synchronized(spots) {
			for(ImplVirtualSpot spot : spots) {
				if(spot.removeDevice(d)) return(true);
			}
		}
		return(false);
	}
	
	public VirtualSpot localspot() { //récupération du spot local virtuel
		if(localspot==null) {
			localspot=newSpot("127.0.0.1");
			localspot.goUp();
		}
		return(localspot);
	}
	//]	
	
	//[Gestion des types d'agent
	public int addTypeAgent(Class<? extends AgentU> classe){ return(samanager.addTypeAgent(classe.getName(), classe)); }
	public int addTypeAgent(String type_name,  Class<? extends AgentU> classe){ return(samanager.addTypeAgent(type_name, classe)); }
	//]
	
	//[Création/suppresion des agents
	public int createAgent(Class<? extends AgentU> classe, String agent_name){ 
		int code_type = addTypeAgent(classe);
		return(samanager.newAgent(code_type, agent_name, null));
	}
	public int createAgent(int code_type, String agent_name){
		return(samanager.newAgent(code_type, agent_name, null));
	}
	public int createAgent(String type_name, String agent_name){ return(samanager.newAgent(type_name, agent_name, null)); }
	public int createAgent(Class<? extends AgentU> classe, String agent_name, Value config){ 
		int code_type = addTypeAgent(classe);
		return(samanager.newAgent(code_type, agent_name, config));
	}
	public int createAgent(int code_type, String agent_name, Value config){
		return(samanager.newAgent(code_type, agent_name, config));
	}
	public int createAgent(String type_name, String agent_name, Value config){ return(samanager.newAgent(type_name, agent_name, config)); }
	public boolean removeAgent(int agent_id){ return(samanager.removeAgent(agent_id)); }
	//]
			
	//[Configuration des agents
	public boolean changeRunModeAgent(int agent_id, byte run_mode){ return(samanager.changeRunModeAgent(agent_id, run_mode)); }
	public byte setSocialPlug(int agent_id, String name_slot, String name_space, String password, String local_id){ return(samanager.setSocialPlug(agent_id, name_slot, name_space, password, local_id)); }
	public int addPhysicalPlugRule(int agent_id, String name_slot, DataReader rule){ 
		PhysicalPlugRule pprule = PhysicalPlugRule.parse(rule);
		rule.dispose();
		return(samanager.addPhysicalPlugRule(agent_id, name_slot, pprule)); 
	}
	public int addPhysicalPlugRule(int agent_id, String name_slot, String rule){ return(addPhysicalPlugRule(agent_id, name_slot, Value.parse(rule))); }
	public int addPhysicalPlugRule(int agent_id, String name_slot, Value rule){ 
		PhysicalPlugRule pprule = PhysicalPlugRule.parse(rule);
		return(samanager.addPhysicalPlugRule(agent_id, name_slot, pprule)); 
	}
	public int addPhysicalPlugRule(int agent_id, String name_slot, PhysicalPlugRule rule){ return(samanager.addPhysicalPlugRule(agent_id, name_slot, rule)); }
	public byte movePhysicalPlugRule(int agent_id, String name_slot, byte rule_id, byte direction){ return(samanager.movePhysicalPlugRule(agent_id, name_slot, rule_id, direction)); }
	public int removePhysicalPlugRule(int agent_id, String name_slot, byte rule_id){ return(samanager.removePhysicalPlugRule(agent_id, name_slot, rule_id)); }
	//]
  //AGENTU]]	
	
	//[A supprimer
	/**
	 * @deprecated remplacé par addVirtualEnv(...)
	 */
	@Deprecated
	public void addVirtaulEnv(VirtualEnv venv) {
		addVirtualEnv(venv);
	}
	/**
	 * @deprecated remplacé par removeVirtualEnv(...)
	 */
	@Deprecated	
	public boolean removeVirtaulEnv(VirtualEnv venv) {
		return(removeVirtualEnv(venv));
	}
	//]
	
}