package skuad.sim.implementation.simamanager;

import skuad.uda.*;
import skuad.agentu.implementation.amanager.*;
import java.util.*;
import skuad.hubudp.*;
import skuad.hubudp.implementation.ValueSMD;
import skuad.sim.*;
import skuad.uda.monitor.*;
import skuad.util.value.Value;

public class ImplVirtualSpot extends VirtualSpot {
	private static int ctarget_count=0;
	private synchronized static int newCTarget() { return(++ctarget_count); }
	
	ImplReadSimOperation i_rsimop;
	Vector<ValueSMD> local_dsmd; 	//définition SMD local brute
	String host_ip;
	DeviceHoster dhoster;	//gestionnaire des devices actif porté par ce virtualspot
	int next_device_id=1;
	AgentUSimContext au_sim_context;
	Vector<IdentifiedDevice> idevices; //listes des devices portés par ce spot
	
	
	protected ImplVirtualSpot(AgentUSimContext vau_sim_context, ImplReadSimOperation vi_rsimop, String vhost_ip) {
		super(newCTarget());
		au_sim_context = vau_sim_context;
		i_rsimop = vi_rsimop;
		local_dsmd = new Vector<ValueSMD>();
		host_ip=vhost_ip;
		dhoster = new DeviceHoster(au_sim_context, this, 0);
		idevices = new Vector<IdentifiedDevice>();
	}

	//[Gestion des devices
	public boolean addDevice(Device device, int type) { synchronized(idevices) {
		IdentifiedDevice idv = getIDevice(device);
		if(idv!=null) return(false); //device déjà enregistré sur ce spot
		int nid = next_device_id++;
		idv = new IdentifiedDevice(nid, device, type);
		idevices.add(idv);
		if(isUp()) idv.setHDevice(dhoster.addDevice(nid, type, idv.name));
		return(true); 
	}}
	private IdentifiedDevice getIDevice(Device device) {
		synchronized(idevices) { for(IdentifiedDevice idv : idevices) if(idv.device==device) return(idv); }
		return(null);
	}
	protected Device getDevice(HostedDevice hdevice) {
		synchronized(idevices) { for(IdentifiedDevice idv : idevices) if(idv.isHDevice(hdevice)) return(idv.device); }
		return(null);
	}
	public boolean removeDevice(Device device) { synchronized(idevices) {
		for(int i=0; i<idevices.size(); i++) {
			IdentifiedDevice idv = idevices.get(i);
			if(idv.device==device) {
				dhoster.removeDevice(idv.id);
				idevices.remove(i);
				return(true);
			}
		}
		return(false);
	}}
	//]
	
	//[ Up/Down
	public synchronized void goUp() {
		if(!is_up) {
			is_up = true;
			fireUp();//on notifie les SpotObserver au niveau des écoutes locales
			i_rsimop.fireUp(this);//et global
			
			//on up les devices
			synchronized(idevices) { for(IdentifiedDevice idv : idevices) idv.setHDevice(dhoster.addDevice(idv.id, idv.type, idv.name)); }	
		}		
	}
	public synchronized void goDown() {
		if(is_up) {
			is_up = false;
			
			//on down des devices
			synchronized(idevices) { 
				for(IdentifiedDevice idv : idevices) idv.setHDevice(null);
				dhoster.removeAllDevice();	
			}
			
			fireDown();//on notifie les SpotObserver au niveau des écoutes locales
			i_rsimop.fireDown(this);//et global
		}
	}

	protected void fireDown(){
		synchronized(observers) { 
			for(SpotObserver so : observers) so.down(this); 
		} 
	}
	protected void fireUp(){
		synchronized(observers) { 
			for(SpotObserver so : observers) so.up(this); 
		} 
	}
	protected void fireUpdateSMD(boolean first_definition){ //Pour chaque méthode mmm de l'interface XXXListener
		synchronized(observers){
			for(SpotObserver so : observers)  so.updateSMD(this, first_definition);
		}
	}
	//]
	
	//[Modification des méta-données de ce spot virtuel
	protected  ValueSMD findCode(int code, int rang){
		int num = 0; ValueSMD cur;
		synchronized(local_dsmd) { for(int i=0; i<local_dsmd.size(); i++) {
			cur = local_dsmd.get(i);
			if(cur.code==code) {
				if(num==rang) return(cur);
				num++;
			}
		}}
		return(null); //code non trouvé
	}
	boolean first_definition=true;
	protected synchronized void needUpdateSMD(){
		if(isUp()) {
			fireUpdateSMD(first_definition);
			i_rsimop.fireUpdateSMD(this, first_definition);
			first_definition=false;
		}
	}
	public void setInt(int code, int value ,boolean mode_add) {	
		boolean need_update=false;
		ValueSMD vhia=null;
		synchronized(local_dsmd) {
			if(!mode_add)  vhia = findCode(code, 0);
			if(vhia==null) {
				vhia = new ValueSMD(code, value);
				local_dsmd.add(vhia);
				need_update = true;
			} else {
				if(vhia.type!=ValueSMD.NUMBER || vhia.num_value!=value) {
					vhia.type=ValueSMD.NUMBER;
					vhia.num_value = value;
					need_update = true;
				}
			}
			if(need_update) needUpdateSMD();
		}
	}	
	public void setString(int code, String value, boolean mode_add) { // si != null la chaine automatiquement tronquée à 250 caractère maximum
		boolean need_update=false;
		ValueSMD vhia=null;
		if(value!=null && value.length()>250) value=value.substring(0,250);
		synchronized(local_dsmd) {
			if(!mode_add)  vhia = findCode(code, 0);
			if(vhia==null) {
				vhia = new ValueSMD(code, value);
				local_dsmd.add(vhia);
				need_update = true;
			} else {
				if(vhia.type!=ValueSMD.STRING || (vhia.st_value!=null && value==null) ||  (vhia.st_value==null && value!=null) || (vhia.st_value!=null && !vhia.st_value.equals(value)) ) {
					vhia.type=ValueSMD.STRING;
					vhia.st_value = value;
					need_update = true;
				}
			}
			if(need_update) {
				switch(code) { //capture de SPOT_NAME et HOST_NAME pour dhoster
				case Spot.SPOT_NAME : dhoster.setSpotName(value); break;
				case Spot.HOST_NAME : dhoster.setNodeName(value); break;
				}
				needUpdateSMD();
			}
		}
	}	

	public void setBuff(int code, byte[] value, boolean mode_add) {	//-> fixe la valeur du code à value au format tableau d'octet, si != null la chaine automatiquement tronquée à 255 éléments maximum	
		boolean need_update=false;
		ValueSMD vhia=null;
		synchronized(local_dsmd) {
			if(!mode_add)  vhia = findCode(code, 0);
			if(vhia==null) {
				vhia = new ValueSMD(code, value);
				local_dsmd.add(vhia);
				need_update = true;
			} else {
				if(!vhia.isSameBuffer(value)) {
					vhia.type=ValueSMD.BUFFER;
					int length = Math.min(value.length, 250);
					vhia.buffer_value = new byte[length];
					for(int i=0; i<length; i++) vhia.buffer_value[i] = value[i];
					need_update = true;
				}
			}
			if(need_update) needUpdateSMD();
		}
	}	
	//suppression
	public void removeAll(int code) {				//	-> supprime toute les valeurs pour le code indiqué
		boolean need_update=false;
		for(int i=0; i<local_dsmd.size(); i++) {
			if(local_dsmd.get(i).code==code){
				local_dsmd.removeElementAt(i); 
				i--;
				need_update=true;
			}
		}
		if(need_update) needUpdateSMD();
	}	

	public boolean removeInt(int code, int value) {	//	-> supprime le couple (code, value) s'il existe
		boolean need_update=false;
		for(int i=0; i<local_dsmd.size(); i++) {
			ValueSMD cur = local_dsmd.get(i); 
			if(cur.code==code && cur.type==ValueSMD.NUMBER && cur.type==value){
				local_dsmd.removeElementAt(i); 
				need_update=true;
				break; //on sort du for
			}
		}
		if(need_update) {
			needUpdateSMD();
			return(true);
		}
		return(false);
	}
	//]
	
	//[Méthode de la classe Spot
	public boolean isLocalHost() { return(false); }
	public Hub getHub() { return(null); }

	public String getIp() { return(host_ip); }

	 //[lecture des méta-donnée SMD (Small Méta-Data)
	public boolean readySMD() { return(is_up); }	
	public int count(int code) {						//	-> retourne le nombre d'occurence existante pour le code SMD indiqué
		int nb=0;
		synchronized(local_dsmd) { for(int i=0; i<local_dsmd.size(); i++) {
			if(local_dsmd.get(i).code==code) nb++;
		}}
		return(nb);
	}		
	public int length(int code, int rang) {				//	-> retourne le nombre d'octet de la valeur du code SMD indiqué
		ValueSMD vhia = findCode(code, rang);
		if(vhia!=null) {
			switch(vhia.type){
			case ValueSMD.NUMBER :
				if(vhia.num_value==0) return(0);
				else if(vhia.num_value<256)  return(1);
				else if(vhia.num_value<65536)  return(2);
				else return(4);
			case ValueSMD.STRING :
				if(vhia.st_value==null)  return(0);
				else return(vhia.st_value.length());
			case ValueSMD.BUFFER :
				if(vhia.buffer_value==null)  return(0);
				else return(vhia.buffer_value.length);
			}
		}
		return(0);
	}	
	public int getInt(int code, int rang) {				//	-> lecture de l'octet numéro 0 de la valeur du code SMD indiqué
		ValueSMD vhia = findCode(code, rang);
		if(vhia!=null) {
			if(vhia.type==ValueSMD.NUMBER) return(vhia.num_value);
			else return(-1); //mauvais type
		}
		return(0); //non trouvé
	}	
	public boolean haveInt(int code , int value) {			//	-> true ssi le code SMD indiqué possède la valeur d'octet value
		ValueSMD cur;
		synchronized(local_dsmd) { for(int i=0; i<local_dsmd.size(); i++) {
			cur = local_dsmd.get(i);
			if(cur.code==code) {
				if(cur.type==ValueSMD.NUMBER && cur.num_value==value) return(true);
			}
		}}
		return(false); //code non trouvé
	}	
	public String getString(int code , int rang) {		//	-> lecture de la valeur du code SMD indiqué sous la forme d'une chaine de caractère
		ValueSMD vhia = findCode(code, rang);
		if(vhia!=null) {
			if(vhia.type==ValueSMD.STRING) return(vhia.st_value);
			else return(null); //mauvais type
		}
		return(null); //non trouvé
	}	
	public byte[] getBuff(int code , int rang) {			//	-> lecture du buffer de la valeur du code SMD indiqué
		ValueSMD vhia = findCode(code, rang);
		if(vhia!=null) {
			if(vhia.type==ValueSMD.BUFFER && vhia.buffer_value!=null) {
				byte[] buff = new byte[vhia.buffer_value.length];
				for(int i=0; i<buff.length; i++) buff[i]=vhia.buffer_value[i];
				return(buff);
			} else return(null); //mauvais type
		}
		return(null); //non trouvé
	}	
	public int readBuff(int code, int pos , int rang) { 	//	-> lecture de l'octet pos de la valeur du code SMD indiqué
		ValueSMD vhia = findCode(code, rang);
		if(vhia!=null && vhia.type==ValueSMD.BUFFER && vhia.buffer_value!=null && pos<vhia.buffer_value.length) return(vhia.buffer_value[pos]);
		return(-1); //non trouvé
	}	
	//]
	
	//[gestion des Tags
	Map<String, Vector<Value>> tags = new HashMap<String, Vector<Value>>(); //table des valeurs des tags
	public static boolean containValue(Vector<Value> tab_value, Value v){
		if(tab_value!=null){ synchronized(tab_value){
			for(Value value : tab_value) if(value.equals(v)) return(true);
		}}
		return(false);
	}
	public void addTag(String name, Value value){	 		//ajoute la valeur 'value' au tag de nom 'name'
		synchronized(tags){ 
			Vector<Value> tab_value = tags.get(name);
			if(tab_value==null){
				tab_value = new Vector<Value>();
				tags.put(name, tab_value);
			}
			synchronized(tab_value){ tab_value.add(value.toConstant()); }
		}
	}
	public boolean haveTag(String name, Value value){	//teste si le tag name possède la valeur value
		return(containValue(tags.get(name), value));
	}
	public boolean addOnce(String name, Value value){	//ajoute la valeur value au tag name s'il ne la possède pas déjà, true si l'ajout est effectif
		if(!containValue(tags.get(name), value)){
			addTag(name, value);
			return(true);
		}
		return(false);
	}
	public boolean removeTag(String name, Value v){ //supprime la valeur 'value' du tag de nom 'name'
		Vector<Value> tab_value = tags.get(name);
		if(tab_value!=null){ synchronized(tab_value){
			for(int i=0; i<tab_value.size(); i++) if(tab_value.get(i).equals(v)){
				tab_value.removeElement(i);
				return(true);
			}
		}}
		return(false);		
	}
	public boolean clearTag(String name){				//efface toutes les valeurs du tag de nom 'name'
		Vector<Value> tab_value = tags.get(name);
		if(tab_value!=null){synchronized(tab_value){
			tab_value.removeAllElements();
			return(true);
		}}
		return(false);
	}
	public void getTag(String name, TagReader callback){ 	//--> déclenche callback(name, tab_value)
		if(callback!=null){
			Vector<Value> tab_value = tags.get(name);
			if(tab_value==null) callback.tagValues(this, name, new Value[0]);
			else callback.tagValues(this, name, (Value[]) tab_value.toArray()); 
		}
	}
	//]
}

class IdentifiedDevice extends MinimalMonitor {
	int id, type;
	private HostedDevice hdevice;
	Device device;
	String name;
	
	public IdentifiedDevice(int vid, Device vdevice, int vtype) {
		id=vid;
		type=vtype;
		hdevice = null;
		device = vdevice;
		name = null;
		device.getParams(this);
	}

	public synchronized void params(Device dev, String dname, int type_device, int nb_sensor, int nb_actuator, boolean can_log){
		name = dname;
		if(hdevice!=null) hdevice.setName(dname);
	}	
	
	public synchronized void setHDevice(HostedDevice vhdevice) {
		hdevice = vhdevice;
		if(hdevice!=null && name!=null) hdevice.setName(name);
	}
	
	public boolean isHDevice(HostedDevice vhdevice) { return(hdevice==vhdevice); }
}