package skuad.sim.implementation.simamanager;

import skuad.agentu.AgentU;
import skuad.util.time.Pulse;
import skuad.agentu.implementation.amanager.*;
import skuad.util.time.temporality.*;
import skuad.sim.implementation.*;

public class SimPulsor extends Pulsor implements PulseTarget {

	SimContext scontext;
	Pulse sim_pulse;
	long next_pulse; //timecode du prochain pulse (pour la gestion des reprises aprés pause) 
	
	public SimPulsor(SimContext vscontext, ExistingAgentU viau, String vlabel, long vdelay,  long vperiode, boolean fireUptodate_or_resetOnResume){ //pulse cyclique
		//si reset_on_resume==true : le cycle de pulse est réinitialisé à la reprise de l'agent (le prochain pulse surviendra aprés 
		//l'écoulement du delay, puis les suivants a chaque période), avec reset_on_resume==false à la reprise aprés pause
		//les cycles de déclenchement reprenne comme si l'agent n'avait jamais été en pause.
		super(vlabel, vdelay, vperiode, fireUptodate_or_resetOnResume);
		scontext = vscontext;		
		sim_pulse = null;
		next_pulse = scontext.curTime()+vdelay;
		lingToExistingAgentU(viau);
	}

	protected void pause(){
		if(sim_pulse!=null){ sim_pulse.cancel(); sim_pulse=null; }
	}
	protected boolean startOrResume(boolean call_remove_pulsor_if_cancel){ //retourne false si le pulsor doit être supprimé	
		if(!active) return(false);
		if(sim_pulse!=null) return(true); //le pulsor est déjà actif et pas en pause
		long cur_time = scontext.curTime();
		long ddelay;
		if(periode==0){ //pulse ponctuel
			//System.out.println("    ---> pulsor ponctuel");				
			if(!resume_reset && cur_time>next_pulse){  //on supprime le pulse
				cancel(call_remove_pulsor_if_cancel);
				return(false);
			}
			ddelay = next_pulse-cur_time;
			if(ddelay<=0) ddelay=0;
			sim_pulse = scontext.newPulse(this, label, ddelay);
		} else { //pulse cyclique
			//System.out.println("    ---> pulsor cyclique");	
			if(resume_reset) ddelay=delay;
			else { ddelay = next_pulse-cur_time; if(ddelay<0) ddelay=0; }
			sim_pulse = scontext.newPulse(this, label, ddelay, periode);
		}
		return(true);
	}
	protected boolean remove(){
		if(super.remove()) {
			if(sim_pulse!=null){ sim_pulse.cancel(); sim_pulse=null; }
			next_pulse=0;
			return(true);
		}
		return(false);
	}
	
	//[Interface PulseTarget
	public void run(Pulse pulse){
//System.out.println("--- run du pulsor : "+label);				
		if(active && eau.getState()==AgentU.RUN) {
//System.out.println("   ---> run confirmé");			
			long tc = scontext.curTime();
			//on met à jour son prochain timecode du prochain déclenchement
			if(periode>0) next_pulse= tc+periode; 
			else next_pulse=0;
			pulseBehavior(tc);
		}
		if(!active || periode==0) cancel(true); //pulse ponctuel -> on le supprime
	}
	//]
}