package skuad.sim.ui;

import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.image.BufferedImage;
import java.net.URL;
import skuad.sim.simop.*;
import skuad.sim.*;

public class SimControlBar extends JToolBar implements ActionListener, SimObserver {
	
	SimPilot simpilot;
	JButton bt_reset, bt_start, bt_pause, bt_stop;
	SimSpeedClock speed_clock;
	
	public SimControlBar(SimPilot vsimpilot, SimSpeedClock ssc) {
		speed_clock = ssc;
		simpilot = vsimpilot;
		setFloatable(false);
		bt_reset= makeButton("reset", "Réinitialiser la simulation", "Reset"); add(bt_reset);
		bt_stop=makeButton("stop", "Stopper la simulation", "Stop"); add(bt_stop);
		bt_start=makeButton("start", "Lancer/reprendre la simulation", "Play"); add(bt_start);
		bt_pause= makeButton("pause", "Mettre la simulation en pause", "Pause"); add(bt_pause);
		
		if(simpilot!=null) {
			simpilot.addSimObserver(this);
			majButton();
		}
	}
	
	public void actionPerformed(ActionEvent event) {
		if(simpilot!=null) switch(event.getActionCommand()) {
		case "reset": simpilot.reset(); speed_clock.unlockSpeedClock(); break;
		case "stop": simpilot.stop(); speed_clock.unlockSpeedClock(); break;
		case "start": simpilot.start(); break;
		case "pause": simpilot.pause(); speed_clock.unlockSpeedClock(); break;			
		}
	}
	
	//[Interface SimObserver
	public void simEvent(SimPilot sim_pilot, int type, ReadSimOperation rsop, long time_code) {
		if(type!=SimPilot.BEGIN_CYCLE && type!= SimPilot.END_CYCLE) majButton();
	}
	public void simLog(SimPilot sim_pilot , long time_code, String mess) {}
	//]
	
	private void majButton() {
		switch(simpilot.getProcessState()){
		case SimPilot.OFF : 
			bt_start.setEnabled(true); 
			bt_reset.setEnabled(false);
			bt_stop.setEnabled(false);
			bt_pause.setEnabled(false);
			break;
		case SimPilot.INITIALIZING : case SimPilot.IN_RUN : case SimPilot.PAUSE_PLANNED :  case SimPilot.RESUMED:
			bt_reset.setEnabled(true);
			bt_stop.setEnabled(true);
			bt_pause.setEnabled(true);
			bt_start.setEnabled(false); 
			break;
		case SimPilot.PAUSED :
			bt_reset.setEnabled(true);
			bt_stop.setEnabled(true);
			bt_start.setEnabled(true);
			bt_pause.setEnabled(false);
			break;
		case SimPilot.STOPPED : case SimPilot.ENDED:
			bt_reset.setEnabled(true);
			bt_start.setEnabled(false); 
			bt_stop.setEnabled(false);
			bt_pause.setEnabled(false);
			break;
		}
	}
	
	protected JButton makeButton(String actionCommand, String toolTipText, String altText) {
		//Create and initialize the button.
		JButton button = new JButton();
		button.setActionCommand(actionCommand);
		button.setToolTipText(toolTipText);
		button.addActionListener(this);

		String imgLocation = "images/"+ actionCommand+".png";
		URL imageURL = SimControlBar.class.getResource(imgLocation);
		if (imageURL != null) button.setIcon(new ImageIcon(imageURL, altText));
		else {                                    
			button.setText(altText);
			System.err.println("Resource not found: " + imgLocation);
		}
		imgLocation = "images/"+ actionCommand+"_d.png";
		imageURL = SimControlBar.class.getResource(imgLocation);
		if (imageURL != null) button.setDisabledIcon(getGray(new ImageIcon(imageURL, altText)));
		else System.err.println("Resource not found: " + imgLocation);
		
		return button;
	}
	
	private Icon getGray(Icon icon) {
	    final int w = icon.getIconWidth();
	    final int h = icon.getIconHeight();
	    GraphicsEnvironment ge =
	        GraphicsEnvironment.getLocalGraphicsEnvironment();
	    GraphicsDevice gd = ge.getDefaultScreenDevice();
	    GraphicsConfiguration gc = gd.getDefaultConfiguration();
	    BufferedImage image =  gc.createCompatibleImage(w, h, Transparency.TRANSLUCENT);
	    Graphics2D g2d = image.createGraphics();
	    icon.paintIcon(null, g2d, 0, 0);
	    Image gray = GrayFilter.createDisabledImage(image);
	    return new ImageIcon(gray);
	}
/*	
	//Lancement autonome
	public static void main(String[] args) {
		JFrame frame=new JFrame("ControlBox");
		frame.setBounds(100, 100, 190, 80);
		frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		
		//mep
		SimControlBar cb = new SimControlBar(new SimAU("Test 1"));
		Container content = frame.getContentPane();
		content.setLayout(new BorderLayout());
		content.add(cb, BorderLayout.CENTER);
		
		frame.setVisible(true);
	}*/
}
