package skuad.sim.venv;

import skuad.sim.*;
import skuad.sim.simop.*;

public class SpeedClock implements VirtualEnv {
	
	int  speed;  //vitesse d'écoulement relatif du temps : tps simu -> tps réel en %
	int delta_speed;
		//-->  Delta simu = (Delta réel / 100) * speed   --> Delta réel = (Delta simu / speed) * 100
	long v_old_tc, r_old_tc;
	double acc_delta_reel;
	
	public SpeedClock() { 
		speed = 100;
		delta_speed=0;
		acc_delta_reel=0;
	}
	
	public synchronized void unlockSpeedClock() {
		delta_speed=0;
		 this.notify();
	}
	
	public synchronized void setSpeed(int vspeed) {
		if(vspeed!=speed) {
			delta_speed += vspeed-speed;
			speed=vspeed;
			if(delta_speed>5) this.notify();
		}
	}
	
	public void start(SimPilot sim_pilot, StartSimOperation ssop) {	//la simulation démarre
		v_old_tc=ssop.curTime();
		r_old_tc=System.currentTimeMillis();
	}
		//elle peut déjà l'être si l'instance VirtualEnv est créée aprés le démarrage de la simulation
	public void stop(SimPilot sim_pilot) {}	//la simulation s'arrête	
	
	//next_tc est le timecode candidat pour le prochain cycle,  <=0 si aucun et simulation infini
	//si next_tc>0 : celui sera effectif si aucun VirtaulEnv ne retourne un nouveau 0<tc<nex_tc durant ce cycle d'appel , sinon c'est le plus petit tc>0 qui sera candidat
	//si next_tc<=0 : la simulation va s'arréter sauf au moins un VirtaulEnv retourne un nouveau 0<tc<nex_tc durant ce cycle d'appel , et si c'est le cas c'est le plus petit tc>0 qui sera candidat
	public long candidateNextTime(SimPilot sim_pilot, StartSimOperation ssop, RunSimOperation rsop, long previous_tc, long next_tc) {
		return(0);
	}
	
	//doit retourner true si un prochain cycle de simulation (aprés cet appel sur cur_tc) est nécessité par cet environnement
	//si aucun retour true est qu'il n'y a plus d'activité pulsable la simulation va s'arréter
	//NB: si la simulation n'est pas infini et que cur_tc>endSim la simulation va s'arréter même si true est renvoyé  
	public synchronized boolean initializeNextTime(SimPilot sim_pilot, StartSimOperation ssop, RunSimOperation rsop, long previous_tc, long cur_tc) {
		long delta_simu = cur_tc - v_old_tc;
		v_old_tc = cur_tc;
		double delta_reel =  (( ((double)delta_simu) / speed ) * 100);
//System.out.println("pause = "+delta_reel);		
		long r_new_tc = System.currentTimeMillis();
		delta_reel = delta_reel - (r_new_tc - r_old_tc)+acc_delta_reel;
		if(delta_reel<20) {
			r_old_tc = r_new_tc;
			acc_delta_reel = delta_reel;
		} else while(delta_reel>=20) {
				delta_speed=0;
				acc_delta_reel = 0;
//				System.out.println("pause = "+delta_reel);
				try{ this.wait((int)delta_reel); } catch(Exception e) {}
				r_old_tc = System.currentTimeMillis();
				if(delta_speed!=0) delta_reel = (delta_reel - (r_old_tc - r_new_tc)) * (1-delta_speed/100);
				else break;
		}
		return(false);
	}
	
	
}