package skuad.ubiquity.implementation;

import java.util.*;
import skuad.ubiquity.space.*;
import skuad.ubiquity.space.activity.*;
import skuad.ubiquity.*;
import skuad.util.databytes.*;
import skuad.util.thread.ThreadStack;
import skuad.util.value.Value;

public class Activity implements WActivity {

	protected static ActivityRunner DistantActivity=new ActivityRunner(){
		public void action(Avatar avatar, String activity, Value args, ProgressFlag progress_flag){}
	};
	
	ImplParticipant thisParticipant;		//participant associé à ce descripteur d'activité
	ImplAvatar avatar;
	ActivityProxyReader rproxy;				//objet proxy de lecture sur cet instance
	Vector<ActivityValue> activities;		//listes des activitées
	Vector<ActivityListener> listener;
	
	protected Activity(ImplParticipant p, ImplAvatar vavatar){
		thisParticipant=p;
		activities=null;
		listener=null;
		rproxy=null;
		avatar = vavatar;
	}
		
	protected ActivityValue findActivity(String activity){
		if(activity!=null && activities!=null) for(ActivityValue cur : activities){
			if(activity.equals(cur.name)) return(cur);
		}
		return(null);		
	}
	protected ActivityValue getActivity(int num){ //consultation de l'alias public de numéro arbitraire num, null si non existant
		return(activities==null || num>=activities.size() || num<0?null:activities.get(num));
	}
	
	//[Interface RActivity
	public Participant participant(){ return(thisParticipant); }	
	public int nbActivities(){ return(activities==null?0:activities.size()); } 
	public String activityName(int num){ //consultation de l'alias public de numéro arbitraire num, null si non existant
		ActivityValue existing=getActivity(num);
		return(existing==null?null:existing.name);	
	}
	public byte activityStatus(int num){ //retourne : NOT_EXIST , NOT_RUN , WAITING, RUN, CANCELED
		ActivityValue existing=getActivity(num);
		return(existing==null?Ubiquity.NOT_EXIST: (byte) existing.status);	
	}
	public byte activityMode(int num){ //retourne : NOT_EXIST , ONCE , CHAIN, REPEAT
		ActivityValue existing=getActivity(num);
		return(existing==null?Ubiquity.NOT_EXIST:  (byte) existing.mode);	
	}
	public int activityProgressMax(int num){
		ActivityValue existing=getActivity(num);
		return(existing==null?0:existing.progressMax);			
	}
	public int activityProgressStep(int num){
		ActivityValue existing=getActivity(num);
		return(existing==null?0:existing.progressStep);			
	}
	public boolean haveActivity(String activity){
		return(findActivity(activity)!=null);
	}
	public byte activityStatus(String activity){ //retourne : NOT_EXIST , NOT_RUN , WAITING, RUN, CANCELED
		ActivityValue existing=findActivity(activity);
		if(existing!=null) return(  (byte) existing.status);
		return(Ubiquity.NOT_EXIST);			
	}
	public byte activityMode(String activity){ //retourne : NOT_EXIST , NONE, ONCE , CHAIN, REPEAT
		ActivityValue existing=findActivity(activity);
		if(existing!=null) return( (byte)  existing.mode);
		return(Ubiquity.NOT_EXIST);		
	}
	public int activityProgressMax(String activity){
		ActivityValue existing=findActivity(activity);
		if(existing!=null) return(existing.progressMax);
		return(0);		
	}
	public int activityProgressStep(String activity){
		ActivityValue existing=findActivity(activity);
		if(existing!=null) return(existing.progressStep);
		return(0);		
	}	
	public synchronized void addListener(ActivityListener il){ 
		if(listener==null) listener=new Vector<ActivityListener>();
		synchronized(listener) { listener.add(il); }
	}
	public void removeListener(ActivityListener il){ if(listener!=null) synchronized(listener) { listener.remove(il); }}
	//]

	//[Interface WActivity
	public RActivity reader(){ 
		if(rproxy==null) rproxy=new ActivityProxyReader(this);
		return(rproxy); 
	}
	//[Manipulation des activitées
	public byte run(String activity,Value arg, long delay){ 		
		if(arg!=null) arg=arg.toConstant(); 
		ActivityValue dactivity=findActivity(activity);
//System.out.println("[LocalActivity - run] A");
		if(dactivity==null) return(Ubiquity.NOT_EXIST);
		synchronized(dactivity) {
//System.out.println("[LocalActivity - run] B");		
			cancel(dactivity);//on annule l'éventuelle planification/exécution en cours 
			setActivityMode(dactivity, Ubiquity.ONCE);//on fixe le mode de l'activité sur ONCE
			dactivity.planning=new PlannedActivity(this,dactivity, arg, delay, 0);
		}
		return(Ubiquity.OK); 
	} 	
	public byte repeat(String activity, Value arg, long delay, long periode){ 
		if(arg!=null) arg=arg.toConstant(); 
		if(periode<=0) return(run(activity, arg, delay));
		ActivityValue dactivity=findActivity(activity);
		if(dactivity==null) return(Ubiquity.NOT_EXIST);
		synchronized(dactivity) {
			cancel(dactivity);//on annule l'éventuelle planification/exécution en cours 
			setActivityMode(dactivity, Ubiquity.REPEAT);//on fixe le mode de l'activité sur REPEAT
			dactivity.planning=new PlannedActivity(this,dactivity, arg, delay, periode);
		}
		return(Ubiquity.OK); 
	}
	public byte chain(String activity, Value arg, long delay){ 
		if(arg!=null) arg=arg.toConstant(); 
		ActivityValue dactivity=findActivity(activity);
		if(dactivity==null) {
			return(Ubiquity.NOT_EXIST);
		}
		synchronized(dactivity) {
			if(dactivity.mode==Ubiquity.REPEAT) cancel(dactivity);//si l'activité est en mode REPEAT, on annule l'éventuelle planification/exécution en cours 
			setActivityMode(dactivity, Ubiquity.CHAIN);//on fixe le mode de l'activité sur CHAIN
			if(dactivity.planning==null) dactivity.planning=new PlannedActivity(this,dactivity, arg, delay, 0);
			else dactivity.planning.chain(arg,delay);
		}
		return(Ubiquity.OK);  
	}
	public byte cancel(String activity){
		ActivityValue dactivity=findActivity(activity);
		if(dactivity==null) return(Ubiquity.NOT_EXIST);
		return(cancel(dactivity));
	}	
	//]
	
	protected void cancelAll(){
		if(activities!=null) for(ActivityValue dactivity:activities) if(dactivity.status!=Ubiquity.NOT_RUN) cancel(dactivity);
	}
	
	protected byte cancel(ActivityValue dactivity){
		synchronized(dactivity) {
			if(dactivity.status!=Ubiquity.NOT_RUN){ //=> WAITING ou RUN
				if(dactivity.planning!=null){
					setActivityMode(dactivity, Ubiquity.CANCELED);
					dactivity.planning.cancel();
					dactivity.planning=null;
				}
				setActivityStatus(dactivity, Ubiquity.NOT_RUN);
			}
		}
		return(Ubiquity.OK);  
	}
	public synchronized byte set(String activity, ActivityRunner runner){
		if(!thisParticipant.exist()) return(Ubiquity.NOT_ACTIF);
		if(activity==null) return(Ubiquity.ERROR);
		if(runner==null) return(remove(activity));
	
		ActivityValue existing=findActivity(activity);
		if(existing!=null) existing.runner=runner; //maj du runner sur cette activité (NB: impactera les prochaines exécutions seulement
		else {
			existing=new ActivityValue(activity, runner);
			if(activities==null) activities=new Vector<ActivityValue>();
			activities.add(existing);
			fireAddActivity(activity);
		}
		return(Ubiquity.OK);
	}	
	
	public synchronized byte remove(String activity){
		if(!thisParticipant.exist()) return(Ubiquity.NOT_ACTIF);
		if(activity==null) return(Ubiquity.ERROR);
		ActivityValue existing=findActivity(activity);
		if(existing!=null){
			activities.remove(existing);
			//Au niveau des écouteurs sur l'agent lui même
			if(listener!=null) synchronized(listener) { for(ActivityListener l : listener) ThreadStack.perform( () -> l.activityRemoved(thisParticipant, activity)); }
			//Au niveau du monitor
			thisParticipant.node.space.notifyRemoveActivity(thisParticipant, activity); 
			return(Ubiquity.OK);		
		}
		return(Ubiquity.NOT_EXIST);			
	}
	//]
	
	protected void fireAddActivity(String activity){	//notification locale
		//Au niveau des écouteurs sur le participant lui même
		if(listener!=null) synchronized(listener) { for(ActivityListener l : listener) ThreadStack.perform( () -> l.activityAdded(thisParticipant, activity)); }
		//Au niveau des écouteurs du space
		thisParticipant.node.space.notifyAddActivity(thisParticipant, activity); //on notifie l'ajout sur le spot local
	}

	
	//[changement de mode des activités
	protected void setActivityMode(String nameActivity, int mode){
		setActivityMode(findActivity(nameActivity), mode);
	}
	protected void setActivityMode(ActivityValue dactivity, int mode){
		if(dactivity!=null && dactivity.mode!=mode){
			int old_mode=dactivity.mode;
			dactivity.mode=mode;
			//Notification du changerement au niveau des écouteurs sur le participant lui même
			if(listener!=null) synchronized(listener) { for(ActivityListener l : listener) ThreadStack.perform( () -> l.activityModeChange(thisParticipant, dactivity.name,  (byte) old_mode,  (byte) mode)); }
			//Notification du changerement au niveau du space
			thisParticipant.node.space.notifyActivityChangeMode(thisParticipant, dactivity.name,  (byte) old_mode,  (byte) mode);				
		}
	}
	protected void setActivityStatus(String nameActivity, int status){
		setActivityStatus(findActivity(nameActivity), status);
	}	
	protected void setActivityStatus(ActivityValue dactivity, int status){
		if(dactivity!=null && dactivity.status!=status){
			int old_status=dactivity.status;
			dactivity.status=status;
			//Notification du changerement au niveau des écouteurs sur le participant lui même
			if(listener!=null) synchronized(listener) { for(ActivityListener l : listener) ThreadStack.perform( () -> l.activityStatusChange(thisParticipant, dactivity.name,  (byte) old_status,  (byte) status)); }
			//Notification du changerement au niveau du space
			thisParticipant.node.space.notifyActivityChangeStatus(thisParticipant, dactivity.name,  (byte) old_status,  (byte) status);
		}
	}
	protected void setActivityProgress(String nameActivity,  int currentStep, int maxStep){
		setActivityProgress(findActivity(nameActivity), currentStep, maxStep);
	}		
	protected void setActivityProgress(ActivityValue dactivity, int currentStep, int maxStep){
		if(dactivity!=null && (dactivity.progressStep!=currentStep || dactivity.progressMax!=maxStep)){
			dactivity.progressStep=currentStep;
			dactivity.progressMax=maxStep;
			
			//Notification du changerement au niveau des écouteurs sur le participant lui même
			if(listener!=null) synchronized(listener) { for(ActivityListener l : listener) ThreadStack.perform( () -> l.activityProgressChange(thisParticipant, dactivity.name,  currentStep, maxStep)); }
			//Notification du changerement au niveau du space
			thisParticipant.node.space.notifyActivityProgress(thisParticipant, dactivity.name,  currentStep, maxStep);
		}
	}	
	//]
	
	//[Pour Identity
	protected synchronized void writeCompleteState(DataWriter w){ //(int: nb_entry) {chaque_valeur: (string: name) [status] [mode] (int:progressMax) (int:progressStp)}
		int nb=(activities==null?0:activities.size());
		w.i(nb);
		if(nb>0) for(ActivityValue ac : activities) w.s(ac.name).b(ac.status).b(ac.mode).i(ac.progressMax).i(ac.progressStep);
	}
	protected synchronized Activity readCompleteState(int nb, DataReader dbrb){ 	//{chaque_valeur: (string: name) [status] [mode] (int:progressMax) (int:progressStp)}
		Vector<ActivityValue> old_activities=activities;
		activities=new Vector<ActivityValue>(); 
		ActivityValue nac;
		for(int i=0; i<nb; i++) {
			String name = dbrb.readString();
			nac=null;
			if(old_activities!=null){
				for(int j=0; j<old_activities.size(); j++) {
					ActivityValue cur = old_activities.get(j);
					if(cur.name.equals(name)) {
						nac=cur;
						old_activities.remove(j);
						break;
					}
				}
			}
			if(nac!=null){
				activities.add(nac);
				setActivityStatus(nac, dbrb.readByte());
				setActivityMode(nac, dbrb.readByte());
				setActivityProgress(nac, dbrb.readInt(), dbrb.readInt());
			} else {
				nac=new ActivityValue(name, DistantActivity);
				activities.add(nac);
				nac.status=dbrb.readByte();
				nac.mode=dbrb.readByte();
				nac.progressMax=dbrb.readInt();
				nac.progressStep=dbrb.readInt();
				fireAddActivity(nac.name);
			}
		}
		if(old_activities!=null) for(ActivityValue cur : old_activities){
			//Au niveau des écouteurs sur l'agent lui même
			if(listener!=null) synchronized(listener) { for(ActivityListener l : listener) ThreadStack.perform( () -> l.activityRemoved(thisParticipant, cur.name)); }
			//Au niveau du monitor
			thisParticipant.node.space.notifyRemoveActivity(thisParticipant, cur.name); 			
		}
		old_activities=null;
		return(this);
	}
	protected synchronized void clear(){
		if(activities!=null) synchronized(activities){
			for(ActivityValue dactivity:activities) {
				if(dactivity.status!=Ubiquity.NOT_RUN) cancel(dactivity);
				//Au niveau des écouteurs sur l'agent lui même
				if(listener!=null) synchronized(listener) { for(ActivityListener l : listener) ThreadStack.perform( () -> l.activityRemoved(thisParticipant, dactivity.name)); }
				//Au niveau du monitor
				thisParticipant.node.space.notifyRemoveActivity(thisParticipant, dactivity.name); 
			}
			activities.removeAllElements();
		}
	}
	//]
}

class ActivityValue {
	public String name;
	public int status; //NOT_RUN , WAITING, RUN
	public int mode; //NONE, ONCE , CHAIN, REPEAT, CANCELED
	public int progressMax;
	public int progressStep;
	public ActivityRunner runner;
	public PlannedActivity planning;
	
	public ActivityValue(String n, ActivityRunner r){
		name=n;
		runner=r;
		status=Ubiquity.NOT_RUN;
		mode=Ubiquity.NONE;
		progressMax=0;
		progressStep=0;
		planning=null;
	}
}

//Coquille de lecture seule
class ActivityProxyReader implements RActivity {
	Activity kernel;

	public ActivityProxyReader(Activity k){ kernel=k; }
	
	public Participant participant(){ return(kernel.participant()); }	
	public int nbActivities(){ return(kernel.nbActivities()); }
	public String activityName(int num){ return(kernel.activityName(num)); } //consultation de l'alias public de numéro arbitraire num, null si non existant
	public byte activityStatus(int num){ return(kernel.activityStatus(num)); } //retourne : NOT_EXIST , NOT_RUN , WAITING, RUN, CANCELED
	public byte activityMode(int num){ return(kernel.activityMode(num)); } //retourne : NOT_EXIST , ONCE , CHAIN, REPEAT
	public int activityProgressMax(int num){ return(kernel.activityProgressMax(num)); }
	public int activityProgressStep(int num){ return(kernel.activityProgressStep(num)); }	
	public boolean haveActivity(String activity){ return(kernel.haveActivity(activity)); }
	public byte activityStatus(String activity){ return(kernel.activityStatus(activity)); } //retourne : NOT_EXIST , NOT_RUN , WAITING, RUN, CANCELED
	public byte activityMode(String activity){ return(kernel.activityMode(activity)); } //retourne : NOT_EXIST , ONCE , CHAIN, REPEAT
	public int activityProgressMax(String activity){ return(kernel.activityProgressMax(activity)); }
	public int activityProgressStep(String activity){ return(kernel.activityProgressStep(activity)); }

	public void addListener(ActivityListener sl){ kernel.addListener(sl); }
	public void removeListener(ActivityListener sl){ kernel.removeListener(sl); }	
}