package skuad.ubiquity.implementation;

import java.util.*;
import skuad.ubiquity.space.*;
import skuad.ubiquity.space.alias.*;
import skuad.ubiquity.*;
import skuad.util.databytes.*;
import skuad.util.thread.ThreadStack;

public class Alias implements WAlias {

	ImplParticipant thisParticipant;	//participant associé à ce descripteur Alias
	AliasProxyReader rproxy;			//objet proxy de lecture sur cet instance
	Vector<AliasValue> alias;			//table des alias
	Vector<AliasListener> listener;
	
	protected Alias(ImplParticipant p){
		thisParticipant=p;
		alias=null;
		listener=null;
		rproxy=null;
	}
	
	//[Interface RAlias	
	public Participant participant(){ return(thisParticipant); }	
	public int nbAlias() { return(alias==null?0:alias.size()); }
	public String aliasName(int num){ //consultation de l'alias public de numéro arbitraire num, null si non existant
		if(alias!=null && num<alias.size() && num>=0) return(alias.get(num).name);
		return(null);
	}
	public int aliasPriority(int num){ //consultation de l'alias public de numéro arbitraire num, null si non existant
		if(alias!=null && num<alias.size() && num>=0) return(alias.get(num).priority);
		return(0);
	}	
	public int priority(String alias_name){
		AliasValue existing=findAlias(alias_name);
		if(existing!=null) return(existing.priority);
		return(0);
	}
	public boolean haveAlias(String alias_name){
		return(findAlias(alias_name)!=null);
	}
	public boolean haveFocus(String alias_name){ return(thisParticipant.node.space.haveFocus(thisParticipant, alias_name)); }
	public synchronized void addListener(AliasListener il){ 
		if(listener==null) listener=new Vector<AliasListener>();
		synchronized(listener) { listener.add(il); }
	}
	public void removeListener(AliasListener il){ if(listener!=null) synchronized(listener) { listener.remove(il); }}
	//]
	
	protected AliasValue findAlias(String alias_name){
		if(alias_name!=null && alias!=null) for(AliasValue cur : alias){
			if(alias_name.equals(cur.name)) return(cur);
		}
		return(null);		
	}
	
	//[Interface WAlias
	public RAlias reader(){ 
		if(rproxy==null) rproxy=new AliasProxyReader(this); 
		return(rproxy); 
	}
	
	/**
	 * Définit ou change la priorité d'un alias de l'agent.<br/>
	 * <br/>
	 * Notez que fixer une priorité à 0 produit le même effet que la suppression de l'alias.<br/>
	 * 
	 * @param alias nom de l'alias concerné.
	 * @param newPriority valeur (ou nouvelle valeur si l'alias existe déjà sur cette agent) de priorité.
	 * @return <ul>
	 * 	<li>{@link ubiquity.Ubiquity#OK} si tout est ok.</li>
	 * 	<li>{@link ubiquity.Ubiquity#NOT_ACTIF} si l'agent est supprimé du space.</li>
	 * </ul>
	 */
	public byte setAlias(String alias_name){
		return(setAlias(alias_name, 10));
	}
	public synchronized byte setAlias(String alias_name, int newPriority){
//System.out.println("AA");		
		if(newPriority>32000) newPriority=32000;
		if(!thisParticipant.exist()) return(Ubiquity.NOT_ACTIF);
		if(alias_name==null) return(Ubiquity.ERROR);
		if(newPriority==0) return(removeAlias(alias_name));
		AliasValue existing=findAlias(alias_name);
		if(existing!=null){ //maj de la priorité sur cet alias
			if(existing.priority==newPriority) return(Ubiquity.OK); //pas de changement
			int oldPriority=existing.priority;
			existing.priority=newPriority;
			fireChangeAliasPriority(alias_name, oldPriority, newPriority);
		} else {
			existing=new AliasValue(alias_name, newPriority);
			if(alias==null) alias=new Vector<AliasValue>();
			alias.add(existing);
			fireAddAlias(alias_name, newPriority);
		}
		return(Ubiquity.OK);
	}

	protected void fireChangeAliasPriority(String alias_name, int oldPriority, int newPriority){	//notification locale
		//Au niveau des écouteurs sur l'agent lui même
		if(listener!=null) synchronized(listener) { for(AliasListener l : listener) ThreadStack.perform( () -> l.aliasPriorityChange(thisParticipant, alias_name, oldPriority, newPriority)); }
		//Au niveau du monitor
		thisParticipant.node.space.notifyPriorityAlias(thisParticipant, alias_name, oldPriority, newPriority); //on notifie l'ajout sur le spot local
	}
	protected void fireAddAlias(String alias_name, int newPriority){
		//Au niveau des écouteurs sur l'agent lui même
		if(listener!=null) synchronized(listener) { for(AliasListener l : listener) ThreadStack.perform( () -> l.aliasAdded(thisParticipant, alias_name, newPriority)); } 	
		//Au niveau du space
		thisParticipant.node.space.notifyAddAlias(thisParticipant, alias_name, newPriority); //on notifie l'ajout dans le space
		
	}
	protected void fireRemoveAlias(String alias_name, int oldPriority){
		//Au niveau des écouteurs sur l'agent lui même
		if(listener!=null) synchronized(listener) { for(AliasListener l : listener) ThreadStack.perform( () -> l.aliasRemoved(thisParticipant, alias_name,oldPriority)); } 	
		//Au niveau du monitor
		thisParticipant.node.space.notifyRemoveAlias(thisParticipant, alias_name, oldPriority); //on notifie la suppression dans le space
	}	

	public synchronized byte removeAlias(String alias_name){
		if(!thisParticipant.exist()) return(Ubiquity.NOT_ACTIF);
		if(alias_name==null) return(Ubiquity.ERROR);
		AliasValue existing=findAlias(alias_name);
		if(existing!=null){
			alias.remove(existing);
			fireRemoveAlias(alias_name,existing.priority);	//notification locale			
		}
		return(Ubiquity.OK);		
	}
	//]
	
	protected void fireAliasFocusChange(String alias, boolean oldFocus, boolean newFocus){ 
		if(listener!=null) synchronized(listener) { for(AliasListener l : listener) ThreadStack.perform( () -> l.aliasFocusChange(thisParticipant, alias, oldFocus, newFocus)); } 	
	}
	
	//[Pour Identity
	protected synchronized void writeCompleteState(DataWriter w){ //(int: nb_entry) {chaque_valeur: (string: name) (int: priority)}
			int nb=(alias==null?0:alias.size());
			w.i(nb);
			if(nb>0) for(AliasValue av : alias) w.s(av.name).i(av.priority);
	}
	protected synchronized Alias readCompleteState(int nb, DataReader dbrb){ // {chaque_valeur: (string: name) (int: priority)}
		Vector<AliasValue> old_alias=alias;
		alias=new Vector<AliasValue>();
		AliasValue nav;
		for(int i=0; i<nb; i++) {
			String alias_name=dbrb.readString();
			int priority= dbrb.readInt();
			nav=null;
			if(old_alias!=null){
				for(int j=0; j<old_alias.size(); j++) {
					AliasValue cur = old_alias.get(j);
					if(cur.name.equals(alias_name)) {
						nav=cur;
						old_alias.remove(j);
						break;
					}
				}
			}
			if(nav!=null){
				int oldPriority = nav.priority;
				if(priority != oldPriority){
					if(priority==0) {
						fireRemoveAlias(alias_name,oldPriority);	//notification locale
					} else {
						nav.priority=priority;
						alias.add(nav);
						fireChangeAliasPriority(alias_name, oldPriority, priority);
					}
				} else alias.add(nav);
			} else {
				nav=new AliasValue(alias_name, priority);
				alias.add(nav);
				fireAddAlias(alias_name, priority);
			}						
		}
		if(old_alias!=null) for(AliasValue cur : old_alias){
			fireRemoveAlias(cur.name, cur.priority);
		}	
		old_alias=null;
		return(this);
	}
	protected synchronized void clear(){
		if(alias!=null) {
			for(AliasValue cur : alias) fireRemoveAlias(cur.name, cur.priority);
			alias.removeAllElements();
		}
	}
	//]
}

class AliasValue {
	public String name;
	public int priority;
	
	public AliasValue(String n, int p){
		name=n;
		priority=p;
	}
}

//Coquille de lecture seule
class AliasProxyReader implements RAlias {
	Alias kernel;

	public AliasProxyReader(Alias k){ kernel=k; }
	public Participant participant() { return(kernel.participant()); }
	public int nbAlias() { return(kernel.nbAlias()); }
	public String aliasName(int num){ return(kernel.aliasName(num)); }
	public int aliasPriority(int num){ return(kernel.aliasPriority(num)); }
	public int priority(String alias_name){ return(kernel.priority(alias_name)); }
	public boolean haveFocus(String alias_name){ return(kernel.haveFocus(alias_name)); }
	public boolean haveAlias(String alias_name) { return(kernel.haveAlias(alias_name)); }
	public void addListener(AliasListener il){ kernel.addListener(il); }
	public void removeListener(AliasListener il){ kernel.removeListener(il); }
}
