package skuad.ubiquity.implementation;

import skuad.hubudp.Spot;
import skuad.util.databytes.*;
import java.util.*;
import skuad.util.thread.*;
import skuad.util.value.Value;
import skuad.SKUAD;

public class DistantNode extends Node {
	
	USpot cspot;							//spot sur lequel ce noeud est relié
	boolean initialized;					//true ssi le spot est pleinement initialisé
	Vector<WaitingAction> waitingActions;	//action en attente de traitement
	Runnable runner;
	
	protected DistantNode(ImplSpace vspace, String id_space, USpot vcspot){
		super(vspace, id_space);
		cspot=vcspot;
		initialized=false;
		vspace.addDistantNode(this);
		waitingActions=null;
		runner=null;
	}
	
	protected synchronized void clear(){
		if(allParticipants!=null){
			for(ImplParticipant p : allParticipants.values()){
				if(p.remove()) space.notifyRemoveParticipant(p); //on supprime le participant
			}
			allParticipants.clear();
			participantsId.removeAllElements();
		}
	}
	
	//[Complément à l'interface Spot
	public String getId(){ return(cspot.spot.getIdGlobal()); }
	public boolean isUp(){ return(cspot.isUp); }
	public boolean isLocal(){ return(false); }
	public Spot spot(){ return(cspot.spot); } 
	//]

	protected synchronized void initialise(DataReader mess){ //mess = { etat_complet_du_spot } = (int: nb_lparticipant) { etat_complet_de_chaque_lparticipant }
			// { etat_complet_d1_lparticipant } = : (string: localId) (int: version) {desc_lparticipant}
		if(initialized) return; //sans effet si le spot est déjà initialisé, ce ne sera normalement jamais le cas
		int nb_participant=mess.readInt(); String localId; ImplParticipant iparticipant;			
		if(nb_participant>0){			
			do{ //on extrait chaque participant: (string: localId) (int: version) {desc_lparticipant}
				localId=mess.readString(); int version=mess.readInt();
				//NB: il est impossible que localId existe déjà dans allParticipants, car les participants distants ne peuvent être instancié uniquement qu'au niveau du code ci-dessous
				iparticipant= new ImplParticipant(localId, this);
				addParticipant(iparticipant);
				iparticipant.readCompleteState(version, mess);
				nb_participant--;
			}while(nb_participant>0);
		}
		initialized=true;
	}

	protected synchronized void updateParticipant(String pid, int pversion, DataReader mess){
		ImplParticipant p = getParticipant(pid);
		if(p!=null) {
			if(pversion==-1)//on doit supprimer ce participant
				internalDeleteParticipant(p, true);
			else p.readCompleteState(pversion, mess); //on doit mettre à jour ce participant
		}
	}
	
	protected synchronized void pushWaintingAction(WaitingAction wa){
		if(waitingActions==null) waitingActions = new Vector<WaitingAction>();
		waitingActions.add(wa);
		if(waitingActions.size()==1){ //on doit programmer le processus d'application des actions
			if(runner==null) runner = new Runnable(){
				public void run(){
					performWaitingAction();
				}
			};
			ThreadStack.perform(runner, Protocole.RETRY_APPLY_ACTION_DELAY);
		}
	}
	protected synchronized void performWaitingAction(){
		ImplParticipant participant;
		if(waitingActions!=null && initialized) for(int i=0; i<waitingActions.size(); i++){
			WaitingAction wa = waitingActions.get(i);
			participant=null;
			switch(wa.cmd){
			case Protocole.NOTIFY_ADD_PARTICIPANT:
				if(!existParticipant(wa.pid)) addParticipant(new ImplParticipant(wa.pid, this));
				wa=null;
				break;
			case Protocole.NOTIFY_REMOVE_PARTICIPANT:
				participant = getParticipant(wa.pid);
				if(participant!=null && participant.isNextVersion(wa.version)){
					internalDeleteParticipant(participant, true);
					wa=null;
				}
				break;
			case Protocole.NOTIFY_ADD_SKILL:
				participant = getParticipant(wa.pid);
				if(participant!=null && participant.isNextVersion(wa.version)){
					participant.addSkill(wa.st_arg);
					wa=null;
				}				
				break;
			case Protocole.NOTIFY_REMOVE_SKILL:
				participant = getParticipant(wa.pid);
				if(participant!=null && participant.isNextVersion(wa.version)){
					participant.removeSkill(wa.st_arg);
					wa=null;
				}				
				break;		
			case Protocole.NOTIFY_ALIAS_PRIORITY:
				participant = getParticipant(wa.pid);
				if(participant!=null && participant.isNextVersion(wa.version)){
					participant.setAlias(wa.st_arg, (short) wa.i_arg);
					wa=null;
				}				
				break;	
			case Protocole.NOTIFY_REMOVE_ALIAS:
				participant = getParticipant(wa.pid);
				if(participant!=null && participant.isNextVersion(wa.version)){
					participant.removeAlias(wa.st_arg);
					wa=null;
				}				
				break;		
			case Protocole.NOTIFY_STATE_CHANGE:
				participant = getParticipant(wa.pid);
				if(participant!=null && participant.isNextVersion(wa.version)){
					participant.stateChange(wa.st_arg, wa.v_arg);
					wa=null;
				}				
				break;	
			case Protocole.NOTIFY_ADD_ACTIVITY:
				participant = getParticipant(wa.pid);
				if(participant!=null && participant.isNextVersion(wa.version)){
					participant.addActivity(wa.st_arg);
					wa=null;
				}				
				break;				
			case Protocole.NOTIFY_REMOVE_ACTIVITY:
				participant = getParticipant(wa.pid);
				if(participant!=null && participant.isNextVersion(wa.version)){
					participant.removeActivity(wa.st_arg);
					wa=null;
				}		
				break;
			case Protocole.NOTIFY_ACTIVITY_CHANGE_MODE:
				participant = getParticipant(wa.pid);
				if(participant!=null && participant.isNextVersion(wa.version)){
					participant.activityChangeMode(wa.st_arg, (byte) wa.i_arg);
					wa=null;
				}		
				break;
			case Protocole.NOTIFY_ACTIVITY_CHANGE_STATUS:
				participant = getParticipant(wa.pid);
				if(participant!=null && participant.isNextVersion(wa.version)){
					participant.activityChangeStatus(wa.st_arg, (byte) wa.i_arg);
					wa=null;
				}		
				break;
			case Protocole.NOTIFY_ACTIVITY_CHANGE_PROGRESS:
				participant = getParticipant(wa.pid);
				if(participant!=null && participant.isNextVersion(wa.version)){
					participant.activityChangeProgress(wa.st_arg, wa.i_arg, wa.i2_arg);
					wa=null;
				}						
				break;
			case Protocole.POST_MESSAGE:
				participant = getParticipant(wa.pid);
				if(participant!=null){
					space.localNode.internalReceiveMessage(participant, wa.st_arg, wa.st2_arg, wa.v_arg, wa.st3_arg);
					wa=null;
				}										
				break;
			case Protocole.SENDING_REPORT: //id_sender, context, id_report, code
				participant = getParticipant(wa.pid);
				if(participant!=null){
					space.localNode.internalSendingReport(wa.st_arg, participant, wa.st2_arg, wa.i_arg);
					wa=null;
				}														
				break;
			}
			if(wa!=null){ //l'action n'a pas été traité
				wa.count++;
				if(wa.count>Protocole.MAX_RETRY_APPLY_ACTION){ //le nombre maximum de tentative d'application a été atteint... on est dans une situation critique de désynchronisation...
					//... on va donc demander l'actualisation de ce participant à l'hote distant
					waitingActions.removeAllElements();
					cspot.spot.send(SKUAD.UBIQUITY, DataBytes.writer().b(Protocole.GET_PARTCIPANT_STATE).s(id_space).s(wa.pid).flushReader()); //[GET_PARTCIPANT_STATE] (string: idSpace_receveur ) (string: localId)
				}
			} else { //l'action a été traité
				waitingActions.remove(i); 
				i=-1; //on recommence le traitement de la liste au début
			}
		}
		if(waitingActions!=null && waitingActions.size()>0) ThreadStack.perform(runner, Protocole.RETRY_APPLY_ACTION_DELAY); //on doit à nouveau programmer le processus d'application des actions
	}
	
	protected synchronized void internalAddParticipant(String local_id){ //fait suite à une notification NOTIFY_ADD_PARTICIPANT du site distant
		if(initialized){ //initialisation ok, donc on ajoute le participant maintenant
			if(!existParticipant(local_id)){ //aucune action si le participant existe déjà
				addParticipant(new ImplParticipant(local_id, this));
			}
		} else { //on doit mettre en attente cette demande de création, le temps que le spot soit initialisé...
			pushWaintingAction(new WaitingAction(-1, Protocole.NOTIFY_ADD_PARTICIPANT, local_id));
		}
	}

	protected synchronized void internalRemoveParticipant(String local_id, int version){ //fait suite à une notification NOTIFY_REMOVE_PARTICIPANT du site distant
		if(initialized){ //initialisation ok, donc on peut agir maintenant
			ImplParticipant p = getParticipant(local_id);
			if(p!=null){
				if(p.isNextVersion(version)){
					internalDeleteParticipant(p, true);
					return;
				} else if(p.notFuturVersion(version)) return;
				//sinon on doit programmer une nouvelle tentative différé
			} else return; //aucun participant avec cet identifiant à supprimer
		}//sinon on doit mettre en attente cette demande de suppression, le temps que le spot soit initialisé...
		
		pushWaintingAction(new WaitingAction(version, Protocole.NOTIFY_REMOVE_PARTICIPANT, local_id));	
	}
	
	protected synchronized void internalAddSkill(String local_id, int version, String name_skill){
		if(initialized){ //initialisation ok, donc on peut agir maintenant
			ImplParticipant p = getParticipant(local_id);
			if(p!=null){
				if(p.isNextVersion(version)){
					p.addSkill(name_skill);
					return;
				} else if(p.notFuturVersion(version)) return;
				//sinon on doit programmer une nouvelle tentative différé
			}
		}//sinon on doit mettre en attente cette demande
		
		pushWaintingAction(new WaitingAction(version, Protocole.NOTIFY_ADD_SKILL, local_id, name_skill));			
	}
	
	protected synchronized void internalRemoveSkill(String local_id, int version, String name_skill){
		if(initialized){ //initialisation ok, donc on peut agir maintenant
			ImplParticipant p = getParticipant(local_id);
			if(p!=null){
				if(p.isNextVersion(version)){
					p.removeSkill(name_skill);
					return;
				} else if(p.notFuturVersion(version)) return;
				//sinon on doit programmer une nouvelle tentative différé
			}
		}//sinon on doit mettre en attente cette demande
		
		pushWaintingAction(new WaitingAction(version, Protocole.NOTIFY_REMOVE_SKILL, local_id, name_skill));			
	}	
	
	protected synchronized void  internalAliasPriority(String local_id, int version, String name_alias, int new_priority){
		if(initialized){ //initialisation ok, donc on peut agir maintenant
			ImplParticipant p = getParticipant(local_id);
			if(p!=null){
				if(p.isNextVersion(version)){
					p.setAlias(name_alias, (short) new_priority);
					return;
				} else if(p.notFuturVersion(version)) return;
				//sinon on doit programmer une nouvelle tentative différé
			}
		}//sinon on doit mettre en attente cette demande
		
		pushWaintingAction(new WaitingAction(version, Protocole.NOTIFY_ALIAS_PRIORITY, local_id, name_alias, new_priority));		
	}
	
	protected synchronized void  internalRemoveAlias(String local_id, int version, String name_alias){
		if(initialized){ //initialisation ok, donc on peut agir maintenant
			ImplParticipant p = getParticipant(local_id);
			if(p!=null){
				if(p.isNextVersion(version)){
					p.removeAlias(name_alias);
					return;
				} else if(p.notFuturVersion(version)) return;
				//sinon on doit programmer une nouvelle tentative différé
			}
		}//sinon on doit mettre en attente cette demande
		
		pushWaintingAction(new WaitingAction(version, Protocole.NOTIFY_REMOVE_ALIAS, local_id, name_alias));		
	}
	
	protected synchronized void internalStateChange(String local_id, int version, String name_field, Value value){
		if(initialized){ //initialisation ok, donc on peut agir maintenant
			ImplParticipant p = getParticipant(local_id);
			if(p!=null){
				if(p.isNextVersion(version)){
					p.stateChange(name_field, value);
					return;
				} else if(p.notFuturVersion(version)) return;
				//sinon on doit programmer une nouvelle tentative différé
			}
		}//sinon on doit mettre en attente cette demande
		
		pushWaintingAction(new WaitingAction(version, Protocole.NOTIFY_STATE_CHANGE, local_id, name_field, value));		
	}
	
	protected synchronized void internalAddActivity(String local_id, int version, String name_activity){
		if(initialized){ //initialisation ok, donc on peut agir maintenant
			ImplParticipant p = getParticipant(local_id);
			if(p!=null){
				if(p.isNextVersion(version)){
					p.addActivity(name_activity);
					return;
				} else if(p.notFuturVersion(version)) return;
				//sinon on doit programmer une nouvelle tentative différé
			}
		}//sinon on doit mettre en attente cette demande
		
		pushWaintingAction(new WaitingAction(version, Protocole.NOTIFY_ADD_ACTIVITY, local_id, name_activity));			
	}
	
	protected synchronized void internalRemoveActivity(String local_id, int version, String name_activity){
		if(initialized){ //initialisation ok, donc on peut agir maintenant
			ImplParticipant p = getParticipant(local_id);
			if(p!=null){
				if(p.isNextVersion(version)){
					p.removeActivity(name_activity);
					return;
				} else if(p.notFuturVersion(version)) return;
				//sinon on doit programmer une nouvelle tentative différé
			}
		}//sinon on doit mettre en attente cette demande
		
		pushWaintingAction(new WaitingAction(version, Protocole.NOTIFY_REMOVE_ACTIVITY, local_id, name_activity));			
	} 
	
	protected synchronized void internalChangeModeActivity(String local_id, int version, String name_activity, byte mode){
		if(initialized){ //initialisation ok, donc on peut agir maintenant
			ImplParticipant p = getParticipant(local_id);
			if(p!=null){
				if(p.isNextVersion(version)){
					p.activityChangeMode(name_activity, mode);
					return;
				} else if(p.notFuturVersion(version)) return;
				//sinon on doit programmer une nouvelle tentative différé
			}
		}//sinon on doit mettre en attente cette demande
		
		pushWaintingAction(new WaitingAction(version, Protocole.NOTIFY_ACTIVITY_CHANGE_MODE, local_id, name_activity, mode));			
	} 
	
	protected synchronized void internalChangeStatusActivity(String local_id, int version, String name_activity,byte status){
		if(initialized){ //initialisation ok, donc on peut agir maintenant
			ImplParticipant p = getParticipant(local_id);
			if(p!=null){
				if(p.isNextVersion(version)){
					p.activityChangeStatus(name_activity, status);
					return;
				} else if(p.notFuturVersion(version)) return;
				//sinon on doit programmer une nouvelle tentative différé
			}
		}//sinon on doit mettre en attente cette demande
		
		pushWaintingAction(new WaitingAction(version, Protocole.NOTIFY_ACTIVITY_CHANGE_STATUS, local_id, name_activity, status));			
	} 				
	
	protected synchronized void internalChangeProgress(String local_id, int version, String name_activity, int current, int max){
		if(initialized){ //initialisation ok, donc on peut agir maintenant
			ImplParticipant p = getParticipant(local_id);
			if(p!=null){
				if(p.isNextVersion(version)){
					p.activityChangeProgress(name_activity, current, max);
					return;
				} else if(p.notFuturVersion(version)) return;
				//sinon on doit programmer une nouvelle tentative différé
			}
		}//sinon on doit mettre en attente cette demande
		
		pushWaintingAction(new WaitingAction(version, Protocole.NOTIFY_ACTIVITY_CHANGE_PROGRESS, local_id, name_activity, current, max));			
	}
	
	protected synchronized void internalReceiveMessage(String id_sender, String id_target, String context, Value mess, String id_report){
		if(initialized){ //initialisation ok, donc on peut agir maintenant
			ImplParticipant sender = getParticipant(id_sender);
			if(sender!=null){
				space.localNode.internalReceiveMessage(sender, id_target, context, mess, id_report);
				return;
			} else cspot.spot.send(SKUAD.UBIQUITY, DataBytes.writer().b(Protocole.GET_PARTCIPANT_STATE).s(id_space).s(id_sender).flushReader()); //[GET_PARTCIPANT_STATE] (string: idSpace_receveur ) (string: localId)
		}//sinon on doit mettre en attente cette demande
		
		pushWaintingAction(new WaitingAction(0, Protocole.POST_MESSAGE, id_sender, id_target, context, mess, id_report));	
	}
	
	protected synchronized void internalSendingReport(String id_sender, String id_target, String id_report, int code){
		if(initialized){ //initialisation ok, donc on peut agir maintenant
			ImplParticipant receiver = getParticipant(id_target);
			if(receiver!=null){
				space.localNode.internalSendingReport(id_sender, receiver, id_report, code);
				return;
			} else cspot.spot.send(SKUAD.UBIQUITY, DataBytes.writer().b(Protocole.GET_PARTCIPANT_STATE).s(id_space).s(id_target).flushReader()); //[GET_PARTCIPANT_STATE] (string: idSpace_receveur ) (string: localId)
		}//sinon on doit mettre en attente cette demande
		
		pushWaintingAction(new WaitingAction(0, Protocole.SENDING_REPORT, id_target, id_sender, id_report, code));			
	}
//]
}

class WaitingAction { //action en attente de traitement
	int count;
	int version;
	int cmd;
	String pid, st_arg, st2_arg, st3_arg;
	int i_arg, i2_arg;
	Value v_arg;
	
	public WaitingAction(int vversion, int vcmd, String vpid){
		count=0;
		version = vversion;
		cmd = vcmd;
		pid = vpid;
	}
	
	public WaitingAction(int vversion, int vcmd, String vpid, String arg){
		count=0;
		version = vversion;
		cmd = vcmd;
		pid = vpid;
		st_arg = arg;
	}

	public WaitingAction(int vversion, int vcmd, String vpid, String arg, int vi_arg){
		count=0;
		version = vversion;
		cmd = vcmd;
		pid = vpid;
		st_arg = arg;
		i_arg = vi_arg;
	}
	
	public WaitingAction(int vversion, int vcmd, String vpid, String arg, String arg2, int vi_arg){
		count=0;
		version = vversion;
		cmd = vcmd;
		pid = vpid;
		st_arg = arg;
		st2_arg = arg2;
		i_arg = vi_arg;
	}
	
	public WaitingAction(int vversion, int vcmd, String vpid, String arg, Value vv_arg){
		count=0;
		version = vversion;
		cmd = vcmd;
		pid = vpid;
		st_arg = arg;
		v_arg = vv_arg;
	}
	public WaitingAction(int vversion, int vcmd, String vpid, String arg,  String arg2, Value vv_arg, String arg3){
		count=0;
		version = vversion;
		cmd = vcmd;
		pid = vpid;
		st_arg = arg;
		v_arg = vv_arg;
		st2_arg = arg2;
		st3_arg = arg3;
	}
	public WaitingAction(int vversion, int vcmd, String vpid, String arg, int vi_arg,  int vi2_arg){
		count=0;
		version = vversion;
		cmd = vcmd;
		pid = vpid;
		st_arg = arg;
		i_arg = vi_arg;
		i2_arg = vi2_arg;
	}
}