package skuad.ubiquity.implementation;

import java.util.Vector;
import skuad.ubiquity.space.*;
import skuad.ubiquity.space.mailbox.*;
import skuad.util.thread.ThreadStack;
import skuad.util.value.Value;

public class ImplMailBox implements MailBox {

	ImplParticipant thisParticipant;				//Participant associé à cette mailBox
	Vector<MessageListener> messageListener;
	Vector<SendReportListener> reportListener;
	Vector<Message> messages;
	
	protected ImplMailBox(ImplParticipant p){
		thisParticipant=p;
		messageListener=null;
		reportListener=null;
		messages=null;
	}	
	
	//[Interface MailBox
	public Participant participant(){ return(thisParticipant); }
	
	public boolean haveMessage(){
		return(messages!=null && !messages.isEmpty());
	}
	public int nbMessage(){
		return(messages==null?0:messages.size());
	}
	public Message get(int num){
		return(messages==null || num<0 || num>=messages.size()?null:messages.get(num));
	}
	public Message pop(){
		if(messages==null ||0==messages.size()) return(null);
		Message mess=messages.firstElement();
		messages.remove(0);
		return(mess);
	}
	public synchronized void addListener(MessageListener il){ 
		if(messageListener==null) messageListener=new Vector<MessageListener>();
		synchronized(messageListener) { messageListener.add(il); }
	}
	public void removeListener(MessageListener il){ if(messageListener!=null) synchronized(messageListener) { messageListener.remove(il); }}	

	public synchronized void addReporter(SendReportListener sl){
		if(reportListener==null) reportListener=new Vector<SendReportListener>();
		synchronized(reportListener) { reportListener.add(sl); }
	}
	public void removeReporter(SendReportListener sl){ if(reportListener!=null) synchronized(reportListener) { reportListener.remove(sl); }}
	//]
	
	//[
	//[Envoie de messages
	/**
	 * Envoi d'un message via une désignation de la cible par identifiant.<br/>
	 * <br/>
	 * Voir l'aide générale de cette classe pour plus de détails.
	 * 
	 * @param target Identifiant de l'agent destinataire du message.
	 * @param skill_context Contexte de compétence relatif à ce message.
	 * @param message Contenu du message
	 * @return <ul>
	 * 	<li>{@link ubiquity.Ubiquity#OK} si tout est ok.</li>
	 * 	<li>{@link ubiquity.Ubiquity#NOT_EXIST}  si l'agent ciblé n'est plus actif.</li> 
	 * </ul>
	 */
	public byte send(Participant target, String skill_context, Value message){
		return(thisParticipant.node.space.send(thisParticipant, target, skill_context, message, null));
	}

	/**
	 * Envoi d'un message via une désignation de la cible par identifiant et un demande de rapport de transmission.<br/>
	 * <br/>
	 * Voir l'aide générale de cette classe pour plus de détails.
	 * 
	 * @param target Identifiant de l'agent destinataire du message.
	 * @param skill_context Contexte de compétence relatif à ce message.
	 * @param message Contenu du message
	 * @param idreport Clé d'identification pour le rapport de transmission.
	 * @return <ul>
	 * 	<li>{@link ubiquity.Ubiquity#OK} si tout est ok.</li>
	 *  <li>{@link ubiquity.Ubiquity#NOT_ACTIF} si l'agent expéditeur est supprimé du space.</li>
	 * 	<li>{@link ubiquity.Ubiquity#NOT_EXIST}  si l'agent ciblé n'est plus actif.</li> 
	 * </ul>
	 * 
	 * @see Ubiquity.FullParticipantAdpator.FullBodyListener#sendingReport(Interactor, RIdentity, String, byte)
	 */
	public byte send(Participant target, String skill_context, Value message, String idreport){
		return(thisParticipant.node.space.send(thisParticipant, target, skill_context, message, idreport));
	}
	
	/**
	 * Envoi d'un message via une désignation de la cible par une chaine de ciblage.<br/>
	 * <br/>
	 * Voir l'aide générale de cette classe pour plus de détails.
	 * 
	 * @param target Chaine de ciblage indiquant le ou les destinataires du message.
	 * @param skill_context Contexte de compétence relatif à ce message.
	 * @param message Contenu du message
	 * @return <ul>
	 * 	<li>{@link ubiquity.Ubiquity#OK} si tout est ok.</li>
	 * 	<li>{@link ubiquity.Ubiquity#NOT_EXIST} si la chaine de ciblage ne désigne aucun agent actif.</li> 
	 * </ul>
	 */	
	public byte send(String target, String skill_context, Value message){
		return(thisParticipant.node.space.send(thisParticipant, target, skill_context, message, null));
	}
	
	/**
	 * Envoi d'un message via une désignation de la cible par une chaine de ciblage et une demande de rapport de transmission.<br/>
	 * <br/>
	 * Voir l'aide générale de cette classe pour plus de détails.
	 * 
	 * @param target Chaine de ciblage indiquant le ou les destinataires du message.
	 * @param skill_context Contexte de compétence relatif à ce message.
	 * @param message Contenu du message
	 * @param idreport Clé d'identification pour le rapport de transmission.
	 * @return <ul>
	 * 	<li>{@link ubiquity.Ubiquity#OK} si tout est ok.</li>
	 * 	<li>{@link ubiquity.Ubiquity#NOT_EXIST} si la chaine de ciblage ne désigne aucun agent actif.</li> 
	 * </ul>
	 * 
	 * @see Ubiquity.FullParticipantAdpator.FullBodyListener#sendingReport(Interactor, RIdentity, String, byte)
	 */
	public byte send(String target, String skill_context, Value message, String idreport){
		return(thisParticipant.node.space.send(thisParticipant, target, skill_context, message, idreport));
	}
	//]	
	
	//[Méthode pour LocalIdentity
	protected void deliverMessage(Message mess){ 
		if(messages==null) messages=new Vector<Message>();
		messages.add(mess);
		
		//Notification des écouteurs
		if(messageListener!=null) synchronized(messageListener) { for(MessageListener l : messageListener) ThreadStack.perform( () -> l.receive(this, mess)); }
	}	
	protected void eraseMessage(Message mess){ 
		if(messages==null) messages.remove(mess);
	}
	protected void notifySendingReport(Avatar hags, Participant receiver, String idReport, byte code){ 
		//Notification des écouteurs
		if(reportListener!=null) synchronized(reportListener) { for(SendReportListener l : reportListener) ThreadStack.perform( () -> l.sendingReport(hags, receiver, idReport, code)); }	
	}
	//]
}
