package skuad.ubiquity.implementation;

import skuad.ubiquity.space.*;
import skuad.ubiquity.space.activity.*;
import skuad.ubiquity.space.alias.*;
import skuad.ubiquity.space.skill.*;
import skuad.ubiquity.space.state.*;
import skuad.hubudp.*;
import skuad.ubiquity.Ubiquity;
import skuad.util.databytes.*;
import skuad.util.thread.ThreadStack;
import skuad.util.value.Value;

import java.util.*;

public class ImplParticipant implements Participant {

	Node node;			//node porteur de ce participant
	Vector<ParticipantListener> listener;
	boolean exist;
	int version;		//version du participant, dans l'interval [0 ; 65500]
	
	String localId;
	Alias alias;
	Skill skill;
	State state;
	Activity activity;
	
	protected ImplParticipant(String lid, Node v_node){
		localId=lid;
		node=v_node;
		listener=null;
		alias=null;
		skill=null;
		state=null;
		activity=null;
		exist=true;
		version=0;
	}
	
	protected synchronized int incAndGetVersion(){ //incrémente le numéro de version et retourne sa nouvelle valeur
		version++;
		if(version>65500) version=1;
		return(version);
	}
	
	protected boolean isNextVersion(int vversion){ //si vversion est bien le prochaine numéro de version, le numéro de version prendra aussi la valeur de vversion 
		if(vversion==(version+1) || (version==65500  && vversion==1)){
			version = vversion;
			return(true);
		} else return(false);
	}
	
	protected boolean notFuturVersion(int vversion){ //true si vversion n'est pas une future version pour ce participant (vversion est une furture version s'il est l'une des 50 valeurs suivant celle de version)
		if((version>65450 && vversion<50) || (vversion>version && (vversion-version)<50)) return(false);
		return(true);
	}
	
	protected WAlias getWAlias(){
		if(alias==null) alias=new Alias(this);
		return(alias);
	}	
	protected WState getWState(){
		if(state==null) state=new State(this);
		return(state); 
	}
	protected WSkill getWSkill(){
		if(skill==null) skill=new Skill(this);
		return(skill); 
	}
	protected WActivity getWActivity(){
		if(activity==null) activity=new Activity(this, null);
		return(activity); 
	}	
	
	//[pour DistantSpot et LocalSpot
	protected boolean remove(){
		if(exist){
			exist=false;
			if(activity!=null) activity.clear();
			if(alias!=null) alias.clear();
			if(skill!=null) skill.clear();
			if(state!=null) state.clear();
			if(listener!=null) synchronized(listener) { for( ParticipantListener l : listener) ThreadStack.perform( () -> l.participantRemoved(this)); } //on informe les écouteurs au niveau participant
			return(true);
		}
		return(false); //agent déjà détruit !
	}
	protected synchronized void writeCompleteState(DataWriter w){ // { etat_complet_d1_lparticipant } = (string: localId) (int: version) {desc_lparticipant}
		w.s(localId).i(version);
	
		//{desc_lparticipant_vg} = [(int: nb_activity) {valeur des activity} [(int: nb_alias) {valeur des alias}  (int: nb_skill) {valeur des skill}  (int: nb_state) {valeur des state}
		if(activity!=null) activity.writeCompleteState(w); else w.i(0); 	//injection des activity
		if(alias!=null) alias.writeCompleteState(w); else w.i(0);			//injection des alias
		if(skill!=null) skill.writeCompleteState(w); else w.i(0);			//injection des skill
		if(state!=null) state.writeCompleteState(w); else w.i(0);			//injection des state
	}
	protected synchronized void readCompleteState(int vversion, DataReader dbrb){  // {desc_agent} = (int: nb_activity) {valeur des activity} (int: nb_alias) {valeur des alias}  (int : nb_skill) {valeur des skill}  (int :nb_state) {valeur des state}		
		version=-1; //afin de bloquer toute action sur ce participant
		int nb_entry;
		nb_entry=dbrb.readInt(); if(nb_entry>0) { //lecture des activity
			if(activity==null) getWActivity();
			activity.readCompleteState(nb_entry, dbrb);  	
		} else if(activity!=null) activity.clear();
		nb_entry=dbrb.readInt(); if(nb_entry>0) { //lecture des alias
			if(alias==null) getWAlias();
			alias.readCompleteState(nb_entry, dbrb); 			
		} else if(alias!=null) alias.clear();
		nb_entry=dbrb.readInt(); if(nb_entry>0) { //lecture des skill
			if(skill==null) getWSkill();
			skill.readCompleteState(nb_entry, dbrb); 			
		} else if(skill!=null) skill.clear();
		nb_entry=dbrb.readInt(); if(nb_entry>0) { //lecture des state
			if(state==null) getWState();
			state.readCompleteState(nb_entry, dbrb); 			
		} else if(state!=null) state.clear();
		version = vversion;
	}
	
	protected void addSkill(String nameSkill){
		if(skill==null) getWSkill();
		skill.add(nameSkill);
	}
	protected void removeSkill(String nameSkill){
		if(skill!=null) skill.remove(nameSkill);
	}
	protected void setAlias(String nameAlias, short priority){
		if(alias==null) getWAlias();
		alias.setAlias(nameAlias, priority);
	}
	protected void removeAlias(String nameAlias){
		if(alias!=null) alias.removeAlias(nameAlias);
	}
	protected void stateChange(String field, Value value){
		if(state==null) getWState();
		state.set(field, value);
	}
	protected void addActivity(String nameActivity){
		if(activity==null) getWActivity();
		activity.set(nameActivity, Activity.DistantActivity);
	}
	protected void removeActivity(String nameActivity){
		if(activity!=null) activity.remove(nameActivity);
	}
	protected void activityChangeMode(String nameActivity, byte newMode){
		if(activity!=null) activity.setActivityMode(nameActivity, newMode); 
	}
	protected void activityChangeStatus(String nameActivity, byte newStatus){
		if(activity!=null) activity.setActivityStatus(nameActivity, newStatus); 
	}
	protected void activityChangeProgress(String nameActivity, int currentStep, int maxStep){
		if(activity!=null) activity.setActivityProgress(nameActivity, currentStep, maxStep); 
	}	
	//]
	
	//[interface Participant
	public Space getSpace(){ //null si le space est protégé par un mot de passe, il faut alors utiliser getSpace(String)
		if(node.space.havePassword()) return(null);
		return(node.space);
	}
	public Space getSpace(String password){
		if(node.space.isThisPassword(password)) return(node.space);
		else return(null);
	}

	public Spot spot(){
		return(node.spot());
	}
	
	/**
	 * Retourne l'état d'existance actuel de l'agent dans le space.<br/>
	 * <br/>
	 * 	<li><b>true</b> si l'agent est enregistré et actif dans le space.</li>
	 * 	<li><b>false</b> si l'agent n'existe pas encore, ou n'existe plus.</li>
	 * </ul>
	 */
	public boolean exist(){ 
		return(exist); 
	}
	
	public boolean ordoredBefore(Participant ia){ return(fullId().compareTo(ia.fullId())<0);}
	public boolean ordoredAfter(Participant ia){ return(fullId().compareTo(ia.fullId())>0);} 
	
	/**
	 * Test si l'agent est un agent local ou un agant distant.<br/>
	 * <br/>
	 * 	<li><b>true</b> si l'agent est présent physiquement dans le contexte local.</li>
	 * 	<li><b>false</b> si l'agent est instancié dans un contexte distant.</li>
	 * </ul>
	 */
	public boolean isLocal(){ return(node.isLocal()); }
	
	/**
	 * Retourne l'identifiant local de l'agent sur son node porteur.<br/>
	 * <br/>
	 * Cette valeur est celle attribuée à l'agent au moment de son enregistrement sur son node local. C'est une valeur qui est unique sur un noeud donnée (deux agents situés
	 * sur un même node ne peuvent avoir la même valeur d'identifiant local).<br/>
	 * 
	 * @return Valeur de la chaîne identifiant l'agent localement sur son node.
	 */
	public String localId(){ return(localId); }

	/**
	 * L'identifiant global de ce participant dans le Space.<br/>
	 * <br/>
	 * Cette chaine est produite de la façon suivante :  <i>id local de l'agent</i>@<i>global_id de l'hote porteur</i>.<br/>
	 * Ainsi, par construction, cette chaine permet de désigner un agent unique sur l'ensemble du space.
	 * 
	 * @return l'identifiant global de l'agent dans le Space.
	 */
	public String fullId(){ return(localId+"@"+node.getId()); }
	
	/**
	 * Retourne le descripteur en lecture de la facette fonctionnalité de l'agent.<br/>
	 * 
	 * @return le descripteur en lecture de la facette fonctionnalité de l'agent.
	 */
	public RSkill getSkill(){ 
		if(!exist) return(null); //on n'est pas dans une zone de visibilité pour cet agent
		if(skill==null) getWSkill(); //on s'assure que l'instance skill existe bien
		return(skill.reader()); 
	}
	
	/**
	 * Retourne le descripteur en lecture de la facette state de l'agent.<br/>
	 * 
	 * @return le descripteur en lecture de la facette state de l'agent.
	 */	
	public RState getState(){ 
		if(!exist) return(null); //on n'est pas dans une zone de visibilité pour cet agent	
		if(state==null) getWState();
		return(state.reader()); 
	}
	
	/**
	 * Retourne le descripteur en lecture de la facette alias de l'agent.<br/>
	 * 
	 * @return le descripteur en lecture de la facette alias de l'agent.
	 */	
	public RAlias getAlias(){ 
		if(!exist) return(null); //on n'est pas dans une zone de visibilité pour cet agent
		if(alias==null) getWAlias();
		return(alias.reader()); 
	}
	
	/**
	 * Retourne le descripteur en lecture de la facette activity de l'agent.<br/>
	 * 
	 * @return le descripteur en lecture de la facette activity de l'agent.
	 */	
	public RActivity getActivity(){ 
		if(!exist) return(null); //on n'est pas dans une zone de visibilité pour cet agent
		if(activity==null) getWActivity();
		return(activity.reader()); 
	}
	
	/**
	 * Ajout d'un écouteur de suppression locale à cet agent.<br/>
	 * L'écouteur sera ici notifié que pour la suppression de cet agent, pour une écoute plus globale (portant sur tous les agents du space),
	 * il faut passer par l'objet monitor et utiliser la méthode {@link Ubiquity.Monitor#addAgentListener(AgentListener)}.
	 * 
	 * @param sl Objet à notifier lorsque l'agent sera supprimé du space.
	 */
	public synchronized void addListener(ParticipantListener sl){ 
		if(listener==null) listener=new Vector<ParticipantListener>();
		synchronized(listener) { listener.add(sl); }
	}
	
	/**
	 * Désenregistrement d'un écouteur de suppression locale à cet agent.<br/>
	 * 
	 * @param sl Objet spécifié sur l'appel {@link #addListener(AgentListener)} qu'il faut annuler ici.
	 */
	public void removeListener(ParticipantListener sl){  if(listener!=null) synchronized(listener) { listener.remove(sl); }}
	//]
}