package skuad.ubiquity.implementation;

import java.util.*;
import skuad.ubiquity.*;
import skuad.ubiquity.space.*;
import skuad.ubiquity.space.activity.*;
import skuad.ubiquity.space.alias.*;
import skuad.ubiquity.space.skill.*;
import skuad.ubiquity.space.state.*;
import skuad.util.thread.*;
import skuad.util.value.Value;
import skuad.util.databytes.*;
import skuad.util.base.*;
import skuad.SKUAD;
import skuad.hubudp.*;

/**
 * Classe de gestion et d'implémentation des spaces.
 * 
 * @author denis
 */
public class ImplSpace implements Space {
	
	String name, password, spaceid; 
	Map<String, Vector<ImplParticipant>> aliases;	//table des alias de ce space (alias -> liste des identity portant cet alias au niveau local ou global, ordonnée par ordre décroissant de priority)

	LocalNode localNode; //node local
	Vector<DistantNode> distant_nodes;	//liste de tous les spots distant relié à ce space
	
	//Les écouteurs
	Vector<ParticipantListener> identityListener=null;
	Vector<AliasListener> aliasListener=null;
	Vector<SkillListener> skillListener=null;
	Vector<ActivityListener> activityListener=null;
	Vector<StateListener> stateListener=null;
	
	public ImplSpace(Hub hub, String vspaceid, String vname, String vpassword){
		spaceid = vspaceid;
		name=vname;
		password=vpassword;
		aliases=new HashMap<String, Vector<ImplParticipant>>();
		distant_nodes = new Vector<DistantNode>();
		localNode = new LocalNode(hub, this, vspaceid);
	}
	public ImplSpace(Spot spot, String vname, String vpassword){ //space purement local virtuellement logé sur le spot "spot"
		spaceid = null;
		name=vname;
		password=vpassword;
		aliases=new HashMap<String, Vector<ImplParticipant>>();
		distant_nodes = new Vector<DistantNode>();
		localNode = new LocalNode(spot, this, null);
	}
	
	public void destroy() {
		aliases=null;
		distant_nodes =null;
		localNode =null;
		identityListener=null;
		aliasListener=null;
		skillListener=null;
		activityListener=null;
		stateListener=null;
	}
	
	//pour DistantNode
	protected void addDistantNode(DistantNode dp){
		synchronized(distant_nodes) { distant_nodes.add(dp); }
	}
		
	//[Pour ImplParticipant et UbiquityManager
	public boolean isThisPassword(String test_password){
		if(password==null) return(test_password==null);
		else return(password.equals(test_password));
	}

	public boolean isThisPassword(String crypted_password, String key){
		if(password==null) return(crypted_password==null);
		else return(Util.cryptMD5(password, key).equals(crypted_password));
	}
	//]
	
	//[Interface Space
		//Information général sur le Space auquel ce spot est rataché
	public String nameSpace(){ return(name); }
	public boolean havePassword() { return(password!=null);} //l'espace dispose d'un mot de passe ou non

	public Avatar newAvatar() { return(newAvatar(null)) ; }
	public Avatar newAvatar(String id){
		if(id==null || id.length()==0 || id.equals("#new")){ //génération automatique d'un nouvelle identifiant d'agent
			int n=2; while(localNode.nbParticipant()>(Math.pow(62, n)*0.66)) n++; //calcul de la longueur d'identifiant nécessaire
			do id=Util.randString(n); while(localNode.existParticipant(id));
		} else if(localNode.existParticipant(id)) return(null); //l'identifiant indiqué est déjà utilisé
		
		//création effective du participant
		LocalParticipant lpart = new LocalParticipant(id, localNode);			
		localNode.addParticipant(lpart);			
		return(lpart.getAvatar()); 
	}
	
		//lecture des informations sur les agents
	public Participant getParticipant(String id){ //id = local_id ou full_id, ou #alias ou &alias
		if(id!=null && id.length()>0){
			char firstLetter=id.charAt(0);
			if(firstLetter=='#' || firstLetter=='&'){ //ciblage indirecte par alias (NB: * et ! aucun sens ici)
				String alias=id.substring(1);
				synchronized(aliases){
					Vector<ImplParticipant> aliasor=aliases.get(alias);
					if(aliasor!=null && aliasor.size()>0){
						if(firstLetter=='#' || aliasor.size()==1) return(aliasor.get(0));
						else return(aliasor.get(Util.randInt(0, aliasor.size()-1)));
					}
				}
			} else {
				int posAlpha=id.indexOf('@');
				if(posAlpha==-1) return(localNode.getParticipant(id)); //id local
				else {
					String siteId=id.substring(posAlpha+1);
					id=id.substring(0, posAlpha);
					Node node=getNode(siteId);
					if(node!=null) return(node.getParticipant(id));
				}
			}
		}
		return(null); 
	}
	protected Node getNode(String node_id){
		if(node_id.equals(localNode.spot().getIdGlobal())) return(localNode);
		else synchronized(distant_nodes) {
			for(DistantNode cur : distant_nodes) if(node_id.equals(cur.getId())) return(cur);
		}
		return(null);
	}
		//ecoute des modifications subit par les agents du space.
	public synchronized void addParticipantListener(ParticipantListener ial){ 
		if(identityListener==null) identityListener=new Vector<ParticipantListener>();
		synchronized(identityListener) { identityListener.add(ial); }
		
		//On va notifier tous les paticipants existants à cet écouteur
		localNode.fireAllExistParticipant(ial);	//Ceux du noeud local
		//ceux des neouds distant
		synchronized(distant_nodes) { for(DistantNode cur : distant_nodes) cur.fireAllExistParticipant(ial); }		
	}
	public void removeParticipantListener(ParticipantListener ial){ if(identityListener!=null) synchronized(identityListener) { identityListener.remove(ial); }}
	public synchronized void addAliasListener(AliasListener ial){ 
		if(aliasListener==null) aliasListener=new Vector<AliasListener>();
		synchronized(aliasListener) { aliasListener.add(ial); }
	}
	public void removeAliasListener(AliasListener ial){ if(aliasListener!=null) synchronized(aliasListener) { aliasListener.remove(ial); } }
	public synchronized void addSkillListener(SkillListener fl){ 
		if(skillListener==null) skillListener=new Vector<SkillListener>();
		synchronized(skillListener) { skillListener.add(fl); }
	}
	public void removeSkillListener(SkillListener fl){ if(skillListener!=null) synchronized(skillListener) { skillListener.remove(fl); }}
	public synchronized void addStateListener(StateListener fl){ 
		if(stateListener==null) stateListener=new Vector<StateListener>();
		synchronized(stateListener) { stateListener.add(fl); }
	}
	public void removeStateListener(StateListener fl){ if(stateListener!=null) synchronized(stateListener) { stateListener.remove(fl); }}
	public synchronized void addActivityListener(ActivityListener ial){ 
		if(activityListener==null) activityListener=new Vector<ActivityListener>();
		synchronized(activityListener) { activityListener.add(ial); }
	}
	public void removeActivityListener(ActivityListener ial){ if(activityListener!=null) synchronized(activityListener) { activityListener.remove(ial);}}
	//]
	
	//[Gestion des messages au sein du space
	protected byte send(ImplParticipant sender, Participant target, String skill_context, Value message, String idreport){
		if(!sender.exist()) return(Ubiquity.NOT_ACTIF);
		if(message==null) return(Ubiquity.EMPTY_MESS);
		if(target==null || !target.exist() || !(target instanceof ImplParticipant)) return(Ubiquity.NOT_EXIST);
		ImplParticipant ptarget = (ImplParticipant) target;
		if(ptarget.node.space!=this) return(Ubiquity.ERROR); //la cible est dans un space différent !
		if(target.isLocal()){ //la cible est sur le spot local
//System.out.println("[ImplSpace - send] distribution de message local");			
			ImplMessage mess=new ImplMessage(((LocalParticipant) target), sender,skill_context, message, idreport);
			return(mess.deliver());
		} else { //la cible est sur un spot distant
//System.out.println("[ImplSpace - send] distribution de message distant : "+message.toString());		
			//POST_MESSAGE](string: idSpaceSite_receveur)(string: id_sender)(string: id_target)(string: skill_context)(value: mess)(string: idreport)   ->idreport=="" si pas de report
			ptarget.node.spot().send(SKUAD.UBIQUITY, DataBytes.writer().b(Protocole.POST_MESSAGE).s(ptarget.node.id_space).s(sender.localId).s(ptarget.localId).s(skill_context).v(message).s(idreport !=null ? idreport : "" ).flushReader());
			return(Ubiquity.OK);			
		}
	}

	private Map<String, CascadingDescriptor> cascadingTargeting=null;
	public final static int CASCADING_TIMEOUT=30*1000; //durée de conservation des flux de transmission de message en cascade
	protected void notifySendingReport(ImplParticipant sender, ImplParticipant receiver, String idReport, byte code){ //pour ImplMessage
		if(cascadingTargeting!=null && idReport!=null && idReport.length()>0 && idReport.charAt(0)=='@'){ //c'est peut-être un ciblage en cascade
			CascadingDescriptor cascDesc=cascadingTargeting.get(idReport);
			if(cascDesc!=null){ //c'est bien un ciblage en cascade
				if(receiver==cascDesc.participant_target){
					if(code==Ubiquity.PERFORMED) synchronized(cascadingTargeting) { cascadingTargeting.remove(idReport); } //arrêt des appels en cascade, NB: l'agent emmetteur sera notifié ci-après
					else if(code==Ubiquity.IGNORED){ //continuation des appels en cascade
						synchronized(aliases){
							Vector<ImplParticipant> aliasor=aliases.get(cascDesc.alias);
							if(aliasor!=null && aliasor.size()>0){
								for(int i=0; i<aliasor.size(); i++) if(aliasor.get(i)==cascDesc.participant_target){
									i++;
									if(i<aliasor.size()){
										cascDesc.participant_target=aliasor.get(i);
										send(sender, cascDesc.participant_target,cascDesc.skill_context, cascDesc.message, idReport);
										return;
									}	
								}//s'il n'y a plus de cible, l'agent emmetteur sera notifié ci-après
							}					
						}
					}
				}
				idReport=cascDesc.idreport;
			}
		}
		if(sender.exist()) {
			if(sender.isLocal()){ //la source est sur le spot local
				((LocalParticipant) sender).notifySendingReport(receiver, idReport, code);
			} else { //la source est sur un spot distant
				//[SENDING_REPORT] (string: idSpaceSite_receveur)(string: id_sender) (string: id_target)(string: id_report)[code_report]
				sender.spot().send(SKUAD.UBIQUITY, DataBytes.writer().b(Protocole.SENDING_REPORT).s(sender.node.id_space).s(sender.localId).s(receiver.localId).s(idReport).b(code).flushReader());
			}
		}
	}	
	protected byte send(ImplParticipant sender, String target, String skill_context, Value message, String idreport){
		if(!sender.exist()) return(Ubiquity.NOT_ACTIF);
		if(message==null) return(Ubiquity.EMPTY_MESS);
		if(target==null || target.length()==0) return(Ubiquity.NOT_EXIST);
		Participant participant_target=null; String alias;
		char firstLetter=target.charAt(0);
		switch(firstLetter){ //analyse du mode de ciblage
		
		//[Gestion du ciblage
		case '#' : case '&' : //ciblage indirecte fort ou faible
			participant_target=getParticipant(target);
			break;
		case '*' : case '!' : //ciblage de groupe ou en casccade
			alias=target.substring(1);
			synchronized(aliases){
				String cascadingId;
				Vector<ImplParticipant> aliasor=aliases.get(alias);
				if(aliasor!=null && aliasor.size()>0){
					if(firstLetter=='*') for(ImplParticipant atarget: aliasor) send(sender, atarget,skill_context, message, idreport);
					else { //ciblage en cascade
						do cascadingId='@'+Util.randString(10); while(cascadingTargeting!=null && cascadingTargeting.get(cascadingId)!=null);
						participant_target=aliasor.get(0);
						CascadingDescriptor cascDesc=new CascadingDescriptor(cascadingId, alias, participant_target,skill_context, message, idreport);
						if(cascadingTargeting==null) {
							cascadingTargeting=new HashMap<String, CascadingDescriptor>();
							//on procéde aussi à l'enregistrement du processus de vidange de la pile cascadingTargeting
							ThreadStack.addWaiter(new Waiter(){
								public void run(long ct){
									if(cascadingTargeting.size()>0){
										synchronized(cascadingTargeting) {
											String to_remove=null;
											for(CascadingDescriptor cd: cascadingTargeting.values()) if(ct-cd.date>CASCADING_TIMEOUT){
//System.out.println("vidange de "+cd.cascadingId);											
												to_remove=cd.cascadingId;
												break;
											}
//else System.out.println("en cascade depuis "+(ct-cd.date));	
											if(to_remove!=null) cascadingTargeting.remove(to_remove);
										}
									}
								}
							});
						}
						synchronized(cascadingTargeting) {cascadingTargeting.put(cascadingId, cascDesc); }
//System.out.println("cascading de "+cascadingId);							
						idreport=cascadingId;
					}
				}
			}
			break;
		//]
		default: 	//ciblage par id	
			int posAlpha=target.indexOf('@');
			String participantId=null; Node node=null;
			if(posAlpha==-1){ //id local
				node=localNode;
				participantId=target;
			} else { 
				String siteId=target.substring(posAlpha+1);
				participantId=target.substring(0, posAlpha);
				node=getNode(siteId);
			}
			if(participantId!=null && node!=null) participant_target=node.getParticipant(participantId);
		}
		if(participant_target!=null) {
//System.out.println("send to "+iagent_target.fullId());	
			return(send(sender, participant_target,skill_context, message, idreport));
		}
		return(Ubiquity.NOT_EXIST);
	}	
	
	//[Notification des changements au sein du space
		//Notification au niveau du monitor local
	protected void notifyAddParticipant(ImplParticipant p){
		//notification des écouteurs locaux
		if(identityListener!=null) synchronized(identityListener) { for(ParticipantListener l : identityListener) ThreadStack.perform( () -> l.participantAdded(p)); }
		//notification des écouteurs distants
		if(p.isLocal() && distant_nodes.size()>0) synchronized(distant_nodes){ 	//[NOTIFY_ADD_PARTICIPANT] (string: idSpaceSite_emetteur) (string : local_participant_id)
			DataReader mess = DataBytes.writer().b(Protocole.NOTIFY_ADD_PARTICIPANT).s(p.node.id_space).s(p.localId).flushReader();
			for(DistantNode dspot : distant_nodes){
				dspot.spot().send(SKUAD.UBIQUITY, mess, false, true);
			}
			mess.dispose();
		}
	}
	
	protected void notifyRemoveParticipant(ImplParticipant p){
		//Mise à jour du hash aliases:
		int nb_alias=(p.alias!=null?p.alias.nbAlias():0);
		if(nb_alias>0) synchronized(aliases){
			for(int i=0; i<nb_alias; i++){
				String alias_name=p.alias.aliasName(i);
				Vector<ImplParticipant> aliasor = aliases.get(alias_name);
				if(aliasor!=null) { //c'est normalement toujours le cas
					int removePosition=-1;
					for(int j=0; j<aliasor.size(); j++){
						if(p==aliasor.get(j)){
							removePosition=j;
							aliasor.removeElementAt(j);
							break;
						}
					}
					//On informe les écouteurs du changement de focus
					if(removePosition==0){
						p.alias.fireAliasFocusChange(alias_name, true, false); //perte du focus
						notifyAliasFocusChange(p, alias_name, true, false);
						if(aliasor.size()>0) {
							ImplParticipant new_focuser=(ImplParticipant)(aliasor.get(0));
							new_focuser.alias.fireAliasFocusChange(alias_name, false, true); //gain du focus
							notifyAliasFocusChange(new_focuser, alias_name, false, true);
						}
					}
				}
			}
		}
		//notification des écouteurs locaux
		if(identityListener!=null) synchronized(identityListener) { for(ParticipantListener l : identityListener) ThreadStack.perform( () -> l.participantRemoved(p)); }
		//notification des écouteurs distants
		if(p.isLocal() && distant_nodes.size()>0) synchronized(distant_nodes){
			DataReader mess = DataBytes.writer().b(Protocole.NOTIFY_REMOVE_PARTICIPANT).s(p.node.id_space).s(p.localId).i(p.incAndGetVersion()).flushReader();
			for(DistantNode dspot : distant_nodes){
				dspot.spot().send(SKUAD.UBIQUITY, mess, false, true);
			}
			mess.dispose();
		}	
	}
	
	protected void notifyAddSkill(ImplParticipant participant, String newSkill){
		//notification des écouteurs locaux
		if(skillListener!=null)  synchronized(skillListener) { for(SkillListener l : skillListener) ThreadStack.perform( () -> l.skillAdded(participant, newSkill)); }
		//notification des écouteurs distants
		if(participant.isLocal() &&distant_nodes.size()>0) synchronized(distant_nodes){ //[NOTIFY_ADD_SKILL] (string: idSpaceSite_emetteur) (string : local_participant_id) (int : version) (string: nameSkill)
			DataReader mess = DataBytes.writer().b(Protocole.NOTIFY_ADD_SKILL).s(participant.node.id_space).s(participant.localId).i(participant.incAndGetVersion()).s(newSkill).flushReader();
			for(DistantNode dspot : distant_nodes){
				dspot.spot().send(SKUAD.UBIQUITY, mess, false, true);
			}
			mess.dispose();
		}	
	}
	
	protected void notifyRemoveSkill(ImplParticipant participant, String lostSkill){
		//notification des écouteurs locaux
		if(skillListener!=null)  synchronized(skillListener) { for(SkillListener l : skillListener) ThreadStack.perform( () ->  l.skillRemoved(participant, lostSkill)); }
		//notification des écouteurs distants
		if(participant.isLocal() &&distant_nodes.size()>0) synchronized(distant_nodes){ //[NOTIFY_REMOVE_SKILL] (string: idSpaceSite_emetteur) (string : local_participant_id) (int : version) (string: nameSkill)
			DataReader mess = DataBytes.writer().b(Protocole.NOTIFY_REMOVE_SKILL).s(participant.node.id_space).s(participant.localId).i(participant.incAndGetVersion()).s(lostSkill).flushReader();
			for(DistantNode dspot : distant_nodes){
				dspot.spot().send(SKUAD.UBIQUITY, mess, false, true);
			}
			mess.dispose();
		}		
	}
	
	protected void notifyPriorityAlias(ImplParticipant participant,String alias_name, int oldPriority, int newPriority){	//on propage le changement de priorité
		//Mise à jour du hash aliases:
		int insertPosition=-1, removePosition=-1;
		synchronized(aliases){
			Vector<ImplParticipant> aliasor = aliases.get(alias_name);
			if(aliasor!=null && aliasor.size()>1) {
				for(int i=0; i<aliasor.size(); i++){
					ImplParticipant cur_iagent=aliasor.get(i);
					if(cur_iagent==participant){ //c'est la position de suppression
						removePosition=i;
						if(insertPosition==-1 && (i==aliasor.size()-1)){ //cas de l'insertion et de la suppression en fin de liste
							insertPosition=i;
							break;
						}
						aliasor.removeElementAt(i);
						if(insertPosition==-1){ //on a pas encore fait la réinsertion
							i--;
							continue;
						} else break; //insertion et suppression effectué
					} else if(insertPosition==-1){ //on regarde si c'est la nouvelle position d'insertion
						int cur_p=cur_iagent.getAlias().priority(alias_name);
						if((cur_p==newPriority && participant.ordoredBefore(cur_iagent)) || cur_p<newPriority){
							insertPosition=i;
							aliasor.insertElementAt(participant, i);
							if(removePosition==-1){ //on a pas encore fait la suppression
								i++;
								continue;
							} else break;  //insertion et suppression effectué
						}
					}
				}
				if(insertPosition==-1){ //cas de la ré-insertion en fin de liste
					insertPosition=aliasor.size();
					aliasor.add(participant);
				}
			}
			if(insertPosition!=removePosition){//On informe les écouteurs du changement de focus éventuel
				if(insertPosition==0){ //gain de focus
					if(aliasor.size()>1) { //tjs vrai
						ImplParticipant old_focuser=(ImplParticipant)(aliasor.get(1));
						old_focuser.alias.fireAliasFocusChange(alias_name, true, false); //perte de l'ancien focus
						notifyAliasFocusChange(old_focuser, alias_name, true, false);
					}
					participant.alias.fireAliasFocusChange(alias_name, false, true); //gain du focus
					notifyAliasFocusChange(participant, alias_name, false, true);
				} else if(removePosition==0){ //perte de focus
					participant.alias.fireAliasFocusChange(alias_name, true, false); //perte du focus
					notifyAliasFocusChange(participant, alias_name, true, false);
					if(aliasor.size()>0) { //tjs vrai
						ImplParticipant new_focuser=(ImplParticipant)(aliasor.get(0));
						new_focuser.alias.fireAliasFocusChange(alias_name, false, true); //gain du focus
						notifyAliasFocusChange(new_focuser, alias_name, false, true);
					}					
				}
			}
		}
		//notification des écouteurs locaux
		if(aliasListener!=null) synchronized(aliasListener) { for(AliasListener l : aliasListener) ThreadStack.perform( () ->  l.aliasPriorityChange(participant, alias_name, oldPriority, newPriority)); }
		//notification des écouteurs distants
		if(participant.isLocal() &&distant_nodes.size()>0) synchronized(distant_nodes){ //[NOTIFY_ALIAS_PRIORITY]  (string: idSpaceSite_emetteur) (string : local_participant_id) (int : version) (string: nameAlias) (int: newPriority)
			DataReader mess = DataBytes.writer().b(Protocole.NOTIFY_ALIAS_PRIORITY).s(participant.node.id_space).s(participant.localId).i(participant.incAndGetVersion()).s(alias_name).i(newPriority).flushReader();
			for(DistantNode dspot : distant_nodes){
				dspot.spot().send(SKUAD.UBIQUITY, mess, false, true);
			}
			mess.dispose();
		}		
	}
	
	protected void notifyAddAlias(ImplParticipant participant,String alias_name, int newPriority){	//on propage l'ajout d'alias, et l'éventuel changement de focus sur cet alias
		//Mise à jour du hash aliases:
		int insertPosition=-1;
		synchronized(aliases){
			Vector<ImplParticipant> aliasor = aliases.get(alias_name);
			if(aliasor==null) {
				aliasor=new Vector<ImplParticipant>();
				aliasor.add(participant); insertPosition=0;
				aliases.put(alias_name, aliasor);
			} else { //on va positionner l'agent dans la liste
				for(int i=0; i<aliasor.size(); i++){
					ImplParticipant cur_iagent=aliasor.get(i);
					int cur_p=cur_iagent.getAlias().priority(alias_name);
					if((cur_p==newPriority && participant.ordoredBefore(cur_iagent)) || cur_p<newPriority){
						insertPosition=i;
						aliasor.insertElementAt(participant, i);
						break;
					} 
				}
				if(insertPosition==-1){ //insertion en fin de liste
					insertPosition=aliasor.size();
					aliasor.add(participant);
				}
			}
			//On informe les écouteurs du changement de focus
			if(insertPosition==0){
				if(aliasor.size()>1) {
					ImplParticipant old_focuser=(ImplParticipant)(aliasor.get(1));
					old_focuser.alias.fireAliasFocusChange(alias_name, true, false); //perte de l'ancien focus
					notifyAliasFocusChange(old_focuser, alias_name, true, false);
				}
				participant.alias.fireAliasFocusChange(alias_name, false, true); //gain du focus
				notifyAliasFocusChange(participant, alias_name, false, true);
			}
		}
		//notification des écouteurs locaux
		if(aliasListener!=null) synchronized(aliasListener) { for(AliasListener l : aliasListener) ThreadStack.perform( () ->  l.aliasAdded(participant, alias_name, newPriority)); }
		//notification des écouteurs distants
		if(participant.isLocal() &&distant_nodes.size()>0) synchronized(distant_nodes){ //[NOTIFY_ALIAS_PRIORITY]  (string: idSpaceSite_emetteur) (string : local_participant_id) (int : version) (string: nameAlias) (int: newPriority)
			DataReader mess = DataBytes.writer().b(Protocole.NOTIFY_ALIAS_PRIORITY).s(participant.node.id_space).s(participant.localId).i(participant.incAndGetVersion()).s(alias_name).i(newPriority).flushReader();
			for(DistantNode dspot : distant_nodes){
				dspot.spot().send(SKUAD.UBIQUITY, mess, false, true);
			}
			mess.dispose();
		}			
	}	
	
	protected void notifyRemoveAlias(ImplParticipant participant,String alias_name, int oldPriority){	//on propage l'ajout d'alias, et l'éventuel changement de focus sur cet alias
		//Mise à jour du hash aliases:
		int removePosition=-1;
		synchronized(aliases){
			Vector<ImplParticipant> aliasor = aliases.get(alias_name);
			if(aliasor!=null) { //c'est normalement toujours le cas
				for(int i=0; i<aliasor.size(); i++){
					ImplParticipant cur_participant=aliasor.get(i);
					if(participant==cur_participant){
						removePosition=i;
						aliasor.removeElementAt(i);
						break;
					}
				}
			}
			//On informe les écouteurs du changement de focus
			if(removePosition==0){
				participant.alias.fireAliasFocusChange(alias_name, true, false); //perte du focus
				notifyAliasFocusChange(participant, alias_name, true, false);
				if(aliasor.size()>0) {
					ImplParticipant new_focuser=(ImplParticipant)(aliasor.get(0));
					new_focuser.alias.fireAliasFocusChange(alias_name, false, true); //gain du focus
					notifyAliasFocusChange(new_focuser, alias_name, false, true);
				}
			}
		}
		//notification des écouteurs locaux
		if(aliasListener!=null) synchronized(aliasListener) { for(AliasListener l : aliasListener) ThreadStack.perform( () ->  l.aliasRemoved(participant, alias_name, oldPriority)); }
		//notification des écouteurs distants
		if(participant.isLocal() &&distant_nodes.size()>0) synchronized(distant_nodes){ //[NOTIFY_REMOVE_ALIAS]  (string: idSpaceSite_emetteur) (string : local_participant_id) (int : version) (string: nameAlias)
			DataReader mess = DataBytes.writer().b(Protocole.NOTIFY_REMOVE_ALIAS).s(participant.node.id_space).s(participant.localId).i(participant.incAndGetVersion()).s(alias_name).flushReader();
			for(DistantNode dspot : distant_nodes){
				dspot.spot().send(SKUAD.UBIQUITY, mess, false, true);
			}
			mess.dispose();
		}			
	}
	
	protected void notifyAliasFocusChange(ImplParticipant participant, String alias, boolean oldFocus, boolean newFocus){ 
		//cette notification ne concerne que les écouteurs locaux
		if(aliasListener!=null) synchronized(aliasListener) { for(AliasListener l : aliasListener) ThreadStack.perform( () ->  l.aliasFocusChange(participant, alias, oldFocus, newFocus)); } 			
	}
	
	protected void  notifyStateChange(ImplParticipant participant, String field, Value oldValue, Value newValue){
		if(stateListener!=null)  synchronized(stateListener) { for(StateListener l : stateListener) ThreadStack.perform( () ->  l.stateChange(participant, field, oldValue, newValue)); }
		//notification des écouteurs distants
		if(participant.isLocal() &&distant_nodes.size()>0) synchronized(distant_nodes){ //[NOTIFY_STATE_CHANGE] (string: idSpaceSite_emetteur) (string : local_participant_id) (int : version) (string: nameField) (value: newValue)
			DataReader mess = DataBytes.writer().b(Protocole.NOTIFY_STATE_CHANGE).s(participant.node.id_space).s(participant.localId).i(participant.incAndGetVersion()).s(field).v(newValue).flushReader();
			for(DistantNode dspot : distant_nodes){
				dspot.spot().send( SKUAD.UBIQUITY, mess, false, true);
			}
			mess.dispose();
		}	
	}
		
	protected void  notifyAddActivity(ImplParticipant participant, String activity){
		if(activityListener!=null)  synchronized(activityListener) { for(ActivityListener l : activityListener) ThreadStack.perform( () ->  l.activityAdded(participant, activity)); }
		//notification des écouteurs distants
		if(participant.isLocal() &&distant_nodes.size()>0) synchronized(distant_nodes){ //[NOTIFY_ADD_ACTIVITY] (string: idSpaceSite_emetteur) (string : local_participant_id) (int : version) (string: nameActivity)
			DataReader mess = DataBytes.writer().b(Protocole.NOTIFY_STATE_CHANGE).s(participant.node.id_space).s(participant.localId).i(participant.incAndGetVersion()).s(activity).flushReader();
			for(DistantNode dspot : distant_nodes){
				dspot.spot().send(SKUAD.UBIQUITY, mess, false, true);
			}
			mess.dispose();
		}	
	}
		
	protected void  notifyRemoveActivity(ImplParticipant participant, String activity){
		if(activityListener!=null)  synchronized(activityListener) { for(ActivityListener l : activityListener) ThreadStack.perform( () ->  l.activityRemoved(participant, activity)); }
		//notification des écouteurs distants
		if(participant.isLocal() &&distant_nodes.size()>0) synchronized(distant_nodes){ //[NOTIFY_REMOVE_ACTIVITY] (string: idSpaceSite_emetteur) (string : local_participant_id) (int : version) (string: nameActivity)
			DataReader mess = DataBytes.writer().b(Protocole.NOTIFY_REMOVE_ACTIVITY).s(participant.node.id_space).s(participant.localId).i(participant.incAndGetVersion()).s(activity).flushReader();
			for(DistantNode dspot : distant_nodes){
				dspot.spot().send(SKUAD.UBIQUITY, mess, false, true);
			}
			mess.dispose();
		}	
	}
	protected void  notifyActivityChangeMode(ImplParticipant participant, String activity, byte oldMode, byte newMode){
		//notification des écouteurs locaux
		if(activityListener!=null)  synchronized(activityListener) { for(ActivityListener l : activityListener) ThreadStack.perform( () ->  l.activityModeChange(participant, activity, oldMode, newMode)); }
		//notification des écouteurs distants
		if(participant.isLocal() && distant_nodes.size()>0) synchronized(distant_nodes){ //[NOTIFY_ACTIVITY_CHANGE_MODE] (string: idSpaceSite_emetteur) (string : local_participant_id) (int : version) (string: nameActivity) [newMode]
			DataReader mess = DataBytes.writer().b(Protocole.NOTIFY_ACTIVITY_CHANGE_MODE).s(participant.node.id_space).s(participant.localId).i(participant.incAndGetVersion()).s(activity).b(newMode).flushReader();
			for(DistantNode dspot : distant_nodes){
				dspot.spot().send(SKUAD.UBIQUITY, mess, false, true);
			}
			mess.dispose();
		}	
	}	
	protected void  notifyActivityChangeStatus(ImplParticipant participant, String activity, byte oldStatus, byte newStatus){
		//notification des écouteurs locaux
		if(activityListener!=null)  synchronized(activityListener) { for(ActivityListener l : activityListener) ThreadStack.perform( () ->  l.activityStatusChange(participant, activity, oldStatus, newStatus)); }
		//notification des écouteurs distants
		//notification des écouteurs distants
		if(participant.isLocal() && distant_nodes.size()>0) synchronized(distant_nodes){ //[NOTIFY_ACTIVITY_CHANGE_STATUS] (string: idSpaceSite_emetteur) (string : local_participant_id) (int : version) (string: nameActivity) [newStatus]
			DataReader mess = DataBytes.writer().b(Protocole.NOTIFY_ACTIVITY_CHANGE_STATUS).s(participant.node.id_space).s(participant.localId).i(participant.incAndGetVersion()).s(activity).b(newStatus).flushReader();
			for(DistantNode dspot : distant_nodes){
				dspot.spot().send(SKUAD.UBIQUITY, mess, false, true);
			}
			mess.dispose();
		}	
	}
	protected void  notifyActivityProgress(ImplParticipant participant, String activity, int currentStep, int maxStep){
		//notification des écouteurs locaux
		if(activityListener!=null) synchronized(activityListener) { for(ActivityListener l : activityListener) ThreadStack.perform( () -> l.activityProgressChange(participant, activity, currentStep, maxStep)); }
		//notification des écouteurs distants
		if(participant.isLocal() && distant_nodes.size()>0) synchronized(distant_nodes){ //[NOTIFY_ACTIVITY_CHANGE_PROGRESS] (string: idSpaceSite_emetteur) (string : local_participant_id) (int : version) (string: nameActivity) (int: currentStep) (int: maxStep)
			DataReader mess = DataBytes.writer().b(Protocole.NOTIFY_ACTIVITY_CHANGE_PROGRESS).s(participant.node.id_space).s(participant.localId).i(participant.incAndGetVersion()).s(activity).i(currentStep).i(maxStep).flushReader();
			for(DistantNode dspot : distant_nodes){
				dspot.spot().send(SKUAD.UBIQUITY, mess, false, true);
			}
			mess.dispose();
		}	
	}
	//]
	
	//[
	protected boolean haveFocus(ImplParticipant iagent, String alias_name){
		synchronized(aliases){
			Vector<ImplParticipant> aliasor = aliases.get(alias_name);
			if(aliasor!=null && aliasor.size()>0) { //c'est normalement toujours le cas
				return(aliasor.get(0)==iagent);
			}
		}
		return(false);
	}	
	//]
}

class CascadingDescriptor {
	public String cascadingId, alias, skill_context, idreport;
	public Participant participant_target;
	public Value message;
	long date;
	public CascadingDescriptor(String vcascadingId, String valias, Participant vparticipant_target,String vskill_context, Value vmessage, String vidreport){
		cascadingId=vcascadingId;
		alias=valias;
		skill_context=vskill_context;
		idreport=vidreport;
		participant_target=vparticipant_target;
		message=vmessage;
		date=System.currentTimeMillis();
	}
}

