package skuad.ubiquity.implementation;

import skuad.ubiquity.Ubiquity;
import skuad.ubiquity.space.*;
import java.util.*;
import skuad.hubudp.*;

public abstract class Node {
	
	protected ImplSpace space;						//space dans lequel ce spot est inscrit	
	String id_space;								//valeur de l'id space sur l'hote porteur du participant
	Vector<String> participantsId;					//vecteur des identifiant des participants de ce spot
	Map<String, ImplParticipant> allParticipants;	//tous les participants , key = idLocal du participant sur le spot
	
	protected Node(ImplSpace vspace, String vid_space){
		space=vspace;
		id_space = vid_space;
		participantsId=new Vector<String>();
		allParticipants=new HashMap<String, ImplParticipant>();	
	}
	public Space getSpace() { return(space); }
	public abstract String getId(); //retourne la valeur global_id de l'hote porteur
	
	public abstract boolean isUp(); 
	public abstract Spot spot(); 
	public abstract boolean isLocal();
				
	//[Pour LocalSpot et DistantSpot
	public int nbParticipant(){ //nombre de participant existant sur ce spot
		return(participantsId.size());
	}
	public boolean existParticipant(String local_id){ return(allParticipants.containsKey(local_id));}
	
	public ImplParticipant getParticipant(String local_id){ 
		return(allParticipants.get(local_id));
	}
	
	protected synchronized void addParticipant(ImplParticipant participant){
		synchronized(allParticipants) { 
			participantsId.add(participant.localId);	
			allParticipants.put(participant.localId, participant); 
		}
		space.notifyAddParticipant(participant);
	}
	
	protected byte internalDeleteParticipant(ImplParticipant p, boolean removeInList){ //seulement pour les classes filles
		if(p.remove()){ //on supprime le participant
			//on informe les écouteurs du space
			space.notifyRemoveParticipant(p);
			if(removeInList) {
				synchronized(allParticipants) { 
					if(allParticipants.remove(p.localId())!=null) participantsId.remove(p.localId());
				}
				
			}
			return(Ubiquity.OK);
		} else return(Ubiquity.NOT_EXIST);
	}
	
	protected void fireAllExistParticipant(ParticipantListener ial) {
		synchronized(allParticipants) { for(Participant p : allParticipants.values()) ial.participantAdded(p); }
	}
	//]
}