package skuad.ubiquity.implementation;

import java.util.*;

import skuad.ubiquity.*;
import skuad.util.thread.*;
import skuad.util.value.Value;
import skuad.ubiquity.space.activity.*;

public class PlannedActivity extends TimerTask {
	
	//[Propriété de classe
	protected static Timer timer = new Timer(); //timer interne
	//]
	
	ExecutorActivity executor;
	
	public PlannedActivity(Activity activity, ActivityValue a, Value v_args, long delay, long periode){
		executor=new ExecutorActivity(activity, a, v_args);
		if(delay<=0 && periode<=0) run(); //l'activité est pontuelle et non périodique, elle doit être traitée maintenant
		else {
			activity.setActivityStatus(a, Ubiquity.WAITING);//on signal que l'activité est en attente d'exécution
			if(periode<=0) timer.schedule(this, delay); //l'activité est ponctuelle
			else timer.schedule(this, delay, periode); //l'activité est périodique
		}
	}
	
	protected void chain(Value arg, long delay){ executor.chain(arg,delay); }
	
	//[interface TimerTask
	public void run(){ ThreadStack.perform(executor); }
	public boolean cancel(){
		executor.cancel();
		return(super.cancel());
	}
	//]	
}

class ExecutorActivity implements Runnable, ProgressFlag {
	Activity activity;
	ActivityValue avalue;
	Value args;
	LinkedList<ChainedActivity> statckArgs;
	boolean canceled;
	
	public ExecutorActivity(Activity a, ActivityValue av, Value v_args){
		activity=a;
		args=v_args;
		canceled=false;
		avalue=av;
		statckArgs=null;
	}
	
	public void run(){
		if(!canceled){
			activity.setActivityStatus(avalue, Ubiquity.RUN);//on signal que l'activité est en cours d'exécution
			boolean run_again=true;
			while(run_again){
				avalue.runner.action(activity.avatar, avalue.name, args, this);
				run_again=false;
				long delay=0;
				synchronized(avalue) {
					if(!canceled) { //on change de mode/status et on notifie
						switch(avalue.mode){ ////NONE, ONCE , CHAIN, REPEAT, CANCELED
						case Ubiquity.REPEAT : activity.setActivityStatus(avalue, Ubiquity.WAITING); break; //on signal que l'activité est en attente d'exécution
						case Ubiquity.CHAIN:
							if(statckArgs!=null && !statckArgs.isEmpty()){
								ChainedActivity ca=statckArgs.poll();
								args=ca.args;
								if(ca.date>0) delay=ca.date-System.currentTimeMillis();
								run_again=true; //on continue l'exécution
								break;
							}
							//break; pas de break volontaire
						default : //l'exécution de l'activité s'arrête la
							activity.setActivityStatus(avalue, Ubiquity.NOT_RUN);
							activity.setActivityMode(avalue, Ubiquity.NONE);
							avalue.planning=null;
						}
					}
				}
				if(delay>0) synchronized(this){ try{ wait(delay);} catch(Exception e){}}
			}
		}
	}
	
	protected void cancel(){ canceled=true; }
	
	protected synchronized void chain(Value arg, long delay){ 
		if(!canceled){
			if(statckArgs==null) statckArgs=new LinkedList<ChainedActivity>();
			statckArgs.add(new ChainedActivity(arg, delay));
		}
	}
	
	//[Interface ProgressFlag
	public boolean is_cancel(){ return(canceled); }
	
	public void setProgress(int currentStep, int maxStep){
		if(!canceled) activity.setActivityProgress(avalue, currentStep, maxStep);
	}	
	//]
	
}

class ChainedActivity {
	Value args;
	long date;
	public ChainedActivity(Value v, long delay){
		if(delay>0) date=System.currentTimeMillis()+delay;
		else date=0;
		args=v;
	}
}