package skuad.ubiquity.implementation;

import java.util.*;
import skuad.ubiquity.space.*;
import skuad.ubiquity.space.skill.*;
import skuad.ubiquity.*;
import skuad.util.databytes.*;
import skuad.util.thread.ThreadStack;

public class Skill implements WSkill { //instance skill pour les agents distants, pour les agents locaux ce sera une instance de la classe LocalSkill

	ImplParticipant thisParticipant;			//identité de l'agent porteur
	SkillProxyReader rproxy;	//objet proxy de lecture sur cette instance
	Vector<String> skills;	//table des skills: private n'existe que sur que si l'agent associé est sur le noeud local et qu'au moins 
								//un skill de visibilité private est définie. publicSkill : sur le noeud local contient les skills de visibilité
								//local et global, sur les autres noeuds il ne contiendra que les skill de visibilité global
	Vector<SkillListener> listener;
	
	protected Skill(ImplParticipant p){
		thisParticipant=p;
		rproxy=null;
		skills=null; 
		listener=null;
	}
	
	//[Interface RSkill
	public Participant participant(){ return(thisParticipant); }	
	public synchronized void addListener(SkillListener sl){ 
		if(listener==null) listener=new Vector<SkillListener>();
		synchronized(listener) { listener.add(sl); }
	}
	public void removeListener(SkillListener sl){ if(listener!=null) synchronized(listener) { listener.remove(sl); }}
	public int nbSkills(){ //nombre de skill public existant
		return(skills==null?0:skills.size());
	}
	public String skillName(int num){ //consultation de la compétance public de numéro arbitraire num, null si non existant	
		return(skills==null || num>=skills.size()?null:skills.get(num));
	}
	public boolean haveSkill(String skill){ return(skills!=null && skills.contains(skill)); }
	//]
	
	//[Interface WRole
	public RSkill reader(){ 
		if(rproxy==null) rproxy=new SkillProxyReader(this);
		return(rproxy); 
	}
	public synchronized byte add(String newSkill){
		if(!thisParticipant.exist()) return(Ubiquity.NOT_ACTIF);
		if(skills==null) skills=new Vector<String>();
		if(!skills.contains(newSkill)){
			skills.add(newSkill);
			fireAddSkill(newSkill);	//notification locale
		}
		return(Ubiquity.OK);
	}
	public synchronized byte remove(String lostSkill){
		if(!thisParticipant.exist()) return(Ubiquity.NOT_ACTIF);
		if(skills!=null && skills.contains(lostSkill)){
			skills.remove(lostSkill);
			fireRemoveSkill(lostSkill);	//notification locale			
		}
		return(Ubiquity.OK);		
	}
	//]
	
	protected void fireAddSkill(String newSkill){	//notification locale
		//Au niveau des écouteurs sur l'agent lui même
		if(listener!=null)  synchronized(listener) { for(SkillListener l : listener) ThreadStack.perform( () -> l.skillAdded(thisParticipant, newSkill)); }
		//Au niveau du monitor
		thisParticipant.node.space.notifyAddSkill(thisParticipant, newSkill); //on notifie l'ajout sur le space
	}
	protected void fireRemoveSkill(String removedSkill){	//notification locale
		//Au niveau des écouteurs sur l'agent lui même
		if(listener!=null)  synchronized(listener) { for(SkillListener l : listener) ThreadStack.perform( () -> l.skillRemoved(thisParticipant, removedSkill)); }
		//Au niveau du monitor
		thisParticipant.node.space.notifyRemoveSkill(thisParticipant, removedSkill); //on notifie l'ajout sur le space
	}
	
	//[Pour Identity
	protected synchronized void writeCompleteState(DataWriter w){ //(int : nb_entry) {chaque_valeur: (string: skill_name)}
		int nb=(skills==null?0:skills.size());
		w.i(nb);
		if(nb>0) for(String s : skills) w.s(s);
	}
	protected synchronized Skill readCompleteState(int nb, DataReader dbrb){//{chaque_valeur: (string: skill_name)}
		Vector<String> old_skills=skills;
		skills=new Vector<String>();
		String ns;
		for(int i=0; i<nb; i++) {
			String newSkill=dbrb.readString();
			ns = null;
			if(old_skills!=null){
				for(int j=0; j<old_skills.size(); j++) {
					if(old_skills.get(i).equals(newSkill)) {
						ns=newSkill;
						old_skills.remove(j);
						break;
					}
				}
			}
			skills.add(newSkill);
			if(ns==null) fireAddSkill(newSkill);	//notification locale							
		}
		if(old_skills!=null) for(String cur : old_skills){
			fireRemoveSkill(cur);
		}		
		old_skills=null;
		return(this);
	}
	protected synchronized void clear(){
		if(skills!=null) {
			for(String cur : skills) fireRemoveSkill(cur);
			skills.removeAllElements();
		}
	}	
	//]
}

//Coquille de lecture seule
class SkillProxyReader implements RSkill {
	Skill kernel;

	public SkillProxyReader(Skill k){ kernel=k; }
	public Participant participant(){ return(kernel.participant()); }
	public int nbSkills(){ return(kernel.nbSkills()); }
	public String skillName(int num){ return(kernel.skillName(num)); }
	public boolean haveSkill(String skill){ return(kernel.haveSkill(skill)); }
	public void addListener(SkillListener sl){ kernel.addListener(sl); }
	public void removeListener(SkillListener sl){ kernel.removeListener(sl); }
}