package skuad.ubiquity.implementation;

import java.util.*;

import skuad.ubiquity.space.*;
import skuad.ubiquity.space.state.*;
import skuad.ubiquity.*;
import skuad.util.thread.ThreadStack;
import skuad.util.value.Value;
import skuad.util.databytes.*;

public class State implements WState {

	ImplParticipant thisParticipant;				//agent associé à ce descripteur Alias
	StateProxyReader rproxy;		//objet proxy de lecture sur cet instance
	Vector<FieldValue> fields;		//listes des champs
	Vector<StateListener> listener;
	
	protected State(ImplParticipant p){
		thisParticipant=p;
		fields=null;
		listener=null;
		rproxy=null;
	}
		
	//[Interface RState
	public Participant participant(){ return(thisParticipant); }	
	public int nbFields(){ return(fields==null?0:fields.size()); } //nombre d'alias public existant
	public String fieldName(int num){ //consultation de l'alias public de numéro arbitraire num, null si non existant
		if(fields!=null && num<fields.size() && num>=0) return(fields.get(num).name);
		return(null);
	}
	public Value fieldValue(int num){ 
		if(fields!=null && num<fields.size() && num>=0) return(fields.get(num).value);
		return(null);
	} 
	public Value get(String field){ 
		FieldValue existing=findField(field);
		if(existing!=null) return(existing.value);
		return(null);
	}
	public boolean haveField(String field){
		return(findField(field)!=null);
	}
	public synchronized void addListener(StateListener il){ 
		if(listener==null) listener=new Vector<StateListener>();
		synchronized(listener) { listener.add(il); }
	}
	public void removeListener(StateListener il){ if(listener!=null) synchronized(listener) { listener.remove(il); }}
	//]
	
	protected FieldValue findField(String field_name){
		if(field_name!=null && fields!=null) for(FieldValue cur : fields){
			if(field_name.equals(cur.name)) return(cur);
		}
		return(null);		
	}
	
	//[Interface WState
	public RState reader(){ 
		if(rproxy==null) rproxy=new StateProxyReader(this);
		return(rproxy); 
	}

	public synchronized byte set(String field, Value value){
		if(!thisParticipant.exist()) return(Ubiquity.NOT_ACTIF);
		if(field==null) return(Ubiquity.ERROR);
		if(value==null) return(remove(field));
		
		FieldValue existing=findField(field);
		if(existing!=null){ //maj de la priorité sur ce champ
			if(value.equals(existing.value)) return(Ubiquity.OK); //pas de changement
			Value oldValue=existing.value;
			existing.value=value.toConstant();
			fireChangeState(field, oldValue, existing.value);
		} else {
			existing=new FieldValue(field, value.toConstant());
			if(fields==null) fields=new Vector<FieldValue>();
			fields.add(existing);
			fireChangeState(field, null, existing.value);
		}
		return(Ubiquity.OK);
	}	
	
	public synchronized byte remove(String field){
		if(!thisParticipant.exist()) return(Ubiquity.NOT_ACTIF);
		if(field==null) return(Ubiquity.ERROR);
		FieldValue existing=findField(field);
		if(existing!=null){
			fields.remove(existing);
			fireChangeState(field, existing.value, null);	//notification locale			
		}
		return(Ubiquity.OK);			
	}
	//]
	
	protected void fireChangeState(String field, Value oldValue, Value newValue){	//notification locale
		//Au niveau des écouteurs sur l'agent lui même
		if(listener!=null) synchronized(listener) { for(StateListener l : listener) ThreadStack.perform( () -> l.stateChange(thisParticipant, field, oldValue, newValue)); }
		//Au niveau du monitor
		thisParticipant.node.space.notifyStateChange(thisParticipant, field, oldValue, newValue); //on notifie l'ajout sur le spot local
	}
	
	//[Pour Identity
	protected synchronized void writeCompleteState(DataWriter w){ //[s: nb_entry] {chaque_valeur: [st: name] [v: value]}
		int nb=(fields==null?0:fields.size());
		w.i(nb);
		if(nb>0) for(FieldValue fv : fields) w.s(fv.name).v(fv.value);
	}
	protected synchronized State readCompleteState(int nb, DataReader dbrb){//{chaque_valeur}
		Vector<FieldValue> old_fields=fields;
		fields=new Vector<FieldValue>();
		FieldValue nfv;
		for(int i=0; i<nb; i++) {
			String field_name=dbrb.readString();
			Value newValue = dbrb.readValue();
			nfv = null;
			if(old_fields!=null){
				for(int j=0; j<old_fields.size(); j++) {
					FieldValue cur = old_fields.get(j);
					if(cur.name.equals(field_name)) {
						nfv=cur;
						old_fields.remove(j);
						break;
					}
				}
			}
			if(nfv!=null){
				Value oldValue = nfv.value;
				nfv.value=newValue;
				fields.add(nfv);
				if(!oldValue.equals(nfv.value)) fireChangeState(field_name, oldValue, newValue);
			} else {
				nfv=new FieldValue(field_name, newValue);
				fields.add(nfv);
				fireChangeState(field_name, null, newValue);
			}							
		}
		if(old_fields!=null) for(FieldValue cur : old_fields){
			 fireChangeState(cur.name, cur.value, null);
		}		
		old_fields=null;
		return(this);
	}
	protected synchronized void clear(){
		if(fields!=null) {
			for(FieldValue cur : fields) fireChangeState(cur.name, cur.value, null);
			fields.removeAllElements();
		}
	}		
	//]
}

class FieldValue {
	public String name;
	public Value value;
	public FieldValue(String n, Value v){
		name=n;
		value=v;
	}
}

//Coquille de lecture seule
class StateProxyReader implements RState {
	State kernel;

	public StateProxyReader(State k){ kernel=k; }
	
	public Participant participant(){ return(kernel.participant()); }		
	public int nbFields(){ return(kernel.nbFields()); }
	public String fieldName(int num){ return(kernel.fieldName(num)); }
	public Value fieldValue(int num){ return(kernel.fieldValue(num)); }
	public Value get(String field){ return(kernel.get(field)); }	
	public boolean haveField(String field){ return(kernel.haveField(field)); }
	public void addListener(StateListener sl){ kernel.addListener(sl); }
	public void removeListener(StateListener sl){ kernel.removeListener(sl); }	
}