package skuad.ubiquity.implementation;

import skuad.util.databytes.*;
import skuad.SKUAD;

import java.util.*;
import skuad.hubudp.*;

public class USpot {
	
	Spot spot;
	Map<String, DistantNode> allNodes;	//tous les spots lié à cet hote qui sont associé à un space connecté localement, key : idSpace de l'hote distant pour le space associé au spot
	String md5Key;				//clé d'encodage des mots de passe de cet hote
	UbiquityManager umanager;
	boolean isUp;
	
	public USpot(UbiquityManager vumanager, Spot vspot, String vmd5Key){
		umanager = vumanager;
		spot = vspot;
		md5Key = vmd5Key;
		allNodes = new HashMap<String, DistantNode>();
		isUp=true;
	}
		
	public void disconnect(){ //l'hote correspondant vient de se déconnecter	
		isUp = false;
		for(DistantNode node : allNodes.values()) node.clear(); //-> purge de tous les spots
		allNodes.clear();
	}
	
	//[Gestion des messages reçus depuis cet hote distant
	protected void perform(int cmd, DataReader mess){
		DistantNode dnode;
		switch(cmd){
		case Protocole.LIST_SPACES :    //[LIST_SPACES] (int : nb_space ) {pour chaque space: (string: idSpace_emetteur) }	
			int nb_space = mess.readInt();
			for(int i=0; i<nb_space; i++) isConnectedToSpace( mess.readString());
			break;
		case Protocole.ADD_SPACE:	//[ADD_SPACE] (string: idSpace_emetteur) 
			isConnectedToSpace(mess.readString());
			break;
		case Protocole.GET_FULL_NODE_STATE : //[GET_FULL_NODE_STATE]	
			ImplSpace ispace = umanager.getSpaceByExternalId(mess.readString(), md5Key);
			if(ispace!=null) ispace.localNode.sendFullSpotState(spot);	//ce space est bien relié à l'hote local
			else { //cette situation ne devrait normalement pas se produire...
				System.out.println("--- Ubiquity/USpot : query GET_FULL_NODE_STATE on not connected space !");
			}
			break;
		case Protocole.FULL_NODE_STATE: //[FULL_NODE_STATE] (string: idSpace_emetteur) (int: nb_lparticipant) { etat_complet_de_chaque_lparticipant } 
//System.out.println("--> FULL_NODE_STATE receive from "+spot.getIdTarget());			
			dnode = isConnectedToSpace(mess.readString());
			if(dnode!=null) dnode.initialise(mess); //ce sera normalement toujours le cas ! (car l'émetteur envoie FULL_NODE_STATE seulement s'il est certain que le destinataire a une connexion locale vers son space)
			break;
		case Protocole.GET_PARTCIPANT_STATE:	//[GET_PARTCIPANT_STATE] (string: idSpace_receveur ) (string: localId)
			dnode = isConnectedToSpace(mess.readString());
			if(dnode!=null){ //-> on doit répondre: [FULL_PARTCIPANT_STATE] (string: idSpace_emetteur) (string: localId) (int: version) {desc_lparticipant}, version ==-1 si le participant n'existe pas
				String pid = mess.readString();
				ImplParticipant p = dnode.space.localNode.getParticipant(pid);
				DataWriter writer = DataBytes.writer().b(Protocole.FULL_PARTCIPANT_STATE).s(dnode.id_space);
				if(p!=null) p.writeCompleteState(writer);
				else writer.s(pid).i(-1); //ce participant n'existe pas
				spot.send(SKUAD.UBIQUITY, writer.flushReader());
			}
			break;
		case Protocole.FULL_PARTCIPANT_STATE:	//[FULL_PARTCIPANT_STATE] (string: idSpace_emetteur) (string: localId) (int: version) {desc_lparticipant}  [FULL_PARTCIPANT_STATE] (string: idSpace_emetteur) (string: localId) (int: version) {desc_lparticipant}, version ==-1 si le participant n'existe pas
			dnode = isConnectedToSpace(mess.readString());
			if(dnode!=null){ //-> on doit répondre: [FULL_PARTCIPANT_STATE] (string: idSpace_emetteur) (string: localId) (int: version) {desc_lparticipant}, version ==-1 si le participant n'existe pas
				dnode.updateParticipant(mess.readString(), mess.readInt(), mess);
			}
			break;
		case Protocole.NOTIFY_ADD_PARTICIPANT: //[NOTIFY_ADD_PARTICIPANT] (string: idSpaceSite_emetteur) (string : local_participant_id)
			dnode = isConnectedToSpace(mess.readString());
			if(dnode!=null) dnode.internalAddParticipant(mess.readString()); //ce sera normalement toujours le cas
			break;
		case Protocole.NOTIFY_REMOVE_PARTICIPANT : //[NOTIFY_REMOVE_PARTICIPANT] (string: idSpaceSite_emetteur) (string : local_participant_id) (int : version)
			dnode = isConnectedToSpace(mess.readString());
			if(dnode!=null) dnode.internalRemoveParticipant(mess.readString(), mess.readInt()); //ce sera normalement toujours le cas			
			break;
		case Protocole.NOTIFY_ADD_SKILL :	//[NOTIFY_ADD_SKILL] (string: idSpaceSite_emetteur) (string : local_participant_id) (int : version) (string: nameSkill)
			dnode = isConnectedToSpace(mess.readString());
			if(dnode!=null) dnode.internalAddSkill(mess.readString(), mess.readInt(), mess.readString()); 						
			break;
		case Protocole.NOTIFY_REMOVE_SKILL :		//[NOTIFY_REMOVE_SKILL] (string: idSpaceSite_emetteur) (string : local_participant_id) (int : version) (string: nameSkill)
			dnode = isConnectedToSpace(mess.readString());
			if(dnode!=null) dnode.internalRemoveSkill(mess.readString(), mess.readInt(), mess.readString()); 						
			break;	
		case Protocole.NOTIFY_ALIAS_PRIORITY : //[NOTIFY_ALIAS_PRIORITY]  (string: idSpaceSite_emetteur) (string : local_participant_id) (int : version) (string: nameAlias) (int: newPriority)
			dnode = isConnectedToSpace(mess.readString());
			if(dnode!=null) dnode.internalAliasPriority(mess.readString(), mess.readInt(), mess.readString(), mess.readInt()); 			
			break;
		case Protocole.NOTIFY_REMOVE_ALIAS : //[NOTIFY_REMOVE_ALIAS]  (string: idSpaceSite_emetteur) (string : local_participant_id) (int : version) (string: nameAlias)
			dnode = isConnectedToSpace(mess.readString());
			if(dnode!=null) dnode.internalRemoveAlias(mess.readString(), mess.readInt(), mess.readString()); 						
			break;
		case Protocole.NOTIFY_STATE_CHANGE ://[NOTIFY_STATE_CHANGE] (string: idSpaceSite_emetteur) (string : local_participant_id) (int : version) (string: nameField) (value: newValue)
			dnode = isConnectedToSpace(mess.readString());
			if(dnode!=null) dnode.internalStateChange(mess.readString(), mess.readInt(), mess.readString(), mess.readValue()); 						
			break;	
		case Protocole.NOTIFY_ADD_ACTIVITY : //[NOTIFY_ADD_ACTIVITY] (string: idSpaceSite_emetteur) (string : local_participant_id) (int : version) (string: nameActivity)
			dnode = isConnectedToSpace(mess.readString());
			if(dnode!=null) dnode.internalAddActivity(mess.readString(), mess.readInt(), mess.readString()); 						
			break;			
		case Protocole.NOTIFY_REMOVE_ACTIVITY : //[NOTIFY_REMOVE_ACTIVITY] (string: idSpaceSite_emetteur) (string : local_participant_id) (int : version) (string: nameActivity)
			dnode = isConnectedToSpace(mess.readString());
			if(dnode!=null) dnode.internalRemoveActivity(mess.readString(), mess.readInt(), mess.readString()); 						
			break;	
		case Protocole.NOTIFY_ACTIVITY_CHANGE_MODE : //[NOTIFY_ACTIVITY_CHANGE_MODE] (string: idSpaceSite_emetteur) (string : local_participant_id) (int : version) (string: nameActivity) [newMode]
			dnode = isConnectedToSpace(mess.readString());
			if(dnode!=null) dnode.internalChangeModeActivity(mess.readString(), mess.readInt(), mess.readString(),  (byte) mess.readByte()); 						
			break;	
		case Protocole.NOTIFY_ACTIVITY_CHANGE_STATUS : //[NOTIFY_ACTIVITY_CHANGE_STATUS] (string: idSpaceSite_emetteur) (string : local_participant_id) (int : version) (string: nameActivity) [newStatus]
			dnode = isConnectedToSpace(mess.readString());
			if(dnode!=null) dnode.internalChangeStatusActivity(mess.readString(), mess.readInt(), mess.readString(), (byte)  mess.readByte()); 							
			break;	
		case Protocole.NOTIFY_ACTIVITY_CHANGE_PROGRESS : //[NOTIFY_ACTIVITY_CHANGE_PROGRESS] (string: idSpaceSite_emetteur) (string : local_participant_id) (int : version) (string: nameActivity) (int: currentStep) (int: maxStep)
			dnode = isConnectedToSpace(mess.readString());
			if(dnode!=null) dnode.internalChangeProgress(mess.readString(), mess.readInt(), mess.readString(), mess.readInt(), mess.readInt()); 						
			break;	
		case Protocole.POST_MESSAGE ://[POST_MESSAGE](string: idSpaceSite_receveur)(string: id_sender)(string: id_target)(string: skill_context)(value: mess)(string: idreport)   ->idreport=="" si pas de report
			dnode = isConnectedToSpace(mess.readString());
			if(dnode!=null) dnode.internalReceiveMessage(mess.readString(), mess.readString(), mess.readString(), mess.readValue(), mess.readString()); 						
			break;		
		case Protocole.SENDING_REPORT : //[SENDING_REPORT] (string: idSpaceSite_receveur)(string: id_sender) (string: id_target)(string: id_report)[code_report]
			dnode = isConnectedToSpace(mess.readString());
			if(dnode!=null) dnode.internalSendingReport(mess.readString(), mess.readString(), mess.readString(), mess.readByte()); 									
			break;
		}	
	}
		
	protected synchronized DistantNode isConnectedToSpace(String idspace){ //analyse si le space est aussi connecté localement, si c'est le cas renvoie l'instance du spot distant pour ce space (le crée si nécessaire)
		//sinon renvoie null
		DistantNode dnode = allNodes.get(idspace);
		if(dnode==null){ //c'est un nouveau space pour cet hote
			ImplSpace ispace = umanager.getSpaceByExternalId(idspace, md5Key);
			if(ispace!=null){	//ce space est relié à l'hote local
				//on enregistre l'instance DistantSpot
				dnode = new DistantNode(ispace, idspace, this);
				allNodes.put(idspace, dnode);
//System.out.println("-- space-connexion : space = "+idspace+" // distant spot = "+spot.getIdTarget());	
				//on va demande le FULL_NODE_STATE de ne noeud
				spot.send(SKUAD.UBIQUITY, DataBytes.writer().b(Protocole.GET_FULL_NODE_STATE).s(idspace).flushReader());
				//on envoie le message [FULL_NODE_STATE] vers l'hote distant
				ispace.localNode.sendFullSpotState(spot); //NB: il peut y avoir un double envoie si une demande GET_FULL_NODE_STATE sera émise par le spot cible, mais si c'est le cas c'est que le spot cible n'était pas encore prés l'ors du 1er envoie
					//réciproquement : cet envoie ci doit être conservé car il se peut que le spot local n'était pas prés lors de la réception de la demande GET_FULL_NODE_STATE et n'y a donc potentiellement pas répondu !
			}
		}
		return(dnode);
	}
	//]
}