/**
	API du service :
 
  ---------------------------------------------------------------------------
  
	Protocole d'échange UBIQUITY (sous protocole HubUdp)
	
	 [cmd_ubiquity] ...
	 
cmd_udaudp :
	
//	[DEVICE_LIST] 		-> demande de la liste des devices exposés sur l'hote receveur

	[ADD_SPACE] (string: name_space) (string: "" ou si password present md5_du_password_crypté_par_md5Key_du_site_receveur)


*/
package skuad.ubiquity.implementation;

import java.util.*;
import skuad.hubudp.*;
import skuad.SKUAD;
import skuad.ubiquity.space.Space;
import skuad.util.databytes.*;
import skuad.util.base.*;

public class UbiquityManager implements MessageReceiver, SpotObserver {
	
	Map<String, ImplSpace> allSpaces;	//table des spaces relié à l'hote local, key = idspace_local
	Map<String, USpot> allUSpot; 		//table des spots connectés au réseau Ubiquity : key = idTarget
	String localMd5Key;
	Hub hub;
	
	public UbiquityManager(Hub vhub){
		hub = vhub;
		allSpaces=new HashMap<String, ImplSpace>();
		allUSpot = new HashMap<String, USpot>();
		localMd5Key = Util.randString(8);
		
		//On s'enregistre comme gestionnaire du protocole UBIQUITY
		hub.addReceiver(SKUAD.UBIQUITY, "Ubiquity", this);
		hub.addSpotObserver(this);
		hub.whenReady( (LocalSpot localspot) -> localspot.setString(SKUAD.SMD_UBIQUITY, localMd5Key) );
	}
	//]
	
	//[API	
	public String cryptPass(String password){ return(Util.cryptMD5(password, localMd5Key)); }
	public synchronized Space getSpace(String name, String password){
		if(password!=null && password.length()==0) password=null;
		if(name==null || name.length()==0) return(new ImplSpace(hub, null, null, password)); //il s'agit d'un space anonyme
		if(name.equals("#new")){ 
			//génération d'un nouveau space avec nom aléatoire
			String pass_part = (password==null ? null : "#"+cryptPass(password));
			do name=Util.randString(6); while(allSpaces.containsKey(pass_part==null?name:name+pass_part));
		} else if(name.contains("#")) return(null); //l'usage du symbole # est interdit dans le nom d'un space

		String idSpace=(password==null?name:name+"#"+cryptPass(password));
		ImplSpace space=allSpaces.get(idSpace);
		if(space==null) { //création de l'objet space
			space=new ImplSpace(hub, idSpace, name, password);
			allSpaces.put(idSpace, space);
			
			//on doit envoyer un message [ADD_SPACE] à tous les hotes de allUSpot
			DataReader mess_add_space=null;
			for(USpot uspot : allUSpot.values()){
				if(mess_add_space==null) mess_add_space=DataBytes.writer().b(Protocole.ADD_SPACE).s(idSpace).flushReader();
				uspot.spot.send(SKUAD.UBIQUITY, mess_add_space, true, false);
			}
			if(mess_add_space!=null) mess_add_space.dispose();
		}
		return(space);	
	}
	//]
	
	//[Interface SpotObserver
	public void up(Spot dspot){}
	public void updateSMD(Spot dspot, boolean first_definition){
		if(dspot.isLocalSpot()) return; //RAS si c'est l'hote local
		String md5Key = dspot.getString(SKUAD.SMD_UBIQUITY);
		if(md5Key!=null) connectSpot(dspot, md5Key); //dhost possède le support UBIQUITY alors il faut l'ajouter à allUSpot
		else disconnectSpot(dspot); //si dhost est dans allUSpot, alors il faut le deconnecter
	}
	public void down(Spot dhost){
		disconnectSpot(dhost); //si dhost est dans allUSpot, alors il faut le deconnecter
	}
	public void idContext(Spot dhost, String ch_context, String id_context){}
	//]
	
	//[Traitement interne
	protected synchronized void disconnectSpot(Spot dhost){ //Suppression de dhost dans allUSpot
		USpot uhost = allUSpot.remove(dhost.getIdTarget());
		if(uhost != null) uhost.disconnect();
	}
	protected synchronized void connectSpot(Spot dhost, String md5Key){ //Ajout de dhost dans allUSpot
		USpot uhost = allUSpot.get(dhost.getIdTarget());
		if(uhost == null){
			uhost = new USpot(this, dhost, md5Key);
			allUSpot.put(dhost.getIdTarget(), uhost);
			
			//on doit envoyer [GET_LIST_SPACES] à dhost
			dhost.send(SKUAD.UBIQUITY, Protocole.GET_LIST_SPACES);
		}
	}
	
	protected synchronized ImplSpace getSpaceByExternalId(String idspace, String md5key){
		int p = idspace.indexOf('#');
		String name, cpassword = null;
		if(p>0){
			name = idspace.substring(0, p);
			cpassword = idspace.substring(p+1);
		} else name = idspace;
		for(ImplSpace space : allSpaces.values()){	
			if(space.name.equals(name) && (
					(cpassword==null && !space.havePassword())
				|| space.isThisPassword(cpassword, md5key)
			)) return(space);
		}
		return(null);
	}
	//]
	
	//[Interface ProtocoleManager	
	public void receive(Spot sender, DataReader message, int mode){
		USpot uhost = allUSpot.get(sender.getIdTarget());
		int cmd = message.readByte();
//System.out.println("=== cmd receive = "+cmd);		
		switch(cmd){ //octet de commande
		case Protocole.GET_LIST_SPACES :		//[GET_LIST_SPACES] 
			DataWriter dwriter = DataBytes.writer();
			dwriter.b(Protocole.LIST_SPACES).i(allSpaces.size());
			for(String idSpace : allSpaces.keySet()) dwriter.writeString(idSpace);
			sender.send(SKUAD.UBIQUITY, dwriter.flushReader());
			break;
		default: //message spécique, qui doit être traité par uhost: LIST_SPACES, ADD_SPACE, FULL_SPOT_STATE, NOTIFY_ADD_PARTICIPANT... 
			if(uhost!=null) uhost.perform(cmd, message);
//else System.out.println("***> no uhost for cmd = "+cmd);				
		}
	    message.dispose();
	}
	//]
}