package skuad.ubiquity.space;

/**
 * Interface d'écoute des ajouts et suppression de <b>body</b> dans un space.<br/>
 * <br/>
 * Cette interface est celle que doivent implémenter les objets que l'on enregistre sur l'écoute des ajouts/suppressions des <b>body</b>.
 * <br/>
 * Il y a 3 niveaux d'écoute possibles :<br/>
 *  - centré sur un boy : ...
 *  - l'écoute locale à un spot : ...
 *  - l'écoute globale au space : ....
 * 
 * @see Ubiquity.Participant.BodyReader
 */
public interface ParticipantListener {
	/**
	 * Notifie qu'un nouveau particpant apparaît dans le space.<br/>
	 * <br/>
	 * Un tel évènement se produit à deux moments : soit quand un agent enregistré un interacteur dans le space au niveau d'un spot (une notification de ce type est alors déclenchée pour ce nouvel agent),
	 * soit quand un spot rejoint le space (dans ce cas une notification de ce type sera déclenchée pour chacun des agents portés par le nouveau noeud à destination de tous ceux à l'écoute dans
	 * le space rejoint, et dans le même temps une notification de ce type sera délcenchée pour chacun des agents présents avant l'insertion du spot à destination de tous les écouteurs d'identité
	 * enregistrés sur le spot qui s'insèrent).<br/>
	 * 
	 * @param identity descripteur d'identité de l'agent qui vient d'apparaitre.
	 */
	public void participantAdded(Participant p);
	
	/**
	 * Notifie qu'un agent vient de quitter le space.<br/>
	 * <br/>
	 * Comme pour l'ajout, ce type d'évènement peut se produire à deux échelles : pour un agent seulement, quand celui-ci est supprimé de son spot porteur; ou pour un ensemble d'agent quand 
	 * c'est un spot qui se retire du space.
	 * @param identity descripteur d'identité de l'agent qui vient de disparaitre.
	 */
	public void participantRemoved(Participant p);
}