//doc : a corriger
package skuad.ubiquity.space.activity;

import skuad.ubiquity.space.*;

/**
 * Interface d'écoute des facettes Activity des <b>body</b>.<br/>
 * <br/>
 * Cette intreface doit être implémentée par les objets qui s'enregistrent comme écouteur des activités exercés depuis un <b>body</b>
 * via les méthodes : {@link Ubiquity.body.activity.RActivity#addListener(ActivityListener)} ou 
 * {@link Ubiquity.Space#addActivityListener(ActivityListener)}.<br/>
 * <br/>
 * Notez que les activitées d'un <b>body</b> sont par défaut définit sur le niveau de visibilité privée, et leur existence
 * et évolution ne sont donc pas propagé dans le système. L'écoute n'est donc possible que sur les activités que l'agent
 * aura choisi explicitement de rendre visible (via le méthode {@link Ubiquity.body.activity.WActivity#visibility(String, byte)}
 * ou {@link Ubiquity.body.activity.WActivity#visibility(String[], byte)}.<br/>
 * <br/>
 * 
 * @see Ubiquity.body.activity.Ractivity, Ubiquity.Space
 */
public interface ActivityListener {

	/**
	 * Notifie qu'une nouvelle activité est enregistré sur le <b>body</b>.<br/>
	 * @param rbody Le descipteur du body concerné par l'ajout.
	 * @param newActivity Le nom de la nouvelle activité qui vient d'être définie.
	 */
	public void activityAdded(Participant rbody, String newActivity);
	
	/**
	 * Notifie qu'une activité vient d'être supprimée d'un <b>body</b>.<br/>
	 * @param rbody Le descipteur du body concerné par cette suppression.
	 * @param lostActivity Le nom de l'activité qui vient de supprimer.
	 */
	public void activityRemoved(Participant rbody, String lostActivity);	
	
	/**
	 * Signale qu'une activité vient de subir un changement de mode.<br/>
	 * <br/>
	 * Une activité peut prendre les valeurs de mode suivants :<br/>
	 * <ul>
	 * 	<li>{@link Ubiquity.SKUAD#ONCE} : l'activité est en exécution simple.</li>
	 * 	<li>{@link Ubiquity.SKUAD#CHAIN} : l'activité est en exécution chainée.</li>
	 * 	<li>{@link Ubiquity.SKUAD#REPEAT} : l'activité est en exécution périodique.</li>
	 * 	<li>{@link Ubiquity.SKUAD#CANCELED} : l'exécution de l'activité a été annulée.</li>
	 * 	<li>{@link Ubiquity.SKUAD#NONE} : l'activité est inexistante.</li>
	 * </ul>
	 * @param rbody Le descipteur du body concerné.
	 * @param activity Le nom de l'activité.
	 * @param old_mode L'ancienne valeur du mode l'activité.
	 * @param new_mode La nouvelle valeur du mode de l'activité.
	 */
	public void activityModeChange(Participant rbody, String activity, byte old_mode, byte new_mode);
	
	/**
	 * Signale que l'activité vient de subir un changement de statut.<br/>
	 * <br/>
	 * Une activité peut prendre les valeurs de statut suivantes :<br/>
	 * <ul>
	 * 	<li>{@link Ubiquity.SKUAD#WAITING} : l'activité est en attente d'une prochaine exécution.</li>
	 * 	<li>{@link Ubiquity.SKUAD#RUN} : l'activité est en cours d'éxécution.</li>
	 * 	<li>{@link Ubiquity.SKUAD#NOT_RUN} : l'activité n'est plus en cours d'exuction, et aucune futur exécution n'est programmée.</li>
	 * </ul>
	 * @param rbody Le descipteur du body concerné.
	 * @param activity Le nom de l'activité.
	 * @param old_status L'ancienne valeur du statut de l'activité.
	 * @param new_status La nouvelle valeur du statut de l'activité.
	 */
	public void activityStatusChange(Participant rbody, String activity, byte old_status, byte new_status);
	
	/**
	 * Signale la progression de l'exécution d'une activité.<br/>
	 * @param rbody Le descipteur du body concerné.
	 * @param activity Le nom de l'activité.
	 * @param currentStep Le numéro de l'étape courante.
	 * @param maxStep Le nombre maximum d'étape dans cette exécution.
	 */
	public void activityProgressChange(Participant rbody, String activity, int currentStep, int maxStep);	
	
}