package skuad.ubiquity.space.activity;

import skuad.util.value.Value;
import skuad.ubiquity.space.*;

/**
 * Interface des unités d'activtés d'un agent.<br/>
 * <br/>
 * Quand un agent définie une activité via son objet {@link Ubiquity.agent.activity.WActivity}, il doit fournir un objet qui 
 * implémente cette interface <b>ActivityRunner</b>. Cette dernière spécifie une méthode 
 * {@link #action(Interactor, String, Value, CancelFlag)} qui est celle devant contenir le code qui sera exécuté quand 
 * l'activité sera déclenchée.
 * <br/>
 * 
 * @see Ubiquity.agent.activity.WActivity 
 */
public interface ActivityRunner {
	
	/**
	 * Méthode exécutée au moment du déclenchement de l'activité.<br/>
	 * <br/>
	 * Cette méthode contient le code qui définie le traitement de l'activité.<br/>
	 * <br/>
	 * Notez que si le traitement réalisé par l'activité est un traitement long, vous devez prendre soin de vérifier régulièrement que
	 * l'activité est toujours active et qu'elle n'a pas été annulée. Pour cela vous devez consulter la valeur renvoyée par 
	 * {@link Ubiquity.agent.activity.ProgressFlag#is_cancel()} de l'objet {@link Ubiquity.agent.activity.ProgressFlag} reçue en argument 
	 * de cette méthode <b>action</b>.<br/>
	 * Notez également que les activités sont exécutées dans leur propre fil d'exécution (thread), il n'y a donc aucun risque à 
	 * utiliser des traitements longs, voir cycliques, si ce n'est de vérifier régulièrement l'état d'annulation afin d'y réagir 
	 * convenablement (arrêt du traitement long, ou de l'enchainement des cycles).<br/>
	 * Pour limiter cette pression dans le cas des traitements cycliques il est préférable d'utiliser le mode <b>repeat</b> pour
	 * déclencher l'activité via la méthode {@link Ubiquity.agent.activity.WActivity#repeat(String, Value, long, long)}. 
	 * En effet, dans ce mode l'état d'annulation de l'activité est géré automatiquement pour l'arret de l'exécution cyclique.
	 * 
	 * @param handle Handle de l'agent pour lequel cette activité est déclenchée.
	 * @param id_activity identifiant associé à l'activité. La valeur de cet identifiant est une chaine unique générée automatiquement 
	 * si l'activité n'a pas été définie avec un identifiant explicite via la méthode 
	 * {@link Ubiquity.agent.activity.Scheduler#define(String, Activity)}.  
	 * @param args Argument associé au lancement de cette activité. Cette argument est celui fourni au moment où l'ordre de déclenchement 
	 * a été donnée via l'une des méthodes <b>run</b>, <b>repeat</b> ou <b>chain</b> de l'objet {@link Ubiquity.agent.activity.WActivity}
	 * de l'agent.
	 * @param progress_flag Objet de progression et d'état d'annulation de l'activité
	 */
	public void action(Avatar avatar, String activity, Value args, ProgressFlag progress_flag);
	
}