package skuad.ubiquity.space.activity;

/**
 *  Moniteur de progression d'activité.<br/>
 *  <br/>
 *  Un objet de cette classe est passé à la méthode {@link Ubiquity.agent.activity.ActivityRunner#action(Ubiquity.agent.AHandle, String, value.Value, ProgressFlag)} 
 *  des activités au moment de leur lancement. Grâce à cet objet, le code de cette méthode peut vérifier à tout moment si l'activité a 
 *  été annulée ou non (méthode {@link #is_cancel()}), et mettre à jour la progression des traitement de l'activité (méthode {@linl #setProgress(int, int)}.<br/>
 */
public interface ProgressFlag {
	
	/**
	 * Test si l'activité est annulée ou non.<br/>
	 * <br/>
	 * La valeur de retour de cette méthode doit être régulièrement consultée dans le code d'exécution des activités, en particulier dans le cas où ces derniers
	 * réalisent un traitement long ou cyclique. Si la valeur obtenue est <b>true</b> le traitement de l'activité doit être stoppé.<br/>
	 * 
	 * @return <b>true</b> si l'activité est annulée, <b>false</b> sinon.
	 * 
	 * @see {@link Ubiquity.agent.activity.ActivityRunner#action(Ubiquity.agent.AHandle, String, value.Value, ProgressFlag)}
	 */
	public boolean is_cancel();
	
	/**
	 * Fixe la progression dans le traitement de l'activité.<br/>
	 * @param currentStep L'indice de l'étape courante.
	 * @param maxStep Le nombre d'étape total à réaliser.
	 */
	public void setProgress(int currentStep, int maxStep);
}