package skuad.ubiquity.space.activity;

import skuad.ubiquity.space.*;

/**
 * Interface de lecture de la facette Activity des agents Ubiquity.<br/>
 * <br/>
 * Cette interface de lecture peut être obtenue en tout point d'un space via la méthode {@link Ubiquity.agent.AIdentity#getActivity()} du descripteur d'identité
 * de l'agent ciblé.<br/> 
 * Elle permet de consulter les activités définies par l'agent et d'observer l'évolution de cette facette en enregistrant un écouteur (méthode {@link #addListener(ActivityListener)}.
 */
public interface RActivity {

	/**
	 * Consultation du coprs auquel est associé cette facette activité.
	 * @return descripteur d'identité du corps auquel appartient cette facette d'activité.
	 */
	public Participant participant();
	
	//[Consultation des activitées publiques
	/**
	 * Nombre d'activité définies sur cette facette.<br/>
	 * @return le nombre d'activité définies sur cette facette
	 */
	public int nbActivities(); 
	/**
	 * Consultation du nom d'une activité.
	 * <br/>
	 * Cette méthode permet de parcourir l'ensemble des noms (identifiants) des activités existantes sur cette façette.<br/>
	 * Ce parcours est à effectuer sur la base d'une numérotation arbitraire (suceptible de changer dans le temps) des actités 
	 * de 0 à (n-1) s'il y a <b>n</b> activités pour ce niveau de visibilité (valeur fournie par {@link #nbActivities()}).<br/>
	 * Cette numérotation est stable tant qu'aucune activité n'est pas ajoutée ou retirée.<br/>
	 * @param num numéro arbitraire de l'activité à consulter.
	 * @return le nom (identifiant) de l'activité de numéro arbitraire <b>num</b>
	 */	
	public String activityName(int num); //consultation de l'alias public de numéro arbitraire num, null si non existant
	public byte activityStatus(int num); //retourne : NOT_EXIST , NOT_RUN , WAITING, RUN, CANCELED
	public byte activityMode(int num); //retourne : NOT_EXIST , ONCE , CHAIN, REPEAT
	public int activityProgressMax(int num);
	public int activityProgressStep(int num);	
	public boolean haveActivity(String activity);
	public byte activityStatus(String activity); //retourne : NOT_EXIST , NOT_RUN , WAITING, RUN, CANCELED
	public byte activityMode(String activity); //retourne : NOT_EXIST , ONCE , CHAIN, REPEAT
	public int activityProgressMax(String activity);
	public int activityProgressStep(String activity);
	//]
	
	/**
	 * Enregistrement d'un écouteur d'activité.<br/>
	 * L'écouteur sera ici notifié que pour les changements subis par l'agent porteur de cette facette d'activité, pour une écoute plus globale (portant sur tous les agents du space),
	 * il faut utiliser la méthode {@link Ubiquity.Space#addActivityListener(ActivityListener)} de l'objet space.
	 * 
	 * @param al Objet à notifier lors des changements subis par cette facette d'activité.
	 */
	public void addListener(ActivityListener al);
	
	/**
	 * Désenregistrement d'un écouteur d'activité.<br/>
	 * 
	 * @param al Objet spécifié sur l'appel {@link #addListener(ActivityListener)} qu'il faut annuler ici.
	 */
	public void removeListener(ActivityListener al);
}