package skuad.ubiquity.space.alias;

import skuad.ubiquity.space.*;

/**
 * Interface du descripteur en lecture de la facette d'identité d'un agent.<br/>
 * <br/>
 * Cette facette identité est "exposée", cela signifie que la consultation de cette facette est possible pour n'importe quel agent du space.<br/>
 * Pour cela, tout agent peut obtenir le descripteur en lecture ({@link Ubiquity.agent.identity.RIdentity}) des identités d'un autre agent de diverses façons. Par exemple via le monitor et sa méthode {@link Ubiquity.Monitor#getAgent(String)}, 
 * ou encore en argument de la méthode utilisée pour livrer un message à l'agent : {@link Ubiquity.agent.FullParticipantAdpator#receive(AHandle , RIdentity , String , Value )} et dans laquelle l'expéditeur est spécifié
 * au travers de son descripteur d'identité.<br/>
 *
 */
public interface RAlias {

	/**
	 * Consultation de l'identité de l'agent auquel appartient cette facette.
	 * @return descripteur d'identité de l'agent auquel appartient cette facette fonctionnalité.
	 */
	public Participant participant();	
		
	/**
	 * Retourne la liste des noms de tous les alias attribués à l'agent<br/>
	 *
	 * @return la liste des noms de tous les alias attribués à l'agent<
	 */
	public int nbAlias(); //nombre d'alias public existant
	public String aliasName(int num); //consultation de l'alias public de numéro arbitraire num, null si non existant
	public int aliasPriority(int num);
	public boolean haveAlias(String alias);

	
	/**
	 * Lecture de la valeur de priorité d'un des alias de l'agent.<br/>
	 * 
	 * @param alias_name nom de l'alias pour lequel on souhaite connaitre la valeur de priorité.
	 * @return la valeur de priorité de alias de nom <i>alias_name</i> pour l'agent, ou 0 si l'agent ne porte pas cette alias.
	 */
	public int priority(String alias_name);
	
	/**
	 * Teste si l'agent est celui qui a le focus sur un alias donné.<br/>
	 * <br/>
	 * Voir {@link Ubiquity.agent.PartcipantListener.IdentityListener#aliasFocusChange(RIdentity, String, boolean, boolean)}.
	 * 
	 * @param alias_name nom de l'alias pour lequel on souhaite savoir si l'agent a ou non le focus.
	 * @return <b>true<b/> si l'agent a le focus, <b>false</b> sinon.
	 */
	public boolean haveFocus(String alias_name);	
		
	/**
	 * Ajout d'un écouteur d'identité local à l'agent auquel cette facette identité appartient.<br/>
	 * L'écouteur sera ici notifié que pour les changements subit par l'agent porteur de cette facette d'identité, pour une écoute plus global (portant sur tous les agents du space),
	 * il faut passer par l'objet monitor et utiliser la méthode {@link Ubiquity.Monitor#addIdentityAgentListener(PartcipantListener)}.
	 * 
	 * @param il Objet à notifier lors des changements subit par cette facette identité.
	 */	
	public void addListener(AliasListener il);
	
	/**
	 * Désenregistrement d'un écouteur d'identité local.<br/>
	 * 
	 * @param il Objet spécifié sur l'appel {@link #addListener(PartcipantListener)} qu'il faut annuler ici.
	 */
	public void removeListener(AliasListener il);
	
}