package skuad.ubiquity.space.alias;

/**
 * Interface du descripteur en écriture de la facette identité d'un agent.<br/>
 * <br/>
 * Comme pour toutes les facettes exposées d'un agent (identity, fonctionnality et state), les opérations en écriture ne sont possibles que pour 
 * l'agent qui porte la facette (ou le code ayant enregistré cet agent sur le spot local).<br/>
 * En effet, ce descripteur en écriture ne peut s'obtenir que sur l'handle de l'agent via la méthode {@link Ubiquity.agent.Interactor#myIdentity()}.<br/>
 * <b>NB :</b> seul l'agent lui-même, et le code ayant procédé à son enregistrement, possèdent une référence sur cet handle.<br/>
 * <br/>
 * Notez que les opérations d'écriture sur les facettes de l'agent ne sont possibles que si celui-ci est actif dans le space.<br/>
 * <br/>
 * Pour un composant agent conçu en héritant de la classe {@link Ubiquity.util.component.BaseAgentComponent}, ce descripteur en écriture est aussi disponible sous l'attribut <b>this.alias</b>.<br/>
 */
public interface WAlias extends RAlias {

	public RAlias reader();
		
	/**
	 * Définit ou change la priorité d'un alias de l'agent.<br/>
	 * <br/>
	 * Notez que fixer une priorité inférieur ou égale à 0 produit le même effet que la suppression de l'alias.<br/>
	 * 
	 * @param alias nom de l'alias concerné.
	 * @param newPriority valeur (ou nouvelle valeur si l'alias existe déjà sur cette agent) de priorité.
	 * @return <ul>
	 * 	<li>{@link Ubiquity.SKUAD#OK} si tout est ok.</li>
	 * 	<li>{@link Ubiquity.SKUAD#NOT_ACTIF} si l'agent est supprimé du space.</li>
	 * </ul>
	 */
	public byte setAlias(String alias); //priorité par défaut = 10
	public byte setAlias(String alias, int newPriority); //int pour éviter les cast (short)

	/**
	 * Supprime un alias de l'agent<br/>
	 * 
	 * @param alias nom de l'alias concerné.
	 * @return <ul>
	 * 	<li>{@link Ubiquity.SKUAD#OK} si tout est ok.</li>
	 * 	<li>{@link Ubiquity.SKUAD#NOT_ACTIF} si l'agent est supprimé du space.</li>
	 * </ul>
	 */
	public byte removeAlias(String alias);
	
}