package skuad.ubiquity.space.skill;

import skuad.ubiquity.space.*;

/**
 * Interface du descripteur en lecture de la facette de fonctionnalités d'un agent.<br/>
 * <br/>
 * Cette facette est "exposée", cela signifie que la consultation de cette facette est possible pour n'importe quel agent du space.<br/>
 * Pour cela, tout agent peut obtenir le descripteur en lecture ({@link Ubiquity.agent.fonctionnality.RFonctionnality}) des fonctionnalités d'un autre agent via la 
 * méthode {@link Ubiquity.agent.identity.RIdentity#getFonctionnality()}.<br/>
 * 
 */
public interface RSkill {
	
	/**
	 * Consultation de l'identité de l'agent auquel appartient cette facette.
	 * @return descripteur d'identité de l'agent auquel appartient cette facette fonctionnalité.
	 */
	public Participant participant();	
	
	/**
	 * Consultation des noms des compétences déclarées pour l'agent.<br/>
	 * 
	 * @return La liste des noms de compétences portées par l'agent.
	 */
	public int nbSkills(); //nombre de skill public existant
	public String skillName(int num); //consultation de la compétance public de numéro arbitraire num, null si non existant
	public boolean haveSkill(String skill);
	
	/**
	 * Ajout d'un écouteur de fonctionnalités locales à l'agent auquel cette facette fonctionnalité appartient.<br/>
	 * L'écouteur sera ici notifié que pour les changements subis par l'agent porteur de cette facette de fontionnalités, pour une écoute plus globale (portant sur tous les agents du space),
	 * il faut passer par l'objet monitor et utiliser la méthode {@link Ubiquity.Monitor#addFonctionnalityAgentListener(FonctionnalityListener)}.
	 * 
	 * @param sl Objet à notifier lors des changements subis par cette facette fonctionnalité.
	 */
	public void addListener(SkillListener sl);
	
	/**
	 * Désenregistrement d'un écouteur de fonctionnalité locale.<br/>
	 * 
	 * @param sl Objet spécifié sur l'appel {@link #addListener(FonctionnalityListener)} qu'il faut annuler ici.
	 */
	public void removeListener(SkillListener sl);
}