package skuad.ubiquity.space.skill;

import skuad.ubiquity.space.*;

/**
 * Interface d'écoute des facettes Fonctionnality des agents.<br/>
 * <br/>
 * Cette interface est celle que doit implémenter les objets que l'on enregistre sur l'écoute des changements subis par les facettes Fonctionnality des agents. Par cette 
 * écoute on reçoit donc une notification quand l'agent ajoute ou retire une compétence (skill) de sa facette Fonctionnality.<br/>
 * Il y a deux niveaux d'écoute possibles :<br/>
 *  - l'écoute locale, centrée sur un agent spécifique : cette écoute se met en oeuvre par l'enregistrement de l'écouteur directement sur le descripteur 
 * {@link Ubiquity.agent.fonctionnality.RFonctionnality} de l'agent ciblé, via la méthode {@link Ubiquity.agent.fonctionnality.RFonctionnality#addListener(FonctionnalityListener)}.
 * Ce descripteur de fonctionnalités s'obtient via la méthode {@link Ubiquity.agent.identity.RIdentity#getFonctionnality()} du descripteur d'identité de ce même agent.<br/>
 *  - l'écoute globale au space : à ce niveau on écoute les fonctionnalités de tous les agents, et pas seulement d'un seul d'entre eux comme précédemment. Pour se faire
 * l'écouteur doit cette fois être enregistré sur le monitor via la méthode {@link Ubiquity.Monitor#addFonctionnalityAgentListener(FonctionnalityListener)}.<br/>
 * 
 * @see Ubiquity.agent.fonctionnality.RFonctionnality
 * @see Ubiquity.Monitor
 */
public interface SkillListener {
	/**
	 * Signale l'ajout d'une compétence sur un agent.<br/>
	 * 
	 * @param agent le descripteur d'identité de l'agent concerné par cette ajout de compétence.
	 * @param newSkill nom de la compétence acquise par l'agent.
	 */
	public void skillAdded(Participant participant, String newSkill);

	/**
	 * Signale la perte d'une compétence sur un agent.<br/>
	 * 
	 * @param agent le descripteur d'identité de l'agent concerné par cette perte de compétence.
	 * @param lostSkill nom de la compétence perdue par l'agent.
	 */
	public void skillRemoved(Participant participant, String lostSkill);	
	
}