package skuad.ubiquity.space.skill;

/**
 * Interface du descripteur en écriture de la facette de fonctionnalité d'un agent.<br/>
 * <br/>
 * Comme pour toutes les facettes exposées d'un agent (identity, fonctionnality et state), les opérations en écriture ne sont possibles que pour 
 * l'agent qui porte la facette (ou le code ayant enregistré cet agent sur le spot local).<br/>
 * En effet, ce descripteur en écriture ne peut s'obtenir que sur l'handle de l'agent via la méthode {@link Ubiquity.agent.Interactor#myFonctionnality()}.<br/>
 * <b>NB :</b> seul l'agent lui même, et le code ayant procédé à son enregistrement, possède une référence sur cet handle.<br/>
 * <br/>
 * Notez que les opérations d'écriture sur les facettes de l'agent ne sont possiblee que si celui-ci est actif dans le space.<br/>
 * <br/>
 * Pour un composant agent conçu en héritant de la classe {@link Ubiquity.util.component.BaseAgentComponent}, ce descripteur en écriture est aussi disponible sous l'attribut <b>this.fonctionnality</b>.<br/>
 */
public interface WSkill extends RSkill {
	
	public RSkill reader();

	/**
	 * Déclare une compétence de l'agent.<br/>
	 * <br>
	 * Si la compétence est déjà déclarée, cette méthode sera sans effet.<br/>
	 * Après l'ajout l'ensemble des écouteurs (locaux ou globaux) des fontionnalités de l'agent porteur seront notifiées de cet ajout.
	 * 
	 * @param newSkill nom de la nouvelle compétence à ajouter.
	 * @return <ul>
	 * 	<li>{@link Ubiquity.SKUAD#OK} si tout est ok.</li>
	 * 	<li>{@link Ubiquity.SKUAD#NOT_ACTIF} si l'agent est supprimé du space.</li>
	 * </ul>
	 */
	public byte add(String newSkill);	
	
	/**
	 * Supprime une compétence de l'agent.<br/>
	 * <br>
	 * Si la compétence n'existe pas, cette méthode sera sans effet.<br/>
	 * Après la suppression l'ensemble des écouteurs (locaux ou globaux) des fontionnalités de l'agent porteur seront notifiées de cette suppression.
	 * 
	 * @param lostSkill compétence à supprimer
	 * @return  <ul>
	 * 	<li>{@link Ubiquity.SKUAD#OK} si tout est ok.</li>
	 * 	<li>{@link Ubiquity.SKUAD#NOT_ACTIF} si l'agent est supprimé du space.</li>
	 * </ul>
	 */
	public byte remove(String lostSkill);
	
}