package skuad.ubiquity.space.state;

import skuad.ubiquity.space.*;
import skuad.util.value.Value;


/**
 * Interface du descripteur en lecture de la facette d'état d'un agent.<br/>
 * <br/>
 * Cette facette est "exposée", cela signifie que la consultation de cette facette est possible pour n'importe quel agent du space.<br/>
 * Pour cela, tout agent peut obtenir le descripteur en lecture ({@link Ubiquity.agent.state.RState}) de l'état d'un autre agent via la 
 * méthode {@link Ubiquity.agent.identity.RIdentity#getState()}.<br/>
 * <br/>
 */
public interface RState {
	
	/**
	 * Consultation de l'identité de l'agent auquel appartient cette facette.
	 * @return descripteur d'identité de l'agent auquel appartient cette facette d'état.
	 */
	public Participant participant();	
	
	/**
	 * Consultation des noms de champs d'état existants de l'agent.<br/>
	 * 
	 * @return La liste des champs d'état existants.
	 */
	public int nbFields(); //nombre d'alias public existant
	public String fieldName(int num); //consultation de l'alias public de numéro arbitraire num, null si non existant
	public Value fieldValue(int num); 
	public Value get(String field);	
	public boolean haveField(String field);
	
	/**
	 * Ajout d'un écouteur d'état local à l'agent auquel cette facette d'état appartient.<br/>
	 * L'écouteur sera ici notifié que pour les changements subis par l'agent porteur de cette facette d'état, pour une écoute plus globale (portant sur tous les agents du space),
	 * il faut passer par l'objet monitor et utiliser la méthode {@link Ubiquity.Monitor#addStateAgentListener(StateListener)}.
	 * 
	 * @param sl Objet à notifier lors des changements subis par cette facette d'état.
	 */	
	public void addListener(StateListener sl);
	
	/**
	 * Désenregistrement d'un écouteur d'état local.<br/>
	 * 
	 * @param sl Objet spécifié sur l'appel {@link #addListener(StateListener)} qu'il faut annuler ici.
	 */	
	public void removeListener(StateListener sl);
}