package skuad.ubiquity.space.state;

import skuad.util.value.Value;
import skuad.ubiquity.space.Participant;

/**
* Interface d'écoute des facettes State des agents.<br/>
* <br/>
* Cette interface est celle que doit implémenter les objets que l'on enregistre sur l'écoute des changements subis par les facettes State des agents. Par cette 
* écoute on reçoit donc une notification quand l'agent modifie (ou ajoute/supprime) la valeur d'un des champs de sa structure d'état.<br/>
* Il y a deux niveaux d'écoute possibles :<br/>
*  - l'écoute locale, centrée sur un agent spécifique : cette écoute se met en oeuvre par l'enregistrement de l'écouteur directement sur le descripteur 
* {@link Ubiquity.agent.state.RState} de l'agent ciblé, via la méthode {@link Ubiquity.agent.state.RState#addListener(StateListener)}.<br/>
* Pour rappel, ce descripteur d'état s'obtient via la méthode {@link Ubiquity.agent.identity.RIdentity#getState()} du descripteur d'identité de ce même agent.<br/>
*  - l'écoute globale au space : à ce niveau on écoute les changements d'état de tous les agents, et pas seulement d'un seul d'entre eux comme précédemment. Pour se faire
* l'écouteur doit cette fois s'enregistrer sur le monitor via la méthode {@link Ubiquity.Monitor#addStateAgentListener(StateListener)}.<br/>
* 
* @see Ubiquity.agent.state.RState
* @see Ubiquity.Monitor
*/
public interface StateListener {
	
	/**
	 * Notifie le changemengt de valeur d'un des champs d'état de l'agent.<br/>
	 * <br/>
	 * Notez que pour un nouveau champ la valeur <i>oldValue</i> sera {@link value.Value#UNDEFINED}, et pour un champ qui est
	 * supprimé ce sera l'argument <i>newValue</i> qui aura pour valeur {@link value.Value#UNDEFINED}.
	 * @param identity descripteur d'identité de l'agent concerné.
	 * @param field nom du champ sur lequel porte cette notification.
	 * @param oldValue ancienne valeur associée à ce champ, ou {@link value.Value#UNDEFINED} si le champ n'existait par aupparavant.
	 * @param newValue nouvelle valeur prise par ce champ, ou {@link value.Value#UNDEFINED} si le champ est supprimé.
	 */
	public void stateChange(Participant participant, String field, Value oldValue, Value newValue);
	
}