package skuad.ubiquity.space.state;

import skuad.util.value.Value;


/**
 * Interface du descripteur en écriture de la facette d'état d'un agent.<br/>
 * <br/>
 * Comme pour toutes les facettes exposées d'un agent (identity, fonctionnality et state), les opérations en écriture ne sont possibles que pour 
 * l'agent qui porte la facette (ou le code ayant enregistré cet agent sur le spot local).<br/>
 * En effet, ce descripteur en écriture ne peut s'obtenir que sur l'handle de l'agent via la méthode {@link Ubiquity.agent.Interactor#myState()}.<br/>
 * <b>NB :</b> seul l'agent lui-même, et le code ayant procédé à son enregistrement, possède une référence sur cet handle.<br/>
 * <br/>
 * Notez que les opérations d'écriture sur les facettes de l'agent ne sont possibles que si celui-ci est actif dans le space.<br/>
 * <br/>
 * Pour un composant agent conçu en héritant de la classe {@link Ubiquity.util.component.BaseAgentComponent}, ce descripteur en écriture est aussi disponible sous l'attribut <b>this.state</b>.<br/>
 */
public interface WState extends RState {
	
	public RState reader();
	
	/**
	 * Fixe la valeur d'un champ d'état de l'agent.<br/>
	 * 
	 * @param field Nom du champ à fixer
	 * @param value Valeur à attribuer au champ, si cette valeur est null le traitement correspondra à celui de la suppression du champ
	 * @return <ul>
	 * 	<li>{@link Ubiquity.SKUAD#OK} si tout est ok.</li>
	 * 	<li>{@link Ubiquity.SKUAD#NOT_ACTIF}  si l'agent est supprimé du space.</li>
	 * </ul>
	 */
	public byte set(String field, Value value);	
	
	/**
	 * Supprime un champ d'état de l'agent.<br/>
	 * 
	 * @param field Nom du champ à supprimer
	 * @return <ul>
	 * 	<li>{@link Ubiquity.SKUAD#OK} si tout est ok.</li>
	 * 	<li>{@link Ubiquity.SKUAD#NOT_ACTIF}  si l'agent est supprimé du space.</li>
	 * </ul>
	 */
	public byte remove(String field);
}