/**
 * Ajoute la gestion des opérations de lecture à la classe MinimalDevice.
 * 
 *  /!\ la méthode ready(true) doit être invoqué à la fin de la configuration du composant !!! ainsi que chaque fois qu'une nouvelle configuration est terminée.
 * 
 * 
 * Méthode importante :
 * 	- Constructeur :  BaseDevice(String vname, int vtype_device, boolean vcan_log)
 *      ou   void setParams(String vname, int vtype_device, boolean vcan_log)
 *  - void setMoveCapacity(int vmovement_type, int vpower_min, int vpower_moy, int vpower_max)
 *  - int addSensor(String name, int type_sensor, int type_value, boolean observable)
 *  - int addActuator(String name, int type_actuator, int type_value)
 *  - void ready(boolean flag)	//si flag==true signale la fin de la configuration du device et demande l'envoie de la notification de l'état ready
 *  
 *  A masquer :
 *  - public boolean readSensor(Monitor mon, int sensor_id){ return(false); }
 *  - public boolean move(int cmd, int power){ return(false);}												=> fireMoved2D  // fireMoved3D
 *  - public boolean action(int actuator_id, int value){ return(action(actuator_id, (float) value)); }		=> fireSensorValue(int sensor_id, int value)
 *  - public boolean action(int actuator_id, float value){ return(false);}
 *  - public boolean action(int actuator_id, byte[] value){ return(false);}
 */
package skuad.uda.device;

import skuad.uda.*;
import java.util.*;

public class BaseDevice extends MinimalDevice {
	
	protected String name;
	protected int type_device;
	protected boolean can_log;
	protected int movement_type, power_min, power_moy, power_max;
	protected Vector<SADescriptor> sensors, actuators;
	protected boolean ready;
	protected boolean new_config_fired; 
	
	public BaseDevice(String vname, int vtype_device, boolean vcan_log){
		name=vname;
		type_device=vtype_device;
		can_log=vcan_log;
		movement_type=UDA.NO_MOVE;
		power_min=0; power_moy=0; power_max=0;
		sensors=null; actuators=null;
		ready=false; new_config_fired=true;
	}
	
	public int nbSensor(){ return(sensors==null ? 0 : sensors.size()); }
	public int nbActuator(){ return(actuators==null ? 0 : actuators.size()); }
	
	//[Interface Device
	//action
	public boolean log(String text){
		if(can_log) System.out.println(text);
		return(can_log);
	}
	
	//lecture
	public boolean getState(Monitor mon){ mon.state(this, ready, false); return(true); }
	public boolean getParams(Monitor mon){ mon.params(this, name, type_device, nbSensor(), nbActuator(), can_log); return(true); }
	public boolean getMoveCapacity(Monitor mon){ mon.moveCapacity(this, movement_type, power_min, power_moy, power_max); return(true);}
	public boolean getSensorParam(Monitor mon, int sensor_id){
		SADescriptor desc = (sensors==null || sensor_id>=sensors.size() ? null : sensors.get(sensor_id));
		if(desc==null) mon.sensorParam(this, sensor_id, "", UDA.NOT_EXIST, 0, false);
		else mon.sensorParam(this, sensor_id, desc.name, desc.type_sa, desc.type_value, desc.observable);
		return(desc!=null);
	}
	public boolean getActuatorParam(Monitor mon, int actuator_id){
		SADescriptor desc = (actuators==null || actuator_id>=actuators.size() ? null : actuators.get(actuator_id));
		if(desc==null) mon.actuatorParam(this, actuator_id, "", UDA.NOT_EXIST, 0);
		else mon.actuatorParam(this, actuator_id, desc.name, desc.type_sa, desc.type_value);
		return(desc!=null);
	}	
	//]
	
	//[Méthode pour les classes filles
		//Fixation des paramètres
	protected void setParams(String vname, int vtype_device, boolean vcan_log){
		name=vname;
		type_device=vtype_device;
		can_log=vcan_log;
		new_config_fired=false;
	}	
	protected void setMoveCapacity(int vmovement_type, int vpower_min, int vpower_moy, int vpower_max){
		movement_type=vmovement_type;
		power_min=vpower_min; power_moy=vpower_moy; power_max=vpower_max;
		new_config_fired=false;
	}
	protected int addSensor(String name, int type_sensor, int type_value, boolean observable){
		if(sensors==null) sensors=new Vector<SADescriptor>();
		int id = sensors.size();
		sensors.add(new SADescriptor(id, name, type_sensor, type_value, observable));
		new_config_fired=false;
		return(id);
	}
	protected int addActuator(String name, int type_actuator, int type_value){
		if(actuators==null) actuators=new Vector<SADescriptor>();
		int id = actuators.size();
		actuators.add(new SADescriptor(id, name, type_actuator, type_value, false));
		new_config_fired=false;
		return(id);
	}	
	protected void clearSensorsAndActuator(){
		if(sensors!=null) { sensors.removeAllElements(); new_config_fired=false; }
		if(actuators!=null) { actuators.removeAllElements(); new_config_fired=false; }
	}
	protected void ready(boolean flag){	//signale la fin de la configuration du device et demande l'envoie de la notification de l'état ready
		if(flag!=ready){
			ready=flag;
			this.fireState(ready, !new_config_fired);
			if(ready) new_config_fired=true;
		}
	}
}