/**
 * Implémentation Wrapper de Device :
 * -> toutes les méthodes de ce connecteur sont routé vers le device encapsulé.
 * 
 * Cette classe peut servir de classe de base pour un connecteur qui aurait besoin d'intercepter certaine méthode
 * du device encapsulé (voir par exemple le conencteur MoveAutoRepeater).
 */
package skuad.uda.device;

import skuad.uda.*;
import skuad.uda.monitor.*;

public class DeviceWrapper implements Device {
	
	Device slave;
	
	public DeviceWrapper(Device vslave){
		slave=vslave;
	}
	
	//[Interface Device
	//action
	public boolean log(String text){ return(slave.log(text));}
	public boolean move(int cmd, int power){ return(slave.move(cmd, power));}
	public boolean action(int actuator_id, int value){ return(slave.action(actuator_id, value));}
	public boolean action(int actuator_id, float value){ return(slave.action(actuator_id, value));}
	public boolean action(int actuator_id, byte[] value){ return(slave.action(actuator_id, value));}
	
	//lecture
	public boolean getState(Monitor mon){ return(slave.getState(mon)); }
	public boolean getParams(Monitor mon){ return(slave.getParams(MaskedMonitor.getMask(this, mon))); }
	public boolean getMoveCapacity(Monitor mon){ return(slave.getMoveCapacity(MaskedMonitor.getMask(this, mon))); }
	public boolean getSensorParam(Monitor mon, int sensor_id){ return(slave.getSensorParam(MaskedMonitor.getMask(this, mon), sensor_id)); }
	public boolean getActuatorParam(Monitor mon, int actuator_id){ return(slave.getActuatorParam(MaskedMonitor.getMask(this, mon), actuator_id)); }
	public boolean readSensor(Monitor mon, int sensor_id){ return(slave.readSensor(MaskedMonitor.getMask(this, mon), sensor_id)); }
	
	//écoute
	public boolean addListener(Monitor mon){ return(slave.addListener(MaskedMonitor.getMask(this, mon))); }
	public boolean removeListener(Monitor mon){ return(slave.removeListener(MaskedMonitor.getMask(this, mon))); }
	//]
}