/**
 * Implémentation minimal d'un Device :
 * 	- aucune action possible
 * 	- aucune lecture possible
 * 
 * -> seul la gestion des Moniteurs écouteurs est réalisée, mais les notifications (méthode fireXXX(....)) d'écoute doivent être déclenché par les sous-classes. 
 */
package skuad.uda.device;

import skuad.uda.*;
import skuad.util.thread.ThreadStack;

import java.util.*;

public class MinimalDevice implements Device {
	
	Vector<Monitor> monitors;
	
	public MinimalDevice(){
		monitors=null;
	}
	
	//[Interface Device
	//action
	public boolean log(String text){ return(false);}
	public boolean move(int cmd, int power){ return(false);}
	public boolean action(int actuator_id, int value){ return(action(actuator_id, (float) value)); }
	public boolean action(int actuator_id, float value){ return(false);}
	public boolean action(int actuator_id, byte[] value){ return(false);}
	
	//lecture
	public boolean getState(Monitor mon){ return(false); }
	public boolean getParams(Monitor mon){ return(false); }
	public boolean getMoveCapacity(Monitor mon){ return(false); }
	public boolean getSensorParam(Monitor mon, int sensor_id){ return(false); }
	public boolean getActuatorParam(Monitor mon, int actuator_id){ return(false); }
	public boolean readSensor(Monitor mon, int sensor_id){ return(false); }
	
	//écoute
	public synchronized boolean addListener(Monitor mon){
		if(mon==null) return(false);
		if(monitors==null) monitors=new Vector<Monitor>();
		else if(monitors.contains(mon)) return(false);
		synchronized(monitors) { monitors.add(mon); }
		return(true);
	}
	public boolean removeListener(Monitor mon){ 
		if(monitors==null) return(false); 
		synchronized(monitors) { return(monitors.remove(mon)); }
	}
	//]
		
		//Diffusion Réponse aux requêtes
	protected void fireSensorValue(int sensor_id, int value){
		if(monitors!=null)  synchronized(monitors) {
			for(Monitor mon : monitors) ThreadStack.perform(() -> mon.sensorValue(this, sensor_id, value));
		}
	}
	protected void fireSensorValue(int sensor_id, float value){
		if(monitors!=null)  synchronized(monitors) {
			for(Monitor mon : monitors) ThreadStack.perform(() -> mon.sensorValue(this, sensor_id, value));
		}
	}
	protected void fireSensorValue(int sensor_id, byte[] value){
		if(monitors!=null)  synchronized(monitors) {
			for(Monitor mon : monitors) ThreadStack.perform(() -> mon.sensorValue(this, sensor_id, value));
		}
	}
	protected void fireMoved2D(float dx, float dy, float da){
		if(monitors!=null)  synchronized(monitors) {
			for(Monitor mon : monitors) ThreadStack.perform(() -> mon.moved2D(this, dx, dy, da));
		}
	}
	protected void fireMoved3D(float dx, float dy, float dz, float ax, float ay, float az){
		if(monitors!=null)  synchronized(monitors) {
			for(Monitor mon : monitors) ThreadStack.perform(() -> mon.moved3D(this, dx, dy, dz, ax, ay, az));
		}
	}
	protected void fireLog(String text){
		if(monitors!=null)  synchronized(monitors) {
			for(Monitor mon : monitors) ThreadStack.perform(() -> mon.log(this, text));
		}
	}
	protected void fireState(boolean ready, boolean config_changed){
		if(monitors!=null)  synchronized(monitors) {
			for(Monitor mon : monitors) ThreadStack.perform(() -> mon.state(this, ready, config_changed));
		}
	}	
	//]
}