/**
 * MoveAutoRepeater est un connecter de Device qui applique une répétition (sur une fréquence spécifié) automatique des ordres de mouvement
 * sur un device slave. Et ce jusqu'à ce que l'ordre UDA.STOP soit formulé. 
 */
package skuad.uda.device;

import skuad.uda.*;
import skuad.util.thread.*;

public class MoveAutoRepeater extends DeviceWrapper implements Runnable {
	
	boolean in_run;
	int cur_move; int cur_force;
	long freq;	//fréquence de répétition des commandes
	
	public MoveAutoRepeater(Device vslave, long vfreq){
		super(vslave);
		cur_move = UDA.STOP;
		cur_force = 0;
		freq = vfreq;
		in_run = false;
	}
	
	public void setFrequency(long vfreq){ freq = vfreq; }
	public long getFrequency(){ return(freq); }
	
	public void run(){
		in_run=true; 
		while(in_run){
			if(cur_move!=UDA.STOP) slave.move(cur_move, cur_force);
			else break; //fin d'exécution du processus de répétition cyclique
			
			try{ Thread.sleep(freq); }catch(Exception e){} //on fait un pause correspondant à la fréquence de répétition
		}
		in_run = false;
	}
	
	public void stop(){ //arrête le processus de répétition cyclique s'il est actif (NB: cette méthode n'applique pas d'action sur le slave)
		cur_move = UDA.STOP;
	}
	
	//[interface Device
	public boolean move(int cmd, int power){
		cur_move = cmd; cur_force = power;
		if(cur_move!=UDA.STOP && !in_run) ThreadStack.perform(this); //on lance le processus de répétition cyclique
		else return(slave.move(cmd, power));
		return(true);
	}
	//]
}