package skuad.uda.gui;

import javax.swing.*;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import skuad.uda.*;
import skuad.uda.device.*;
import skuad.uda.monitor.*;
import skuad.util.gui.XYLayout.*;
import skuad.util.gui.*;

@SuppressWarnings("serial")
public class ControlPane extends JPanel implements MouseListener, ChangeListener, HaveGUI {
	
	Device device;
	JButton b_up,b_left, b_right, b_down;
	JSlider slider; int force;
	Monitor paramMonitor;
	boolean move_captured, device_ready;
	
	public ControlPane(Device vdevice){	
		device = vdevice;
		move_captured = false; device_ready=false;
		force = 0;
		//composant
		b_up=new JButton(ImagePanel.loadIcon(this,"image/up.png"));
		b_left=new JButton(ImagePanel.loadIcon(this,"image/left.png"));
		b_right=new JButton(ImagePanel.loadIcon(this,"image/right.png"));
		b_down=new JButton(ImagePanel.loadIcon(this,"image/down.png"));
		
		//mise en page
		setLayout(new XYLayout());
		setPreferredSize(new Dimension(250+80,250+10));
		add(b_up, Pos.d(88, 10, 74, 74));
		add(b_down, Pos.d(88, 166, 74, 74));
		add(b_left, Pos.d(10, 88, 74, 74));
		add(b_right, Pos.d(166, 88, 74, 74));

		slider = new JSlider(JSlider.VERTICAL, 0, 100, force);
		slider.setMajorTickSpacing(10);
		slider.setPaintTicks(true);
		slider.setPaintLabels(true);
		add(slider, Pos.d(260, 10,50,240));
		slider.addChangeListener(this);
		
		b_up.addMouseListener(this);
		b_left.addMouseListener(this);
		b_right.addMouseListener(this);
		b_down.addMouseListener(this);
		
		paramMonitor = new MinimalMonitor(){
			public void moveCapacity(Device dev, int movement_type, int power_min, int power_moy, int power_max){
//System.out.println("mtype = "+movement_type+" pmin = "+power_min+" pmoy = "+power_moy+" pmax = "+power_max);				
				slider.setMinimum(power_min);
				slider.setMaximum(power_max);
				if(force==0 || force>power_max || force<power_min) {
					force=power_moy;
					slider.setValue(power_moy);	
				}
				slider.setMajorTickSpacing(power_max>100 ? 20 : 10);		
				move_captured=true;
				setEnabled(device_ready && move_captured);
			}
			public void state(Device dev, boolean ready, boolean config_changed){
//System.out.println("ControlPane receive STATE = "+ready+" changed = "+config_changed);   				
				if(ready && (config_changed || !move_captured)) device.getMoveCapacity(paramMonitor); 
				device_ready=ready;
				setEnabled(device_ready && move_captured);
			}
//			public void params(Device dev, String dname, int type_device, int nb_sensor, int nb_actuator, boolean can_log){System.out.println("TEST params : name = "+dname+" // type_device = "+type_device+" // nb_sensor = "+nb_sensor+" // nb_actuator = "+nb_actuator+ " // can_log = "+can_log);}
		};
		device.getState(paramMonitor);
		device.addListener(paramMonitor);
	}
	
	public JComponent getGUI(){ return(this); }
/*	public void refresh(){ refreshParams(); }
	public void refreshParams(){ 
//		device.getParams(paramMonitor); 
		device.getMoveCapacity(paramMonitor); 
	}
*/	
	public void setEnabled(boolean flag){
		super.setEnabled(flag);
		b_up.setEnabled(flag);
		b_left.setEnabled(flag);
		b_right.setEnabled(flag);
		b_down.setEnabled(flag);
		slider.setEnabled(flag);
	}
	
	public void showBorder(boolean flag){
		if(flag) setBorder(BorderFactory.createLineBorder(Color.BLACK));
		else setBorder(null);
	}
	
	public void mouseClicked(MouseEvent arg0) {}
	public void mouseEntered(MouseEvent arg0) {}
	public void mouseExited(MouseEvent arg0) {}

	@Override
	public void mousePressed(MouseEvent e) {
		JButton b_target=(JButton) e.getSource();
		if(b_target==b_up) device.move(UDA.FORWARD,force);
		else if(b_target==b_left) device.move(UDA.TURN_LEFT,force);
		else if(b_target==b_right) device.move(UDA.TURN_RIGHT,force);
		else if(b_target==b_down) device.move(UDA.BACK,force);
	}

	@Override
	public void mouseReleased(MouseEvent e) { device.move(UDA.STOP,0); }

	@Override
	public void stateChanged(ChangeEvent e) {
		force = slider.getValue();
	}
	
	//[Fonction utiles
	public static JFrame openInFrame(String title, Device device){
		JFrame frame = new JFrame(title);
		frame.setLocation(100,100);
		Container pane = frame.getContentPane();
		pane.setLayout(new BorderLayout());
			
		ControlPane control_pane = new ControlPane(device);
		pane.add(BorderLayout.CENTER, control_pane);
	
		frame.pack();
		frame.setVisible(true);
		return(frame);
	}
		//Test
	public static void main(String[] args){
		LogDevice lgd = new LogDevice("Robot1", UDA.LOGGER, true);
		lgd.setMoveCapacity(UDA.MOVE2D, 0, 100, 200);
		openInFrame("Test CF", new MoveAutoRepeater(lgd, 600)).setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
	}
	//]
}