/**
 * Panneau d'affichage de l'état de connexion 
 */
package skuad.uda.gui.net;

import skuad.uda.net.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.border.Border;
import javax.swing.border.CompoundBorder;

@SuppressWarnings("serial")
public class ConnectionPane extends JPanel implements ConnectionListener , ActionListener {

	//Les constantes
	public final static Color ERROR = new Color(255,173,187);
	public final static Color SUCCESS = new Color(173,255,194);
	public final static Color DISABLE = new Color(181,186,182);
	public final static Color NEUTRE = new Color(250,238,187);
	private final static Border margin = BorderFactory.createEmptyBorder(3,3,3,3);
	
	Connectable connectable;
	JPanel state_pane;
	JLabel state_label;
	JButton connect_bt;
	JTextField tf_ip, tf_port;
	
	public ConnectionPane(Connectable vc){
		connectable = vc;
		setLayout(new GridLayout(2,1));

		JPanel bt_pane=new JPanel();
		bt_pane.setLayout(new FlowLayout());
		
		tf_ip = new JTextField(9);
		tf_port = new JTextField(3);
		bt_pane.add(new JLabel("IP :")); bt_pane.add(tf_ip);
		bt_pane.add(new JLabel(" PORT :")); bt_pane.add(tf_port);
		connect_bt = new JButton();
		bt_pane.add(connect_bt);
		add(bt_pane);
		
		connect_bt.addActionListener(this);

		
		state_pane = new JPanel();
		state_pane.setLayout(new FlowLayout());
		state_label =new JLabel();
		state_label.setFont(new Font("Arial", Font.PLAIN, 18));
		state_pane.add(state_label);
		add(state_pane);
		
		setBorder(margin);
		setPreferredSize(new Dimension(380,80));
		
		vc.addConnectionListener(this);
		if(connectable.isConnected()) newConnection(null,null,0); 
		else lostConnection(null,null,0); 
	}
	
	public void showBorder(boolean flag){
		if(flag) { 
			Border border = BorderFactory.createLineBorder(Color.BLACK);
			setBorder(new CompoundBorder( margin, new CompoundBorder(border, margin)));
		} else setBorder(margin);
	}
	
	private void display(Color level, String mess){ //affiche un message dans la zone d'état
		state_label.setText(mess);
		state_pane.setBackground(level);
	}
	
	public void actionPerformed(ActionEvent e){
		if(connectable.isConnected()) connectable.disconnet();
		else {
			String ip=tf_ip.getText();
			int port = 0;
			try{ port=Integer.parseInt(tf_port.getText());	}
			catch(Exception ex){ display(ERROR, "Port non valide"); return; }
			connectable.connect(ip, port);
		}
	}
	
	//[interface ConnectionListener
	public void newConnection(String label, String ip, int port){
		display(SUCCESS, "Connecté au serveur");
		connect_bt.setText("disconnect"); 
		tf_ip.setEnabled(false); tf_port.setEnabled(false);
	}
	public void lostConnection(String label, String ip, int port){
		display(DISABLE, "non connecté");
		connect_bt.setText("connect"); 
		tf_ip.setEnabled(true); tf_port.setEnabled(true);	
	}
	public void errorConnection(String mess){
		System.out.println("ConnectionPane - errorConnection : "+mess);
	}
	public void warningConnection(String mess){
		System.out.println("ConnectionPane - warningConnection : "+mess);
	}
	//]
}