/**
 * Panneau d'affichage de l'état de connexion 
 */
package skuad.uda.gui.net;

import skuad.uda.*;
import skuad.uda.device.*;
import skuad.util.gui.*;
import skuad.uda.net.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.border.Border;
import javax.swing.border.CompoundBorder;

@SuppressWarnings("serial")
public class ServerPane extends JPanel implements ConnectionListener , ActionListener, HaveGUI {

	//Les constantes
	public final static Color ERROR = new Color(255,173,187);
	public final static Color SUCCESS = new Color(173,255,194);
	public final static Color DISABLE = new Color(181,186,182);
	public final static Color NEUTRE = new Color(250,238,187);
	private final static Border margin = BorderFactory.createEmptyBorder(3,3,3,3);
	
	Server server;
	JPanel state_pane;
	JLabel state_label;
	JButton start_bt;
	JTextField tf_port;
	
	public ServerPane(Server vs){
		server = vs;
		setLayout(new GridLayout(2,1));

		JPanel bt_pane=new JPanel();
		bt_pane.setLayout(new FlowLayout());
		
		tf_port = new JTextField(4);
		bt_pane.add(new JLabel(" PORT :")); bt_pane.add(tf_port);
		start_bt = new JButton();
		bt_pane.add(start_bt);
		add(bt_pane);
		
		start_bt.addActionListener(this);

		
		state_pane = new JPanel();
		state_pane.setLayout(new FlowLayout());
		state_label =new JLabel();
		state_label.setFont(new Font("Arial", Font.PLAIN, 18));
		state_pane.add(state_label);
		add(state_pane);
		
		setBorder(margin);
		setPreferredSize(new Dimension(300,80));
		vs.addConnectionListener(this);
		if(vs.isStarted()) newConnection("server", "", vs.getServerPort());
		else lostConnection("server", "", 0);
	}
	
	public JComponent getGUI(){ return(this); }
	public void refresh(){ }
	
	public void showBorder(boolean flag){
		if(flag) { 
			Border border = BorderFactory.createLineBorder(Color.BLACK);
			setBorder(new CompoundBorder( margin, new CompoundBorder(border, margin)));
		} else setBorder(margin);
	}
	
	private void display(Color level, String mess){ //affiche un message dans la zone d'état
		state_label.setText(mess);
		state_pane.setBackground(level);
	}
	
	public void actionPerformed(ActionEvent e){
		if(server.isStarted()) server.stop();
		else {
			int port = 0;
			try{ port=Integer.parseInt(tf_port.getText());	}
			catch(Exception ex){ display(ERROR, "Port non valide"); return; }
			server.start(port);
		}
	}
	
	//[interface ConnectionListener
	public void newConnection(String label, String ip, int port){
		if(label.equals("server")){
			 display(NEUTRE, "... serveur demarré ..."); 
			 tf_port.setText(""+port);
		} else if(label.equals("client")){
			display(SUCCESS, "Client plug: "+ip+":"+port);
		}
		 start_bt.setText("stop"); 
		 tf_port.setEnabled(false);
	}
	public void lostConnection(String label, String ip, int port){
		if(label.equals("server")){
			display(DISABLE, "serveur arrêté"); 
			start_bt.setText("start");
			tf_port.setEnabled(true);
		} else if(label.equals("client")){
			 display(NEUTRE, "... serveur demarré ...");
		}		
	}
	public void errorConnection(String mess){
		System.out.println("ServerPane - errorConnection : "+mess);
	}
	public void warningConnection(String mess){
		System.out.println("ServerPane - warningConnection : "+mess);
	}
	//]
	
	//[Test
	public static void main(String[] args){
		LogDevice device = new LogDevice("Server Side Device", UDA.LOGGER, true);
		device.setMoveCapacity(0, 0, 20, 60);
		HaveGUI.FrameFor(new ServerPane(new TcpServerDevice(device)), "Test SF", true).setResizable(false);
	}
	//]
}