package skuad.uda.implementation.udaudp;

import java.util.Vector;
import skuad.hubudp.Spot;
import skuad.hubudp.SpotObserver;
import skuad.util.databytes.DataBytes;
import skuad.util.databytes.DataReader;
import skuad.util.databytes.DataWriter;
import skuad.SKUAD;
import skuad.uda.Device;
import skuad.uda.Monitor;
import skuad.uda.UDA;

public class ExposedDevice implements Monitor, SpotObserver {
	UdaUdpManager manager;
	
	Device device;
	int type;
	int id;
	Vector<Spot> observers;	
	boolean actif;
	
	public ExposedDevice(UdaUdpManager vmanager, int i, Device d){
		manager = vmanager;
		id = i;
		device = d;
		type = UDA.UNKNOW;
		actif=true;
		observers=new Vector<Spot>();
		
		//on va surveiller les évolutions du device (en particulier pour surveiller les changements de valeur de device_type)
		device.addListener(this);
		device.getState(this);
	}
	
	protected void addObserver(Spot sender){
		synchronized(observers){ if(!observers.contains(sender)) {
			observers.add(sender);
			sender.addSpotObserver(this);
		}}
	}
	protected void removeObserver(Spot sender){
		synchronized(observers){ observers.remove(sender); }
		sender.removeSpotObserver(this);
	}

	protected void unexpose(){ //UNEXPOSE_DEVICE{int: device_id}
		if(actif){
			actif=false;
			DataReader cmd=null;
			for(Spot spot : observers){
				spot.removeSpotObserver(this);
				if(cmd==null) cmd = DataBytes.writer().b(UdaUdpManager.UNEXPOSE_DEVICE).i(id).flushReader();
				spot.send(SKUAD.UDA,cmd, false, true);
			}
			if(cmd!=null) cmd.dispose();
			device.removeListener(this);
			observers=null;
		}
	}
	
	//[nterface SpotObserver 
	public void up(Spot dspot){}
	public void updateSMD(Spot dspot, boolean first_definition){}
	public void down(Spot dspot){ //auto-purge du vecteur observers 
		synchronized(observers){ observers.remove(dspot); }
	}
	public void idContext(Spot dspot, String ch_context, String id_context){}
	//]
	
	//[Décodage des commandes TO_DEVICE reçu sur le device géré
	protected synchronized void performToDevice(Spot sender, DataReader message, int response_id){ //[TO_DEVICE] [device_id]{...paquet uda...} : -> le paquet est une commande à destination du device local device_id
		int cmd_uda = message.readByte();
		Monitor mon=null;
		if(UDA.needMonitor(cmd_uda)) mon=new FromDevice(manager, sender, id, response_id);
		int code_perform = UDA.decodeDevice(device, cmd_uda, message, mon);
		if(response_id>0 && (mon==null || code_perform!=UDA.OK)) manager.internalResponseDevice(sender, id, response_id, code_perform, null); 
	}
	//]
	
	//[Interface moniteur
		//Réponse aux requêtes
	public void params(Device dev, String name, int type_device, int nb_sensor, int nb_actuator, boolean can_log){ 
		if(actif && type_device!=type){ //changement de type
			type=type_device;
			manager.updateHiaDevice();
		}
	}
	public void moveCapacity(Device dev, int movement_type, int power_min, int power_moy, int power_max){}
	public void sensorParam(Device dev, int sensor_id, String name, int type_sensor, int type_value, boolean observable){}
	public void actuatorParam(Device dev, int actuator_id, String name, int type_actuator, int type_value){}

	//Ecoute
	public void state(Device dev, boolean ready, boolean config_changed){ if(!actif) return;
		if(config_changed || (ready && (type == UDA.UNKNOW))) device.getParams(this); //pour le cas ou le type_device aurait changé
		DataReader cmd=null;
		synchronized(observers){ for(Spot spot : observers){
			if(cmd==null) {
				DataWriter wcmd=DataBytes.writer().b(UdaUdpManager.FROM_DEVICE).i(id);
				UDA.encodeState(wcmd, ready, config_changed);
				cmd =  wcmd.flushReader();
			}
			spot.send(SKUAD.UDA,cmd, false, true);	
		}}
		if(cmd!=null) cmd.dispose();
	}
	public void sensorValue(Device dev, int sensor_id, int value){ if(!actif) return;
		DataReader cmd=null;
		synchronized(observers){ for(Spot spot : observers){
			if(cmd==null) {
				DataWriter wcmd=DataBytes.writer().b(UdaUdpManager.FROM_DEVICE).i(id);
				UDA.encodeSensorValue(wcmd, sensor_id, value);
				cmd =  wcmd.flushReader();
			}
			spot.send(SKUAD.UDA,cmd, false, true);	
		}}
		if(cmd!=null) cmd.dispose();		
	}
	public void sensorValue(Device dev, int sensor_id, float value){ if(!actif) return;
		DataReader cmd=null;
		synchronized(observers){ for(Spot spot : observers){
			if(cmd==null) {
				DataWriter wcmd=DataBytes.writer().b(UdaUdpManager.FROM_DEVICE).i(id);
				UDA.encodeSensorValue(wcmd, sensor_id, value);
				cmd =  wcmd.flushReader();
			}
			spot.send(SKUAD.UDA,cmd, false, true);	
		}}
		if(cmd!=null) cmd.dispose();		
	}
	public void sensorValue(Device dev, int sensor_id, byte[] value){ if(!actif) return;
		DataReader cmd=null;
		synchronized(observers){ for(Spot spot : observers){
			if(cmd==null) {
				DataWriter wcmd=DataBytes.writer().b(UdaUdpManager.FROM_DEVICE).i(id);
				UDA.encodeSensorValue(wcmd, sensor_id, value);
				cmd =  wcmd.flushReader();
			}
			spot.send(SKUAD.UDA,cmd, false, true);	
		}}
		if(cmd!=null) cmd.dispose();		
	}
	public void moved2D(Device dev, float dx, float dy, float da){ if(!actif) return;
		DataReader cmd=null;
		synchronized(observers){ for(Spot spot : observers){
			if(cmd==null) {
				DataWriter wcmd=DataBytes.writer().b(UdaUdpManager.FROM_DEVICE).i(id);
				UDA.encodeMoved2D(wcmd, dx, dy, da);
				cmd =  wcmd.flushReader();
			}
			spot.send(SKUAD.UDA,cmd, false, true);		
		}}
		if(cmd!=null) cmd.dispose();		
	}
	public void moved3D(Device dev, float dx, float dy, float dz, float ax, float ay, float az){ if(!actif) return;
		DataReader cmd=null;
		synchronized(observers){ for(Spot spot : observers){
			if(cmd==null) {
				DataWriter wcmd=DataBytes.writer().b(UdaUdpManager.FROM_DEVICE).i(id);
				UDA.encodeMoved3D(wcmd, dx, dy, dz, ax, ay, az);
				cmd =  wcmd.flushReader();
			}
			spot.send(SKUAD.UDA,cmd, false, true);		
		}}
		if(cmd!=null) cmd.dispose();			
	}
	public void log(Device dev, String text){ if(!actif) return;
		DataReader cmd=null;
		synchronized(observers){ for(Spot spot : observers){
			if(cmd==null) {
				DataWriter wcmd=DataBytes.writer().b(UdaUdpManager.FROM_DEVICE).i(id);
				UDA.encodeLog(wcmd,text);
				cmd =  wcmd.flushReader();
			}
			spot.send(SKUAD.UDA,cmd, false, true);		
		}}
		if(cmd!=null) cmd.dispose();			
	}
	//]	
}
