/**
   Gestionnaire UdaUdp spécifique à un hub.
   
   NB: dans un contexte d'exécution donné on a au maximum 1 seul gestionnaire UdaUdp par hub
    
 
	Protocole d'échange HubUDP UDA
	
	 [cmd_udaudp] ...
	 
cmd_udaudp :
	
	[DEVICE_LIST] 		-> demande de la liste des devices exposés sur l'hote receveur
	[EXPOSED_DEVICE]...	-> réponse à une demande DEVICE_LIST, syntaxe : 	[EXPOSED_DEVICE]{int: nb_device}{int: device_id1}{int: type_dev1}...{int: device_idN}{int: type_devN}		
				
		--> Opérations sur les devices locaux	
	[TO_DEVICE] {int: device_id}{...paquet uda...} : -> le paquet est une commande à destination du device local device_id
		
	[FROM_DEVICE] {int: device_id}{... paquet uda...} : ->  le paquet est un message du device distant (porté par l'hote émetteur du message)
			--> ce peut être une réponse à une requête [TO_DEVICE] ou une notification d'écoute pour les hote enregistré via [ADD_OBSERVER]
	
	[QUERY_DEVICE]{int: device_id}(int: response_id){...paquet uda...} : -> requête sur le device locale ciblé avec une attente de réponse via RESPONSE_DEVICE si response_id!=0
	[RESPONSE_DEVICE]{int: device_id}(int: response_id)[code_perform][bool: have_paquet_response]{...paquet uda...} --> réponse à un précédent envoie QUERY_DEVICE	
	
	[ADD_OBSERVER] {int: device_id}    --> l'hote émetteur s'inscrit comme écouteur sur le device local indiqué
	
	[REMOVE_OBSERVER] {int: device_id} --> supprime l'écoute de l'hote émetteur sur le device local indiqué

	[UNEXPOSE_DEVICE]{int: device_id}  --> émis quand le device porté par l'hote émetteur devient unexposé, ce message est envoyé à tous les hotes qui se sont inscrit en [ADD_OBSERVE] sur ce device
	
	[OBSERVER_LIST] --> retourne la liste complète des observers enregistré sur l'ensemble des devices gérés
	<-- [LIST_OBSERVER](int: nb_observer)(int: max_observer){int: device_id1}(int: nb_observer)(string: observer1_idglobal)...(string: observerN_idglobal)...{int: device_idM}...   max_observer==0 si pas de limite d'observer

	[LINKED_LIST] --> retourne la liste complète des devices distants observés
	<-- [LIST_LINKED](int: nb_linked)(string: linked1_idglobal){int: device_id1}...(string: linked2_idglobal){int: device_id2}...

*/
package skuad.uda.implementation.udaudp;

import skuad.hubudp.*;
import skuad.SKUAD;
import skuad.uda.*;
import skuad.util.databytes.*;
import java.util.*;
import skuad.uda.net.*;

public class UdaUdpManager implements MessageReceiver {
	
	//[Constante du protocole UdaUdp
	public final static byte DEVICE_LIST     = 20; 	
	public final static byte EXPOSED_DEVICE  = 21;  

	public final static byte TO_DEVICE       = 22;
	public final static byte FROM_DEVICE     = 23;

	public final static byte ADD_OBSERVER    = 24;
	public final static byte REMOVE_OBSERVER = 25;
	public final static byte UNEXPOSE_DEVICE = 26;
	public final static byte OBSERVER_LIST	= 27;
	public final static byte LIST_OBSERVER	= 28;

	public final static byte QUERY_DEVICE	= 29;
	public final static byte RESPONSE_DEVICE	= 30;	

	public final static byte LINKED_LIST		= 31;
	public final static byte LIST_LINKED		= 32;
	//]
	
	int nb_device_exposed=0;		//nombre de device exposé
	int next_id=1;					//id du prochain device à exposer
	int hi_device_version=0;		//numéro de version du code SMD.DEVICE
	Map<Device, Integer> managedDtoId = new  HashMap<Device, Integer>();		//table Device->Id des devices locaux exposés
	Map<Integer, ExposedDevice> managedIdtoD = new  HashMap<Integer, ExposedDevice>();	//table Id->ExposedDevice des devices locaux exposés
	
	Map<String, VirtualDevice> linkedDevice = new  HashMap<String, VirtualDevice>();		//table des devices distants linké : key = id_spot+":"+device_id	
	Vector<ExposedDeviceWaiter> edWaiters= new Vector<ExposedDeviceWaiter>();	//<- liste des demandeurs exposedDevices().
	int next_response_id=1;		//prochain identifiant response_id
	Map<Integer, ResponseDevice> reqWaiters = new  HashMap<Integer, ResponseDevice>();	//table des demandeurs RESQUEST_DEVICE ,  RESQUEST_DEVICE[response_id] = cb
	
	Hub hub;
	
	//[c'est une classe singleton
	public UdaUdpManager(Hub vhub){
		hub = vhub;
		//On s'enregistre comme gestionnaire du protocole UDAUDP
		hub.addReceiver(SKUAD.UDA, "UdaUdp", this);
	}
	//]
	
	//[API public
	public int expose(Device d){ //<- inscrit le device sur le gestionnaire UdaUdp du Hub indiqué, retourne l'id attribué au device d, 0 en cas d'erreur
		Integer existing_id=managedDtoId.get(d);
		if(existing_id!=null) return(existing_id.intValue()); //device déjà enregistré
		if(nb_device_exposed>=UdaUdp.MAX_EXPOSED_DEVICE) return(0);  //inscription impossible, nombre maximum de device gérable atteind 
		//on va enregistré ce device
		while(managedIdtoD.containsKey(next_id)){ //recherche du prochain id disponible
			next_id++;
			if(next_id<=0) next_id=1;
		}
		int device_id=next_id; ++next_id;
		//enregistrement du device
		managedDtoId.put(d, device_id);
		ExposedDevice edevice = new ExposedDevice(this, device_id, d);
		managedIdtoD.put(device_id,edevice);
		nb_device_exposed++;
		
		updateHiaDevice(); //mise à jour de la définition SMD_DEVICE
		
		return(device_id);
	}

	public int unexposeDevice(Device d){	//<- désinscrit le device sur le gestionnaire UdaUdp, retourne l'id qui avait été attribué au device d
		Integer existing_id=managedDtoId.remove(d);
		if(existing_id!=null){
			ExposedDevice edevice = managedIdtoD.remove(existing_id);
			nb_device_exposed--;

			updateHiaDevice(); //mise à jour de la définition SMD.DEVICE
			
			edevice.unexpose(); //on envoie les messages UNEXPOSE_DEVICE à tous les observers de ce device
			
			return(existing_id.intValue());
		}
		return(0);
	}

	public Device linkDevice(Spot target, int device_id){	//<- crée un lien vers le device distant géré par le spot target sous l'identifiant device_id
		String vid = target.getIdLocal()+":"+device_id;
		VirtualDevice vdevice = linkedDevice.get(vid);
		if(vdevice==null){ //c'est une nouvelle demande de link
			vdevice = new VirtualDevice(this, vid, target, device_id);
			linkedDevice.put(vid, vdevice);
		} else vdevice.incLink();
		return(vdevice);
	}
	
	protected void remove(VirtualDevice vdevice) { //pour la classe VirtualDevice
		linkedDevice.remove(vdevice.id);
	}
	
	public void exposedDevices(Spot target, ExposedDeviceResponse cb){	//<- invoce cb.exposedDevice(DynIpPort source, int[][] exposed_device) avec la liste des devices exposés par l'hote target 
				//(il peut s'agir de l'hote local): exposed_device[i][0]=device_id , exposed_device[i][1]=device_type
		if(target.isLocalSpot()){ //c'est l'hote local
			int[][] rep = new int[nb_device_exposed][2];
			int i=0;
			for(ExposedDevice edevice : managedIdtoD.values()){
				rep[i][0]=edevice.id;
				rep[i][1]=edevice.type;
				i++;
			}
			 cb.exposedDevice(target, rep);
		} else { //c'est un hote distant actif, on va regarder s'il a déclaré des devices sur son code SMD
			if(target.getInt(SKUAD.SMD_DEVICE)==0) cb.exposedDevice(target, new int[0][0]); //aucun device sur ce spot
			else {
				//il y a des device, on va envoyer une requete DEVICE_LIST
				edWaiters.add(new ExposedDeviceWaiter(target, cb));
				target.send(SKUAD.UDA, DEVICE_LIST);
			}
		}
	}
	
	public boolean requestDevice(Spot target, int device_id, DataReader paquet_request, ResponseDevice cb_response , boolean auto_dispose){ //<- lance une requête paquet_request sur le device ciblé, 
		//cb_response.deviceResponse(DynIpPort target, int device_id, int code_perform, DataReader paquet_response) sera déclenché en réponse --> paquet_response est une insatnce de DataReader qu'il faudra diposé aprés lecture ( .dispose())	
		int response_id=0;
		if(cb_response!=null){
			response_id = next_response_id++;	
			if(next_response_id>60000) next_response_id=1;
			reqWaiters.put(response_id,cb_response);
		}
		
		//-> [QUERY_DEVICE][device_id](int: response_id){...paquet uda...} : -> requête sur le device locale ciblé avec une attente de réponse via RESPONSE_DEVICE si response_id!=0
		DataWriter wcmd= DataBytes.writer();
		wcmd.writeByte(QUERY_DEVICE);
		wcmd.writeInt(device_id);
		wcmd.writeInt(response_id);
		wcmd.writeData(paquet_request);
		target.send(SKUAD.UDA, wcmd.flushReader());
		if(auto_dispose) paquet_request.dispose();
		return(true);
	}
	
	//]
	
	protected void updateHiaDevice(){ //on doit mettre à jour la définition SMD.DEVICE
		if(nb_device_exposed==0) {
			LocalSpot localspot = hub.localSpot();
			if(localspot!=null) localspot.removeAll(SKUAD.SMD_DEVICE);
		} else {
			hi_device_version++; if(hi_device_version>65000) hi_device_version=1;
			hub.whenReady( (LocalSpot localspot) -> localspot.setInt(SKUAD.SMD_DEVICE, hi_device_version));
		}
	}

	//[Interface ProtocoleManager	
	public synchronized void receive(Spot sender, DataReader message, int protocole){	//message.cmd == code protocole du message
				// /!\ aprés lecture, on doit appliquer un .dispose() sur le message
		int cmd = message.readByte(); DataWriter wcmd;
		int device_id, response_id;
		ExposedDevice edevice; VirtualDevice vdevice;
	    switch(cmd){ //octet de commande
	    case DEVICE_LIST :    //-> demande de la liste des devices gérés localement, réponse : [EXPOSED_DEVICE]...	
	    	wcmd=DataBytes.writer();
	    	wcmd.writeByte(EXPOSED_DEVICE);
	    	wcmd.writeInt(nb_device_exposed);
	    	for(ExposedDevice expdevice : managedIdtoD.values()){
	    		wcmd.writeInt(expdevice.id);
	    		wcmd.writeInt(expdevice.type);
	    	}
	    	sender.send(SKUAD.UDA, wcmd.flushReader());
	    	break;
	    case EXPOSED_DEVICE: //réponse à une demande [DEVICE_LIST] , syntaxe : 	[EXPOSED_DEVICE][nb_device]{int: device_id1}{int: type_dev1}...{int: device_idN}{int: type_devN}	
//System.out.println("(UdaUdp) device list for spot port = "+sender.getPort());	
	    	//on regarde si des ExposedDeviceWaiter sont en attente
	    	int [][] device_list=null;
	    	if(edWaiters!=null && !edWaiters.isEmpty()){
	    		for(int nedw=0; nedw<edWaiters.size(); nedw++){
	    			ExposedDeviceWaiter waiter = edWaiters.get(nedw); 
//System.out.println("\t finder for port = "+waiter.spot.getPort());	    			
	    			if(waiter.spot == sender){
//System.out.println("\t --> = is it !");	 	    				
	    				if(device_list==null){ //extraction de la liste
	    					int nd = message.readInt();
	    					device_list = new int[nd][2];
	    					for(int i=0; i<nd; i++){
	    						device_list[i][0]=message.readInt();
	    						device_list[i][1]=message.readInt();
	    					}
	    				}
	    				waiter.cb.exposedDevice(waiter.spot, device_list);
	    				edWaiters.removeElementAt(nedw); nedw--;
	    			}
	    		}
	    	}
	    	device_list=null;
	    	break;
	    case TO_DEVICE :      //[TO_DEVICE] {int: device_id}{...paquet uda...} : -> le paquet est une commande à destination du device local device_id
	    	device_id = message.readInt();
	    	edevice = managedIdtoD.get(device_id);
	    	if(edevice!=null) edevice.performToDevice(sender, message, 0);
	    	break;
	    case FROM_DEVICE : 	//[FROM_DEVICE] {int: device_id}{... paquet uda...} : ->  le paquet est un message du device distant (porté par l'hote émetteur du message)
	    			//--> ce peut être une réponse à une requête [TO_DEVICE] ou une notification d'écoute pour les hote enregistré via [ADD_OBSERVER]
	    	device_id = message.readInt();
	    	vdevice = linkedDevice.get(sender.getIdLocal()+":"+device_id);
	    	if(vdevice!=null) vdevice.performFromDevice(sender, message);
	    	break;	
	    case ADD_OBSERVER :   //[ADD_OBSERVER] {int: device_id}    --> l'hote émetteur s'inscrit comme écouteur sur le device local indiqué
	    	device_id = message.readInt();
	    	edevice = managedIdtoD.get(device_id);
	    	if(edevice!=null) edevice.addObserver(sender);
	    	break;
	    case REMOVE_OBSERVER :   //[REMOVE_OBSERVER] {int: device_id} --> supprime l'écoute de l'hote émetteur sur le device local indiqué
	    	device_id = message.readInt();
	    	edevice = managedIdtoD.get(device_id);
	    	if(edevice!=null) edevice.removeObserver(sender);
	    	break;
	    case UNEXPOSE_DEVICE :		//[UNEXPOSE_DEVICE]{int: device_id}  --> émis quand le device porté par l'hote émetteur devient unexposé, ce message est envoyé à tous les hotes qui se sont inscrit en [ADD_OBSERVE] sur ce device
	    	device_id = message.readInt();
			vdevice = linkedDevice.get(sender.getIdLocal()+":"+device_id);
	    	if(vdevice != null) {
	    		linkedDevice.remove(vdevice.id);
	    		vdevice._down(false);
	    	}
	    	break;
	    case OBSERVER_LIST : 	//[OBSERVER_LIST] --> retourne la liste complète des observers enregistré sur l'ensemble des devices gérés
	    		//<-- [LIST_OBSERVER](int: nb_observer)(int: max_observer){int: device_id1}(int: nb_observer)(string: observer1_idglobal)...(string: observerN_idglobal)...{int: device_idM}...   max_observer==0 si pas de limite d'observer
	    	wcmd=DataBytes.writer();
	    	wcmd.writeByte(LIST_OBSERVER);
	    	//Calcul de nombre d'observer total
	    	int nb_obs=0; for(ExposedDevice expdevice : managedIdtoD.values()) nb_obs+=expdevice.observers.size();
	    	wcmd.writeInt(nb_obs);
	    	wcmd.writeInt(0); //pas de limite sur le nombre total d'observer supporté
	    	for(ExposedDevice expdevice : managedIdtoD.values()){
	    		wcmd.writeInt(expdevice.id);
	    		wcmd.writeInt(expdevice.observers.size());
	    		for(Spot obs :  expdevice.observers) wcmd.writeString(obs.getIdGlobal());
	    	}
	    	sender.send(SKUAD.UDA, wcmd.flushReader());	
	    	break;
	    case QUERY_DEVICE: //[QUERY_DEVICE]{int: device_id}(int: response_id){...paquet uda...} : -> requête sur le device locale ciblé avec une attente de réponse via RESPONSE_DEVICE si response_id!=0
	    	device_id = message.readInt();
	    	edevice = managedIdtoD.get(device_id);
	    	response_id = message.readInt();
	    	if(edevice!=null) edevice.performToDevice(sender, message, response_id);
	    	else if(response_id>0) internalResponseDevice(sender, device_id, response_id, UDA.NOT_EXIST, null); //le device cible n'existe pas
	    	break;
	    case RESPONSE_DEVICE: //[RESPONSE_DEVICE]{int: device_id}(int: response_id)[code_perform][bool: have_paquet_response]{...paquet uda...} --> réponse à un précédent envoie QUERY_DEVICE
	    	device_id = message.readInt();
	    	response_id = message.readInt();
	    	ResponseDevice respd = reqWaiters.remove(response_id);
	    	if(respd!=null){ //--> cb_response(spot, device_id, code_perform, paquet_response) sera déclenché en réponse
	    		int code_perform = message.readByte();
	    		boolean have_paquet_response = message.readBool();
	    		if(have_paquet_response) respd.deviceResponse(sender, device_id, code_perform, message.cloneRest());
	    		else  respd.deviceResponse(sender, device_id, code_perform, null);
	    	}
	    	break;	   
	    case LINKED_LIST : 		//[LINKED_LIST] --> retourne la liste complète des devices distants observés
	    	//<-- [LIST_LINKED](int: nb_linked)(string: linked1_idglobal){int: device_id1}...(string: linked2_idglobal){int: device_id2}...
	    	wcmd=DataBytes.writer();
	    	wcmd.writeByte(LIST_LINKED);
	    	Collection<VirtualDevice> list = linkedDevice.values(); 
	    	wcmd.writeInt(list.size());
	    	for(VirtualDevice vdev : list){
    			wcmd.writeString(vdev.target.getIdGlobal());
    			wcmd.writeInt(vdev.device_id);
	    	}
	    	sender.send(SKUAD.UDA, wcmd.flushReader());		
	    	break;	    	
	    }
	    message.dispose();
	}
	
	protected void internalResponseDevice(Spot spot, int device_id, int response_id, int code_perform, DataReader paquet){
		DataWriter wc=DataBytes.writer().b(RESPONSE_DEVICE).i(device_id).i(response_id).b(code_perform);
		if(paquet!=null){
			wc.writeBool(true);
			wc.writeData(paquet);
			paquet.dispose();
		} else wc.writeBool(false);
		spot.send(SKUAD.UDA, wc.flushReader());		
	}
	//]
}

class ExposedDeviceWaiter {
	ExposedDeviceResponse cb;
	Spot spot;
	
	public ExposedDeviceWaiter(Spot h, ExposedDeviceResponse c){
		spot = h;
		cb = c;
	}
}