package skuad.uda.implementation.udaudp;

import skuad.hubudp.Spot;
import skuad.hubudp.SpotObserver;
import skuad.util.databytes.DataBytes;
import skuad.util.databytes.DataReader;
import skuad.util.databytes.DataWriter;
import skuad.SKUAD;
import skuad.uda.Device;
import skuad.uda.Monitor;
import skuad.uda.UDA;
import skuad.uda.monitor.MonitorManager;

public class VirtualDevice implements Device, SpotObserver {  //<- device lié à un device distant
	UdaUdpManager manager;
	
	Spot target;
	int device_id;
	String id;
	boolean actif;
	MonitorManager monitorManager;
	int linkCount;		//compteur de lien sur ce device, à 0 le device sera libéré
	
	public VirtualDevice(UdaUdpManager vmanager, String bid, Spot t, int d){
		manager = vmanager;
		
		target = t;
		device_id = d;
		id = bid;
		actif=true;
		monitorManager = new MonitorManager();
		linkCount=1;
		
		//on surveille le passage de l'hote distant à l'état down
		target.addSpotObserver(this);	
		
		//On envoie le message ADD_OBSERVER[device_id] à l'hote porteur
		DataWriter wcmd=DataBytes.writer();
		wcmd.writeByte(UdaUdpManager.ADD_OBSERVER);
		wcmd.writeInt(device_id);
		target.send(SKUAD.UDA, wcmd.flushReader());				
	}	
	
	public boolean unlink() {
		if(actif && decLink()){
			manager.remove(this);
			_down(true);
			return(true);
		}
		return(false);
	}
	
	protected void incLink(){ linkCount++; }
	protected boolean decLink() { if(linkCount>0) linkCount--; return(linkCount==0);} //retourne true si plus aucun lien n'existe
	
	//[interface SpotObserver
	public void up(Spot dspot){} 
	public void updateSMD(Spot dspot, boolean first_definition){}
	public void down(Spot dspot){
		_down(false);
	}
	public void idContext(Spot dspot, String ch_context, String id_context){}
	//]
	public void _down(boolean send_remove_observer){ //le device n'est plus exposé sur son site porteur ou délinké localement
		if(actif){
			actif=false; 
			target.removeSpotObserver(this);
			
			 //On envoie le message REMOVE_OBSERVER[device_id] à l'hote porteur
			if(send_remove_observer) target.send(SKUAD.UDA, DataBytes.writer().b(UdaUdpManager.REMOVE_OBSERVER).i(device_id).flushReader());
		
			//On déclenche un state(false, false) sur l'ensemble des moniteurs écouteurs
			monitorManager.state(this, false, false);
		}
	}
	
	protected void performFromDevice(Spot sender, DataReader message){ if(!actif) return; //[FROM_DEVICE] [device_id]{... paquet uda...} : ->  le paquet est un message du device distant encapsulé par ce VirtualDevice
		//--> ce peut être une réponse à une requête ([TO_DEVICE]) ou une notification d'écoute
		int cmd_uda = message.readByte();
		UDA.decodeMonitor(monitorManager, cmd_uda, message, this);
   	}
		
	//[Interface Device	
		//action
	public boolean log(String text){ if(!actif) return(false);
		DataWriter wcmd=DataBytes.writer().b(UdaUdpManager.TO_DEVICE).i(device_id);
		UDA.encodeLog(wcmd, text);
		target.send(SKUAD.UDA, wcmd.flushReader());
		return(true);
	}
	public boolean move(int cmd, int power){ if(!actif) return(false);
		DataWriter wcmd=DataBytes.writer().b(UdaUdpManager.TO_DEVICE).i(device_id);
		UDA.encodeMove(wcmd, cmd, power);
		target.send(SKUAD.UDA, wcmd.flushReader());
		return(true);
	}
	public boolean action(int actuator_id, int value){ if(!actif) return(false);
		DataWriter wcmd=DataBytes.writer().b(UdaUdpManager.TO_DEVICE).i(device_id);
		UDA.encodeAction(wcmd, actuator_id, value);
		target.send(SKUAD.UDA, wcmd.flushReader());
		return(true);		
	}
	public boolean action(int actuator_id, float value){ if(!actif) return(false);
		DataWriter wcmd=DataBytes.writer().b(UdaUdpManager.TO_DEVICE).i(device_id);
		UDA.encodeAction(wcmd, actuator_id, value);
		target.send(SKUAD.UDA, wcmd.flushReader());
		return(true);		
	}
	public boolean action(int actuator_id, byte[] value){ if(!actif) return(false);
		DataWriter wcmd=DataBytes.writer().b(UdaUdpManager.TO_DEVICE).i(device_id);
		UDA.encodeAction(wcmd, actuator_id, value);
		target.send( SKUAD.UDA, wcmd.flushReader());
		return(true);		
	}
	
		//lecture
	public boolean getState(Monitor mon){ if(!actif || mon==null) return(false);
		monitorManager.addStateWaiters(mon);
		DataWriter wcmd=DataBytes.writer().b(UdaUdpManager.TO_DEVICE).i(device_id);
		UDA.encodeGetState(wcmd);
		target.send(SKUAD.UDA, wcmd.flushReader());
		return(true);		
	}
	public boolean getParams(Monitor mon){ if(!actif || mon==null) return(false);
		monitorManager.addParamsWaiters(mon);
		DataWriter wcmd=DataBytes.writer().b(UdaUdpManager.TO_DEVICE).i(device_id);
		UDA.encodeGetParams(wcmd);
		target.send(SKUAD.UDA, wcmd.flushReader());
		return(true);				
	}
	public boolean getMoveCapacity(Monitor mon){ if(!actif || mon==null) return(false);
		monitorManager.addMoveCapacityWaiters(mon);
		DataWriter wcmd=DataBytes.writer().b(UdaUdpManager.TO_DEVICE).i(device_id);
		UDA.encodeGetMoveCapacity(wcmd);
		target.send( SKUAD.UDA, wcmd.flushReader());
		return(true);	
	}
	public boolean getSensorParam(Monitor mon, int sensor_id){ if(!actif || mon==null) return(false);
		monitorManager.addSensorParamWaiters(mon, sensor_id);
		DataWriter wcmd=DataBytes.writer().b(UdaUdpManager.TO_DEVICE).i(device_id);
		UDA.encodeGetSensorParam(wcmd, sensor_id);
		target.send( SKUAD.UDA, wcmd.flushReader());		
		return(true);	
	}
	public boolean getActuatorParam(Monitor mon, int actuator_id){ if(!actif || mon==null) return(false);
		monitorManager.addActuatorParamWaiters(mon, actuator_id);
		DataWriter wcmd=DataBytes.writer().b(UdaUdpManager.TO_DEVICE).i(device_id);
		UDA.encodeGetActuatorParam(wcmd, actuator_id);
		target.send(SKUAD.UDA, wcmd.flushReader());		
		return(true);	
	}
	public boolean readSensor(Monitor mon, int sensor_id){ if(!actif || mon==null) return(false);
		monitorManager.addReadSensorWaiters(mon, sensor_id);
		DataWriter wcmd=DataBytes.writer().b(UdaUdpManager.TO_DEVICE).i(device_id);
		UDA.encodeReadSensor(wcmd, sensor_id);
		target.send(SKUAD.UDA, wcmd.flushReader());	
		return(true);	
	}
	
		//écouteurs
	public boolean addListener(Monitor mon){  if(!actif || mon==null) return(false);
		return(monitorManager.addListener(mon));
	}
	public boolean removeListener(Monitor mon){
		return(monitorManager.removeListener(mon));
	}
	//]
}