/**
 * Implémentation d'un Monitor masqué :
 * -> toutes les méthodes de ce connecteur sont forcé sur un Device précis à la place du véritable Device remonté par les méthodes.
 * 
 * (voir par exemple le conencteur MoveAutoRepeater).
 */
package skuad.uda.monitor;

import skuad.uda.*;
import java.util.*;

public class MaskedMonitor implements Monitor {
	
	Monitor slave;
	Device force_device;
	
	public MaskedMonitor(Monitor vslave, Device vforce_device){
		slave = vslave;
		force_device = vforce_device;
	}
	
	//[Interface Monitor 
	//Réponse aux requêtes
	public void params(Device dev, String name, int type_device, int nb_sensor, int nb_actuator, boolean can_log){
		slave.params(force_device, name, type_device, nb_sensor, nb_actuator, can_log);
	}
	public void moveCapacity(Device dev, int movement_type, int power_min, int power_moy, int power_max){
		slave.moveCapacity(force_device, movement_type, power_min, power_moy, power_max);
	}
	public void sensorParam(Device dev, int sensor_id, String name, int type_sensor, int type_value, boolean observable){ //nb : type_value sur 1 octet
		slave.sensorParam(force_device, sensor_id, name, type_sensor, type_value, observable);
	}
	public void actuatorParam(Device dev, int actuator_id, String name, int type_actuator, int type_value){	//nb : type_value sur 1 octet
		slave.actuatorParam(force_device, actuator_id, name, type_actuator, type_value);
	}

	//Ecoute
	public void state(Device dev, boolean ready, boolean config_changed){ slave.state(force_device, ready, config_changed); }
	public void sensorValue(Device dev, int sensor_id, int value){ slave.sensorValue(force_device, sensor_id, value); } 
	public void sensorValue(Device dev, int sensor_id, float value){ slave.sensorValue(force_device, sensor_id, value); }
	public void sensorValue(Device dev, int sensor_id, byte[] value){ slave.sensorValue(force_device, sensor_id, value); }
	public void moved2D(Device dev, float dx, float dy, float da){ slave.moved2D(force_device, dx, dy, da); }
	public void moved3D(Device dev, float dx, float dy, float dz, float ax, float ay, float az){ slave.moved3D(force_device, dx, dy, dz, ax, ay, az); }
	public void log(Device dev, String text){ slave.log(force_device, text); }
	//]
	
	//[
	static HashMap<Device, HashMap<Monitor, MaskedMonitor>> cache=null;
	public static MaskedMonitor getMask(Device device, Monitor mon){ //retourne une instance MaskedMonitor pour le couple (device, mon) indiqué avec réutilisation des instances déjà crée pour ce couple.
		if(cache==null) cache = new HashMap<Device, HashMap<Monitor, MaskedMonitor>>();
		HashMap<Monitor, MaskedMonitor> deviceCache = cache.get(device);
		if(deviceCache==null){
			deviceCache = new HashMap<Monitor, MaskedMonitor>();
			cache.put(device, deviceCache);
		}
		MaskedMonitor mask = deviceCache.get(mon);
		if(mask==null){
			mask = new MaskedMonitor(mon, device);
			deviceCache.put(mon, mask);
		}
		return(mask);
	}
	//]
}