package skuad.uda.monitor;

import skuad.uda.*;
import java.util.*;

public class MonitorManager implements Monitor{
	
	Vector<Monitor> listeners;
	Vector<Monitor> stateWaiters;
	Vector<Monitor> paramsWaiters;
	Vector<Monitor> moveCapacityWaiters;
	Vector<WaitersSA> saParamWaiters;
	Vector<WaitersSA> sensorWaiters;
	
	public MonitorManager(){
		stateWaiters=null;
		paramsWaiters=null;
		moveCapacityWaiters=null;
		saParamWaiters=null;
		sensorWaiters=null;
		listeners=null;
	}
	
	//[Enregistrement des moniteurs
		//Les listeners
	public synchronized boolean addListener(Monitor l){
		if(listeners==null) listeners=new Vector<Monitor>();
		else if(listeners.contains(l)) return(false);
		listeners.add(l);
		return(true);		
	}
	public boolean removeListener(Monitor l){
		return(listeners!=null && listeners.remove(l));
	}	
		//Les waiters
	public void addStateWaiters(Monitor mon){
		if(stateWaiters==null) stateWaiters=new Vector<Monitor>();
		synchronized(stateWaiters) { stateWaiters.add(mon); }
	}	
	public void addParamsWaiters(Monitor mon){
		if(paramsWaiters==null) paramsWaiters=new Vector<Monitor>();
		synchronized(paramsWaiters) { paramsWaiters.add(mon); }
	}
	public void addMoveCapacityWaiters(Monitor mon){
		if(moveCapacityWaiters==null) moveCapacityWaiters=new Vector<Monitor>();
		synchronized(moveCapacityWaiters) { moveCapacityWaiters.add(mon); }
	}
	public void addSensorParamWaiters(Monitor mon, int sensor_id){
		WaitersSA desc = new WaitersSA(0, mon, sensor_id);
		if(saParamWaiters==null) saParamWaiters=new Vector<WaitersSA>();
		synchronized(saParamWaiters) { saParamWaiters.add(desc); }
	}	
	public void addActuatorParamWaiters(Monitor mon, int actuator_id){
		WaitersSA desc = new WaitersSA(1, mon, actuator_id);
		if(saParamWaiters==null) saParamWaiters=new Vector<WaitersSA>();
		synchronized(saParamWaiters) { saParamWaiters.add(desc); }
	}	
	public void addReadSensorWaiters(Monitor mon, int sensor_id){
		WaitersSA desc = new WaitersSA(0, mon, sensor_id);
		if(sensorWaiters==null) sensorWaiters=new Vector<WaitersSA>();
		synchronized(sensorWaiters) { sensorWaiters.add(desc); }
	}	
	//]
	
	//[interface Monitor
		//Réponse aux requêtes
	public void state(Device dev, boolean ready, boolean config_changed){
		if(stateWaiters!=null && !stateWaiters.isEmpty()) {
			synchronized(stateWaiters){
				for(Monitor mon : stateWaiters) mon.state(dev, ready, config_changed);
				stateWaiters.removeAllElements();
			}
		}		
		if(listeners!=null) for(Monitor mon : listeners) mon.state(dev,  ready, config_changed);
	}
	public void params(Device dev, String name, int type_device, int nb_sensor, int nb_actuator, boolean can_log){
		if(paramsWaiters!=null && !paramsWaiters.isEmpty()) {
			synchronized(paramsWaiters){
				for(Monitor mon : paramsWaiters) mon.params(dev, name, type_device, nb_sensor, nb_actuator, can_log);
				paramsWaiters.removeAllElements();
			}
		}		
	}
	public void moveCapacity(Device dev, int movement_type, int power_min, int power_moy, int power_max){
		if(moveCapacityWaiters!=null && !moveCapacityWaiters.isEmpty()) {
			synchronized(moveCapacityWaiters){
				for(Monitor mon : moveCapacityWaiters) mon.moveCapacity(dev, movement_type, power_min, power_moy, power_max);
				moveCapacityWaiters.removeAllElements();
			}
		}			
	}
	public void sensorParam(Device dev, int sensor_id, String name, int type_sensor, int type_value, boolean observable){ //nb : type_value sur 1 octet 
		if(saParamWaiters!=null && !saParamWaiters.isEmpty()) {
			synchronized(saParamWaiters){
				for(int i=0; i<saParamWaiters.size();  i++){
					WaitersSA cur = saParamWaiters.get(i);
					if(cur.type==0 && cur.saId==sensor_id){ //candidat
						cur.mon.sensorParam(dev, sensor_id, name, type_sensor, type_value, observable);
						saParamWaiters.removeElementAt(i); 
						i--;
					}
				}
			}
		}
	}
	public void actuatorParam(Device dev, int actuator_id, String name, int type_actuator, int type_value){ //nb : type_value sur 1 octet
		if(saParamWaiters!=null && !saParamWaiters.isEmpty()) {
			synchronized(saParamWaiters){
				for(int i=0; i<saParamWaiters.size();  i++){
					WaitersSA cur = saParamWaiters.get(i);
					if(cur.type==1 && cur.saId==actuator_id){ //candidat
						cur.mon.actuatorParam(dev, actuator_id, name, type_actuator, type_value);
						saParamWaiters.removeElementAt(i); 
						i--;
					}
				}
			}
		}	
	}
	public void sensorValue(Device dev, int sensor_id, int value){
		if(sensorWaiters!=null && !sensorWaiters.isEmpty()) {
			synchronized(sensorWaiters){
				for(int i=0; i<sensorWaiters.size();  i++){
					WaitersSA cur = sensorWaiters.get(i);
					if(cur.saId==sensor_id){ //candidat
						cur.mon.sensorValue(dev, sensor_id, value);
						sensorWaiters.removeElementAt(i); 
						i--;
					}
				}
			}
		}
		if(listeners!=null) for(Monitor mon : listeners) mon.sensorValue(dev, sensor_id, value);
	}
	public void sensorValue(Device dev, int sensor_id, float value){
		if(sensorWaiters!=null && !sensorWaiters.isEmpty()) {
			synchronized(sensorWaiters){
				for(int i=0; i<sensorWaiters.size();  i++){
					WaitersSA cur = sensorWaiters.get(i);
					if(cur.saId==sensor_id){ //candidat
						cur.mon.sensorValue(dev, sensor_id, value);
						sensorWaiters.removeElementAt(i); 
						i--;
					}
				}
			}
		}
		if(listeners!=null) for(Monitor mon : listeners) mon.sensorValue(dev, sensor_id, value);
	}
	public void sensorValue(Device dev, int sensor_id, byte[] value){
		if(sensorWaiters!=null && !sensorWaiters.isEmpty()) {
			synchronized(sensorWaiters){
				for(int i=0; i<sensorWaiters.size();  i++){
					WaitersSA cur = sensorWaiters.get(i);
					if(cur.saId==sensor_id){ //candidat
						cur.mon.sensorValue(dev, sensor_id, value);
						sensorWaiters.removeElementAt(i); 
						i--;
					}
				}
			}
		}		
		if(listeners!=null) for(Monitor mon : listeners) mon.sensorValue(dev, sensor_id, value);
	}
	public void moved2D(Device dev, float dx, float dy, float da){ //pas de waiter possible sur move2D
		if(listeners!=null) for(Monitor mon : listeners) mon.moved2D(dev,  dx, dy, da);
	}
	public void moved3D(Device dev, float dx, float dy, float dz, float ax, float ay, float az){ //pas de waiter possible sur moved3D
		if(listeners!=null) for(Monitor mon : listeners) mon.moved3D(dev,  dx, dy, dz, ax, ay, az);
	}
	public void log(Device dev, String text){ //pas de waiter possible sur log
		if(listeners!=null) for(Monitor mon : listeners) mon.log(dev,  text);
	}
	//]
}

class WaitersSA {
	public WaitersSA(int t,Monitor m, int id){
		mon = m;
		type = t;
		saId = id;
	}
	public Monitor mon;
	int type; //0 pour Sensor, 1 pour Actuator
	int saId;
}