package skuad.uda.net;

import skuad.uda.*;
import skuad.uda.device.*;
import skuad.uda.monitor.*;
import skuad.util.databytes.*;
import skuad.util.thread.*;
import java.io.*;
import java.net.*;
import java.util.*;

public class TcpClientDevice extends MinimalDevice implements Runnable, Connectable {
	
	Socket socket_to_server;
	InputStream from_server;
	OutputStream to_server;
	Thread conectProcess; String server_ip; int server_port; //processus de connexion, ==null si non connecté
	boolean in_run;
	Vector<ConnectionListener> connection_listeners;
	MonitorManager waiters;
	InternalMoveAutoRepeater imaRepeater; // !=null si le mode de répétition automatique des ordres de mouvement est activé
	StreamDecoder stream_decoder;
	StreamWriter writer;
	
	public TcpClientDevice(){
		in_run = false;
		socket_to_server = null;
		conectProcess=null;
		connection_listeners=new Vector<ConnectionListener>();
		waiters = new MonitorManager();
		imaRepeater = null;
		stream_decoder = new StreamDecoder();
		writer=null;
	}	
	
	public void enableAutoRepeatMove(long freq){ //active le mode de répétition automatique des ordres de mouvement avec une fréquence freq
		if(imaRepeater==null) imaRepeater = new InternalMoveAutoRepeater(this, freq);
		else imaRepeater.setFrequency(freq);	
	}
	public void disableAutoRepeatMove(){ //désactive le mode de répétition automatique des ordres de mouvement
		if(imaRepeater!=null){
			imaRepeater.stop();
			imaRepeater=null;
		}
	}
	public long getAutoRepeatMoveFrequency(){ // -1 si le mode de répétition automatique des ordres de mouvement n'est pas activé
		if(imaRepeater!=null) return(imaRepeater.getFrequency());
		else return(-1);
	}

	//[Gestion des écouteurs de connection
	public boolean addConnectionListener(ConnectionListener cl){  synchronized(connection_listeners) { return(connection_listeners.add(cl)); }}
	public boolean removeConnectionListener(ConnectionListener cl){  synchronized(connection_listeners) { return(connection_listeners.remove(cl)); }}
	protected void fireNewConnection(String label, String ip, int port){
		synchronized(connection_listeners) { for(ConnectionListener cl : connection_listeners) ThreadStack.perform( () -> cl.newConnection(label, ip, port)); }
	}
	protected void fireLostConnection(String label, String ip, int port){
		synchronized(connection_listeners) { for(ConnectionListener cl : connection_listeners) ThreadStack.perform( () -> cl.lostConnection(label, ip, port)); }
	}
	protected void fireErrorConnection(String mess){
		synchronized(connection_listeners) { for(ConnectionListener cl : connection_listeners) ThreadStack.perform( () -> cl.errorConnection(mess)); }
	}
	protected void fireWarningConnection(String mess){
		synchronized(connection_listeners) { for(ConnectionListener cl : connection_listeners) ThreadStack.perform( () -> cl.warningConnection(mess)); }
	}	
	//]
	
	//[Interface Connectable
	public boolean isConnected() { return(conectProcess!=null); }
	public synchronized boolean disconnet(){ //coupe la connection avec le serveur
//System.out.println("Disconnect demandé !");		
		if(conectProcess!=null){ //on est bien connecté	
			in_run = false; //arret du processus
			closeServer();
			return(true);
		} 
		return(false); //on n'est pas connecté		
	}
	public boolean connect(String vip, int vport){ //tentative de connexion avec le serveur vip:vport
		if(conectProcess==null){ //on n'est pas encore connecté	
			server_ip = vip; server_port = vport;
			//on lance le processus de connexion
			conectProcess = new Thread(this); conectProcess.start();
			return(true);
		} 
		return(false); //on est déjà connecté
	}
	//]
	
	public void run(){ //processus de connexion
		in_run = true;
		
		//on va essayer de se connecter
		try {         
			socket_to_server=new Socket(server_ip,server_port);   
			from_server = socket_to_server.getInputStream();
			to_server = socket_to_server.getOutputStream();
		} catch(Exception e) { 
			String label_server = server_ip+":"+server_port;
			server_ip = null;
			closeServer(); 
			conectProcess=null;
			fireErrorConnection("Impossible de se connecter au serveur indiqué ("+label_server+")");			
			return;
		} 
		
		StreamReader reader = DataBytes.reader(from_server);
		writer = DataBytes.writer(to_server);
		synchronized(to_server) { sendVERSION(to_server); } //on commence par envoyer le numéro de version au serveur			
		fireNewConnection("to server", server_ip, server_port);

		stream_decoder.force_config_change=true;
		getState(null); //on impulse une lecture de l'état du device distant
		
		//Boucle de lecture des commandes du serveur
		int cmd, size;
listen_server :	while(in_run){
			//Extraction de la commande et de sa taille d'encodage
			cmd = reader.readByte();
//System.out.println("cmd = "+cmd+" inrun = "+in_run);
			if(cmd<=0) {
//				continue; //test
//System.out.println("BOUM 0!!! "+cmd);	
				break; //rupture du flux de lecture :(
			}
			size = reader.readShort();			
			switch(cmd){
            case UDA.CLOSE: break listen_server; //fin de traitement de ce client
            case UDA.VERSION: 	//[VERSION]{2:2}[H_VERSION][L_VERSION]
            	int byte_hight = reader.readByte();
            	int byte_low = reader.readByte();
//System.out.println("Version recu : "+byte_hight+"."+byte_low);            		
            	if(!UDA.isCompatibleVersion(byte_hight, byte_low)){
//System.out.println("BOUM 1!!!");	
            		
            		log("not compatible protocole version");
            		break listen_server;
            	}
            	break;			
			default:
				//decode(Monitor monitor, int cmd, DataReader reader, Device device_source)
				if(UDA.decodeMonitor(stream_decoder, cmd, reader, this)==UDA.CMD_UNKNOW){
//System.out.println("BOUM 2!!!");					
					
					fireWarningConnection("Réception d'une commande serveur inconnue ("+cmd+" : "+size+"), la commande est ignorée !");
					try { from_server.skip(size); } catch(Exception e){ break listen_server; } //fin de traitement de l'écoute du serveur en cas d'erreur
				}
			}
		}
//System.out.println("-fermeture- inrun = "+in_run);		
		closeServer();
		conectProcess = null;
	}
	
	//[Interface Device
		//action
	public boolean log(String text){
		if(to_server!=null) synchronized(to_server){
			int size = UDA.sizeToEncodeLog(text)-1; //calcul de la taille du segment
			if(!writer.writeByte(UDA.LOG)) return(false);	
			writer.writeShort(size);
			UDA.encodeLog(writer, text, false);
			return(true);			
		}
		return(false);
	}
	public boolean move(int cmd, int power){ //[MOV]{2:x+1}[commande_move]{x-int: value}
		if(imaRepeater!=null) return(imaRepeater.move(cmd, power));
		else return(internalMove(cmd, power));
	}
	protected boolean internalMove(int cmd, int power){ //[MOV]{2:x+1}[commande_move]{x-int: value}
		if(to_server!=null) synchronized(to_server){
			int size = UDA.sizeToEncodeMove(cmd,power)-1; //calcul de la taille du segment
			if(!writer.writeByte(UDA.MOVE)) return(false);	
			writer.writeShort(size);
			UDA.encodeMove(writer, cmd, power, false);
			return(true);
		}
		return(false);
	}
	public boolean action(int actuator_id, int value){ //[ACTION]{2:x+y}{x-int: actuator_id}{y-val: value}
		if(to_server!=null) synchronized(to_server){
			int size = UDA.sizeToEncodeAction(actuator_id, value)-1; //calcul de la taille du segment
			if(!writer.writeByte(UDA.ACTION)) return(false);
			writer.writeShort(size);
//System.out.println("Action --> envoyé size = "+size+" value = "+value);			
			UDA.encodeAction(writer, actuator_id, value, false);				
			return(true);
		}
		return(false);		
	}
	public boolean action(int actuator_id, float value){ //[ACTION]{2:x+y}{x-int: actuator_id}{y-val: value}
		if(to_server!=null) synchronized(to_server){
			int size = UDA.sizeToEncodeAction(actuator_id, value)-1; //calcul de la taille du segment
			if(!writer.writeByte(UDA.ACTION)) return(false);
			writer.writeShort(size);
			UDA.encodeAction(writer, actuator_id, value, false);				
			return(true);
		}
		return(false);			
	}
	public boolean action(int actuator_id, byte[] value){//[ACTION]{2:x+y}{x-int: actuator_id}{y-val: value}
		if(to_server!=null) synchronized(to_server){
			int size = UDA.sizeToEncodeAction(actuator_id, value)-1; //calcul de la taille du segment
			if(!writer.writeByte(UDA.ACTION)) return(false);
			writer.writeShort(size);
			UDA.encodeAction(writer, actuator_id, value, false);				
			return(true);
		}
		return(false);		
	}
	
		//lecture
	public boolean getState(Monitor mon){ 
		if(to_server!=null) synchronized(to_server){ //[GET_STATE]{2:0} 
			if(mon!=null) waiters.addStateWaiters(mon);
			if(to_server!=null) synchronized(to_server){
				int size = UDA.sizeToEncodeGetState()-1; //calcul de la taille du segment
				if(!writer.writeByte(UDA.GET_STATE)) return(false);
				writer.writeShort(size);
				UDA.encodeGetState(writer, false);				
				return(true);
			}
			return(false);	
		} 
		//non pret
		if(mon!=null) mon.state(this, false, false); 
		return(true);		
	}
	public boolean getParams(Monitor mon){
		if(to_server!=null) synchronized(to_server){ //[GET_PARAMS]{2:0} 
			waiters.addParamsWaiters(mon);
			if(to_server!=null) synchronized(to_server){
				int size = UDA.sizeToEncodeGetParams()-1; //calcul de la taille du segment
				if(!writer.writeByte(UDA.GET_PARAMS)) return(false);
				writer.writeShort(size);
				UDA.encodeGetParams(writer, false);				
				return(true);
			}
		}
		return(false);
	}
	public boolean getMoveCapacity(Monitor mon){ 
		if(to_server!=null) synchronized(to_server){ //[GET_MOVE_CAPACITY]{2:0} 
			waiters.addMoveCapacityWaiters(mon);
			if(to_server!=null) synchronized(to_server){
				int size = UDA.sizeToEncodeGetMoveCapacity()-1; //calcul de la taille du segment
				if(!writer.writeByte(UDA.GET_MOVE_CAPACITY)) return(false);
				writer.writeShort(size);
				UDA.encodeGetMoveCapacity(writer, false);				
				return(true);
			}			
		}
		return(false);
	}
	public boolean getSensorParam(Monitor mon, int sensor_id){
		if(to_server!=null) synchronized(to_server){ //[GET_SENSOR_PARAMS]{2:x}{x-int: sensor_id}
			waiters.addSensorParamWaiters(mon, sensor_id);
			if(to_server!=null) synchronized(to_server){
				int size = UDA.sizeToEncodeGetSensorParam(sensor_id)-1; //calcul de la taille du segment
				if(!writer.writeByte(UDA.GET_SENSOR_PARAMS)) return(false);
				writer.writeShort(size);
				UDA.encodeGetSensorParam(writer, sensor_id, false);				
				return(true);
			}				
		}
		return(false);
	}
	public boolean getActuatorParam(Monitor mon, int actuator_id){
		if(to_server!=null) synchronized(to_server){ //[GET_ACTUATOR_PARAMS]{2:x}{x-int: actuator_id}	
			waiters.addActuatorParamWaiters(mon, actuator_id);
			if(to_server!=null) synchronized(to_server){
				int size = UDA.sizeToEncodeGetActuatorParam(actuator_id)-1; //calcul de la taille du segment
				if(!writer.writeByte(UDA.GET_ACTUATOR_PARAMS)) return(false);
				writer.writeShort(size);
				UDA.encodeGetActuatorParam(writer, actuator_id, false);				
				return(true);
			}
		}
		return(false);
	}
	public boolean readSensor(Monitor mon, int sensor_id){ 
		if(to_server!=null) synchronized(to_server){ //[READ_SENSOR]{2:x}{x-int: sensor_id}
			waiters.addReadSensorWaiters(mon, sensor_id);
			if(to_server!=null) synchronized(to_server){
				int size = UDA.sizeToEncodeReadSensor(sensor_id)-1; //calcul de la taille du segment
				if(!writer.writeByte(UDA.READ_SENSOR)) return(false);
				writer.writeShort(size);
				UDA.encodeReadSensor(writer, sensor_id, false);				
				return(true);
			}
		}
		return(false);
	}
		//écoute
	//public boolean addListener(Monitor mon){ return(monitors.add(mon)) ;}
	//public boolean removeListener(Monitor mon){ return(monitors.remove(mon)) ;}
	//]
	
	//[Méthodes internes
	protected synchronized void closeServer(){
		//fermeture des flux et de la socket
		if(to_server!=null) try{ synchronized(to_server) { sendClose(to_server); to_server.close(); to_server=null; }}catch(Exception e){}
		if(from_server!=null) try{ from_server.close(); from_server=null;}catch(Exception e){}
		if(socket_to_server!=null) try{ socket_to_server.close(); socket_to_server=null;}catch(Exception e){}
		writer=null;
		
		if(server_ip != null){ 
			fireLostConnection("to server", server_ip, server_port);
			server_ip = null;
			fireState(false, false);
		}
	}
	public static boolean writeShort(OutputStream out, int data){try{    //entier sur 2 octets
		out.write((data >> 8) & 0xFF);
		out.write((data & 0xFF));		
	} catch(Exception e){ return(false); } return(true); }
	public static boolean sendClose(OutputStream target){ //[CLOSE]{2:0}
		 try {   
			 target.write(UDA.CLOSE);
			 writeShort(target, 0);
		 } catch(Exception e) { return(false); } //erreur de transmission  		
		 return(true);
	}	
	public static boolean sendVERSION(OutputStream target){ //[VERSION]{2:2}[ VERSION_H , VERSION_L ]
		 try {   
			 target.write(UDA.VERSION);
			 writeShort(target, 2);
			 target.write(UDA.H_VERSION);
			 target.write(UDA.L_VERSION);
			 return(true);
		 } catch(Exception e) { return(false); } //erreur de transmission  		
	}	
	//]
	
	class StreamDecoder implements Monitor { //classe interne de décodage des commandes reçu par le flux stream	
		boolean force_config_change=true;
		
		//Réponse aux requêtes
		public void state(Device dev, boolean ready, boolean config_changed){
        	waiters.state(TcpClientDevice.this, ready, force_config_change || config_changed);
        	fireState(ready, force_config_change || config_changed);
        	force_config_change=false;
		}		
		public void params(Device dev, String name, int type_device, int nb_sensor, int nb_actuator, boolean can_log){
			waiters.params(TcpClientDevice.this, name, type_device, nb_sensor, nb_actuator, can_log);
		}
		public void moveCapacity(Device dev, int movement_type, int power_min, int power_moy, int power_max){
			waiters.moveCapacity(TcpClientDevice.this, movement_type, power_min, power_moy, power_max);
		}
		public void sensorParam(Device dev, int sensor_id, String name, int type_sensor, int type_value, boolean observable){ //nb : type_value sur 1 octet
			waiters.sensorParam(TcpClientDevice.this, sensor_id, name, type_sensor, type_value, observable);
		}
		public void actuatorParam(Device dev, int actuator_id, String name, int type_actuator, int type_value){	//nb : type_value sur 1 octet
			waiters.actuatorParam(TcpClientDevice.this, actuator_id, name, type_actuator, type_value);			
		}
			//Ecoute
		public void sensorValue(Device dev, int sensor_id, int value){
			waiters.sensorValue(TcpClientDevice.this, sensor_id, value);
			fireSensorValue(sensor_id, value); 
		}
		public void sensorValue(Device dev, int sensor_id, float value){
			waiters.sensorValue(TcpClientDevice.this, sensor_id, value);
			fireSensorValue(sensor_id, value); 
		}
		public void sensorValue(Device dev, int sensor_id, byte[] value){
			waiters.sensorValue(TcpClientDevice.this, sensor_id, value);
			fireSensorValue(sensor_id, value); 
		}
		public void moved2D(Device dev, float dx, float dy, float da){ fireMoved2D(dx, dy, da); }
		public void moved3D(Device dev, float dx, float dy, float dz, float ax, float ay, float az){ fireMoved3D(dx, dy, dz, ax, ay, az); }
		public void log(Device dev, String text){ fireLog(text); }
	}
}
class InternalMoveAutoRepeater implements Runnable {
	
	boolean in_run;
	int cur_move; int cur_force;
	long freq;	//fréquence de répétition des commandes
	TcpClientDevice slave;
	
	public InternalMoveAutoRepeater(TcpClientDevice vslave, long vfreq){
		slave = vslave;
		cur_move = UDA.STOP;
		freq = vfreq;
		in_run = false;
	}
	
	public void setFrequency(long vfreq){ freq = vfreq; }
	public long getFrequency(){ return(freq); }
	
	public void run(){
		in_run=true; 
		while(in_run){
			if(cur_move!=UDA.STOP) slave.internalMove(cur_move, cur_force);
			else break; //fin d'exécution du processus de répétition cyclique
			
			try{ Thread.sleep(freq); }catch(Exception e){} //on fait un pause correspondant à la fréquence de répétition
		}
		in_run = false;
	}
	
	public void stop(){ //arrête le processus de répétition cyclique s'il est actif (NB: cette méthode n'applique pas d'action sur le slave)
		cur_move = UDA.STOP;
	}

	public boolean move(int cmd, int power){
		cur_move = cmd; cur_force = power;
		if(cur_move!=UDA.STOP && !in_run) ThreadStack.perform(this); //on lance le processus de répétition cyclique
		else return(slave.internalMove(cmd, power));
		return(true);
	}
}