package skuad.uda.net;

import skuad.uda.*;
import java.util.*;

import skuad.util.databytes.*;
import skuad.util.thread.ThreadStack;

import java.io.*;
import java.net.*;

public class TcpServerDevice implements Runnable, Server {
	
	Device device;
	
	Thread serverProcess; String local_ip; int port; //processus d'écoute serveur, ==null si non connecté
	String client_ip; int client_port; //info sur le client actuellement connecté
	ServerSocket socket_server;
	
	boolean in_run;
	Socket socket_client;
	InputStream from_client;
	OutputStream to_client;
	Vector<ConnectionListener> connection_listeners;
	StreamEncode stream_encoder;
	StreamWriter writer;
	
	public TcpServerDevice(Device vdevice){	
		device = vdevice;
		serverProcess=null;
		connection_listeners=new Vector<ConnectionListener>();
		stream_encoder = new StreamEncode();
		device.addListener(stream_encoder);
	}	

	//[Gestion des écouteurs de connection
	public boolean addConnectionListener(ConnectionListener cl){  synchronized(connection_listeners) { return(connection_listeners.add(cl)); }}
	public boolean removeConnectionListener(ConnectionListener cl){  synchronized(connection_listeners) { return(connection_listeners.remove(cl)); }}
	protected void fireNewConnection(String label, String ip, int port){
		synchronized(connection_listeners) { for(ConnectionListener cl : connection_listeners) ThreadStack.perform( () -> cl.newConnection(label, ip, port)); }
	}
	protected void fireLostConnection(String label, String ip, int port){
		synchronized(connection_listeners) { for(ConnectionListener cl : connection_listeners) ThreadStack.perform( () -> cl.lostConnection(label, ip, port)); }
	}
	protected void fireErrorConnection(String mess){
		synchronized(connection_listeners) { for(ConnectionListener cl : connection_listeners) ThreadStack.perform( () -> cl.errorConnection(mess)); }
	}
	protected void fireWarningConnection(String mess){
		synchronized(connection_listeners) { for(ConnectionListener cl : connection_listeners) ThreadStack.perform( () -> cl.warningConnection(mess)); }
	}	
	//]
	
	//[Intreface Server
	public synchronized boolean start(int vport){ //démarre le serveur sur l'écoute du port indiqué
		if(serverProcess==null){ //on n'est pas encore connecté	
			port = vport;
			//on lance le processus d'écoute
			serverProcess = new Thread(this); serverProcess.start();
			return(true);
		} 
		return(false); //on est déjà connecté
	}
	
	public synchronized boolean stop(){ //arrête l'écoute du serveur
		if(serverProcess!=null){ //on est bien connecté	
			in_run = false; //arret du processus
			closeClient(); //on ferme le dialogue avec l'éventuel client actuel
			if(socket_server!=null) try{ socket_server.close(); socket_server=null; } catch(Exception e){} //on ferme la socket serveur
			return(true);
		} 
		return(false); //on n'est pas connecté		
	}
	
	public boolean isStarted(){ return(serverProcess!=null); }
	public int nbConnectedClient(){
		if(client_ip!=null) return(1);
		else return(0);
	}
	public int getServerPort(){ return(port); }
	//]
	
	public void run(){
		in_run = true;

		try { 	//Ouverture du port d'écoute
			socket_server=new ServerSocket(port);
		} catch(Exception e) {
			fireErrorConnection("Impossible d'ouvrir le port d’écoute indiqué !");
			serverProcess=null;
			return; //fin d'exécution du serveur
		}
		port = socket_server.getLocalPort(); local_ip = socket_server.getInetAddress().getHostAddress();
		client_ip = null; client_port=0;
		fireNewConnection("server", local_ip, port);

		while(in_run){
			closeClient(); //on purge les ressources du précédent client éventuel

			//Attente de la réception d'un client			
			try { 
				socket_client=socket_server.accept();
				from_client = socket_client.getInputStream();
				to_client = socket_client.getOutputStream();
			} catch(Exception e) {
				if(in_run) fireWarningConnection("Erreur lors de l'attente de réception des clients !");
				continue; //on attend le prochain client
			}

			client_port = socket_client.getLocalPort();
			client_ip = socket_client.getInetAddress().getHostAddress();

			StreamReader reader = DataBytes.reader(from_client);
			writer = DataBytes.writer(to_client);
			
			synchronized(to_client) { sendVERSION(to_client); } //on commence par envoyer le numéro de version au client
			fireNewConnection("client", client_ip, client_port);
			
			//Boucle de lecture des commandes du clients
			int cmd, size;
listen_client : while(in_run){
				//Extraction de la commande et de sa taille d'encodage
				cmd = reader.readByte();
//System.out.println("cmd = "+cmd);				
				if(cmd<=0) break; //rupture du flux de lecture :(
				size = reader.readShort();			
				switch(cmd){
				case UDA.CLOSE: break listen_client; //fin de traitement de ce client
				case UDA.VERSION: 	//[VERSION]{2:2}[H_VERSION][L_VERSION]
	            	int byte_hight = reader.readByte();
	            	int byte_low = reader.readByte();
	//System.out.println("Version recu : "+byte_hight+"."+byte_low);            		
	            	if(!UDA.isCompatibleVersion(byte_hight, byte_low)){
	            		System.out.println("MINCE 1");  
	            		stream_encoder.log(device, "not compatible protocole version");
	            		break listen_client;
	            	}
	            	break;	            	
				default :
					if(UDA.decodeDevice(device, cmd, reader, stream_encoder)==UDA.CMD_UNKNOW){
						System.out.println("MINCE 2"); 
						fireWarningConnection("Réception d'une commande cliente inconnu ("+cmd+"), la commande est ignorée !");
						try { from_client.skip(size); } catch(Exception e){ break listen_client; } //fin de traitement de ce client  en cas d'erreur
					};
				}
			}			
		}
//System.out.println("=== fermeture ==="); 
		closeClient(); //on purge les ressources du précédent client éventuel
		if(socket_server!=null) try{ socket_server.close(); socket_server=null; } catch(Exception e){}
		fireLostConnection("server", local_ip, port);
		serverProcess = null;
	}
	
	class StreamEncode implements Monitor { //classe interne de décodage des commandes reçu par le flux stream	
		//Réponse aux requêtes
		public void state(Device dev, boolean ready, boolean config_changed){
			if(to_client!=null) synchronized(to_client){				
				int size = UDA.sizeToEncodeState(true, true)-1;  //calcul de la taille du segment
				writer.writeByte(UDA.STATE);
				writer.writeShort(size);
				UDA.encodeState(writer, ready, config_changed, false);
			}	
		}		
		public void params(Device dev, String name, int type_device, int nb_sensor, int nb_actuator, boolean can_log){
			if(to_client!=null) synchronized(to_client){
				int size = UDA.sizeToEncodeParams(name, type_device, nb_sensor, nb_actuator, can_log)-1;  //calcul de la taille du segment
				writer.writeByte(UDA.PARAMS);
				writer.writeShort(size);
				UDA.encodeParams(writer, name, type_device, nb_sensor, nb_actuator, can_log, false);
			}	
		}
		public void moveCapacity(Device dev, int movement_type, int power_min, int power_moy, int power_max){
			if(to_client!=null) synchronized(to_client){
				int size = UDA.sizeToEncodeMoveCapacity(movement_type, power_min, power_moy, power_max)-1;  //calcul de la taille du segment
				writer.writeByte(UDA.MOVE_CAPACITY);
				writer.writeShort(size);
				UDA.encodeMoveCapacity(writer, movement_type, power_min, power_moy, power_max, false);
			}	
		}
		public void sensorParam(Device dev, int sensor_id, String name, int type_sensor, int type_value, boolean observable){ //nb : type_value sur 1 octet
			if(to_client!=null) synchronized(to_client){
				int size = UDA.sizeToEncodeSensorParam(sensor_id, name, type_sensor, type_value, observable)-1;  //calcul de la taille du segment
				writer.writeByte(UDA.SENSOR_PARAMS);
				writer.writeShort(size);
				UDA.encodeSensorParam(writer, sensor_id, name, type_sensor, type_value, observable, false);
			}
		}
		public void actuatorParam(Device dev, int actuator_id, String name, int type_actuator, int type_value){	//nb : type_value sur 1 octet
			if(to_client!=null) synchronized(to_client){
				int size = UDA.sizeToEncodeActuatorParam(actuator_id, name, type_actuator, type_value)-1;  //calcul de la taille du segment
				writer.writeByte(UDA.ACTUATOR_PARAMS);
				writer.writeShort(size);
				UDA.encodeActuatorParam(writer, actuator_id, name, type_actuator, type_value, false);
			}			
		}
			//Ecoute
		public void sensorValue(Device dev, int sensor_id, int value){
			if(to_client!=null) synchronized(to_client){
				int size = UDA.sizeToEncodeSensorValue(sensor_id, value)-1;  //calcul de la taille du segment
				writer.writeByte(UDA.SENSOR);
				writer.writeShort(size);
				UDA.encodeSensorValue(writer, sensor_id,value, false);
			}
		}
		public void sensorValue(Device dev, int sensor_id, float value){
			if(to_client!=null) synchronized(to_client){
				int size = UDA.sizeToEncodeSensorValue(sensor_id, value)-1;  //calcul de la taille du segment
				writer.writeByte(UDA.SENSOR);
				writer.writeShort(size);
				UDA.encodeSensorValue(writer, sensor_id,value, false);
			}
		}
		public void sensorValue(Device dev, int sensor_id, byte[] value){
			if(to_client!=null) synchronized(to_client){
				int size = UDA.sizeToEncodeSensorValue(sensor_id, value)-1;  //calcul de la taille du segment
				writer.writeByte(UDA.SENSOR);
				writer.writeShort(size);
				UDA.encodeSensorValue(writer, sensor_id,value, false);
			}
		}
		public void moved2D(Device dev, float dx, float dy, float da){ 	
			if(to_client!=null) synchronized(to_client){
				int size = UDA.sizeToEncodeMoved2D(dx, dy, da)-1;  //calcul de la taille du segment
				writer.writeByte(UDA.MOVED2D);
				writer.writeShort(size);
				UDA.encodeMoved2D(writer, dx, dy, da, false);
			}
		}
		public void moved3D(Device dev, float dx, float dy, float dz, float ax, float ay, float az){
			if(to_client!=null) synchronized(to_client){
				int size = UDA.sizeToEncodeMoved3D(dx, dy, dy, ax, ay, az)-1;  //calcul de la taille du segment
				writer.writeByte(UDA.MOVED3D);
				writer.writeShort(size);
				UDA.encodeMoved3D(writer, dx, dy, dy, ax, ay, az, false);
			}
		}
		public void log(Device dev, String text){ 
			if(to_client!=null) synchronized(to_client){
				int size = UDA.sizeToEncodeLog(text)-1;  //calcul de la taille du segment
				writer.writeByte(UDA.LOG);
				writer.writeShort(size);
				UDA.encodeLog(writer, text, false);
			}
		}
	}
	//]
	
	//Méthodes internes
	protected void closeClient(){
		//fermeture des flux et de la socket cliente
		if(to_client!=null) try{ synchronized(to_client) { sendClose(to_client); to_client.close(); to_client=null; }} catch(Exception e){}
		if(from_client!=null) try{ from_client.close(); from_client=null; } catch(Exception e){}
		if(socket_client!=null) try{ socket_client.close(); socket_client=null;}catch(Exception e){}
		if(client_ip != null){ 
			fireLostConnection("client", client_ip, client_port);
			client_ip = null;
		}
		writer=null;
	}
	public static boolean writeShort(OutputStream out, int data){try{    //entier sur 2 octets
		out.write((data >> 8) & 0xFF);
		out.write((data & 0xFF));		
	} catch(Exception e){ return(false); } return(true); }
	public static boolean sendClose(OutputStream target){ //[CLOSE]{2:0}
		 try {   
			 target.write(UDA.CLOSE);
			 writeShort(target, 0);
		 } catch(Exception e) { return(false); } //erreur de transmission  		
		 return(true);
	}
	public static boolean sendVERSION(OutputStream target){ //[VERSION]{2:2}[ VERSION_H , VERSION_L ]
		 try {   
			 target.write(UDA.VERSION);
			 writeShort(target, 2);
			 target.write(UDA.H_VERSION);
			 target.write(UDA.L_VERSION);
			 return(true);
		 } catch(Exception e) { return(false); } //erreur de transmission  		
	}	
	//]
	
}