package skuad.util.base;

import java.net.Inet4Address;
import java.net.InetAddress;
import java.net.InterfaceAddress;
import java.net.NetworkInterface;
import java.util.*;

/**
 * Classe statique utilitaire pour les informations réseaux.<br/>
 * <br/>
 * Vous n'aurez probablement jamais besoin d'utiliser directement cette classe.
 */
public class Network {
	
	public final static String LOCALHOST="127.0.0.1";
	public final static InetAddress localAddress=ipToIAddress("127.0.0.1");

	private static String last_external_ip = null;
	
	public static String getExternalIp(){
		if(last_external_ip==null) checkInterface(false);		
		return(last_external_ip!=null ? last_external_ip : LOCALHOST);
	}
	
	public static boolean isLocalAddress(String ip){
		try { return(LOCALHOST.equals(ip) || isLocalAddress(InetAddress.getByName(ip))); }
		catch (Exception e) { return(false); }
	}
	public static boolean isLocalAddress(InetAddress addr) {
		// Check if the address is a valid special local or loop back
		if (addr.isAnyLocalAddress() || addr.isLoopbackAddress()) return true;

		try { // Check if the address is defined on any interface
			return(NetworkInterface.getByInetAddress(addr) != null);
		} catch (Exception e) { return(false); }
	}
	
	public static InetAddress ipToIAddress(String st_ip){
		InetAddress res=null;
		try{ 
			res=InetAddress.getByName(st_ip);
		} catch(Exception e){ res=null; }
		return(res);
	} 
	
	public static String addressToIp(InetAddress addr){ //127.0.0.1 si c'est une adresse locale
		String res=null;
		if(isLocalAddress(addr)) res=LOCALHOST;
		else res=addr.getHostAddress();
		return(res);
	} 
	
	public static boolean isSameIp(String ip1, String ip2){
		if(ip1==null) return(ip2==null);
		if(ip2!=null) {
			if(Network.isLocalAddress(ip2)) return(Network.isLocalAddress(ip1));
			else return(ip2.equals(ip1));
		}
		return(false);
	}
	
	/**
	 * Fonction d'analyse des interfaces réseaux disponible.<br/>
	 * <br/>
	 * Cette fonction analyse l'ensemble des interfaces réseaux ipv4 disponible et retourne celle qui semble être l'interface principale.
	 * @param display_list flag indiquant si la liste des interfaces détectées doit être affichées ou non dans la console.
	 * @return l'interface réseau qui semble être l'interface principale, <b>null</b> si une erreur s'est produite ou si aucune interface n'a été détectée.
	 */
	public static InterfaceAddress checkInterface(boolean display_list){
		InterfaceAddress candidat=null;
		try { for(Enumeration< NetworkInterface > interfaces =  NetworkInterface.getNetworkInterfaces( );  interfaces.hasMoreElements( ); ){
			NetworkInterface cur = interfaces.nextElement();
			if(cur.isLoopback()) continue; //c'est l'adresse loopback

			for(InterfaceAddress addr : cur.getInterfaceAddresses()) {
				InetAddress inet_addr = addr.getAddress();
				if(inet_addr instanceof Inet4Address) { //on ne traite que les adresses ipv4    	
					String interface_name=cur.getName();
					InetAddress broadcast = addr.getBroadcast();
					if (broadcast == null) continue; //cette adresse ne dispose pas de broadcast, on évite donc de l'utiliser
					if(candidat==null || interface_name.equals("eth0")) candidat=addr; //on va privilégier l'interface de nom eth0 (pour les environnement linux)
					if(display_list) System.out.println("[Network.checkInterface]  interface: "+cur.getName()+" ip: "+ inet_addr.getHostAddress()+" broadcast: "+broadcast.getHostAddress());
				}
			}
		}}catch(Exception e){ e.printStackTrace(); } 
		last_external_ip = candidat.getAddress().getHostAddress();
		return(candidat);
	}	
	
	private static long last_vbroadcast_update=0;
	private static Vector<InetAddress> broadcasts=new Vector<InetAddress>();
	public static Vector<InetAddress> allBroadcastAddress(){ //retourne la liste de toutes les adresses broadcast (actualisé au moins 1 fois toutes les 3 secondes)
		long cur_time = System.currentTimeMillis();
		if(cur_time-last_vbroadcast_update>3000) synchronized(broadcasts) { //il faut actualiser la liste des interfaces broadcasts
			last_vbroadcast_update = cur_time;
			broadcasts.removeAllElements();
			try { for(Enumeration< NetworkInterface > interfaces =  NetworkInterface.getNetworkInterfaces( );  interfaces.hasMoreElements( ); ){
				NetworkInterface cur = interfaces.nextElement();
				if(cur.isLoopback()) continue; //c'est l'adresse loopback
//String interface_name=cur.getName();
//System.out.println("Examen de ["+interface_name+"]");
				for(InterfaceAddress addr : cur.getInterfaceAddresses()) {
					InetAddress inet_addr = addr.getAddress();
					if(inet_addr instanceof Inet4Address) { //on ne traite que les adresses ipv4    	
						InetAddress broadcast = addr.getBroadcast();
//System.out.println("Candidate ["+interface_name+"] "+inet_addr.isSiteLocalAddress()+" broadcast = "+broadcast);							
						if (broadcast == null) continue; //cette adresse ne dispose pas de broadcast, on évite donc de l'utiliser					
						broadcasts.add(broadcast);
					}
				}
			}}catch(Exception e){ e.printStackTrace(); }		
//			InterfaceAddress candidat = checkInterface(true);
//			if(candidat!=null) broadcasts.add(candidat.getBroadcast());
			if(broadcasts.isEmpty()) { //aucune adresse broadcast détectée, on va ajouter au moins l'addresse local
					broadcasts.add(localAddress);
			}
		}
		return(broadcasts);
	}
}