package skuad.util.base;

import java.io.*;
import java.math.BigInteger;
import java.security.MessageDigest;
import java.util.*;

/**
 * Classe utilitaire.<br/> 
 * <br/>
 * Cette classe utilitaire est une classe purement statique (elle ne s'instancie pas),
 * elle n'expose que des méthodes statiques utiles.<br/>
 */
public class Util {
	
	//[Constante de niveau de qualification (d=0)
	/**
	 * Constante pour qualifier un niveau informatif.<br/>
	 * <br/>
	 * Cette valeur exprime généralement qu'un message est de niveau informatif.
	 */
	public final static byte LOGGER_INFO=0;
	
	/**
	 * Constante pour qualifier une alerte.<br/>
	 * <br/>
	 * Cette valeur exprime généralement qu'un message signale un problème non critique.
	 */	
	public final static byte LOGGER_WARNING=2;
	
	/**
	 * Constante pour qualifier un niveau critique.<br/>
	 * <br/>
	 * Cette valeur exprime l'échec d'une oprétation, ou un message signalant un problème d'importance critique. 
	 */
	public final static byte LOGGER_ERROR=3;
	//]	
	
	private Util(){} //cette classe ne s'instancie pas
	
	private static PrintStream _ps = null;	
	
	public static boolean same(String a, String b){ //true ssi a est identique à b
		if(a==null) return(b==null);
		else return(b!=null && a.equals(b));
	}
	
	/**
	 * Affiche un message log formaté sur une ligne entière.<br/>
	 * <br/>
	 * Le formatage réalisé est le suivant:  <b>tag</b> <b>categorie</b> <b>tag</b> <b>message</b>
 	 * @param tag chaine du <b>tag</b>
	 * @param categorie chaine de la <b>categorie</b>
	 * @param message chaine du <b>message</b>
	 */
	public static void log(String tag, String categorie, String message){
		if(_ps==null) try { _ps = new PrintStream(System.out, true, "utf8");} catch (UnsupportedEncodingException error) { _ps=System.out; }
		_ps.print(tag+" "); _ps.print(categorie); _ps.print(" "+tag);
		if(message!=null) _ps.println(" "+message);
		else  _ps.println();
	}
	/**
	 * Affiche un message log sur une portion de ligne.<br/>
	 * @param line texte à afficher.
	 * @param br flag indiquant si la ligne est terminée ou non. Si elle est terminée, le prochain affichage se fera sur la ligne suivante.
	 */
	public static void linelog(String line, boolean br){
		if(_ps==null) try { _ps = new PrintStream(System.out, true, "utf8");} catch (UnsupportedEncodingException error) { _ps=System.out; }
		if(br) _ps.println(line);
		else _ps.print(line);
	}	
	
	
	public static Vector<String> copyStringIterable(Set<String> l){
		Vector<String> cop = new Vector<String>();
		for(String cur: l) cop.add(cur);
		return(cop);
	}
	
	public static Vector<Integer> copyIntegerIterable(Set<Integer> l){
		Vector<Integer> cop = new Vector<Integer>();
		for(Integer cur: l) cop.add(cur);
		return(cop);
	}
	/**
	 * Génération aléatoire d'une chaîne alphanumérique de n lettres.<br/>
	 * <br/>
	 * La chaine produite est formée de n lettres alphanumériques sensibles à la casse (62 possibilités par lettres).<br/>
	 * Cette fonction est en particulier utilisée pour produire des identifiants de site aléatoires de 6 lettres (soient plus de 56,8 milliard de possibilités). 
	 * Mais elle peut être employée partout où la génération d'une chaine aléatoire à grande combinatoire est nécessaire.
	 * 
	 * @param n Nombre de lettres que devra contenir le mot généré.
	 * @return Chaîne de <b>n</b> lettres générées aléatoirement.
	 */
	public static String randString(int n){ //->chaîne alphanumérique de N lettres sensibles à la casse (soit 62^N possibilités)
		String rep="";
		for(int i=0; i<n; i++){
			int d=(int)(Math.random()*62.0); //d € [0,62[
			if(d<10) rep+=(char)('0'+d);
			else if(d<36) rep+=(char)('a'+(d-10));
			else rep+=(char)('A'+(d-36));
		}
		return(rep);
	}	
	
	/**
	 * Génération aléatoire d'un nombre entier dans un interval donné.<br/>
	 * <br/>
	 * @param min Valeur minimale (incluse) à générer
	 * @param max Valeur maximale (incluse) à générer
	 * @return
	 */
	public static int randInt(int min, int max){ //->chaîne alphanumérique de N lettres sensibles à la casse (soit 62^N possibilités)
		return(((int)(Math.random()*(max-min+1)))+min);
	}	
	
	/**
	 * Fusionne deux tableau de chaine de caractère
	 * @param a le premier tableau
	 * @param b	le deuxième tableau
	 * @return Un tableau produit par la fusion du tableau a et du tableau b
	 */
	public static String[] mergeArray(String[] a, String[] b){
		String[] rep=new String[(a==null?0:a.length)+(b==null?0:b.length)];
		int i=0;
		if(a!=null) for(int j=0; j<a.length; j++, i++) rep[i]=a[j];
		if(b!=null) for(int j=0; j<b.length; j++, i++) rep[i]=b[j];
		return(rep);
	}
	
	public static Vector<String> explode(String ch, char[] seps){
		Vector<String> res = new Vector<String>();
		int deb = 0;
		
		for(int i=0; i<ch.length(); i++){
			char cur = ch.charAt(i);
			boolean isSep = false;
			for(int j=0; !isSep && j<seps.length; j++) isSep = (cur == seps[j]);
			if(isSep){
				res.add(ch.substring(deb, i).trim());
				deb = i+1;
			}
		}
		if(deb<ch.length()){
			String last = ch.substring(deb).trim();
			if(last.length()>0) res.add(last);
		}
		return(res);
	}
	
	/**
	 * Traduit une chaine de caractère en entier.<br/>
	 * @param st La chaine de caractère à traduire.
	 * @param error_value La valeur à retourner si la traduction est impossible.
	 * @return La valeur traduite, ou error_value si la traduction n'est pas possible.
	 */
	public static int toInt(String st, int error_value){
		try{ return(Integer.parseInt(st));}
		catch(Exception e){ return(error_value); }
	}
	
	public static boolean isInt(String st){
		try{ Integer.parseInt(st); }
		catch(Exception e){ return(false); }
		return(true);
	}

	public static boolean isFloat(String st){
		try{ Float.parseFloat(st); }
		catch(Exception e){ return(false); }
		return(true);
	}
	
	public static float toFloat(String st, float error_value){
		try{ return(Float.parseFloat(st));}
		catch(Exception e){ return(error_value); }
	}
	
	  /**
	   * Traduction d'un code de qualification de niveau en une chaine de caractère.<br/>
	   * <br/> 
	   * Un code de qualification de niveau est l'une des constantes Logger exprimant une qualification de niveau.
	   * Il s'agit donc des valeurs : {@link utility.Logger#INFO}, {@link utility.Logger#WARNING}, et  {@link utility.Logger#ERROR}.
	   * 
	   * @param level Code de qualification de niveau à traduire.
	   * @return La chaîne de caractère correspondant au code de niveau fournit.
	   */
	  public static String levelString(byte level){
		  switch(level){
		  case LOGGER_INFO: return("Information");
		  case LOGGER_WARNING: return("Warning");
		  case LOGGER_ERROR: return("Error");		  
		  }
		  return("* unknow *");
	  }
	  
	  /**
	   * Fonction statique utilitaire permettant de tester si une chaine n'est composée que de caractères alphanumériques.<br/>
	   * 
	   * @param st chaine à analyser.
	   * @return <b>true</b> si la chaine ne contient que des caractères alphanumériques, <b>flase</b> sinon.
	   */
	  public static boolean isAlphanumeric(String st){
		  if(st!=null){
			  for(int i=0; i< st.length(); i++) {
				  char cur=st.charAt(i);
				  if(cur!='_' && !Character.isLetterOrDigit(cur)) return(false);
			  }
			  return(true);
		  } else return(false);
	  }
	  
	  public static MessageDigest _digesterMD5=null;
	  public static String cryptMD5(String ch){
		  if(_digesterMD5==null) 
			  try{ _digesterMD5=MessageDigest.getInstance("MD5"); } 
		  	  catch(Exception E){ System.out.println("[hubudp.util.Util - cryptMD5] : Erreur interne -> impossible d'optenir l'encodeur MD5"); return(null); }
	    	
		  _digesterMD5.update((ch).getBytes());
		  return(new BigInteger(1, _digesterMD5.digest()).toString(16));
	  }
	  public static String cryptMD5(String ch, String key){
		  if(_digesterMD5==null) 
			  try{ _digesterMD5=MessageDigest.getInstance("MD5"); } 
		  	  catch(Exception E){ System.out.println("[hubudp.util.Util - cryptMD5] : Erreur interne -> impossible d'optenir l'encodeur MD5"); return(null); }
	    	
		  _digesterMD5.update((ch+key).getBytes());
		  return(new BigInteger(1, _digesterMD5.digest()).toString(16));
	  }
	  
	  //cryptage/décryptage
	  public static String decode(String clef, String pass){
		  String cryptclef = cryptSHA1(clef); //Cryptage de la clef
		  byte[] data = Base64.getDecoder().decode(pass.getBytes()); // Décryptage code 64
	      //Décodeur de chaîne
		  int kc;
		  for (int i = 0; i < data.length; i++) {
			  if (i == 0) kc = cryptclef.charAt(cryptclef.length() - 1);
			  else kc = cryptclef.charAt((i % cryptclef.length()) - 1);
			  data[i] = (byte) (((int) data[i]) - ((int) kc));
		  }
		  // Résultat
		  String decryptpass = new String(data);
		  return(decryptpass);
	  }
	  public static String encode(String clef, String pass){
		  String cryptclef = cryptSHA1(clef); //Cryptage de la clef
		  //Encodeur de chaîne
		  byte[] data = new byte[pass.length()];
		  int kc;
		  for (int i = 0; i < data.length; i++) {
			  if (i == 0) kc = cryptclef.charAt(cryptclef.length() - 1);
			  else kc = cryptclef.charAt((i % cryptclef.length()) - 1);
			  data[i] = (byte) ((int) kc + (int) pass.charAt(i));
		  }
		  // Cryptage code 64
		  String cryptpass = new String(Base64.getEncoder().encode((data)));
		  // Résultat
		  return(cryptpass);
	  }
	  public static MessageDigest _digesterSHA1=null;
	  public static String cryptSHA1(String text){
		  if(_digesterSHA1==null) 
			  try{ _digesterSHA1=MessageDigest.getInstance("SHA-1"); } 
		  	  catch(Exception E){ System.out.println("[hubudp.util.Util - cryptSHA1] : Erreur interne -> impossible d'optenir l'encodeur SHA-1"); return(null); }
	    	
		  _digesterSHA1.reset();
		  return(new String(_digesterSHA1.digest(text.getBytes())));
	  }
	  
		//[Traitement utile pour la détermination de l'OS
		private static String OS_NAME=null;
	    public static boolean isWindows() { 
	    	if(OS_NAME==null) OS_NAME = System.getProperty("os.name").toLowerCase();
	    	return (OS_NAME.indexOf("win") >= 0); 
	    }
	    public static boolean isMac() {
	    	if(OS_NAME==null) OS_NAME = System.getProperty("os.name").toLowerCase();
	    	return (OS_NAME.indexOf("mac") >= 0); 
	    }
	    public static boolean isLinux() {
	    	if(OS_NAME==null) OS_NAME = System.getProperty("os.name").toLowerCase();
	    	return (OS_NAME.indexOf("nix") >= 0 || OS_NAME.indexOf("nux") >= 0 || OS_NAME.indexOf("aix") > 0 ); 
	    }
	    //]
	    
	    public static int bytesToInt(byte[] tab, int offset) {
	    	int int_val = 0;
	    	if(tab!=null) {
	    		if(tab.length>offset) int_val=tab[offset];
	    		offset++;
	    		if(tab.length>offset) int_val = (int_val << 8) | (0xFF & tab[offset]);
	    		offset++;
	    		if(tab.length>offset) int_val = (int_val << 8) | (0xFF & tab[offset]);
	    		offset++;
	    		if(tab.length>offset) int_val = (int_val << 8) | (0xFF & tab[offset]);
	    	}
			return(int_val);
	    }
	    public static byte[] intToBytes(int val, byte[] tab, int offset) {
	    	if(tab==null || tab.length<(offset+4)) {
	    		tab= new byte[4+offset];
	    	}
	    	tab[offset++]=(byte) ((val >> 24) & 0xff);
	    	tab[offset++]=(byte) ((val >> 16) & 0xff);
	    	tab[offset++]=(byte) ((val >> 8) & 0xFF);
	    	tab[offset]=(byte) ((val & 0xFF));
			return(tab);
	    }
	    public static byte[] floatToBytes(float val, byte[] tab, int offset) {
	    	int bits = Float.floatToIntBits(val);
	    	return(intToBytes(bits, tab, offset));
	    }
	    public static float bytesToFloat(byte[] tab, int offset) {
	    	int bits = bytesToInt(tab, offset);
	    	return(Float.intBitsToFloat(bits));
	    }
}