package skuad.util.base;

import java.io.*;
import javax.xml.parsers.*;
import javax.xml.transform.*;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import org.w3c.dom.*;

public class XmlManager {

	//[Element de traitement XML interne
	//Constante
	private static String HEAD_XML_DOC="<?xml version=\"1.0\" encoding=\"UTF-8\"?>";	
	private static DocumentBuilder builder=null;	
	private static Transformer transformer=null;

	private static void createBuilder() throws Exception {
		if(builder==null){
			DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
			builder = factory.newDocumentBuilder();
		} 
	}	  
	private static void createTransformer() throws Exception { 
		if(transformer==null){
			TransformerFactory tFactory = TransformerFactory.newInstance();
			transformer = tFactory.newTransformer();
			//On change le jeu de caractères pour supporter les accents
			java.util.Properties properties = transformer.getOutputProperties();
			properties.setProperty(OutputKeys.ENCODING,"UTF-8");
			properties.setProperty(OutputKeys.INDENT,"yes");
			//properties.setProperty(OutputKeys.METHOD,"xml");
			transformer.setOutputProperties(properties);
			transformer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "5");
		}
	}	 
	public static Document newXMLDocument(String racine_tag){
		Document doc=null;
		try{
			createBuilder();  	
			doc = builder.parse(new ByteArrayInputStream((HEAD_XML_DOC+"<"+racine_tag+"/>").getBytes("UTF-8")));
		} catch (Exception e) {
			e.printStackTrace();
			return(null);
		}
		return(doc);
	}
	protected static void setXMLDescription(Element element, String description){
		element.getParentNode().insertBefore(element.getOwnerDocument().createComment(description), element);
	}
	//]
	
	/**
	 * Lecture d'un fichier XML
	 * @param file_name
	 * @return noeud racine du fichier
	 */
	public static Element readFile(File file) {
		try{ createBuilder(); 	
		Document doc=null;
		InputStream in = new FileInputStream(file);
		doc = builder.parse(in);
		in.close();
		return(doc.getDocumentElement());
		}catch(Exception e){ e.printStackTrace(); return(null); }
	}
	
	/**
	 * Ecriture d'un fichier XML.<br/>
	 */
	public static boolean writeFile(Element node_racine, File file) {
		try{
			createTransformer();  
			Source source=new DOMSource(node_racine);
			OutputStream out = new FileOutputStream(file);
			StreamResult result = new StreamResult(out);
			transformer.transform(source, result); 
			out.close();
		} catch (Exception e) {
			e.printStackTrace();
			return(false);
		}
		return(true);
	}
	
	//[Lecture d'attribut
	public static float getFloatAtt(Element el, String attname, float err_value){
		String val=el.getAttribute(attname);
		if(val!=null){
			try{ return(Float.parseFloat(val));}
			catch(Exception e){}
		}
		return(err_value);
	}
	public static int getIntAtt(Element el, String attname, int err_value){
		String val=el.getAttribute(attname);
		if(val!=null){
			try{ return(Integer.parseInt(val));}
			catch(Exception e){}
		}
		return(err_value);
	}
	
	public static void setIntAtt(Element el, String attname, int value){
		el.setAttribute(attname, String.valueOf(value));
	}
	
	public static void setAttIfNotNull(Element el, String attname, String value){
		if(value!=null) el.setAttribute(attname,value);
	}

	public static String getAtt(Element el, String name, String defaultValue){
		  String val=el.getAttribute(name);
		  if(val==null || val.length()==0) return(defaultValue);
		  else return(val);
	}
	
	public static Element newChild(Element parent, String tag_child){
		Element child = parent.getOwnerDocument().createElement(tag_child); 
		parent.appendChild(child);
		return(child);
	}
}