/**
 * Cette classe spécifie les constantes du format d'encodage binaire DataBytes
 * 
 *  * Légende de l'encodage des données 
 
 		LABEL			-> le LABEL désigne une constante DataBytes
 		[label] 		-> un octet pour encoder l'élément désigné par label  
 		{X: label} 		-> X octets pour encoder l'élément désigné par label, X peut être : un nombre, un label spécifié en amont, ou le symbole ? pour indiquer une valeur variable
 		(name: label)	-> label encodé selon le format de nom name
 
 * Syntaxe de l'encodage des données  : 
 		- octet : 	[label]
 		- entier : 	(int: label) = [type_int]{ ? : valeur} , type_int peut être : ZERO (?=0), ONE (?=1), BYTE (?=1), NBYTE (?=1), SHORT (?=2), NSHORT (?=2), WORD (?=4), NWORD (?=4)
 																						-> sens des octets : poids fort vers poids faible.
 		- décimal : (float: label) = { 4 : valeur }
 		- tableau d'octet: (tab: label) = (int: size){size : contenu} 
 		- chaine de caractère: (string: label) = (int: size){size : contenu}
  		- valeur quelconque typée : (val : valeur) = [type]{ ? : valeur }, type peut être un entier (ZERO, ONE, BYTE, NBYTE, SHORT, NSHORT, WORD, NWORD) ou : FLOAT, TBYTES, STRING
  		
NB: les valeurs numériques sur plusieurs octets sont toujours codé de l'octet de poids fort vers l'octet de poid faible.
 */
package skuad.util.databytes;

import java.io.*;

import skuad.util.databytes.implementation.*;

public class DataBytes {
	
		//Code des types des entier bas niveau (0 à 30)
	public final static int ZERO 		= 0;   	//-> entier 0
	public final static int ONE 		= 1;   	//-> entier 1
	public final static int TINY_INT	= 16;	//=> tout code <TINY_INT code l'entier positif de sa valeur (0 à 15)
	public final static int BYTE 		= 17;  	//-> entier positif sur 1 octets
	public final static int NBYTE 		= 18;   //-> entier négatif sur 1 octets
	public final static int SHORT 		= 19;   //-> entier positif sur 2 octets
	public final static int NSHORT 		= 20;   //-> entier négatif sur 2 octets
	public final static int WORD 		= 21;   //-> entier positif sur 4 octets
	public final static int NWORD 		= 22;   //-> entier négatif sur 4 octets
	 	//Code des types haut niveau (31 à 40) 
	public final static int INT 		= 31;  	//-> entier
	public final static int FLOAT 		= 32;  	//-> float
	public final static int TBYTES 		= 33;   //-> tableau de byte
	public final static int STRING 		= 34;  	//-> chaine de caractère
	public final static int VALUE 		= 35;  	//-> type value

	//[API
	public static DataWriter writer(){ return(new BdbWriter()); }
	public static DataWriter writer(int initiale_size){ return(new BdbWriter(initiale_size)); }
	public static StreamWriter writer(OutputStream out){ return(new StreamWriter(out)); }
	
	public static DataReader reader(byte[] buff, int start, int length){ return(BDB.seed(buff, start, length).newReader(false)); }
	public static DataReader reader(InputStream in, int length){ return(BDB.seed(in, length).newReader(false)); }
	public static StreamReader reader(InputStream in){ return(new StreamReader(in)); }
	//]
}