/**
	Interface d'écriture des données encodé dans le format DataBytes
 */

package skuad.util.databytes;

import java.net.DatagramPacket;
import java.net.InetAddress;

import skuad.util.value.Value;

public interface DataWriter extends IWriter {
	
	//Gestion de l'instance
	public void reset();
	public boolean dispose();				//vérouille l'instance et libère les ressources (plus aucune action ne sera possible)
	public byte[] flushBuff();				//retourne une copie du buffer + dispose
	public DataReader flushReader();		//retourne une instance de lecture + dispose l'instance d'écriture courante
	public DataReader copyToReader();		//retourne une instance de lecture sur une copie des donnée (l'instance courante n'est pas disposé)
	public byte[] copyToBuff();				//retourne une copie du buffer
	public DatagramPacket toDatagramPacket(InetAddress iaddr, int port);

	//Méthodes d'écriture spécifique DataWriter
	public int size();				//taille total des données disponible
	public boolean setSize(int size);	//fixe la taille totale des données disponibles (les données situées audela seront perdu, celle situé entre l'ancienne taille et la nouvelle sont arbitraires)
									//le pointeur d'écriture est placé à la fin de cet espace.
	public boolean setByte(int pos, int value); //écrit l'octet value, à la position indiqué (sans déplacer le pointeur d'écriture), renvoie false et sans effet si pos>=position_courante_d_ecriture
	public int getByte(int pos);			
	public int setDataString(String msg);	//remplie l'intégralité du buffer data avec la chaine de caractère passé en argument (le précédent contenu de data est perdu)

	//+ les méthodes de l'interface IWriter
	
	//[Méthode d'écriture courte pour appel en cascade (pas de vérification d'erreur)
	public DataWriter b(int value); 	//écriture d'un byte
	public DataWriter b(boolean value); //écriture d'un byte
	public DataWriter b(byte[] value); 	//écriture d'un tableau de bytes (<-> copyBytes)

	public DataWriter sh(int value); 	//écriture d'une valeur sur 2 octets (writeShort)
	public DataWriter i(int value); 	//écriture d'un entier
	public DataWriter s(String value); 	//écriture d'une  chaine de catactère (string: val)
	public DataWriter w(int value); 	//écriture d'un mot (4 octets)
	public DataWriter d(long value); 	//écriture d'un double word (8 octets)
	public DataWriter v(Value val);		//écrit une valeur (value : val)
	//]
}