/**
	Interface de lecture des données encodé dans le format DataBytes
 */
package skuad.util.databytes;

import java.net.InetAddress;

import skuad.util.value.Value;

public interface IReader {	
	public boolean canRead(); 				//true ssi il est encore possible de lire 1 octet dans le paquet en cours de lecture
	public boolean canRead(int length); 	//true ssi il est encore possible de lire length octets dans le paquet en cours de lecture
	public int avaible();					//taille des données restant à lire
	public void skip(int d);				//avance le pointeur de lecture de d octet(s) 
	
	public int readByte();					//lecture du prochain octet, 0 si impossible ou erreur
	public int readInt();					//lecture du prochain entier : (int: label) = [type_int]{ ? : valeur}
	public int readInt(int type);			//lecture du prochain entier : { type_int : valeur}
	public boolean readBool();				//lecture du prochain boolean : 0 // 1
	public float readFloat();				//lecture (float: label)
	public byte[] readTbytes();				//lecture de (tab: label)
	public String readString();				//lecture de la prochaine chaine de caractère : (string: label)
	public Value readValue();				//lecture de la prochaine valeur : (value: v)
	public long readDoubleWord(); 			//leccture d'un double word (8 octets)
	
	//Méthodes de lecture bas niveau
	public String toString(); 							//retourne l'ensemble du buffer de donnée sous forme de chaine de caractère (ne déplace pas le pointeur de lecture)
	public byte[] readBytes(byte[] buff, int length); 	//lecture de length octets
	public byte[] readBytes(byte[] buff, int length, int to_start_offset); //comme readBytes(byte[] buff, int length) mais commence à écrire dans buff à partir de la position to_start_offset
	public int readShort();								//lecture du prochain entier sur 2 octet (non signé)
	public int readWord();								//lecture du prochain entier sur 4 octet (non signé)
	public String readString(int length); 				//lecture de la prochaine chaine de caractère : {letter : length}
	public String readStringIp();
	public InetAddress readIp(InetAddress sender);
	public String readStringIp(InetAddress sender);
}
