/**
	Interface d'écriture des données encodé dans le format DataBytes
 */

package skuad.util.databytes;

import skuad.util.value.Value;

public interface IWriter {
	
	//[Méthodes d'écritures de valeur non typé
	public boolean writeByte(int val); 		//écrit une valeur octet
	public boolean writeBool(boolean val);	//écrit une valeur 0 ou 1 en fonction de val
	public boolean writeShort(int value); 	//écrit la valeur value sur 2 octets (poids faible, puis poids fort) 
	public boolean writeWord(int value);	//écrit la valeur value sur 4 octets (poids faible -> poids fort) 
	public boolean writeBruteString(String msg);	//copie les octets de la chaine dans le buffer, sans la faire précédé de sa taille
	public boolean copyBytes(byte[] data); //copie le buffer data
/*	public boolean copyIp(IpPort target){ //copie le buffer data
		if(target.isLocalAddress()) return(copyTBytes(IpPort.byteLocalIp));
		else return(copyTBytes(target.byteIp()));
	}*/		
	public boolean copyBytes(byte[] data, int start, int length); //copie le buffer data[start, start+length[
	public int writeData(DataReader content); //écrit les données de content (depuis sa position de lecture courante), ==> ne change pas la position de lecture de content
	public int writeData(DataReader content, int size); //écrit les sizes prochaines données de content (depuis sa position de lecture courante),  ==> ne change pas la position de lecture de content
	//]
	
	//[Méthode d'écriture de valeur typé
	public boolean writeInt(int val);		//écrit la valeur (int: label)
	public boolean writeFloat(float val);	//écrit la valeur (float: label)
	public boolean writeString(String val); //écrit le codage (string: val)
	public boolean writeTbytes(byte[] val); //écrit le codage (tab: val)
	public boolean writeValue(Value val);	//écrit une valeur (value : val)
	//]
	
	//[Méthode d'écriture courte pour appel en cascade (pas de vérification d'erreur)
	public IWriter b(int value); 	//écriture d'un byte
	public IWriter b(boolean value); //écriture d'un byte
	public IWriter b(byte[] value); 	//écriture d'un tableau de bytes (<-> copyBytes)

	public IWriter sh(int value); 	//écriture d'une valeur sur 2 octets (writeShort)
	public IWriter i(int value); 	//écriture d'un entier
	public IWriter s(String value); 	//écriture d'une  chaine de catactère (string: val)
	public IWriter w(int value); 	//écriture d'un mot (4 octets)
	public IWriter d(long value); 	//écriture d'un double word (8 octets)
	public IWriter v(Value val);		//écrit une valeur (value : val)
	//]
}