package skuad.util.databytes;
import java.io.*;
import java.net.InetAddress;
import java.nio.charset.StandardCharsets;
import skuad.util.base.Network;
import skuad.util.value.Value;
import skuad.util.value.ValueWriter;

public class StreamReader implements IReader {

	InputStream source;
	boolean error;
	
	public StreamReader(InputStream s){
		source=s;
		error=false;
	}
	public boolean haveError() { return(error); }
	public boolean dispose() {
		try{ source.close(); source=null; }catch(Exception e){}
		return true;
	}
	
	//[Interface IReader
	public int readByte(){ 		//lecture du prochain octet, 0 si impossible ou erreur
		try{
			return(source.read()); 
		} catch(Exception e){ error=true; return(0); }
	}		
	public int readInt() { return(readInt(0)); } //lecture du prochain entier : (int: label) = [type_int]{ ? : valeur}
	public int readInt(int type) {try{ //si type==0, la valeur du type sera lue sur le 1er octet du flux source
		if(type==0) type = source.read(); 
		int val=0;
		if(type<DataBytes.TINY_INT) return(type);
		switch(type){
		case DataBytes.NBYTE: case DataBytes.BYTE: val = source.read(); if(type==DataBytes.NBYTE) val = -val; break;
		case DataBytes.NSHORT: case DataBytes.SHORT: val = readShort(); if(type==DataBytes.NSHORT) val = -val; break;
		case DataBytes.NWORD: case DataBytes.WORD: val = readWord(); if(type==DataBytes.NWORD) val = -val; break;
		}
		return(val);
	} catch(Exception e){ error=true; return(0); }}		
	public boolean readBool(){try{		//lecture du prochain entier : (int: label)
		return(source.read()!=0);
	} catch(Exception e){ error=true; return(false); }}	
	public float readFloat(){try{		//lecture (float: label)
		int rep;
		rep = source.read();			
		rep = rep << 8 | source.read();
		rep = rep << 8 | source.read();
		rep = rep << 8 | source.read();
		return(Float.intBitsToFloat(rep));
	} catch(Exception e){ error=true; return(0); }}	
	public byte[] readTbytes(){try{	//lecture de (tab: label)
		int size = readInt();
		byte[] buff = new byte[size];
		int offset = 0;
		while(size>0){
			int l = source.read(buff, offset, size);
			if(l<0) return(null);
			size -= l; offset+=l;
		}
		return(buff);
	} catch(Exception e){ error=true; return(null); }}	
	public String readString(){try{	//lecture de la prochaine chaine de caractère : (string: label) 
		String st=null;
		byte[] buff = readTbytes();
		st = new String(buff, StandardCharsets.US_ASCII);
		return(st);			
	} catch(Exception e){ error=true; return(null); }}
	public Value readValue(){ 
		int type = readByte();
		int nb_element;
		switch(type){
		case Value.STRING:	return(Value.s(readString()));
		case Value.INT: return(Value.i(readInt()));
		case Value.FLOAT: return(Value.f(readFloat()));
		case Value.BOOL:return(Value.b(readBool()));
		case Value.ARRAY:
			nb_element = readInt();
			ValueWriter va = Value.writer();
			for(int i=0; i<nb_element; i++) va.add(readValue());
			return(va.toConstant());
		case Value.BYTES:
			nb_element = readInt();
			byte[] bs =readBytes(null, nb_element,0);
			return(Value.bs(bs));			
		case Value.HASH:
			nb_element = readInt();
			ValueWriter vh = Value.writer();
			for(int i=0; i<nb_element; i++) vh.set(readString(), readValue());
			return(vh.toConstant());
		case Value.NULL: return(Value.NULL_VALUE);
		case Value.UNDEFINED: return(Value.UNDEFINED_VALUE);
		}
		return(Value.ERROR_VALUE);
	}
	
	//[Primitive de lecture des valeurs brutes :
	public int readShort(){ try{    //lecture d'un short (2 octets)
		int rep;
		rep = source.read() << 8; 
		rep = rep | source.read();
		return(rep);
	} catch(Exception e){ error=true; return(0); }}
	
	public int readWord(){ try{    //entier sur 4 octets
		int rep;
		rep = source.read();			
		rep = rep << 8 | source.read();
		rep = rep << 8 | source.read();
		rep = rep << 8 | source.read();
		return(rep);
	} catch(Exception e){ error=true; return(0); }}

	public long readDoubleWord(){ 	//leccture d'un double word (8 octets)
		long rep;
		rep = readWord(); 
		rep = rep << 32 | readWord();
		return(rep);
	}

	public int avaible() {
		try { return(source.available()); } catch(Exception e){ return(0); }
	}

	public boolean canRead() {
		try { return(source.available()>0); } catch(Exception e){ return(false); }
	}
	public boolean canRead(int length) {
		try { return(source.available()>=length); } catch(Exception e){ return(false); }
	}

	public void skip(int d) {
		try{ source.skip(d); } catch(Exception e){ error=true; }
	}

	public byte[] readBytes(byte[] buff, int length) {
		if(buff==null) buff=new byte[length];
		try{ source.read(buff, 0, length); } catch(Exception e){ error=true; }
		return(buff);
	}
	public byte[] readBytes(byte[] buff, int length, int offset_start) {
		if(buff==null) buff=new byte[length+offset_start];
		try{ source.read(buff, offset_start, length); } catch(Exception e){ error=true; }
		return(buff);
	}

	public String readString(int length) {
		String rep;
		if(length>0) {
			byte[] buff = readBytes(null, length);
			rep = new String(buff, 0, length, StandardCharsets.US_ASCII);
		} else rep ="";
		return(rep);
	}
	
	public InetAddress readIp(InetAddress sender){
		byte[] byte_ip = readBytes(null, 4);
		if(byte_ip[0]==127 && byte_ip[1]==0 && byte_ip[2]==0 && byte_ip[3]==1) return(sender);
		try { return(InetAddress.getByAddress(byte_ip)); } catch(Exception e){ return(null);}
	}
	
	public String readStringIp(InetAddress sender){
		return(Network.addressToIp(readIp(sender)));
	}

	public String readStringIp() {
		return(readByte()+"."+readByte()+"."+readByte()+"."+readByte());
	}
	//]
}