package skuad.util.time.temporality;

import skuad.util.time.*;

public class InternalTemporalRule implements Pulse {

 //Attribut
 TemporalityScheduler scheduler;
 String label;		
 long delay, periode;
 boolean actif;
 int compteur;
 TimeFilter time_filter;
 
 PulseTarget target;
 TimeSlot timeSlot;	//prochain timeSlot d'exécution
 
 public InternalTemporalRule(TemporalityScheduler vscheduler, PulseTarget vtarget, String vlabel, long vdelay, long vperiode) {
	 scheduler = vscheduler;
	 target=vtarget;
	 label=vlabel;
	 delay=(vdelay>0 ? vdelay : 0);
	 periode=vperiode;
	 
	 compteur=0;
	 time_filter=null;
	 actif=scheduler.register(this); //va conduire à fixer timeSlot
 }
 
 //[Interface Pulse
 public boolean is(String vlabel) {
	 return(label!=null ? vlabel==null : label.equals(vlabel));
 }
 public String label() { return(label); }
 public boolean isLoop() { return(periode>0); } 	//true ssi c'est une pulsation périodique 
 
 public boolean isActive() { return(actif); }
 public boolean cancel() { //annule ce pulse
	 if(actif) {
		 actif=false;
		 scheduler.unregister(this);;
		 return(true);
	 }
	 return(false);
 }
 public int count() { return(compteur); }
 
 public void setFilter(TimeFilter tf) {	//change le filtre d'activation de ce pulse, si tf=null cela revient à supprimer le filtre
	 time_filter=tf;
 }
 //]
 
 //[Méthode de consultation deverses
 public long getPeriod(){ return(periode); }//période de la règle (-1 si règle ponctuel)  
 public long getDelay(){ return(delay); }
 public int cycleCount() {return(compteur); }  //nombre de fois que la rèele a été déclenché
 public long getNextTime(){	//date de prochain déclenchement de cette rêgle (-1 si non connnu)
	 return(timeSlot!=null ? timeSlot.getTime() : -1);
 } 
 //]
 
 //[Méthode pour le scheduler
 protected void setTimeSlot(TimeSlot ts){ timeSlot=ts; }
 protected TimeSlot getTimeSlot() { return(timeSlot); }
 
 /**
  * Active cette rêgle temporelle (=> execute le comportement associé)
  */
 protected boolean run(long tc){
	 if(actif && (time_filter==null || time_filter.validTime(tc))) {
		 compteur++;
		 target.run(this);
		 if(periode<=0) actif=false;
	 }
	 return(actif);
 }
}