package skuad.util.time.temporality;

public class TemporalityScheduler {
  
 TimeSlot curTimeSlot;	//prochain slot temporel a exécuter	
 long start_time;	//date du 1er cycle de la simulation
 long end_time; 	//==-1 si la simulation n'a pas de date de fin
 long cur_time;		//date du cycle courant
 long ping;	 //délais ajouté à chaque nouvel InternalTemporalRule engristré durant le traitement de runCurTime (valeur fixée au début de cette méthode, et remis à 0 à la fin)
 long minimal_step; //pas de temps minimum
 
 public TemporalityScheduler(){
	 reset(0, -1);
 }
 
 /**
  * Initialisation du Scheduler
  */
 public long reset(long start,long end){ // /!\ détruit toute les règles temporelle déjà enregistrée
	 ping=0;
	 start_time=(start>0 ? start : 0);
	 end_time=(end<0 ? -1 : end);
	 minimal_step=0;
	 return(reset());
 }
 public long reset(){ // /!\ détruit toute les règles temporelle déjà enregistrée
	 curTimeSlot=null; 
	 cur_time = start_time;
	 return(cur_time);
 } 

 public long timeStart(){ return(start_time);}
 public long timeEnd(){ return(end_time);}
 public long curTime() { return(cur_time); }

 /**
  * Test si l'exécution est terminée
  */
 public boolean terminated(){ return(end_time>=0 && cur_time>end_time); }
 public boolean empty() { return(curTimeSlot==null); }
 public void setPing(long vping) { ping=vping; } //vping sera automatiquement ajouté au délais de chaque nouvel InternalTemporalRule engristré 
 public long getPing() { return(ping); } 
 public void setMinimalStep(long vmstep) { minimal_step=vmstep; } 
 public long getMinimalStep() { return(minimal_step); }
 
 public void runCurTime() { 
	 if(curTimeSlot!=null && curTimeSlot.getTime()==cur_time) {
//		 curTimeSlot.trace(10);
		 
		 TimeSlot current_slot = curTimeSlot;
		 if(current_slot.run()) { //on doit replacer ce slot
			 TimeSlot next_slot = current_slot.getNextSlot();
			 
			 long next_time=current_slot.getTime();
			 if(end_time<0 || next_time<=end_time) {
				 if(next_slot!=null && next_time>=next_slot.getTime()) {
					 current_slot.setNextSlot(null); //on détruit le next_slot du current_slot à replacer
					 curTimeSlot=next_slot;
					 curTimeSlot.place(current_slot,next_time);
				 } //else //on laisse current_slot sur curTimeSlot sans toucher son next_slot
			 } else curTimeSlot=next_slot;
		 } else { 
			 curTimeSlot=current_slot.getNextSlot();
		 }
 	}
 }

 public long nextTC() { //détermination du prochain TimeCode candidat:  <=0 si aucun et simulation infini
	 return(curTimeSlot==null ? end_time+1 : curTimeSlot.getTime());
 }
 
 public long goStartTime() { cur_time=start_time; return(start_time); }
 public void setCurTime(long tc) { cur_time=tc; }
 
 /**
  * Enregistre une nouvelle règle temporelle
  */ 
 public synchronized boolean register(InternalTemporalRule tr){

	 long delay = tr.getDelay();
	 long periode=tr.getPeriod(); 

	 long begin= cur_time+Math.max(delay, ping); //1er déclenchement de cette règle
	 
	 if(end_time>=0 && begin>end_time) return(false); //règle périmé

	 if(curTimeSlot==null) curTimeSlot=new TimeSlot(this, null, tr, begin, periode);
	 else {
		 if(begin<curTimeSlot.getTime()) curTimeSlot=new TimeSlot(this, curTimeSlot, tr, begin, periode);
		 else curTimeSlot.place(tr,begin,periode);
	 }
//System.out.println("Registred rule count = "+curTimeSlot.trList.size());	 
	 return(true);
 }
 
 /**
  * Désenregistre une règle temporelle
  */
 public void unregister(InternalTemporalRule tr){
	 TimeSlot ts=tr.getTimeSlot();
	 if(ts!=null) ts.remove(tr);
 }
 
 protected void replaceSlot(TimeSlot ts){ //ici on a forcément curTimeSlot!=null , curTimeSlot!=ts et ts.time>=curTimeSlot.time
	 long next_time=ts.getTime();
	 if(end_time<0 || next_time<=end_time) curTimeSlot.place(ts,next_time);
 }
}