package skuad.util.time.temporality;

import java.util.*;

/*
 * Représente un timecode
 */
public class TimeSlot {

 //[Attribut
 TemporalityScheduler scheduler;	
 long time;		//date du slot temporel
 long periode;	//période des règles inscrite dans ce TimeSlot
 //]
 
 //[Gestion de la liste chainée des TimeSlot (L'ensemble des TimeSlot instancié forme une liste chainée)
 TimeSlot sameSlot;	//autre TimeSlot a exécuter dans le même slot temporel
 TimeSlot nextSlot;	//prochain slot temporel a être exécuté (le prochain dans la liste chainée)
 //]
 
 //[Liste des règles temproelles addossées à ce TimeSlot
 Vector<InternalTemporalRule> trList, tmpAddTrList, tmpRemoveTrList; //liste des rêgles inscrite dans ce TimeSlot
 boolean lock_trlist;
 //]	
 
 //Méthode
 public TimeSlot(TemporalityScheduler vscheduler, TimeSlot nts,  InternalTemporalRule frule, long v_time,long v_periode) {
	 scheduler=vscheduler;
	 time=v_time;
	 periode=v_periode;
	 trList=new Vector<InternalTemporalRule>();
	 tmpAddTrList=new Vector<InternalTemporalRule>();
	 tmpRemoveTrList=new Vector<InternalTemporalRule>();
	 trList.add(frule);
	 sameSlot=null;
	 nextSlot=nts;
	 lock_trlist=false;	 
 }
 //]

 public void add(InternalTemporalRule tr){
	 tr.setTimeSlot(this);
	 synchronized(tmpAddTrList) { 
		 if(lock_trlist) tmpAddTrList.add(tr); 
		 else trList.add(tr);	 
	 }
 }
 public void remove(InternalTemporalRule tr){
	 synchronized(tmpRemoveTrList) { 
		 if(lock_trlist) tmpRemoveTrList.add(tr); 
		 else if(trList.remove(tr)) tr.setTimeSlot(null);
	 }
 }
 
 
 /**
  * réaffecte l'ensemble des règles portées ce slot sur le slot ts
  */
 protected void replaceAllInSlot(TimeSlot ts){
	 System.out.println("test = "+(ts==this));
	 for(InternalTemporalRule cur: trList) ts.add(cur);
 }
 
 protected void setNextSlot(TimeSlot ts) { //Attention on considere que ts.nextSlot==NULL !
	 if(ts!=null) ts.nextSlot=nextSlot;
	 nextSlot=ts;
 }
 
 public synchronized void place(TimeSlot t,long atTime){
	 if(Math.abs(atTime-time)<=scheduler.getMinimalStep()){ //on va le placer sur le même slot temporelle
		 if(t.periode==0 || periode==0 || t.periode==periode) t.replaceAllInSlot(this); //même périodicité, on fusionne les listes de règle temporelle
		 else {
			 if(sameSlot!=null) sameSlot.place(t,atTime); //on parcours en récursif pour traiter une éventuelle fusion
			 else sameSlot=t;
		 }
	 } else {
		 if(nextSlot==null) nextSlot=t;
		 else if(atTime<nextSlot.getTime()){ //on insére t entre ici et le prochain slot
			 t.setNextSlot(nextSlot);
			 nextSlot=t;
		 } else if(t!=nextSlot) nextSlot.place(t,atTime);
	 }
 }
 
 /**
  * traite le placement d'une régle (TemporalRule) dans le TimeSlot approprié
  */
 public synchronized void place(InternalTemporalRule tr,long atTime,long tr_periode){
	 if(atTime>time){ //celle régle doit être placée dans les futurs slots temporels
		 if(nextSlot==null) nextSlot=new TimeSlot(scheduler, null, tr, atTime, tr_periode);  //on crée un nouveau slot pour la période de tr
		 else { //on regarde si on doit créer un slot intermédiaire entre celui-ci est le prochain
			 long dif=nextSlot.getTime()-atTime;
			 if(dif>0) nextSlot=new TimeSlot(scheduler, nextSlot, tr, atTime, tr_periode);  //oui
			 else nextSlot.place(tr, atTime, tr_periode);
		 }
	 } else if(periode==0) {  periode=tr_periode; add(tr); }	//on peut accueillir localement cette règle
	 else if(tr_periode==0 || tr_periode==periode) add(tr); 	//on peut accueillir localement cette règle
	 else { //on doit la placer dans un slot SameSlot
		 if(sameSlot==null) sameSlot=new TimeSlot(scheduler, null, tr,atTime,tr_periode);  //on crée un nouveau slot pour la période de tr
		 else sameSlot.place(tr,atTime,periode);
	 }
 }
 
 public synchronized boolean run(){
	 lock_trlist=true;
	 InternalTemporalRule cur;
//System.out.println("Run slot on : "+UT.format(time)+" for count client = "+trList.size());	 
	 for(int i=0; i<trList.size(); i++){
		 cur=trList.get(i);
		 if(!cur.run(time)) {
			 cur.setTimeSlot(null); //cette règle n'est plus valide
			 trList.removeElementAt(i);
			 i--;
		 }
	 }
	 lock_trlist=false;
	 
	 //on traite les règles en attente d'insertion.suppresion
	 synchronized(tmpRemoveTrList) { for(InternalTemporalRule tr : tmpRemoveTrList) if(trList.remove(tr)) tr.setTimeSlot(null); }
	 synchronized(tmpAddTrList) { for(InternalTemporalRule tr : tmpAddTrList) trList.add(tr); }
	 
	 if(sameSlot!=null) { //on execute tous les TimeSlots situés sur le même slot
		 TimeSlot cur_same_slot = sameSlot;
		 sameSlot = null;
		 if(cur_same_slot.run()) scheduler.replaceSlot(cur_same_slot); 
		 else System.out.println("TimeSlot");
	 }
	 
	 //prochain temps d'exécution
	 if(periode>0) {
		 time+=periode;		 
		 return(!trList.isEmpty()); //doit être repositionné
	 } else return(false);  //time slot terminé
 }
 
 public long getTime(){ return(time); }
 public int ruleCount(){
	 return(trList.size());
 }
 public TimeSlot getNextSlot(){ return(nextSlot); }
  
 public void trace(int n){
	 System.out.println("time="+time+" priode="+periode);
	 if(n>0 && sameSlot!=null) sameSlot.traceSame(n-1);
	 if(n>0 && nextSlot!=null) nextSlot.trace(n-1);
 }
 public void traceSame(int n){
	 System.out.println(" -same- time="+time+" priode="+periode);
	 if(n>0 && sameSlot!=null) sameSlot.traceSame(n-1);
 }
}