package skuad.util.value.implementation;

import skuad.util.value.Value;
import skuad.util.base.*;

public class ValueParser {
	
	public final static Value parse(String desc){
		if(desc==null) return(Value.NULL_VALUE);
		desc = desc.trim();
		return(parse(desc, new PosPointer(), (char) 0,(char) 0));
	}
	
	//[Méthode interne
	public final static Value parse(String desc, PosPointer p, char end1, char end2){ 
		//Parse la prochaine valeur dans la chaine desc[p.pos, ....] et fixe dans p.end_char le caractère de fin de valeur détecté
		// => p.end_char pourra être soit end1, soit end2  (==0 si la fin de chaine est un caractère de fin possible)
		//-> les espaces/tab et retour chariot avant et aprés la valeur extraite sont absorbés
		//-> en fin d'extraction p.pos pointera aprés le caractère p.end_char (si !=0)
		//-> si une valeur est extraite et que le caractère de fin (0, end1 ou end2) n'est pas trouvé : => ERROR_VALUE est retournée
		//-> si aucune valeur n'est extraite avant le caractère de fin : => NULL_VALUE est retournée
		Value rep=Value.NULL_VALUE;
		while(p.pos<desc.length()){
//System.out.println("--- loop ---");			
			char cur_char = desc.charAt(p.pos);
			if(cur_char==end1 || cur_char==end2) {
				p.end_char = cur_char;
				p.inc();
//System.out.println("--- ok end loop ---");					
				return(rep); //fin de traitement conforme
			} else if(cur_char==' ' || cur_char=='\t' || cur_char=='\n') p.inc();
			else { //on est sur un caractère !=' ' et !=end1&2 , donc le début d'une nouvelle valeur
				if(rep!=Value.NULL_VALUE) p.error(); //on a déjà extrait une valeur, donc on est en situation d'erreur !
				else switch(cur_char){ //analyse de la valeur à extraire
				case '[': //--> array
					rep = _array(desc, p.inc());
					break;
				case '{': //--> hash
					rep = _hash(desc, p.inc());
					break;
				case '"':
					String val = _string(desc, p.inc(), '"', false); //--> valeur string
					if(!p.error) rep= new CAtome(val);
					break;
				default : //autre type de valeur (au moins 1 caractère): int / float / bool / string (qui ne commence pas par ")
					int pos_end = p.pos+1; //recherche de la position de fin de la valeur					
					while(pos_end<desc.length()){
						char c = desc.charAt(pos_end);
						if(c!=end1 && c!=end2) pos_end++;
						else break;
					}
					String cur = desc.substring(p.pos, pos_end).trim();
					if(Util.isInt(cur)) rep=CIntValue.of(Util.toInt(cur, 0));
					else if(Util.isFloat(cur)) rep = new CFloatValue(Util.toFloat(cur, 0));
					else if(cur.equalsIgnoreCase("true")) rep = Value.TRUE;
					else if(cur.equalsIgnoreCase("false")) rep = Value.FALSE;
					else if(cur.equalsIgnoreCase("null")) rep=Value.NULL_VALUE;
					else rep = new CAtome(cur);
					p.pos = pos_end;
				}
				if(p.error) return(Value.ERROR_VALUE); //error de parsing
			}
		}
//System.out.println("--- end loop ---");			
		if(end1!=0 && end2!=0) p.error(); //caractère de fin explicite non trouvé et non consommé !!!
		return(rep);
	}

	protected final static Value _array(String desc, PosPointer p){ //=>  desc =  val1, val2, .... ,valN ]
		WArray warray = new WArray();
		while(!p.error){
			Value v = parse(desc, p, ',', ']');
			if(!p.error){
				warray.add(v);
				if(p.end_char == ']') return(warray.toConstant()); //fin d'extraction du tableau
			}
		}
		return(Value.ERROR_VALUE); //error de parsing
	}
	
	protected final static Value _hash(String desc, PosPointer p){ //=> desc = name1 : val1,  name2 : val2, .... , nameN : valN }
		WHash whash = new WHash();
		while(!p.error){
			char cur_char;
			while(p.pos<desc.length()){ //absorbtion des espaces
				cur_char = desc.charAt(p.pos);
				if(cur_char!=' ' && cur_char!='\t' && cur_char!='\n') break; //fin d'absorption
				p.inc();
			}
			if(p.pos<desc.length()){
				cur_char = desc.charAt(p.pos);
				if(cur_char=='}'){ //fin d'extraction
					p.inc();
					return(whash.toConstant());
				}
				String name; Value value;
				//extraction du nom du champ
				if(cur_char=='"') { 
					p.inc(); 
					name = _string(desc, p, '"', false);
					if(!p.error){ //extraction du symbole ':'						
						while(p.pos<desc.length()){ //absorbtion des espaces
							cur_char = desc.charAt(p.pos);
							if(cur_char!=' ' && cur_char!='\t' && cur_char!='\n') break; //fin d'absorbtion
							p.inc();
						}
						if(p.pos>=desc.length() || cur_char!=':') p.error();
						else p.inc();
					}
				} else name = _string(desc, p, ':', true);		
				if(!p.error){ //extraction de la valeur
					value = parse(desc, p, ',', '}');						
					if(!p.error){
						whash.set(name, value);
						if(p.end_char == '}') {							
							return(whash.toConstant()); //fin d'extraction du hash
						}
					}
				}
			} else p.error();
		}
		return(Value.ERROR_VALUE); //error de parsing
	}
	
	protected final static String _string(String desc, PosPointer p, char end, boolean trim){ //=> desc = chaine'end' , dans chaine \ doit s'écrire \\ et " doit s'écrire \"
		StringBuffer st_buff=new StringBuffer();
		while(p.pos<desc.length()){
			char cur = desc.charAt(p.pos);
			if(cur==end) { //fin d'extraction
				p.inc();
				String rep = st_buff.toString();
				if(trim) rep = rep.trim();
				return(rep);
			}
			if(cur=='\\'){
				p.inc();
				cur = desc.charAt(p.pos);
				if(cur!='\\' && cur!='"') st_buff.append('\\'); // '\' non absorbé
			}
			p.inc();
			st_buff.append(cur);
		}
		p.error(); //caractère end de fin de chaine non trouvé --> erreur
		return("");
	}
}

class PosPointer {
	int pos;
	char end_char;
	boolean error;
	public PosPointer(){ pos=0; error = false; end_char=0; }
	public PosPointer inc(){ pos++; return(this); }
	public void error(){ error = true; }
}