/*
 * NB: le contenu d'un array sont des values constantes !
 */
package skuad.util.value.implementation;

import skuad.util.value.Value;
import skuad.util.value.ValueReader;

/**
 * Interface d'édition des objets value de type {@link value.Value#ARRAY}.<br/> 
 */
public class WArray extends CArray  {
	CArray last_constant=null;

	/**
	 * Renseigne la descpription (ou le commentaire) de cette valeur.<br/>
	 * <br/>
	 * Cette chaine de description (ou commentaire) est celle renvoyée par la méthode {@link value.Value#description()}.
	 * @param desc La nouvelle chaine de description (ou commentaire) à associer à cette valeur.
	 */
	public void setDescription(String desc){ description=desc; };
	
	/**
	 * Ajoute une valeur à la fin du tableau.
	 * 
	 * @param a Valeur à ajouter.
	 * 
	 * @return l'objet lui-même (pour permettre les appels en cascade)
	 */
	public WArray add(ValueReader a) { synchronized(content) { content.add(a.toConstant()); last_constant=null; } return(this); }

	/**
	 * Supprime la valeur située à l'indice <i>i</i>.<br/>
	 * <br/>
	 * Cette méthode ne produira aucun effet si la position indiquée n'est pas valide.<br/>
	 * Notez que toutes les valeurs situées après la position <i>i</i> verront leur indice de position décrémenté de 1 
	 * (le tableau ne comporte pas de trou dans sa numérotation)
	 * 
	 * @param i Position dans le tableau où se trouve la valeur à supprimer.
	 * 
	 * @return l'objet lui-même (pour permettre les appels en cascade)
	 */	
	public WArray remove(int i) { synchronized(content) { if(content.remove(i)!=null) last_constant=null; }  return(this); }
	
	public WArray removeAll() { synchronized(content) { if(content.size()>0){ last_constant=null; content.removeAllElements();} } return(this); }
	
	/**
	 * Insère une valeur à la position d'indice <i>i</i>.<br/>
	 * <br/>
	 * Cette méthode ne produira aucun effet si la position indiquée n'est pas valide.<br/>
	 * Notez que toutes les valeurs situées à partir de la position <i>i</i> avant l'insertion verront leur indice de position incrémenté de 1 
	 *
	 * @param a Valeur à insérer.
	 * @param i Position dans le tableau où doit être insérée la valeur.
	 * 
	 * @return l'objet lui-même (pour permettre les appels en cascade)
	 */	
	public WArray insert(Value a, int i) { synchronized(content) { content.insertElementAt(a.toConstant(),i); last_constant=null; } return(this); }
		//écriture rapide
	
	/**
	 * Ajout rapide d'une valeur chaîne de caractère à la fin du tableau.
	 * 
	 * @param a La chaîne de caractères à ajouter.
	 * @return l'objet lui-même (pour permettre les appels en cascade)
	 */
	public WArray add(String a){ return(add(new CAtome(a))); }
	
	/**
	 * Ajout rapide d'une valeur entière à la fin du tableau.
	 * 
	 * @param v La valeur entière à ajouter.
	 * @return l'objet lui-même (pour permettre les appels en cascade)
	 */
	public WArray add(int v){ return(add(CIntValue.of(v))); }
	
	/**
	 * Ajout rapide d'une valeur booléenne à la fin du tableau.
	 * 
	 * @param b La valeur booléenne à ajouter.
	 * @return l'objet lui-même (pour permettre les appels en cascade)
	 */
	public WArray add(boolean b){ return(add( b ?CBoolValue.TRUE : CBoolValue.FALSE )); }
	
	/**
	 * Ajout rapide d'une valeur décimale à la fin du tableau.
	 * 
	 * @param f La valeur décimale à ajouter.
	 * @return l'objet lui-même (pour permettre les appels en cascade)
	 */
	public WArray add(float f){ return(add(new CFloatValue(f))); }
	
	public WArray add(byte[] bs){ return(add(new CAtome(bs))); }
	//]
	
	public WArray set(int pos, ValueReader a) { 
		synchronized(content) { if(pos>=0 && pos<content.size()){
			content.set(pos, a.toConstant()); 
			last_constant=null; 
		}}
		return(this); 
	}
	public WArray set(int pos, String a){ return(set(pos, new CAtome(a))); }
	public WArray set(int pos, int v){ return(set(pos,CIntValue.of(v))); }
	public WArray set(int pos, boolean b){ return(set(pos, b ?CBoolValue.TRUE : CBoolValue.FALSE )); }
	public WArray set(int pos, float f){ return(set(pos, new CFloatValue(f))); }
	public WArray set(int pos, byte[] bs){ return(set(pos, new CAtome(bs))); }
	
	/**
	 * Transforme l'objet WArray en un objet value non modifiable.
	 * 
	 * @return une version non modifiable de l'objet WArray
	 */
	public Value toConstant(){
		if(last_constant==null) last_constant=new CArray(content, description());
		return(last_constant); 
	}
}