/*
 * NB: le contenu des entrées d'un Hash sont des values constantes !
 */
package skuad.util.value.implementation;

import java.util.*;

import skuad.util.value.Value;

/**
 * Interface d'édition des objets value de type {@link value.Value#HASH}.<br/> 
 */
public class WHash extends CHash {
	CHash last_constant=null;
	
	public WHash(){}
	public WHash(Value v){ //WritableHash avec copy du contenu de v si c'est un Hash
		if(v.isHash()) for(String k : v.keys()) content.put(k, v.get(k));
	}
	
	/**
	 * Renseigne la descpription (ou le commentaire) de cette valeur.<br/>
	 * <br/>
	 * Cette chaine de description (ou commentaire) est celle renvoyée par la méthode {@link value.Value#description()}.
	 * @param desc La nouvelle chaine de description (ou commentaire) à associer à cette valeur.
	 */
	public void setDescription(String desc){ description=desc; };

	
	//[Méthode d'écriture
	/**
	 * Ajoute une valeur à associer à l'étiquette indiquée.
	 *
	 * @param p Etiquette à attribuer à la valeur.
	 * @param a Valeur à ajouter.
	 * 
	 * @return l'objet lui-même (pour permettre les appels en cascade)
	 */
	public WHash set(String p,Value a){  
		synchronized(content) { 
			content.put(p, a.toConstant());
			last_constant=null;
		}
		return(this);
	}
	
	/**
	 * Supprime la valeur correspondante à l'étiquette <i>p</i>.<br/>
	 * <br/>
	 * Cette méthode ne produira aucun effet si l'étiquette indiquée n'est pas connue.<br/>
	 * 
	 * @param p Etiquette de la valeur à supprimer.
	 * 
	 * @return l'objet lui-même (pour permettre les appels en cascade)
	 */	
	public WHash remove(String p){ synchronized(content) { if(content.remove(p)!=null) last_constant=null; } return(this); }
		//écriture rapide
	
	public WHash removeAll(){ synchronized(content) { if(content.size()>0){ content.clear(); last_constant=null; } } return(this); }
	
	/**
	 * Ajout rapide d'une valeur chaîne de cractère à associer à l'étiquette indiquée.<br/>
	 * <br/>
	 * Si une valeur est déjà associée à l'étiquette spécifiée, alors cette valeur sera remplacée par celle fournie dans cet appel.
	 * 
	 * @param p Etiquette à attribuer à la valeur.
	 * @param a La chaîne de caractères à ajouter.
	 * @return l'objet lui-même (pour permettre les appels en cascade)
	 */	
	public WHash set(String p,String a){ return(set(p, new CAtome(a))); }
	
	/**
	 * Ajout rapide d'une valeur entière à associer à l'étiquette indiquée.<br/>
	 * <br/>
	 * Si une valeur est déjà associée à l'étiquette spécifiée, alors cette valeur sera remplacée par celle fournie dans cet appel.
	 * 
	 * @param p Etiquette à attribuer à la valeur.
	 * @param v Valeur entière à ajouter.
	 * @return l'objet lui-même (pour permettre les appels en cascade)
	 */	
	public WHash set(String p,int v){ return(set(p, CIntValue.of(v))); }
	
	/**
	 * Ajout rapide d'une valeur booléenne à associer à l'étiquette indiquée.<br/>
	 * <br/>
	 * Si une valeur est déjà associée à l'étiquette spécifiée, alors cette valeur sera remplacée par celle fournie dans cet appel.
	 * 
	 * @param p Etiquette à attribuer à la valeur.
	 * @param b La valeur booléenne à ajouter.
	 * @return l'objet lui-même (pour permettre les appels en cascade)
	 */	
	public WHash set(String p,boolean b){ return(set(p, (b?Value.TRUE:Value.FALSE))); }
	
	/**
	 * Ajout rapide d'une valeur décimale à associer à l'étiquette indiquée.<br/>
	 * <br/>
	 * Si une valeur est déjà associée à l'étiquette spécifiée, alors cette valeur sera remplacée par celle fournie dans cet appel.
	 * 
	 * @param p Etiquette à attribuer à la valeur.
	 * @param f La valeur décimale à ajouter.
	 * @return l'objet lui-même (pour permettre les appels en cascade)
	 */	
	public WHash set(String p,float f){ return(set(p, new CFloatValue(f))); }
	
	public WHash set(String p,byte[] bs){ return(set(p, new CAtome(bs))); }
	//]
	
	/**
	 * Transforme l'objet WHash en un objet value non modifiable.
	 * 
	 * @return une version non modifiable de l'objet WHash
	 */	
	public Value toConstant(){
		if(last_constant==null) {
			Hashtable<String,Value> for_clone=new Hashtable<String,Value>();
			synchronized(content) { for(String k : content.keySet()) for_clone.put(k, content.get(k).toConstant()); }
			last_constant=new CHash(for_clone, description());
		}
		return(last_constant); 
	}
}