package skuad.util.value.implementation;

import skuad.util.value.Value;
import skuad.util.value.ValueWriter;

public class WValue extends ValueWriter {
		
	Value vbool; 	//pour le type BOOL
	WHash whash; 	//pour le type HAS
	WArray warray; 	//pour le type ARRAY
	WAtome watome;	//pour les types STRING, INT, FLOAT et BYTES
	
	//[reader
	protected byte type;
	protected String description;
	
	public WValue(){ 
		type=Value.NULL; 
	}
	
	public void setDescription(String vd){	
		description = vd;
	}
	
	//[Analyse du type 
	public byte getType(){ return(type); }
	public boolean isAtome(){ return(type==Value.STRING || type==Value.INT || type==Value.BOOL || type==Value.FLOAT ); }
	public boolean isArray(){ return(type==Value.ARRAY); }
	public boolean isHash(){ return(type==Value.HASH); }
	public String description(){ return(description); }
	//]
	
	//[Lecture de valeur
		//[si type est un ATOME (STRING, INT, BOOL)
	public String toString(){ 
		switch(type){
		case Value.BOOL: return(vbool.toString());
		case Value.STRING: case Value.INT: case Value.FLOAT: case Value.BYTES: return(watome.toString());
		default : return(null); 
		}
	}
	public int toInt(){ 		
		switch(type){
		case Value.BOOL: return(vbool.toInt());
		case Value.STRING: case Value.INT: case Value.FLOAT:  case Value.BYTES: return(watome.toInt());
		default : return(0); 
		}
	}
	public boolean toBool(){ 		
		switch(type){
		case Value.BOOL: return(vbool.toBool());
		case Value.STRING: case Value.INT: case Value.FLOAT:  case Value.BYTES: return(watome.toBool());
		default : return(false); 
		}
	}
	public float toFloat(){ 		
		switch(type){
		case Value.BOOL: return(vbool.toFloat());
		case Value.STRING: case Value.INT: case Value.FLOAT:  case Value.BYTES: return(watome.toFloat());
		default : return(0); 
		}
	}
	public byte[] toBytes() {
		switch(type){
		case Value.BOOL: return(vbool.toBytes());
		case Value.STRING: case Value.INT: case Value.FLOAT:  case Value.BYTES: return(watome.toBytes());
		default : return(null); 
		}
	}
		//]
		//[si type==ARRAY
	public boolean exist(int p){ 
		if(type!=Value.ARRAY) return(false); 
		else return(warray.exist(p));
	}	
	public int size(){ 
		switch(type){
		case Value.ARRAY: return(warray.size());
		case Value.HASH: return(whash.size());
		default : return(0); 
		}
	}
	public Value get(int p){ 
		if(type!=Value.ARRAY) return(Value.ERROR_VALUE); 
		else return(warray.get(p));
	}
			//accesseur ARRAY rapide
	public String getString(int p){ 
		if(type!=Value.ARRAY) return(null); 
		else return(warray.getString(p));
	}
	public int getInt(int p){ 
		if(type!=Value.ARRAY) return(0); 
		else return(warray.getInt(p));
	}
	public boolean getBool(int p){ 
		if(type!=Value.ARRAY) return(false); 
		else return(warray.getBool(p));
	}
	public float getFloat(int p){ 
		if(type!=Value.ARRAY) return(0); 
		else return(warray.getFloat(p));
	}
	public byte[] getBytes(int p){ 
		if(type!=Value.ARRAY) return(null); 
		else return(warray.getBytes(p));
	}
		//]	
		//[si type==HASH	
	//public int size();
	public Value get(String p){ 
		if(type!=Value.HASH) return(Value.ERROR_VALUE); 
		else return(whash.get(p));
	}
	public Iterable<String> keys(){ 
		if(type!=Value.HASH) return(null); 
		else return(whash.keys());
	}
	public boolean exist(String p){ 
		if(type!=Value.HASH) return(false); 
		else return(whash.exist(p));
	}
			//accesseur HASH rapide
	public String getString(String p){ 
		if(type!=Value.HASH) return(null); 
		else return(whash.getString(p));
	}
	public int getInt(String p){ 
		if(type!=Value.HASH) return(0); 
		else return(whash.getInt(p));
	}
	public boolean getBool(String p){ 
		if(type!=Value.HASH) return(false); 
		else return(whash.getBool(p));
	}
	public float getFloat(String p){ 
		if(type!=Value.HASH) return(0); 
		else return(whash.getFloat(p));
	}
	public byte[] getBytes(String p){ 
		if(type!=Value.HASH) return(null); 
		else return(whash.getBytes(p));
	}
		//]
	//]

	public Value toConstant(){
		switch(type){
		case Value.BOOL: return(vbool);
		case Value.STRING: case Value.INT: case Value.FLOAT: return(watome.toConstant());
		case Value.HASH: return(whash.toConstant());
		case Value.ARRAY: return(warray.toConstant());
		case Value.BYTES: return(watome.toConstant());
		case Value.NULL : return(Value.NULL_VALUE);
		default : return(Value.ERROR_VALUE); 
		}		
	}
	//]
	
	//[writer
		//Ecriture de la valeur Atomique
	public ValueWriter set(String val){
		if(watome==null) watome=new WAtome();
		type=Value.STRING;
		watome.set(val);
		return(this);
	}
	public ValueWriter set(int val){
		if(watome==null) watome=new WAtome();
		type=Value.INT;
		watome.set(val);
		return(this);
	}
	public ValueWriter set(float val){
		if(watome==null) watome=new WAtome();
		type=Value.FLOAT;
		watome.set(val);
		return(this);
	}
	public ValueWriter set(byte[] val){
		if(watome==null) watome=new WAtome();
		type=Value.BYTES;
		watome.set(val);
		return(this);
	}
	public ValueWriter set(boolean val){
		type=Value.BOOL;
		if(val) vbool = Value.TRUE;
		else vbool = Value.FALSE;
		return(this);
	}

		//[Ecriture de la valeur ARRAY 
	public ValueWriter add(Value val){
		if(warray==null) warray=new WArray();
		else if(type!=Value.ARRAY) warray.removeAll();
		type=Value.ARRAY;
		warray.add(val);
		return(this);
	}
	public ValueWriter add(String val){
		if(warray==null) warray=new WArray();
		else if(type!=Value.ARRAY) warray.removeAll();
		type=Value.ARRAY;
		warray.add(val);
		return(this);
	}
	public ValueWriter add(int val){
		if(warray==null) warray=new WArray();
		else if(type!=Value.ARRAY) warray.removeAll();
		type=Value.ARRAY;
		warray.add(val);
		return(this);
	}
	public ValueWriter add(float val){
		if(warray==null) warray=new WArray();
		else if(type!=Value.ARRAY) warray.removeAll();
		type=Value.ARRAY;
		warray.add(val);
		return(this);
	}
	public ValueWriter add(byte[] val){
		if(warray==null) warray=new WArray();
		else if(type!=Value.ARRAY) warray.removeAll();
		type=Value.ARRAY;
		warray.add(val);
		return(this);
	}
	
	public ValueWriter add(boolean val){
		if(warray==null) warray=new WArray();
		else if(type!=Value.ARRAY) warray.removeAll();
		type=Value.ARRAY;
		warray.add(val);
		return(this);
	}
	
	public ValueWriter set(int pos, Value val){
		if(warray==null) warray=new WArray();
		else if(type!=Value.ARRAY) warray.removeAll();
		type=Value.ARRAY;
		warray.set(pos, val);
		return(this);
	}
	public ValueWriter set(int pos, String val){
		if(warray==null) warray=new WArray();
		else if(type!=Value.ARRAY) warray.removeAll();
		type=Value.ARRAY;
		warray.set(pos, val);
		return(this);
	}
	public ValueWriter set(int pos, int val){
		if(warray==null) warray=new WArray();
		else if(type!=Value.ARRAY) warray.removeAll();
		type=Value.ARRAY;
		warray.set(pos, val);
		return(this);
	}
	public ValueWriter set(int pos, float val){
		if(warray==null) warray=new WArray();
		else if(type!=Value.ARRAY) warray.removeAll();
		type=Value.ARRAY;
		warray.set(pos, val);
		return(this);
	}
	public ValueWriter set(int pos, byte[] val){
		if(warray==null) warray=new WArray();
		else if(type!=Value.ARRAY) warray.removeAll();
		type=Value.ARRAY;
		warray.set(pos, val);
		return(this);
	}
	public ValueWriter set(int pos, boolean val){
		if(warray==null) warray=new WArray();
		else if(type!=Value.ARRAY) warray.removeAll();
		type=Value.ARRAY;
		warray.set(pos, val);
		return(this);
	}
	
	public ValueWriter remove(int pos){
		if(type==Value.ARRAY) warray.remove(pos);
		return(this);
	}
	public ValueWriter removeAll(){
		if(type==Value.ARRAY) warray.removeAll();
		else if(type==Value.HASH) whash.removeAll();
		return(this);
	}
		//]	
		//[Ecriture de la valeur HASH	
	public ValueWriter set(String etq, Value val){
		if(whash==null) whash=new WHash();
		else if(type!=Value.HASH) whash.removeAll();
		type=Value.HASH;
		whash.set(etq, val);
		return(this);
	}
	public ValueWriter set(String etq, String val){
		if(whash==null) whash=new WHash();
		else if(type!=Value.HASH) whash.removeAll();
		type=Value.HASH;
		whash.set(etq, val);
		return(this);
	}
	public ValueWriter set(String etq, int val){
		if(whash==null) whash=new WHash();
		else if(type!=Value.HASH) whash.removeAll();
		type=Value.HASH;
		whash.set(etq, val);
		return(this);
	}
	public ValueWriter set(String etq, float val){
		if(whash==null) whash=new WHash();
		else if(type!=Value.HASH) whash.removeAll();
		type=Value.HASH;
		whash.set(etq, val);
		return(this);
	}
	public ValueWriter set(String etq, byte[] val){
		if(whash==null) whash=new WHash();
		else if(type!=Value.HASH) whash.removeAll();
		type=Value.HASH;
		whash.set(etq, val);
		return(this);
	}
	
	public ValueWriter set(String etq, boolean val){
		if(whash==null) whash=new WHash();
		else if(type!=Value.HASH) whash.removeAll();
		type=Value.HASH;
		whash.set(etq, val);
		return(this);
	}
	public ValueWriter remove(String etq){
		if(type==Value.HASH) whash.remove(etq);
		return(this);
	}
		//]
	//]

	public boolean equals(Value v) {
		switch(type){
		case Value.NULL: return(v==null || v.getType()==Value.NULL);
		case Value.BOOL: return(vbool.equals(v));
		case Value.INT: case Value.STRING: case Value.FLOAT: return(watome.equals(v));
		case Value.HASH: return(whash.equals(v));
		case Value.ARRAY: return(warray.equals(v));
		}
		return false;
	}
}