import java.io.*;
import java.lang.reflect.Method;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.*;
import java.util.jar.*;
import java.util.zip.*;

public class start extends URLClassLoader { //OXO: extends URLClassLoader --> pour passer à public la méthode addURL ! (pour éviter les messages d'alerte de Java 9)
	
	public final static String version = "1.03"; //version du lanceur
	
	//[Description de la librairie maitre
	private static Class<?> MASTER_CLASS 	= null;
	private static String MASTER_NAME		= null;
	private static String MASTER_CLASS_NAME = null;
	//]
		
	public final static void main(String[] args) {		
		if(args.length==0) {
			System.out.println("Erreur : vous devez spécifier le nom du programme à exécuter !\n"
					+ "Usage :   >java start NOM_PROG [ARGS1] ... [ARGSN]");			
			return;
		}

		Properties config=new Properties();
		
		//Chargement des libriairies du dossier ./lib et de leurs configurations
		start_loader.loadLib(new File("./lib"), config, true, true);

        //[Ajout du dossier bin dans le path (s'il existe)
		File bin_dir = new File("./bin");
		if(bin_dir.exists() && bin_dir.isDirectory()) {  
			//Chargement du fichier de configuration du dossier bin
			Properties local_config=null;
			File config_bin = new File("./bin/config.ini");
			if(config_bin.exists()) try {
				local_config=new Properties();
				InputStream bin_config = new FileInputStream(config_bin);
				local_config.load(bin_config);
				bin_config.close();
			}catch(Exception e) {
				System.out.println("Erreur : impossible de lire le fichier de configuration situé dans le dossier ./bin");
			}
			if(start_loader.loadConfig(config, bin_dir, local_config)) 	start_loader.processLoadLib(bin_dir);
		}
		bin_dir=null;
		//]
        		
		start_loader.loadDeffered(config); //On applique les chargements différés
        
		//[Chargement de la classe principale
        if(MASTER_NAME==null) {
			System.out.println("Erreur : impossible d'identifier la librairie principale, vérifiez que son archive est bien dans le dossier lib !");
			return;        	
        }
		if(MASTER_CLASS==null) try{ 	
			//MASTER_CLASS = Class.forName(MASTER_CLASS_NAME);
			MASTER_CLASS = start_loader.loadClass(MASTER_CLASS_NAME);
		}catch(Exception e) {
			System.out.println("Erreur : impossible de charger la librairie principale "+MASTER_NAME+", vérifiez que l'archive correspondante est bien dans le dossier lib !");
			return;
		}
		//]

		//[Chargement du fichier de configuration de la libriairie principale
		try {
			InputStream master_config = MASTER_CLASS.getResourceAsStream("config.ini");
			config.load(master_config);
			master_config.close();
		}catch(Exception e) {
			System.out.println("** warning : impossible de lire le fichier de configuration de la librairie principale **");
		}
		//]
		
		//Chargement du fichier de configuration local
		File file = new File("config.ini");  
		if(file.exists()) {
			try {
				FileInputStream in = new FileInputStream(file);
				config.load(in);
				in.close();
			}catch(Exception e) {
				System.out.println("Erreur : impossible de lire le fichier de configuration local (config.ini)");
			}
		}
		file=null;
		//]		
		
		//[Analyse de la commande de lancement
		String prog = args[0];
		HashMap<String, Vector<String>> aliases;
		switch(prog) {
		case "-h" : case "-help":
			System.out.println(MASTER_NAME+", Lanceur v"+version+"\n\n"
					+ "- Pour lancer un programme :   >java start NOM_PROG [ARGS1] ... [ARGSN] \n"
					+ "   Avec NOM_PROG le nom de la classe à exécuter ou le nom de l'alias à exécuter (voir ci-dessous).\n\n"
					+ "- Pour gérer l'installation : >java start OPTION \n"
					+ "   Valeur OPTION possible (non cumulable):\n"
					+ "\t-h\t: Affiche l'aide et le numéro de version de ce lanceur.\n"
					+ "\t-v\t: Affiche le numéro de version de la librairie "+MASTER_NAME+".\n"
					+ "\t-l\t: Affiche la liste des alias de lancement configurés.\n"
					+ "\t-c\t: Affiche la liste des catégories d'alias configurées.\n"
					+ "\t-lc CAT\t: Affiche la liste des alias de la catégorie de nom CAT.\n"
					+ "\t-m\t: Affiche la liste des modules détectés et leur numéro de version.\n"
					+ "\t-i MODULE_ZIP\t: Procède à l'installation du module porté par l'archive zip indiquée.\n"
					+ "\t-ir MODULE_ZIP\t: idem que -i mais supprime l'archive zip à la fin de l'installation.\n"
					+ "\t-irt MODULE_ZIP\t: idem que -ir mais supprime aussi le fichier readme du module installé.\n"
					+ "\t-clean\t: supprime les librairies obsolètes présentes dans le dossier lib.\n\n"
					+ "- Pour lancer un programme qui utilise beaucoup de mémoire :\n"
					+ "   Il faur utiliser les options :\n"
					+ "\t-Xms[SIZE] : pour spécifier la taille disponible au lancement du programme.\n"
					+ "\t-Xmx[SIZE] : pour spécifier la taille miximum possible durant l'exécution.\n"
					+ " --> dans l'exemple ci-dessous on va spécifier 256Mo au lancement, et 2Go maximum possible :\n"
					+ "  >java -Xms256m -Xmx2g start NOM_PROG [ARGS1] ... [ARGSN] \n"
					+ "\n"
					+ "NB: l'ensemble des librairies placées dans le dossier lib seront autimatiquement chargées."
					);
			return;
		case "-v" : case "-version" : //affichage du numéro de version de la librairie maitre
			System.out.println(MASTER_NAME+" version "+masterVersion());
			return;
		case "-l" : case "-list":  //affichage de la liste des alias configuées
			aliases = extractAliases(config);
			System.out.println("Liste des alias configurés :");
			for(String cat : aliases.keySet()) {
				Vector<String> lcat = aliases.get(cat);
				System.out.println("\n ****** "+cat+" ******");
				for(String alias : lcat) {
					String label = config.getProperty(alias+"@label", "-");
					System.out.println("   "+alias+"\t: "+label);
				}
			}
			return;
		case "-c" : case "-cat":  //affichage de la liste des catégories configuées
			aliases = extractAliases(config);
			System.out.println("Liste des catégories d'alias configurées :\n");
			for(String cat : aliases.keySet()) {
				System.out.println("\t"+cat);
			}
			return;
		case "-m" : case "-module":  //affichage de la liste des modules détectés
			System.out.println("Liste des modules détectés :");
			HashMap<String, Object[]> mod_list = start_loader.getModules();
	        for(String mod_name : mod_list.keySet()) {
	        	Object[] defl = mod_list.get(mod_name); // 0-> version, 1-> cfg, 2-> File
	        	String err = start_loader.checkModule(mod_name, false);
	        	System.out.println("   "+mod_name+"\t: v"+(String) defl[0]+ (err!=null ? "\t/!\\ "+err : ""));
	        }
			return;			
		case "-lc" : case "-lcat":  //affichage de la liste des alias de la catégorie indiquée
			if(args.length<2) {
				System.out.println("Erreur : vous devez indiquer le nom de la catégorie à afficher");
				return;
			}
			String cat = args[1];
			for(int i = 2; i<args.length; i++) cat+=" "+args[i];
			
			aliases = extractAliases(config);		
			System.out.println("Liste des alias de la catégorie : "+cat+"\n");
			Vector<String> lcat = aliases.get(cat);
			if(lcat==null) System.out.println("\t- catégorie inexistante-\n");
			else for(String alias : lcat) {
				String label = config.getProperty(alias+"@label", "-");
				System.out.println("   "+alias+"\t: "+label);
			}
			return;	
		case "-i": case "-ir": case "-irt":
			if(args.length<2) {
				System.out.println("Erreur : vous devez indiquer le nom de l'archive zip contenant le module à installer");
				return;				
			}
			File file_zip = new File(args[1]);
			if(!file_zip.exists()) {
				System.out.println("Erreur : l'archive zip indiqué est introuvable ("+args[1]+")");
				return;								
			}
			if(unzip(file_zip, ".", prog.equals("-irt"))) {
				if(prog.startsWith("-ir")) {
					System.out.println("- suppression de l'archive zip : "+args[1]+" -");
					file_zip.delete();
				}				
			};
			return;
		case "-clean":
			System.out.println("Purge des libairies obsolètes...");
			for(File obs_file : start_loader.getObsoletes()) {
				if(obs_file.isFile()) {
					System.out.println("- suppression du fichier librairie : "+obs_file.getName()+" -");
					obs_file.delete();
				}
			}
			System.out.println("... purge terminée.");
			return;
		}		
		//On regarde si le programme indiqué est un alias de configuration
		String alias = config.getProperty(prog, null);
		if(alias!=null) {
			//on regarde si les prérequis du module associé sont validé
			String mod = config.getProperty(prog+"@module", null);
			if(mod!=null && start_loader.checkModule(mod, true)!=null) {
				System.out.println(" --> impossible de lancer le programme "+prog+" ("+alias+")");
				return; //on ignore l'exécution
			}
			prog = alias;
		}
		//]

		//[Nétoyage des données du lanceur
		MASTER_CLASS = null; MASTER_NAME = null; MASTER_CLASS_NAME = null;
		config=null;
		start_loader.clear();
		//]
		
		//[Extraction des arguments du programme et lancement du programme
		String[] args2 = new String[args.length-1];
		for(int i=0; i<args2.length; i++) args2[i]=args[i+1];	
		args=null;
		try {
			 //Method meth = Class.forName(prog).getMethod("main", String[].class);
			 Method meth = start_loader.loadClass(prog).getMethod("main", String[].class);
			 meth.invoke(null, (Object) args2);
		}catch(Exception e) {
			System.out.println("Erreur durant le lancement de la classe "+prog);
			e.printStackTrace();
		}
		//]
	}
	
	//[Gestion du fichier de configuration
	private static HashMap<String, Vector<String>> extractAliases(Properties config){
		HashMap<String, Vector<String>> aliases = new HashMap<String, Vector<String>>();
		Enumeration<?> lprop = config.propertyNames();
		while (lprop.hasMoreElements()) {
			String key = (String) lprop.nextElement();
			if(key.indexOf('@')==-1) { //c'est bien une clé alias
				String cat = config.getProperty(key+"@cat", "---");
				
				Vector<String> lcat = aliases.get(cat);
				if(lcat==null) {
					lcat = new Vector<String>();
					aliases.put(cat, lcat);
				}
				lcat.add(key);
			}
		}				
		return(aliases);
	}
	//]
	
	private static String masterVersion() {
		String version="???";
		try {
			version=MASTER_CLASS.getField("VERSION").get(null).toString();
		}catch(Exception e) {}
		return(version);
	}
	private static String relativePath(File file) {
		String path=file.getAbsolutePath();
		int dot = path.indexOf("\\.\\");
		if(dot!=-1) return(path.substring(dot+1));
		else return(path);
	}
	
	//[Instance de gestion du chargement dynamique des libriaires
	private static start start_loader=new start();
	//[Methode d'instance
	URL[] allready_loaded;
	HashMap<String, Object[]> modules; //desc[] =  0-> version (string), 1-> cfg (Properties), 2-> fichier (File), 3->valid (Boolean.TRUE si ok/ String si erreur check/null (si non checked)
	Vector<File> obsoletes;
	protected start() {
		super(new URL[] {});
		modules = new HashMap<String, Object[]>();
		obsoletes = new Vector<File>();
		
		ClassLoader sysClassLoader = ClassLoader.getSystemClassLoader();
		if(sysClassLoader instanceof java.net.URLClassLoader) allready_loaded = ((java.net.URLClassLoader) sysClassLoader).getURLs(); //java < 9
		else  allready_loaded = null; //java >= 9
	}
	private void clear() { modules=null; obsoletes=null; allready_loaded=null; } 
	private HashMap<String, Object[]> getModules(){ return(modules); }
	private Vector<File> getObsoletes(){ return(obsoletes); }
	private void loadLib(File file, Properties cfg, boolean first_call, boolean recurse) { //charge la librairie indiqué, 
		//si path est un dossier toutes les librairies présente dans ce dossier seront chargé, et si recurse==true ce traitement sera récursif
        if(file.isFile()) internalLoadLib(file, cfg);
        else if(file.isDirectory() && (first_call || recurse)) {
        	for(File ssfile : file.listFiles()) loadLib(ssfile, cfg, false, recurse);	
        }        
        
	}
	private void loadDeffered(Properties config) { //On applique les chargements différés
		for(String mod_name : modules.keySet()) {
			Object[] defl = modules.get(mod_name); // 0-> version, 1-> cfg, 2-> File, 3->valid
			mergeConfig(config, (Properties) defl[1], mod_name);
			processLoadLib((File) defl[2]);
		}
	}
	private void internalLoadLib(File file, Properties cfg) { //charge la librairie indiqué
		Properties local_config=null;
		if(file.getName().toLowerCase().endsWith(".jar")) { //on va charger le fichier de configuration éventuellement présent dans cette archive
			try {
				JarFile fj = new JarFile(file);
				JarEntry config_entry = fj.getJarEntry("config.ini");
				if(config_entry!=null) {
					local_config=new Properties();
					local_config.load(fj.getInputStream(config_entry));
				}
				fj.close();
			} catch (Exception e){ e.printStackTrace(); } 
		}
		if(loadConfig(cfg, file, local_config)) processLoadLib(file);
	}
	private boolean loadConfig(Properties global_config, File libfile, Properties local_config){ //charge la configuration indiqué, retourne true si le module associé peut être chargé immédiatement
		if(local_config==null) return(true);
		
		String mod_name = local_config.getProperty("@module", null);
		if(mod_name!=null) { //c'est un module identifiable
			String version = local_config.getProperty("@version", null);
			String masterc = local_config.getProperty("@master", null);
			
			String max_jv = local_config.getProperty("@max_jv", null);
			if(max_jv!=null && isUpJV(max_jv)) System.out.println("** warning : le module "+mod_name+" n'est pas compatible avec la version active de java (v"+getJavaVersion()+"), vous devez utiliser une version <="+max_jv);
			
			if(masterc!=null) { //c'est la librairie maitre
				if(MASTER_NAME!=null && !MASTER_NAME.equals(mod_name)) { //on a déjà une librairie maître...
					System.out.println("** Erreur : 2 librairies maîtres ont été trouvées : "+MASTER_NAME+" et "+mod_name+" , vous devez en supprimer une.*");
					System.exit(0);
					return(false);
				}
				if(MASTER_NAME==null && MASTER_CLASS==null) { //on s'assure que la librairie n'est pas déjà chargée
					try{ MASTER_CLASS = loadClass(masterc); }
					catch(Exception e) { }
				}
				MASTER_NAME=mod_name;
				MASTER_CLASS_NAME=masterc;
				if(MASTER_CLASS!=null) {
					System.out.println("** warning : le chargement du fichier "+relativePath(libfile)+" a été ignoré car la librairie "+mod_name+" (v"+masterVersion()+") est déjà chargée**");
					return(false);
				}
			}	
			Object[] defl = modules.get(mod_name); // 0-> version, 1-> cfg, 2-> File,  3->valid

			if(defl!=null) {
				if(compareVersion((String) defl[0], version)) { //un module de même nom existe déjà, et possède un plus fort numéro de version
					obsoletes.add(libfile);
					System.out.println("** warning : le chargement du fichier "+relativePath(libfile)+" a été ignoré car un module de même nom plus récent ("+relativePath((File) defl[2])+" --> v"+(String) defl[0]+") est aussi présent**");
					return(false);
				} else {
					obsoletes.add((File) defl[2]);
					System.out.println("** warning : le chargement du fichier "+relativePath((File) defl[2])+" a été ignoré car un module de même nom plus récent ("+relativePath(libfile)+" --> v"+version+") est aussi présent**");
				}
			}
			
			//System.out.println("On retient le module "+mod_name+" dans sa version "+version);					
			modules.put(mod_name, new Object[] {version, local_config, libfile, null});
			return(false);
		}
		//Ce n'est pas un module identifiable, on merge sa configuration et on continue
		mergeConfig(global_config, local_config, null);
		return(true);
	}
	
	private void processLoadLib(File file) {
		try {
			java.net.URL url = file.toURI().toURL();
			if(allready_loaded!=null) {
				for(java.net.URL it : allready_loaded){ //on verrifie que cette librairie n'est pas déjà chargée 
					if (it.equals(url)) return;
				}
			}
			//System.out.println("chargement de la librairie : "+url);
			addURL(url);
		} catch (Exception e){ e.printStackTrace(); }
	}
	//]
		
	//[functions utiles
	private static float java_version=-1;
	private static float getJavaVersion() { //retourne la version de java sous la forme d'un nombre décimal X.Y
		if(java_version==-1) {
			String version = System.getProperty("java.version");
			java_version=Float.parseFloat(version.substring(0,3));
		}
		return(java_version);
	}
	private static boolean isUpJV(String max_jv){ //true ssi la version active de java est > à max_jv
		float active_v = getJavaVersion() ;
		float fmax_jv = Float.parseFloat(max_jv.substring(0,3));
		return(active_v>fmax_jv);
	}
	
	private static void mergeConfig(Properties global, Properties local, String mod_name) {
		if(mod_name!=null) for(Object key : local.keySet()) {
			String st_key = (String) key;
			switch(st_key.indexOf('@')) {
			case 0 : break; //on ignore cette clé
			case -1 :
				String key_mod=st_key+"@module";
				String mod = local.getProperty(key_mod, null);
				if(mod==null) global.put(key_mod, mod_name);
				//break; pas de break volontaire
			default :
				global.put(st_key, local.getProperty(st_key));
			}
		}
	}
	private String checkModule(String mod_name, boolean trace) { //vérifie (récursivement) si les prérequis du module sont validé, null si ok ou !=null message d'erreur
		Object[] defl = modules.get(mod_name);  //0-> version (string), 1-> cfg (Properties), 2-> fichier (File), 3->valid (Boolean.TRUE si ok/ String si erreur check/null (si non checked)/Boolean.FALSE si check en cours)
		return(checkModule(mod_name, defl, trace));
	}
	private String checkModule(String mod_name, Object[] defl, boolean trace) { //vérifie (récursivement) si les prérequis du module sont validé, null si ok ou !=null message d'erreur
		if(defl!=null) {
			if(defl[3]==Boolean.TRUE || defl[3]==Boolean.FALSE) return(null); //tout est ok pour ce module, ou son analyse est déjà en cours
			else if(defl[3]!=null && !trace) return((String) defl[3]); //analyse déjà réalisé et il y a un problème
			else {
				defl[3]=Boolean.FALSE; //on doit analyser ce module
				Properties mod_config=(Properties) defl[1];
				
				//version max pour Java
				String max_jv = mod_config.getProperty("@max_jv", null);
				if(max_jv!=null && isUpJV(max_jv)) {
					defl[3]="non compatible avec la version active de java (v"+getJavaVersion()+"), version a utiliser <="+max_jv;
					if(!trace) return((String) defl[3]);
					else System.out.println("* Erreur * : le module "+mod_name+" n'est pas compatible avec la version active de java (v"+getJavaVersion()+"), vous devez utiliser une version <="+max_jv);
				}
	
				//dépendance
				String depends = mod_config.getProperty("@depends", null); //depends= mod1, mod2, ..., modN vXX.XX, ... 
				if(depends!=null) {
					String[] dmods = depends.split(",");
					for(String dm : dmods) {
						String version=null;
						dm=dm.trim();
						int ver_pos=dm.indexOf(" v"); int dec=2;
						if(ver_pos==-1) { ver_pos=dm.indexOf(' '); dec=1;}
						if(ver_pos!=-1) { //un numéro de version est spécifié
							version=dm.substring(ver_pos+dec).trim();
							dm=dm.substring(0, ver_pos).trim();
						}
//System.out.println("mod_name = "+mod_name+" depends = ["+dm+"]");						
						Object[] desc_dm = modules.get(dm);
						if(desc_dm==null) {
							if(defl[3]==Boolean.FALSE) defl[3]="nécessite le module "+dm+" qui n'est pas présent";
							if(!trace) return((String) defl[3]);
							else System.out.println("* Erreur * : le module "+mod_name+" nécessite la présence du module "+dm);							
						} else {
//System.out.println("mod_name = "+mod_name+" depends = "+dm+" v"+version+" present v"+(String) desc_dm[0]+" test ="+compareVersion(version, (String) desc_dm[0]));							
							if(version!=null) { //cohérence de version
								if(compareVersion(version, (String) desc_dm[0])) {
									if(defl[3]==Boolean.FALSE) defl[3]="nécessite le module "+dm+" de version >="+version+", mais seule la version "+((String) desc_dm[0])+" est présente";
									if(!trace) return((String) defl[3]);
									else System.out.println("* Erreur * : le module "+mod_name+" nécessite le module "+dm+" dans sa version "+version+", vous devez mettre "+dm+" à jour (version présente: "+((String) desc_dm[0])+")");										
								}
							}
							if(trace) checkModule(dm, desc_dm, true);
						}
					}
				}
				
				//tout est ok
				if(defl[3]==Boolean.FALSE) {
					defl[3]=Boolean.TRUE;
					return(null);
				} else return((String) defl[3]);
			}
		}
		//le module demandé n'existe pas
		String err = "impossible de trouver le module "+mod_name; 
		if(trace) System.out.println("* Erreur * : "+err);
		return(err); 
	}
	private static boolean compareVersion(String va, String vb) { //true ssi la version va est strictement supérieure à la version vb
//System.out.println("compare va = "+va+" // vb= "+vb);		
		if(va==null) return(false);
		if(vb==null) return(true);
		//Un numéro de version est de la forme "[a|b]XXX[a|b].[a|b]ZZZ[a|b].[a|b]YYY[a|b]....", (NB: le symbole '_' peut être utilisé à la place du point)
		//les versions alpha (a) sont plus faible que les béta (b) qui sont plus faible que les finales (sans lettre)
		  
		va = va.replace('_', '.');
		vb = vb.replace('_', '.');
		
		String[] tab_va = va.split("\\.");
		String[] tab_vb = vb.split("\\.");
//System.out.println("tab_va.length = "+tab_va.length+" // tab_vb.length= "+tab_vb.length);	  
		int i=0, numA, numB;
		boolean is_betaA,is_betaB, is_alphaA, is_alphaB;
		while(true){
			is_betaA=false; is_betaB=false; is_alphaA=false; is_alphaB=false;
			if(i>=tab_va.length) return(false);
			if(i>=tab_vb.length) return(true);
			  
			String tokenA = tab_va[i];
//System.out.println("tokenA("+i+") = ["+tokenA+"]");			
			if(tokenA.startsWith("a")){ is_alphaA=true; tokenA=tokenA.substring(1);}
			else if(tokenA.startsWith("b")){ is_betaA=true; tokenA=tokenA.substring(1);}
			else if(tokenA.endsWith("a")){ is_alphaA=true; tokenA=tokenA.substring(0,tokenA.length()-1);}
			else if(tokenA.endsWith("b")){ is_betaA=true; tokenA=tokenA.substring(0,tokenA.length()-1);}
			numA = Integer.parseInt(tokenA);  

			String tokenB = tab_vb[i];
			if(tokenB.startsWith("a")){ is_alphaB=true; tokenB=tokenB.substring(1);}
			else if(tokenB.startsWith("b")){ is_betaB=true; tokenB=tokenB.substring(1);}
			else if(tokenB.endsWith("a")){ is_alphaB=true; tokenB=tokenB.substring(0,tokenB.length()-1);}
			else if(tokenB.endsWith("b")){ is_betaB=true; tokenB=tokenB.substring(0,tokenB.length()-1);}
			numB = Integer.parseInt(tokenB);  
//System.out.println("numA = "+numA+" // numB= "+numB);
			if(numA>numB) return(true);
			if(numB>numA) return(false);

			//donc ici: numA == numB
			  
			//gestion des versions alpha et beta
			if(is_alphaB && !is_alphaA) return(true);
			if(is_alphaA && !is_alphaB) return(false);
			if(is_betaB && !is_betaA) return(true);
			if(is_betaA && !is_betaB) return(false);

			i++;
		}	
	}
	private static boolean unzip(File zipFilePath, String destDir, boolean remove_readme) {
		System.out.println("Installation du composant : "+zipFilePath.getName());
        File dir = new File(destDir);
        // create output directory if it doesn't exist
        if(!dir.exists()) dir.mkdirs();
        FileInputStream fis;
        File readme_file=null;
        //buffer for read and write data to file
        byte[] buffer = new byte[1024];
        try {
            fis = new FileInputStream(zipFilePath);
            ZipInputStream zis = new ZipInputStream(fis);
            ZipEntry ze = zis.getNextEntry();
            while(ze != null){
                String fileName = ze.getName();
                File newFile = new File(destDir + File.separator + fileName);
                //System.out.println("Unzipping to "+newFile.getAbsolutePath());
                //create directories for sub directories in zip
                if(ze.isDirectory()) newFile.mkdirs();
                else {
                	new File(newFile.getParent()).mkdirs(); 
                	FileOutputStream fos = new FileOutputStream(newFile);
                	int len;
                	while ((len = zis.read(buffer)) > 0) { fos.write(buffer, 0, len); }
                	fos.close();
                	
                	if(readme_file==null && fileName.startsWith("readme") && fileName.endsWith(".txt")) readme_file=newFile;
                }
                zis.closeEntry();
                ze = zis.getNextEntry();
            }
            zis.closeEntry();
            zis.close();
            fis.close();
        } catch (Exception e) {
            e.printStackTrace();
            return(false);
        }
        String info_install=null;
        if(readme_file!=null) {
        	try {
        		BufferedReader breader= new  BufferedReader(new  InputStreamReader(new FileInputStream(readme_file), "UTF-8"));
        		info_install = breader.readLine();
        		breader.close();
        		if(info_install.charAt(0)>512) info_install = info_install.substring(1); //Hack pour supprimer un caractère mystérieux situé en début de fichier...
 
        		if(remove_readme) readme_file.delete();
        	}catch(Exception e) {}
        }
        System.out.println("... installation terminée "+(info_install == null ? "(aucune information disponible sur l'élément installé)" : ": "+info_install));
        return(true);
    }
	//]
}