/** 
   Programme de test : AgentHelloBye
 
   Ce programme permet de tester le lancement, la mise en pause et la reprise de l'exécution d'un agent. 

   Pour plus de détails lancer ce programme avec l'option -h
*/
package test.agentu;

import skuad.hubudp.Hub;
import skuad.hubudp.HubUdp;
import skuad.ubiquity.space.Avatar;
import skuad.ubiquity.space.Participant;
import skuad.ubiquity.space.ParticipantListener;
import skuad.ubiquity.space.alias.AliasListener;
import skuad.ubiquity.space.mailbox.MailBox;
import skuad.ubiquity.space.mailbox.Message;
import skuad.uda.Device;
import skuad.util.base.LaunchOption;
import skuad.util.time.Pulse;
import skuad.util.value.Value;
import skuad.agentu.*;
import skuad.agentu.util.*;

public class AgentHelloBye extends AgentUAdapter  implements AliasListener  {
	
	public final static String physical_specif = ""; 		//aucun slot physique
	public final static String social_specif = "space : world;"; 		//aucun slot social
	
	String space_id=null;
	public AgentHelloBye(){}

	//Initialisation de l'agent
	public void start(AgentUDescriptor aud, int ctx){	//invoqué au démarrage de l'agent
		//System.out.println("Hello ! I am "+aud.getAvatar("space").participant().localId());
	}

	public void notifyPlugAvatar(AgentUDescriptor aud, String name, Avatar avatar, int contexte){
		space_id = avatar.participant().localId();
		System.out.println("Hello ! I am "+space_id);
		
		Value config = aud.getConfig();
		if(config!=null && config.s().equals("master")) { //c'est l'agent maitre
			avatar.alias().setAlias("OPAL RT");
			avatar.participant().getSpace().addAliasListener(this);
		} else {
			aud.loopPulse("sendToMaster", randInt(500,3000), randInt(2000,4000), false);
			avatar.alias().setAlias("alias_"+space_id);
		}
	}		
	
	public void stop(AgentUDescriptor aud, int ctx){	//invoqué au démarrage de l'agent
		System.out.println("I was "+space_id+", Bye !");
	}
	
	public void behavior(AgentUDescriptor aud, Pulse pulse){
		if(pulse.is("sendToMaster")) {
			aud.getAvatar("space").mailBox().send("#OPAL RT", "test", Value.s("{  \"cs\": \"1#005#001\",  \"ds\": \"53.5774258456613\",  \"id_src\": \"102\",  \"alias\": \"WaterHeater_102\",  \"id_order\": \"1\",  \"resCode\": \"0\",  \"resDesc\": \"HW Temp returned\"}"));
		}
	}
	
	
	public void  event(AgentUDescriptor aud, int code, int ctx) {
		switch(code) {
		case AgentU.PAUSED:
			System.out.println("I am "+aud.getName()+" and i go sleep, Good night !");
			break;
		case AgentU.RESUMED:
			System.out.println("Good morning ! I am "+aud.getName()+" and i wake up.");
			break;		
		}
	}
	
	public static int randInt(int min, int max) {
		return((int) (Math.random()*(max-min+1)+min));
	}
	
	public void receive(AgentUDescriptor aud, Avatar avatar, MailBox box, Message mess){
		System.out.println("["+space_id+"] message recu de "+mess.sender().localId()+" contenu : {"+mess.value().toString()+"}");
	}
	
	//[Programme test
	public static void main(String[] hub_config){
		LaunchOption options = new LaunchOption(hub_config);
		if(options.existField("h")) { //affichage des options de lancement
			String tab="    ";
			System.out.println(
						" Programme AgentHelloBye : permet de tester le lancement, la mise en pause et la reprise de l'exécution d'un agent.\n"
					+ 	"  Comportement du programme : \n"
					+tab+ "- ce programme se connecte sur le hub spécifié dans les arguments de lancement.\n"
					+tab+ "- un agent de type AgentHelloBye est créé et son exécution est lancée.\n"
					+tab+ "- le comportement d'un agent de type AgentHelloBye est le suivant :\n"
					+tab+ "  -> dès son lancement il écrit le message \"Hello, I am XXX.\", avec XXX le nom de l'agent.\n"
					+tab+ "  -> quand il est mis en pause il écrit \"I am XXX and i go sleep, Good night !\".\n"
					+tab+ "  -> quand son exécution est relancé il écrit \"Good morning ! I am XXX and i wake up.\".\n"
					+tab+ "  -> à l'arrêt de son exécution il est écrit \"I was XXX, Bye !\".\n"
						+   "\n"
					+ 	" Arguments de lancement possibles (tous facultatifs, sauf mention contraire):\n"
					+tab+ "-trace : active le mode de traçage des logs et des changements d'état de l'agent\n"						
					+tab+ "-m : crées 60 agent de type AgentHelloBye\n"
					+tab+ "-mp : crées 60 agent de type AgentHelloBye et fait une pause de 300ms entre chaque création.\n"
			);
			System.out.println(HubUdp.hubOptions(tab, true));			
		} else {
			System.out.println("[AgentHelloBye] programme lancé");	
			Hub hub = HubUdp.joinHub(options, true);
			ServerAU servau = ServerAU.newServer(hub);
			
			if(options.existField("opal")) {
				int agent_id = servau.createAgent(AgentHelloBye.class, null, Value.writer().set("master").toConstant());
				if(options.existField("trace")) new AgentLogger(servau, AgentLogger.LOG | AgentLogger.STATE).observeAgent(agent_id);
				servau.runAgent(agent_id); //on lance l'exécution de l'agent
			}
			
			if(options.existField("m") || options.existField("mp")) {
				for(int i=0; i<60; i++) {
					int id= servau.createAgent(AgentHelloBye.class, null);
					servau.runAgent(id);
					if(options.existField("mp")) try { Thread.sleep(300); } catch(Exception e) {}
				}
			}
		}
	}
	//]

	@Override
	public void aliasAdded(Participant identity, String newAlias, int priority) {
		// TODO Auto-generated method stub	
	}

	@Override
	public void aliasRemoved(Participant identity, String lostAlias, int oldPriority) {
		// TODO Auto-generated method stub
		System.out.println("* Retrait de l'alias : " 
				+"participant_id = " + identity.localId()
				+" // host_global_id = "+identity.spot().getIdGlobal()+ " alias = {"+lostAlias+"}"
		);			
	}

	@Override
	public void aliasPriorityChange(Participant identity, String alias, int oldPriority, int newPriority) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void aliasFocusChange(Participant identity, String alias, boolean oldFocus, boolean newFocus) {
		// TODO Auto-generated method stub
		
	}
}