/** 
   Programme de test : AgentOnOff
 
   Ce programme permet de tester le bon fonctionnement de l'interaction d'un agent avec les devices pluggué sur ses slots physique. 
   Il s'utilise en particulier avec le programme test.uda.DLightUDP.
   
   Pour plus de détails lancer ce programme avec l'option -h
*/
package test.agentu;

import skuad.hubudp.Hub;
import skuad.hubudp.HubUdp;
import skuad.util.base.LaunchOption;
import skuad.agentu.*;
import skuad.uda.*;
import skuad.agentu.util.*;
import skuad.util.time.Pulse;

public class AgentLightBlink extends AgentUAdapter {
	
	public final static String physical_specif = "lumiere : LIGHT";
	public final static String social_specif = ""; 		//aucun slot social
		
	Device laLumiere;
	boolean light_state;
	public AgentLightBlink(){
		laLumiere=null;
		light_state=false;
	}

	//Initialisation de l'agent
	public void start(AgentUDescriptor aud, int ctx){	//invoqué au démarrage de l'agent
		aud.log("started");
		aud.loopPulse("blink light", 3000,  3000, true); //Définition du comportement temporel de l'agent
	}

	public void behavior(AgentUDescriptor aud, Pulse pulse){ //exécute le comportement de l'agent suite à l'activation d'un pulse, pulse = le descripteur de la pulsation qui a initié cet appel		
		if(laLumiere!=null && pulse.is("blink light")){
			light_state = !light_state;
			aud.log("blink -> "+light_state);
			laLumiere.action(0, (light_state ? 1 : 0));
		}
	}
	
	//[Capture des changements de plugs
	public void notifyPlugDevice(AgentUDescriptor aud, String slot, Device device){
		aud.log("nouveau plug sur le slot "+slot);
		laLumiere=device;		
	}
	public void notifyUnplugDevice(AgentUDescriptor aud, String slot){
		aud.log("suppresion du plug sur le slot "+slot);
		laLumiere=null;		
	}
	public void notifyChangeDevice(AgentUDescriptor aud, String slot, Device device){
		aud.log("changement de plug sur le slot "+slot);
		laLumiere=device;		
	}	
	//]

	//[Programme test
	public static void main(String[] hub_config){
		LaunchOption options = new LaunchOption(hub_config);
		if(options.existField("h")) { //affichage des options de lancement
			String tab="    ";
			System.out.println(
						" Programme AgentLightBlink : l'interaction d'un agent avec les devices pluggués sur ses slots physiques.\n"
					+ 	"  NB: ce programme s'utilise avec un device de type LIGHT sur le réseau (ce peut être par exemple\n"
					+	"	   le programme tools.uda.DLightUDP).\n"
					+ 	"  Comportement du programme : \n"
					+tab+ "- ce programme se connecte sur le hub spécifié dans les arguments de lancement.\n"
					+tab+ "- un agent de type AgentLightBlink est créé et son exécution est lancée.\n"
					+tab+ "- une règles de plug physique lui est affecté à l'agent afin de cabler automatiquement un device\n"
					+tab+ "- de type LIGHT sur son slot physique \"lumière\".\n"
					+tab+ "- le comportement d'un agent de type AgentLightBlink est le suivant :\n"
					+tab+ "  -> dès son lancement un pulse temporel périodique (3 secondes) est définie.\n"
					+tab+ "  -> à chaque déclenchement de ce pulse, si le slot \"lumière\" est pluggué une action de changement\n"
					+tab+ "     de valeur (on/off) est envoyée au device correspondant.\n"
						+   "\n"
					+ 	" Arguments de lancement possibles (tous facultatifs, sauf mention contraire):\n"
					+tab+ "-trace : active le mode de traçage des logs et des changements d'état de l'agent"						
			);
			System.out.println(HubUdp.hubOptions(tab, true));			
		} else {
			System.out.println("[AgentLightBlink] programme lancé");

			Hub hub = HubUdp.joinHub(options, true);
			ServerAU servau = ServerAU.newServer(hub);
			int agent_id = servau.createAgent(AgentLightBlink.class, null);

			if(options.existField("trace")) new AgentLogger(servau, AgentLogger.LOG | AgentLogger.STATE).observeAgent(agent_id);
			servau.runAgent(agent_id); //on lance l'exécution de l'agent
		
			//on va plugger automatiquement un device lumière présent sur le réseau
			servau.addPhysicalPlugRule(agent_id, "lumiere", "{ device : [ TYPE , LIGHT ] }");	
		}
	}
	//]
}