/** 
   Programme de test : AgentOnOff
 
   Ce programme permet de tester le bon fonctionnement des plugs des slots physiques des agents. 
   Il s'utilise en particulier avec les programmes test.uda.DButtonUDP et test.uda.DLightUDP.
   
    Pour plus de détails lancer ce programme avec l'option -h
*/
package test.agentu;

import skuad.agentu.*;
import skuad.uda.*;
import skuad.hubudp.*;
import skuad.util.databytes.*;
import skuad.util.base.LaunchOption;
import skuad.agentu.util.*;

public class AgentOnOff extends AgentUAdapter {
		
	public final static String physical_specif = 
				"bouton 	: 	BUTTON	;"
			+	"lumiere	: 	LIGHT	;";

	public final static String social_specif = ""; 		//aucun slot social
	
	Device leBouton, laLumiere;
	public AgentOnOff(){
		leBouton=null;
		laLumiere=null;
	}

	public void event(AgentUDescriptor aud, int code, int ctx){ //code = PAUSED / RESUMED / ALL_PLUGGED / MISSING_PLUG
		switch(code){
		case ALL_PLUGGED :		
			leBouton=aud.getDevice("bouton");
			laLumiere=aud.getDevice("lumiere");
			aud.log("full plug !");
			break;
		case MISSING_PLUG :
			leBouton=null;
			laLumiere=null;			
			aud.log("break plug !");
			break;
		}
	}
	//[Trace
	public void notifyPlugDevice(AgentUDescriptor aud, String slot, Device device){
		aud.log("nouveau plug sur le slot "+slot);
	}
	public void notifyUnplugDevice(AgentUDescriptor aud, String slot){
		aud.log("suppresion du plug sur le slot "+slot);
	}
	public void notifyChangeDevice(AgentUDescriptor aud, String slot, Device device){
		aud.log("changement de plug sur le slot "+slot);
	}
	//]
	
	public void sensorChange(AgentUDescriptor aud, String slot, Device dev, int sensor_id, int value){
		//aud.log("sensorValueChange on slot = "+slot+ " sensor_id = "+sensor_id);
		if(dev==leBouton && sensor_id==0) laLumiere.action(0, value); //nb: ici on a forcément laLumiere!=null
	}
		
	//[Programme test
	public static void main(String[] config){
		LaunchOption options = new LaunchOption(config);
		if(options.existField("h")) { //affichage des options de lancement
			String tab="    ";
			System.out.println(
						" Programme AgentOnOff : les plugs des slots physiques des agents.\n"
					+ 	"  NB: ce programme s'utilise avec un device de type LIGHT (ce peut être par exemple le programme\n"
					+	"	   tools.uda.DLightUDP) et un device de type BUTTON (par exemple tools.uda.DButtonUDP).\n"
					+ 	"  Comportement du programme : \n"
					+tab+ "- ce programme se connecte sur le hub spécifié dans les arguments de lancement.\n"
					+tab+ "- un agent de type AgentOnOff est créé et son exécution est lancée.\n"
					+tab+ "- une règles de plug physique lui est affecté à l'agent afin de cabler automatiquement un device\n"
					+tab+ "- de type LIGHT sur son slot physique \"lumière\".\n"
					+tab+ "- une règles de plug physique lui est affecté à l'agent afin de cabler automatiquement un device\n"
					+tab+ "- de type BUTTON sur son slot physique \"bouton\".\n"					
					+tab+ "- le comportement d'un agent de type AgentOnOff est le suivant :\n"
					+tab+ "  -> à chaque fois quand changement de valeur se produit sur le capteur de son slot \"bouton\"\n"
					+tab+ "      une action de changement de valeur (on/off) est envoyée vers le device du slot \"lumiere\".\n"
						+   "\n"
					+ 	" Arguments de lancement possibles (tous facultatifs, sauf mention contraire):\n"
					+tab+ "-trace : active le mode de traçage des logs et des changements d'état de l'agent\n"	
					+tab+ "-lgid SPOT_ID_L : la règle de plug appliquée sur le slot \"lumiere\" ciblera spécifiquement\n"
					+tab+ " le spot dont l'identifiant global a la valeur SPOT_ID_L\n"	
					+tab+ "-bgid SPOT_ID_B : la règle de plug appliquée sur le slot \"bouton\" ciblera spécifiquement\n"
					+tab+ " le spot dont l'identifiant global a la valeur SPOT_ID_B"										
			);
			System.out.println(HubUdp.hubOptions(tab, true));			
		} else {
			System.out.println("[AgentOnOff] programme lancé");

			Hub hub = HubUdp.joinHub(options, true);
			ServerAU servau = ServerAU.newServer(hub);
			int agent_id = servau.createAgent(AgentOnOff.class, "agent o/n");

			if(options.existField("trace")) new AgentLogger(servau, AgentLogger.LOG | AgentLogger.STATE).observeAgent(agent_id);
			servau.runAgent(agent_id); //on lance l'exécution de l'agent
		
			//on regarde si l'option -lgid est fournie en argument de lancement
			String lgid=options.val("lgid", null);
			String bgid=options.val("bgid", null);
		
			//on va plugger automatiquement un device lumière présent sur le réseau
			DataWriter wdesc_rule = DataBytes.writer()
				.b(PhysicalPlugRule.ALL);					//node_key
			if(lgid==null) wdesc_rule.b(PhysicalPlugRule.ALL);	//spot_key
			else wdesc_rule.b(PhysicalPlugRule.ID).s(lgid);
			wdesc_rule.b(PhysicalPlugRule.TYPE).i(UDA.LIGHT);		//device_key
			PhysicalPlugRule rule = PhysicalPlugRule.parse(wdesc_rule.flushReader());
			servau.addPhysicalPlugRule(agent_id, "lumiere", rule);
		
			//on va plugger automatiquement un device bouton présent sur le réseau
			wdesc_rule = DataBytes.writer()
				.b(PhysicalPlugRule.ALL)	;				//node_key
			if(bgid==null) wdesc_rule.b(PhysicalPlugRule.ALL);	//spot_key
			else wdesc_rule.b(PhysicalPlugRule.ID).s(bgid);
			wdesc_rule.b(PhysicalPlugRule.TYPE).i(UDA.BUTTON)	;	//device_key

			rule = PhysicalPlugRule.parse(wdesc_rule.flushReader());
			servau.addPhysicalPlugRule(agent_id, "bouton", rule);
		}
	}
	//]	
}