/** 
   Programme de test : MessageSender
 
   Ce programme permet de tester le bon fonctionnement de l'envoie de message en mode broadcast et send.
   L'interaction avec l'utilisateur (texte du message et choix de la cible) se fait via une interface graphique,
   ce programme ne peut donc pas s'utiliser entièrement en mode console.
   
   Pour plus de détails lancer ce programme avec l'option -h
 */
package test.hubudp;

import java.awt.*;
import javax.swing.*;
import java.awt.event.*;
import skuad.hubudp.Hub;
import skuad.hubudp.HubUdp;
import skuad.hubudp.Spot;
import skuad.hubudp.SpotObserver;
import skuad.util.databytes.*;
import skuad.util.base.LaunchOption;

public class MessageSender extends JFrame implements SpotObserver, ActionListener {
	
	JTextArea log_pane; //panneau d'affichage des logs (dont les messages reçu)
	JTextField jt_text;
	JComboBox<String> jcb_target; 
	Hub hub;
	
	
	public MessageSender(LaunchOption options) {
		super("MessageSender v1.0");
		setSize(420,300);
		setLocation(200,100);
		setDefaultCloseOperation(EXIT_ON_CLOSE);
		
		log_pane=new JTextArea();
		log_pane.setEditable(false);
		JScrollPane scroll_log = new JScrollPane(log_pane);
	
		//[Mise en page du panneau d'envoie de message
		JPanel send_panel=new JPanel();
		send_panel.setBorder(BorderFactory.createLineBorder(Color.BLACK));

		jt_text = new JTextField(20);
		JPanel phaut=new JPanel();
		phaut.add(new JLabel("Texte à envoyer :"));
		phaut.add(jt_text);
		
		jcb_target = new JComboBox<String>();
		jcb_target.setEditable(false);
		
		jcb_target.addItem("*tous+moi*" );
		jcb_target.addItem("*tous-moi*" );
		
		JPanel pbas=new JPanel();
		pbas.add(new JLabel("Cible :"));
		pbas.add(jcb_target);
		JButton bt_send=new JButton("ENVOYER");
		bt_send.addActionListener(this);
		pbas.add(bt_send);

		send_panel.setLayout(new GridLayout(2,1));
		send_panel.add(phaut);
		send_panel.add(pbas);
		//]
		
		Container content = getContentPane();
		content.setLayout(new BorderLayout());
		content.add(send_panel, BorderLayout.NORTH);
		content.add(scroll_log, BorderLayout.CENTER);
		setVisible(true);
		
		//Cablage avec le hub	
		hub = HubUdp.joinHub(options, true);
		if(hub!=null) {
			hub.addReceiver(0, "MessageSender",
				(Spot sender, DataReader message, int channel) -> {  //Code de traitement des messages reçus
					String text_message = message.readString();
					log("Réception", "message #"+text_message+"# en provenance de "+sender.getIdTarget());
					message.dispose();
				}					
			);	
			hub.addSpotObserver(this);
		} else System.out.println("... impossible de se connecter au hub !");
	}
	
	public static void main(String[] args) {  
		LaunchOption options = new LaunchOption(args);
		if(options.existField("h")) { //affichage des options de lancement
			String tab="    ";
			System.out.println(
						" Programme MessageSender : permet de tester l'envoie de message en mode broadcast et send.\n"
					+ 	"  NB: - ce programme peut s'utiliser avec les programmes de test : PingPong, Romeo,\n"
					+ 	"        Juliette ou d'autres programmes MessageSender.\n"
					+	"      - il affiche une fenêtre, aussi il ne peut pas s'utiliser entièrement en mode console.\n"
					+ 	"  Comportement du programme : \n"
					+tab+ "- ce programme se connecte sur le hub spécifié dans les arguments de lancement.\n"
					+tab+ "- il affiche une une interface graphique qui permet d'écrire le texte d'un message\n"
					+tab+ "  et de sélectionner un distinataire (un spot, ou tous).\n"
					+tab+ "- il observe les spots présents pour les ajouter dans la liste des cibles possibles.\n"
					+tab+ "- un receveur de message sur le canal 0 est enregistré, et tout les messages \n"
					+tab+ "  reçues sont affichés dans une zone de texte de l'interface graphique. \n"
					+   "\n"
					+ 	" Arguments de lancement possibles (tous facultatifs, sauf mention contraire):"
			);
			System.out.println(HubUdp.hubOptions(tab, true));
		} else new MessageSender(options); 
	}
	
	//[Interface ActionListener
	public void actionPerformed(ActionEvent e) {
		String text_message = jt_text.getText();
		String idTarget = jcb_target.getSelectedItem().toString();	
		DataReader mess = DataBytes.writer().s(text_message).flushReader();
		
		if(idTarget.equals("*tous+moi*")) {
			log("->Envoi ->", "cible : *broadcast* + SERVE_LOCAL , message #"+text_message+"#");
			hub.broadcast(0, mess, Hub.NEED_ACK | Hub.AUTO_DISPOSE | Hub.SERVE_LOCAL);
		} else if(idTarget.equals("*tous-moi*")) {
			log("->Envoi ->", "cible : *broadcast* NO_SERVE_LOCAL , message #"+text_message+"#");
			hub.broadcast(0, mess, Hub.NO_SERVE_LOCAL);
		} else {
			Spot target = hub.getSpotByIdTarget(idTarget);
			if(target==null) log("** Erreur", "cible : "+idTarget+" inconnue !");
			else {
				log("->Envoi ->", "cible : "+idTarget+" , message #"+text_message+"#");
				target.send(0, mess);
			}
		}
	}
	//]
	
	//[Interface SpotObserver
	public void up(Spot spot){		
		if(spot.isLocalSpot()) {
			log("-- Info --", "spot local activé, idTarget = "+spot.getIdTarget());
			setTitle("MessageSender v1.0 : "+spot.getIdTarget());
		}
		else log("-- Info --", "up de idTarget = "+spot.getIdTarget());
		jcb_target.addItem(spot.getIdTarget());
	}
	public void down(Spot spot){
		jcb_target.removeItem(spot.getIdTarget());
	}
	public void idContext(Spot spot, String ch_context, String id_context) {}
	public void updateSMD(Spot spot, boolean first_definition){}
	//]
	
	public void log(String type, String text) {
		log_pane.append("["+type+"] : "+text+" \n");
	}
}