/** 
   Programme de test : PingPongHub
 
   Ce programme permet de tester la communication de message via la méthode Spot.send(...)
   
   Pour plus de détails lancer ce programme avec l'option -h
 */
package test.hubudp;

import skuad.hubudp.*;
import skuad.util.databytes.DataBytes;
import skuad.util.databytes.DataReader;
import skuad.util.base.LaunchOption;

public class PingPongHub implements SpotObserver {
	
	public static void main(String[] args) {
		LaunchOption options = new LaunchOption(args);
		if(options.existField("h")) { //affichage des options de lancement
			String tab="    ";
			System.out.println(
						" Programme PingPongHub : permet de tester la communication de message via la méthode Spot.send(...).\n"
					+ 	"  NB: ce programme s'utilise avec d'autres programmes PingPongHub.\n"
					+ 	"  Comportement du programme : \n"
					+tab+ "- ce programme se connecte sur le hub spécifié dans les arguments de lancement. \n"
					+tab+ "- un message \"ping\" est envoyé à tous les spots détectés sur le hub.\n"
					+tab+ "- un receveur de message le canal 0 est enregistré, et tout les messages \n"
					+tab+ "  reçues sont affichés dans la console. \n"
					+tab+ "- quand un message \"ping\" est reçu, un message \"pong\" est envoyé à son expéditeur.\n"
					+tab+ "NB : dans ce PingPongHub l'échange de message ping/pong est pontuel (un seul échange par binôme)\n"
					+   "\n"
					+ 	" Arguments de lancement possibles (tous facultatifs, sauf mention contraire):"
			);
			System.out.println(HubUdp.hubOptions(tab, true));
		} else {
			System.out.println("[PingPongHub] programme lancé");
			PingPongHub pp = new PingPongHub();
		
			Hub hub = HubUdp.joinHub(options, true);
		
			if(hub!=null) {
				hub.addReceiver(0, "PingPongHub",  //pour pouvoir recevoir les messages "ping" et "pong", canal utilisé : 0
					(Spot sender, DataReader message, int channel) -> {
						String text_message = message.readString();
						System.out.println("[PingPongHub] réception du message #"+text_message+"# en provenance de "+sender.getIdTarget());

						if(text_message!=null && text_message.equals("ping")) { //on doit répondre "pong" au sender
							sender.send(0, DataBytes.writer().s("pong").flushReader());
						}

						message.dispose();
					}		
				);	
				hub.addSpotObserver(pp); //pour pouvoir détecter la présence des autres spot
			} else System.out.println("... impossible de se connecter au hub !");
		}
	}
	
	//[Interface SpotObserver
	public void up(Spot spot){		
		//System.out.println("( up de : idTarget = "+spot.getIdTarget()+" )");
		if(spot.isLocalSpot()) System.out.println("[PingPongHub] spot local activé, idTarget = "+spot.getIdTarget());
		else { //si ce n'est pas le spot local, on lui envoie le message "ping"
			spot.send(0, DataBytes.writer().s("ping").flushReader());
		}
	}
	public void down(Spot spot){}
	public void idContext(Spot spot, String ch_context, String id_context) {}
	public void updateSMD(Spot spot, boolean first_definition){}
	//]
}