/** 
   Programme de test : Romeo
 
   Ce programme permet de tester les fonctionnalités SMD, il s'utilise avec le programme Juliette.
   
    Pour plus de détails lancer ce programme avec l'option -h
 */
package test.hubudp;

import skuad.hubudp.*;
import skuad.util.databytes.DataBytes;
import skuad.util.databytes.DataReader;
import skuad.util.base.LaunchOption;

public class Romeo implements SpotObserver {
	
	public static void main(String[] args) {
		LaunchOption options = new LaunchOption(args);
		if(options.existField("h")) { //affichage des options de lancement
			String tab="    ";
			System.out.println(
						" Programme Romeo : permet de tester les fonctionnalités SMD. \n"
					+ 	"  NB: ce programme s'utilise avec le programme de test Juliette.\n"
					+ 	"  Comportement du programme : \n"
					+tab+ "- ce programme se connecte sur le hub spécifié dans les arguments de lancement. \n"
					+tab+ "- il fixe la chaine 'romeo' sous le code SMD 19 du spot local. \n"
					+tab+ "- il observe les notification updateSMD des autres spots. \n"
					+tab+ "- quand un spot possédant la chaine 'juliette' sous le code SMD 19 est découvert, \n"
					+tab+ "  un message 'ma Juliette !' lui est envoyé. \n"
					+tab+ "- un receveur de message le canal 0 est enregistré, et tout les messages \n"
					+tab+ "  reçues sont affichés dans la console. \n"
					+   "\n"
					+ 	" Arguments de lancement possibles (tous facultatifs, sauf mention contraire):"
			);
			System.out.println(HubUdp.hubOptions(tab, true));
		} else {
			System.out.println("[Romeo] programme lancé");
			Romeo romeo = new Romeo();

			Hub hub = HubUdp.joinHub(options, true);
			if(hub!=null) {
				hub.addReceiver(0, "romeo",
					(Spot sender, DataReader message, int channel) -> {  //Code de traitement des messages reçus
						String text_message = message.readString();
						System.out.println("[Romeo] réception du message #"+text_message+"# en provenance de "+sender.getIdTarget());

						message.dispose();
					}		
				);	
				hub.addSpotObserver(romeo);
				hub.whenReady((LocalSpot localspot) -> {
					localspot.setString(19, "romeo");	
				});
			} else System.out.println("... impossible de se connecter au hub !");
		}
	}
	
	//[Interface SpotObserver
	public void up(Spot spot){		
		if(spot.isLocalSpot()) System.out.println("[Romeo] spot local activé, idTarget = "+spot.getIdTarget());
		else System.out.println("( up de : idTarget = "+spot.getIdTarget()+" )");
	}
	public void down(Spot spot){}
	public void idContext(Spot spot, String ch_context, String id_context) {}
	public void updateSMD(Spot spot, boolean first_definition){
		//on regarde si le spot porte la chaine "juliette" sous le code SMD 19
		String name = spot.getString(19);
		if(name!=null && name.equals("juliette")) {
			//trouvé !
			spot.send(0, DataBytes.writer().s("ma Juliette !").flushReader());
		}
	}
	//]
}