/** 
   Programme de test : HelloSim
 
   Ce programme permet de tester le fonctionnement de base de simskaud
   
   Pour plus de détails lancer ce programme avec l'option -h
 */
package test.sim;

import skuad.util.base.LaunchOption;
import skuad.sim.*;
import skuad.sim.simop.*;
import skuad.sim.ui.SimFrame;
import skuad.agentu.AgentU;
import skuad.util.time.Pulse;
import skuad.util.time.*;
import test.sim.agent.*;

public class HelloSim {
	
	public static void main(String[] arg) {
		LaunchOption options = new LaunchOption(arg);
		if(options.existField("h")) { //affichage des options de lancement
			String tab="    ";
			System.out.println(
						" Programme SimHello : ce programme permet de tester le fonctionnement de base de SimSkuad.\n"
					+ 	"  NB: - ce programme s'utilise uniquement en mode graphique.\n"
					+ 	"  Comportement du programme : \n"
					+tab+ "- ce programme instancie un noyau de simulation SimAU de nom \"Hello Sim\".\n"
					+tab+ "- il ajoute dans cette simulation un initialiseur qui :\n"
					+tab+ "  -> défini une plage de simulation de 5 jours\n"
			   		+tab+ "  -> enregistre un objet pulsable HelloPulsable\n"
			   		+tab+ "  -> crée un agent HelloAgent et fixe son mode d'exécution sur ALLWAYS_RUN\n"
			   		+tab+ "- il termine en créant et en affichant une fenêtre de controle de la simulation.\n"
			   		+tab+ "NB: - l'objet pulsable HelloPulsable effectue un logSim \"Coucou Pulsable !\" toutes les 3h.\n"
			   		+tab+ "    - l'agent HelloAgent effectue un log \"Hello Agent !\" toutes les 24h.\n"
			   		+tab+ "    - le niveau de vitesse de la simulation est réglé initialement à la valeur 400,\n"
			   		+tab+ "      ce qui donne que 1h de temps simulé durera environs 1 secondes en temps réel.\n"
						+   "\n"
					+ 	" Arguments de lancement possibles : Aucun"
			);
		} else {
			System.out.println("[Hello Sim] programme lancé");

			SimAU hello_sim = new SimAU("Hello Sim");

			//Configuration de la simulation
			hello_sim.addInitializer((SimPilot sim_pilot, InitSimOperation isop, StartSimOperation ssop)->{
				isop.setTimes(UT.realCurTime(), UT.PING, UT.realCurTime()+UT.day(5));
				ssop.addPulsable(new HelloPulsable());
				int agent_id=ssop.createAgent(HelloAgent.class, "helloA01");
				ssop.changeRunModeAgent(agent_id, AgentU.ALLWAYS_RUN);
			});

			//fenêtre de controle de la simulation
			SimFrame sframe=new SimFrame(hello_sim, true);
			sframe.setLevelSpeed(400);
			sframe.setVisible(true);
		}
	}
}

//Exemple d'objet pulsable
class HelloPulsable implements Pulsable {
	int n;
	
	public void start(SimPilot sim_pilot, PulsableDescriptor pdesc, StartSimOperation ssop) {
		n=0;
		pdesc.loopPulse("say coucou", 0, UT.hour(3));
	}
	public void stop(SimPilot sim_pilot) {}	//fait suite à l'arrêt de la simulation, ou le désenregistrement du pulsable
	public void behavior(SimPilot sim_pilot, PulsableDescriptor pdesc, Pulse pulse, StartSimOperation ssop, RunSimOperation rsop) {	
		ssop.log("Coucou Pulsable ! ("+(n++)+")");
	}
}
