/** 
   Programme de test : PingPong
 
   Ce programme permet de tester :
   		- l'observation de skill dans un space
   		- la communication de message via la mailbox des avatars
   		- le chainement des activités d'un avatar
   
   Pour plus de détails lancer ce programme avec l'option -h
 */
package test.ubiquity;

import skuad.ubiquity.*;
import skuad.ubiquity.space.*;
import skuad.ubiquity.space.activity.*;
import skuad.ubiquity.space.mailbox.*;
import skuad.hubudp.*;
import skuad.util.value.Value;
import skuad.util.base.*;

public class PingPong extends FullParticipantListener implements ActivityRunner {

	Avatar myAvatar;

	public PingPong(Hub hub, Space space){	
		myAvatar=space.newAvatar();
		log("==[PingPong]", "ADDED");

		space.addSkillListener(this);					//On écoute les attributions de compétence (-> skillAdded(...))
		myAvatar.mailBox().addListener(this);
		myAvatar.activity().set("sendPingPong", this);	//On enregistre l'activité d'envoi de message Ping/Pong (son comportement sera la méthode action(...))
		myAvatar.skill().add("ping-pong");				//On déclare la compétence ping-pong
		//myAvatar.alias().setAlias("Bob");
	}

	public void skillAdded(Participant participant, String newSkill) { //notification d'une attribution de compétance dans le space
		if(newSkill.equals("ping-pong") && myAvatar.participant().ordoredBefore(participant)){ //c'est une compétance PingPong et on a l'initiative du ping
			log("=[PingPong - skillAdded]","je découvre le partenaire : '"+participant.fullId()+"'");
			Value arg = Value.parse("{ target : "+participant.fullId()+" , message : ping }"); //préparation de l'argument d'action pour l'activité qui va être programmée
			
			myAvatar.activity().chain("sendPingPong", arg, 3000); //on prévoit d'envoyer ping dans 3 secondes
		}
	}

	public void action(Avatar body, String activity, Value args, ProgressFlag progress_flag) { //comportement d'envoie ping/pong (décrit par args)
		if(activity.equals("sendPingPong")) {
			String target=args.getString("target"); //la cible
			Value message=args.get("message");		//message ping ou pong
			log("=[PingPong - action]","j'envoie le message '"+message.toString()+"' au partenaire : '"+target+"'");
			body.mailBox().send(target, "pingpong", message); //on envoie le message
		}
	}

	public void receive(MailBox box, Message mess) { //on recois un message  : "ping"  // { nom : "ping" }
		String vmess=mess.value().toString(); //valeur du message
		String sender=mess.sender().fullId(); //l'expéditeur (à qui on va répondre ping ou pong)
		log("=[PingPong - receive]","je reçois : '"+vmess+"' de l'avatar : '"+sender+"'");

		if(vmess!=null && (vmess.equals("ping") || vmess.equals("pong"))) {
			Value arg = Value.parse("{ target : "+sender+" , message : "+("ping".equals(vmess)?"pong":"ping")+" }"); //préparation de l'argument d'action
			myAvatar.activity().chain("sendPingPong", arg, 3000); //on prévoit d'envoyer ping dans 3 secondes
		}
		mess.consume(true); //on a traité ce message (donc on le purge)
	}

	public void log(String head, String mess){
		Util.log(":", head+" je suis l'avatar '"+myAvatar.participant().fullId()+"'", mess);
	}
	
	public static void main(String[] hub_config){
		LaunchOption options = new LaunchOption(hub_config);
		if(options.existField("h")) { //affichage des options de lancement
			String tab="    ";
			System.out.println(
						" Programme PingPong : permet de tester l'observation de skill, la communication et le chainage d'activité.\n"
					+ 	"  NB: ce programme s'utilise avec un autre programme PingPong.\n"
					+ 	"  Comportement du programme : \n"
					+tab+ "- ce programme se connecte sur le hub spécifié dans les arguments de lancement.\n"
					+tab+ "- nous allons appeler space S le space indiqué en argument.\n"					
					+tab+ "- un avatar est créé dans le space S.\n"
					+tab+ "- le programme se met à l'écoute des modifications de skill dans le space S.\n"
					+tab+ "- quand l'ajout d'un skill de nom \"ping-pong\" est détecté, le programme analyse s'il doit ou non\n"
					+tab+ "  prendre l'initiative d'envoyer un message \"ping\" à destination du participant porteur de ce skill.\n"
					+tab+ "  -> cette décision est prise en fonction de la relation d'ordre entre les participants donnée par\n"
					+tab+ "     la méthode Participant.ordoredBefore(...) qui garantie qu'à chaque couple de participant ayant\n"
					+tab+ "     skill \"ping-pong\" qui se voient l'un et l'autre, seulement l'un d'entre eux prendra l'initiative\n"
					+tab+ "     de l'envoie du message vers l'autre.\n"
			   		+tab+ "  -> si l'initiative est prise, le message sera envoyé aprés un délai de 3 secondes via le chainage\n"
			   		+tab+ "     d'une activité à déclenchement retardé.\n"
			   		+tab+ "- tous les messages reçue via la mailbox est affiché sur la console.\n"
			   		+tab+ "- quand un message \"ping\" est reçu, un message \"pong\" est envoyé à son expéditeur.\n"
			   		+tab+ "- quand un message \"pong\" est reçu, un message \"ping\" est envoyé à son expéditeur.\n"
			   		+tab+ "NB : dans ce PingPong l'échange de message ping/pong est cyclique (perpétuel).\n"
						+   "\n"
					+ 	" Arguments de lancement possibles (tous facultatifs, sauf mention contraire):\n"
					+tab+ "-space NAME_SPACE : nom du space à utiliser (\"world\" par défaut)"
			);
			System.out.println(HubUdp.hubOptions(tab, true));			
		} else {
			System.out.println("[PingPong] programme lancé");

			String space_name = options.val("space", "world");
			Hub hub = HubUdp.joinHub(options, true);
			hub.whenReady( (LocalSpot lp) -> {
				Space space = Ubiquity.getSpace(hub, space_name);
				new PingPong(hub, space);
				System.out.println("pingpong OK");
			});
		}
	}
}