/** 
   Programme de test : TraceParticipant
 
   Ce programme permet de tester l'observation de l'ajout et du retrait de participant dans un space.
   
   Pour plus de détails lancer ce programme avec l'option -h
*/
package test.ubiquity;

import skuad.ubiquity.space.*;
import skuad.hubudp.Hub;
import skuad.hubudp.HubUdp;
import skuad.util.base.LaunchOption;
import skuad.ubiquity.Ubiquity;

public class TraceParticipant implements ParticipantListener {
	
	public static void main(String[] hub_config){
		LaunchOption options = new LaunchOption(hub_config);
		if(options.existField("h")) { //affichage des options de lancement
			String tab="    ";
			System.out.println(
						" Programme TraceParticipant : permet de tester l'observation de l'ajout et du retrait de participant dans un space.\n"
					+ 	"  Comportement du programme : \n"
					+tab+ "- ce programme se connecte sur le hub spécifié dans les arguments de lancement.\n"
					+tab+ "- un observateur de participant est enregistré sur le space indiqué en argument.\n"
					+tab+ "- le programme se met à l'écoute des modifications de skill dans le space S.\n"
					+tab+ "- quand l'ajout ou le retrait d'un participant est détecté, un message détaillant cet évènement est\n"
					+tab+ "  affiché dans la console.\n"
						+   "\n"
					+ 	" Arguments de lancement possibles (tous facultatifs, sauf mention contraire):\n"
					+tab+ "-space NAME_SPACE : nom du space à utiliser (\"world\" par défaut)"
			);
			System.out.println(HubUdp.hubOptions(tab, true));			
		} else {
			System.out.println("[TraceParticipant] programme lancé");

			String space_name = options.val("space", "world");
			Hub hub = HubUdp.joinHub(options, true);
		
			//On enregistre une instance de cette classe (un ParticipantListener)
			//pour écouter l'ajout/retrait des participants
			Ubiquity.getSpace(hub, space_name).addParticipantListener(new TraceParticipant());
		}
	}
	
	//Interface ParticipantListener
	int nb=0;
		//Notifie qu'un nouveau participant apparaît dans le space
	public synchronized void participantAdded(Participant p){
		nb++;
		System.out.println("* Ajout du participant : " 
				+"participant_id = " + p.localId()   
				+" // host_global_id = "+p.spot().getIdGlobal()+ " ["+nb+"]"
		);
	}
	
		//Notifie qu'un participant vient de quitter le space
	public synchronized void participantRemoved(Participant p){
		nb--;
		System.out.println("* Retrait du participant : " 
				+"participant_id = " + p.localId()
				+" // host_global_id = "+p.spot().getIdGlobal()+ " ["+nb+"]"
		);		
	}
}