/** 
   Programme de test : LocalCoupling
 
   Ce programme permet de tester le bon fonction du coupleur de device (uda.tools.Coupleur) en mode local (pas interaction sur le réseau, tout
   se passe dans le même contexte d'exécution).
   
   Comportement du programme :
   - un device bouton virtuel (classe uda.gui.device.DButton) est créé
   - un device lumière virtuel (classe uda.gui.device.DLight) est créé
   - ces 2 devices sont couplés localement l'un à l'autre à l'aide du coupleur uda.tools.Coupleur
     
 
    Argument de lancement : aucun
  
    Exemple de lancement depuis le dossier bin du projet :
    windows: 	> java  -classpath .;../lib/hubudp_v0.1.a11.jar  test.uda.LocalCoupling
    	mac et linux: 	> java  -classpath .:../lib/hubudp_v0.1.a11.jar   test.uda.LocalCoupling
    	
    	NB: - même si la foncitonnélité HubUDP n'est pas utilisé par ce programme, il est quand même nécessaire
    		de place cette librairie dans le classpath car le sous-package hubudp.util est utilisé à de nombreuse
    		reprise par le code d'implémentation de skuad.uda (en particulier la classe hubudp.util.thread.ThreadStack).
    		- le nom de l'archive hubudp doit être adapté en fonction de son numéro de version.
*/
package test.uda;

import skuad.uda.gui.device.*;
import skuad.uda.tools.Coupler;
import javax.swing.*;
import skuad.util.base.LaunchOption;
import skuad.util.gui.*;

public class LocalCoupling {
	
	public static void main(String[] args){
		LaunchOption options = new LaunchOption(args);
		if(options.existField("h")) { //affichage des options de lancement
			String tab="    ";
			System.out.println(
						" Programme LocalCoupling : permet de tester le bon fonction du coupleur de device (uda.tools.Coupleur)\n"
					+	" en mode local (pas d'interaction sur le réseau, tout se passe dans le même contexte d'exécution).\n"
					+ 	"  Comportement du programme : \n"
					+tab+ "- un device bouton virtuel (classe uda.gui.device.DButton) est créé.\n"
					+tab+ "- un device lumière virtuel (classe uda.gui.device.DLight) est créé.\n"
					+tab+ "- ces 2 devices sont couplés localement l'un à l'autre à l'aide du coupleur uda.tools.Coupleur.\n"
						+   "\n"
					+ 	" Arguments de lancement possibles (tous facultatifs, sauf mention contraire):\n"
					+tab+ "  (aucun)"
			);
		} else {
			System.out.println("[LocalCoupling] programme lancé");

			//Les devices
			DButton dButton = new DButton("Le bouton");
			DLight dLight = new DLight("La lumière");
		
			//Couplage
			Coupler.plug(dButton, 0, dLight, 0);
		
			//Affichage des devices
			JFrame f1 = HaveGUI.FrameFor(dButton, "Le bouton",true);
			JFrame f2 = HaveGUI.FrameFor(dLight, "La lumière",true);
		
			f1.setSize(250, 250); f1.setLocation(100, 100);
			f2.setSize(250, 250); f2.setLocation(400, 100);
		}
	}
	
}