/** 
   Programme de test : UdpCoupling
 
   Ce programme permet de tester le pontage UdaUdp de device. 
   Il s'utilise en particulier avec les programmes test.uda.DButtonUDP et test.uda.DLightUDP.
   
   Comportement du programme :
   - Ce programme se connecte sur le hub spécifié dans les arguments de lancement.
   - les notifications SMD de ce hub son écouté afin de détecter les devices sur lesquels le couplage doit être réalisée.
   - cette détection est réalisé en regardant si le spot dispose d'un SMD de code, et si oui alors cela voudra dire que ce spot expose un device via UdaUdp et si :
    	- la valeur de ce SMD est "bouton1" : alors le device d'id 1 exposé par ce spot sera utilisé comme device sensor pour l'opérationd de couplage.
    	- la valeur de ce SMD est "lumiere1" : alors le device d'id 1 exposé par ce spot sera utilisé comme device actuator pour l'opérationd de couplage.     
   - une fois que les 2 devices cibles sont trouvés, le couplage est réalisé à la de la classe uda.tools.Coupler
  
 
    Argument de lancement :
      	test.uda.UdpCoupling 					-> c'est le hub local par défaut qui sera rejoint
	  	test.uda.UdpCoupling -hubname NAME		-> c'est le hub local de nom NAME qui sera rejoint
	  	test.uda.UdpCoupling -portal IP PORT	-> ce sera le hub connecté au portail IP:PORT
  
    Exemple de lancement depuis le dossier bin du projet :
    windows: 	> java  -classpath .;../lib/hubudp_v0.1.a11.jar   test.uda.UdpCoupling
    	mac et linux: 	> java  -classpath .:../lib/hubudp_v0.1.a11.jar   test.uda.UdpCoupling
    	
    	NB: le nom de l'archive hubudp doit être adapté en fonction de son numéro de version.
*/
package test.uda;

import skuad.uda.*;
import skuad.uda.net.UdaUdp;
import skuad.uda.tools.Coupler;
import skuad.hubudp.*;
import skuad.util.base.LaunchOption;

public class UdpCoupling implements SpotObserver {
	
	//[Attribut à régler
		//Le device capteur
	public final static int SensorDeviceID = 1;
	public final static int ActuatorDeviceID = 1;
	
	//Attributs internes
	Device sensor; Spot spot_sensor;
	Device actuator; Spot spot_actuator;
	boolean coupled;
	int find_code;
	String val_sensor, val_actuator; 
	
	public UdpCoupling(Hub hub, int fcode, String vsensor, String vactuator){
		sensor=null;
		actuator=null;
		coupled=false;
		find_code = fcode;
		val_sensor = vsensor;
		val_actuator = vactuator;
		hub.addSpotObserver(this);
	}
	
	private void performCoupling(){
		if(!coupled && sensor!=null && actuator!=null){
			coupled = true;
		
			//Couplage
			Coupler.plug(sensor, 0, actuator, 0);
			System.out.println("UdpCoupling : Couplage établi !");
		}
	}
	
	private void breakCoupling(){
		if(coupled){
			coupled = false;
		
			Coupler.unplug(sensor, 0, actuator, 0);
			System.out.println("UdpCoupling : Couplage rompu !");
		}
	}
	
	//[Interface SpotObserver 	
	public void up(Spot dspot){}
	public void updateSMD(Spot dspot, boolean first_definition){
		String smd_val = dspot.getString(find_code);
		if(smd_val!=null) {
			if(sensor==null && smd_val.equals(val_sensor)) {
				System.out.println("UdpCoupling : sensor ok !");
				spot_sensor = dspot;
				sensor=UdaUdp.linkDevice(dspot, SensorDeviceID);  
			} else 	if(actuator==null && smd_val.equals(val_actuator)) {
				spot_actuator = dspot;
				System.out.println("UdpCoupling : actuator ok !");
				actuator=UdaUdp.linkDevice(dspot, ActuatorDeviceID);  
			} else return;
			performCoupling();	
		}
	}
	public void down(Spot dspot){ //on doit supprimer tous les devices portés par ce noeud
//System.out.println("DownSpot "+dspot.getIp()+":"+dspot.getPort());		
		if(sensor!=null && spot_sensor==dspot){ breakCoupling(); UdaUdp.unlinkDevice(sensor); sensor=null; }
		if(actuator!=null && spot_actuator==dspot) { breakCoupling(); UdaUdp.unlinkDevice(actuator); actuator=null; }
	}
	public void idContext(Spot dspot, String ch_context, String id_context){}
	//]
	
	public static void main(String[] hub_config){
		LaunchOption options = new LaunchOption(hub_config);
		if(options.existField("h")) { //affichage des options de lancement
			String tab="    ";
			System.out.println(
						" Programme UdpCoupling : permet de tester le pontage UdaUdp de device.\n"
					+   " NB: ce programme s'utilise en particulier avec les programmes test.uda.DButtonUDP et test.uda.DLightUDP."
					+ 	"  Comportement du programme : \n"
					+tab+ "- ce programme se connecte sur le hub spécifié dans les arguments de lancement.\n"
					+tab+ "- les notifications SMD de ce hub sont écoutées afin de détecter les devices sur lesquels\n"
					+tab+ "  le couplage doit être réalisée. La détection se base sur la présence du smd CODE (valeur\n"
					+tab+ "  de l'argument obligatoire -find_code) et de sa valeur VALUE_SENSOR ou VALUE_ACTUATOR (voir\n"
					+tab+ "  les arguments facultatifs -sensor et -actuator).\n"
					+tab+ "- une fois que les 2 devices cibles sont trouvés, le couplage est réalisé à l'aide de la\n"
					+tab+ "  classe uda.tools.Coupler\n"
					+   "\n"
					+ 	" Arguments de lancement possibles (tous facultatifs, sauf mention contraire):\n"
					+tab+ "-find_code CODE : code smd sur lequel établir la détection (obligatoire)\n"
					+tab+ "-sensor VALUE_SENSOR : valeur du smd CODE à détecter pour identifier le capteur (par défaut \"bouton\")\n"
					+tab+ "-actuator VALUE_ACTUATOR : valeur du smd CODE à détecter pour identifier l'actuateur (par défaut \"lumiere\")"
					);
			System.out.println(HubUdp.hubOptions(tab, true));			
		} else {
			System.out.println("[UdpCoupling] programme lancé");

			int fsmd = options.intVal("find_code", -1);
			if(fsmd<=0) {
				System.out.println("Erreur : vous devez spécifier l'argument -find_code CODE\n"
								+  "         Utilisez l'option -h pour avoir plus d'information.");
			} else {
				Hub hub = HubUdp.joinHub(options, true);
				if(hub!=null) new UdpCoupling(hub, fsmd, options.val("sensor", "bouton"), options.val("actuator", "lumiere"));
				else System.out.println("... impossible de se connecter au hub !");	
			}
		}
	}
}