/** 
   Programme utilitaire : wit (Who Is There)
 
   Ce programme affiche la liste des spots présents sur le hub indiqué dans les arguments de lancement.
   Si aucun nom de hub est précisé, le hub par défaut sera utilisé.
   
    Pour plus de détails lancer ce programme avec l'option -h
 * */

package tools.hubudp;

import skuad.hubudp.*;
import skuad.util.base.LaunchOption;
import skuad.util.thread.*;
import java.util.*;

public class wit implements SpotObserver, Runnable, SpotExplorer {
	
	//Constante pour l'affichage
	public final static String begin_tab = " ";
	public final static String[] WTitle = { "Ip:Port" , "idLocal", "IdGlobal", "Info" };
	public final static int[] WRow = { 19, 8, 10, 30 }; //Constantes des largeur de colonne pour l'affichage (4 colonnes)
	
	boolean no_exit, first_display;
	TimerTask timer_refresh_list;
	int duree;
	Hub hub;
	
	public wit(int v_duree, boolean v_no_exit, LaunchOption config){
		first_display = true;
		duree = v_duree;
		no_exit = v_no_exit;
		hub=null;

		System.out.println("== début du scan du hub :   ...");
		timer_refresh_list=ThreadStack.perform(this, 50); //premier affichage dans 50ms
		
		hub = HubUdp.joinHub(config, true);
		if(hub!=null) hub.addSpotObserver(this);
		else {
			System.out.println("... impossible de se connecter au hub !");
			System.exit(0);
		}
	}
	
	public static void main(String args[]){
		LaunchOption options = new LaunchOption(args);
		if(options.existField("h")) { //affichage des options de lancement
			String tab="    ";
			System.out.println(
						" Programme wit : analyse les spots présents dans le hub.\n"
					+ 	" Arguments de lancement possibles (tous facultatifs, sauf mention contraire):\n"
					+tab+ "-t DUREE : l'analyse du hub durera DUREE seconde(s) (4 par défaut) [ignoré si -no_exit]\n"
					+tab+ "-no_exit : l'analyse du hub sera permanente"
			);
			System.out.println(HubUdp.hubOptions(tab, true));			  
		} else {
			boolean no_exit= options.existField("no_exit");
			int duree= options.intVal("t", 4);
			new wit(duree, no_exit,  options);
		}
	}
		
	//[Interface SpotObserver
	public void idContext(Spot spot, String ch_context, String id_context) {}
	
	public synchronized void up(Spot spot){		
		printRow(spot.getIpPort(), spot.getIdLocal()+"", "???", ( spot.isLocalSpot() ? "(me)" : ""), true);
	}
	public void updateSMD(Spot spot, boolean first_definition){
		String name_host = spot.getString(Spot.SPOT_NAME);		
		if(timer_refresh_list==null && (first_definition || (name_host!=null && name_host.length()>0))) timer_refresh_list=ThreadStack.perform(this, 2000); //nouvel affichage dans 2 secondes
	}
	public synchronized void down(Spot spot){
		System.out.println("-- down --> "+spot.getIdTarget());
		System.out.flush();
		if(timer_refresh_list==null) timer_refresh_list=ThreadStack.perform(this, 2000); //nouvel affichage dans 2 secondes
	}
	//]
	
	//[Interface HostExplorer
	public void spot(Spot spot){
		String idglobal = spot.getIdGlobal(); if(idglobal==null) idglobal="";
		//Info:  spotName@hostname OS/SOFT/HARDWARE (me) 
		String spot_name = spot.getString(Spot.SPOT_NAME); if(spot_name==null) spot_name=""; else spot_name+="@";
		String host_name = spot.getString(Spot.HOST_NAME); if(host_name==null) host_name=""; else if(spot_name.length()==0) spot_name="@";
		String os=Spot.decodeOs(spot.getInt(Spot.OS), "");
		String soft=Spot.decodeSoft(spot.getInt(Spot.SOFT), "");
		String hard=Spot.decodeHardware(spot.getInt(Spot.HARDWARE), "");
		String info = spot_name+host_name+" "+os+"/"+soft+"/"+hard+( spot.isLocalSpot() ? " (me) " : "");
		
		printRow(spot.getIpPort(), spot.getIdLocal()+"", idglobal, info, false);
	}
	//]
	
	public synchronized void run(){
		String hub_name="";
		if(hub!=null) {
			hub_name = hub.getHubName();
			if(hub_name == null) hub_name=""; 
			else hub_name = "(Hub = "+hub_name+(hub.isPrivate() ? "$" : "")+") ";
		}
		
		if(first_display) {
			first_display = false;
			System.out.println("\n  Liste des spots actifs "+hub_name+":");
			printRow(WTitle[0], WTitle[1], WTitle[2], WTitle[3], false); System.out.println();
			
			if(!no_exit) ThreadStack.perform(new Runnable(){ //on va s'arréter dans duree seconde
					public void run(){
						System.out.println("\n... fin du scan ==");
						System.exit(0);
					}
			}, duree*1000);
		} else {
			System.out.println("\n\n\n  Nouvelle liste des spots actifs "+hub_name+" :");
			printRow(WTitle[0], WTitle[1], WTitle[2], WTitle[3], false); System.out.println();
			if(hub!=null) hub.explore(this);
			System.out.flush();
		}

		timer_refresh_list=null;		
	}

	public void printRow(String col1, String col2, String col3, String col4, boolean flush) {
		System.out.print(begin_tab); printCell(col1, WRow[0]);
		System.out.print(" "); printCell(col2, WRow[1]);
		System.out.print(" "); printCell(col3, WRow[2]);
		System.out.print(" "); printCell(col4, WRow[3]);
		System.out.println();
		if(flush) System.out.flush();
	}
	public void printCell(String val, int length) {
		int n = length - val.length();
		if(n<0) { val=val.substring(0, length-3)+"..."; n=0; }
		System.out.print(val);
		for(; n>0; n--) System.out.print(" "); 	
	}
}