package skuad.agentu.implementation;

import skuad.hubudp.*;
import skuad.uda.*;
import skuad.agentu.*;
import skuad.util.databytes.*;
import skuad.uda.net.*;
import skuad.SKUAD;
import java.util.*;
import skuad.agentu.implementation.amanager.*;

public class DeviceDetector implements SpotObserver, ExposedDeviceResponse, ResponseDevice {
	
	Vector<DeviceHoster> hosters; //liste des noeuds qui portent des devices
	DataReader query_name_paquet;
	RunAgentContext racontext;
	
	public DeviceDetector(RunAgentContext vracontext, Hub hub){
		racontext=vracontext;
		hosters = new Vector<DeviceHoster>();
		//création du paquet pour la demande du nom des devices
		DataWriter writer = DataBytes.writer();
		UDA.encodeGetParams(writer);
		query_name_paquet=writer.flushReader();
				
		hub.addSpotObserver(this); //l'instance s'enregistre comme écouteur des changements d'information des noeuds
	}
	
	//[interface SpotObserver
	public void up(Spot dspot){
//		if(already_exist) updateSMD(dspot, true);
	}
	public void updateSMD(Spot dspot, boolean first_definition){
//System.out.println("updateSMD for spot port = "+dspot.getPort());
		DeviceHoster dhoster = findHoster(dspot.getIdGlobal());
		//on regarde si l'hote porte des devices
		int dversion =  dspot.getInt(SKUAD.SMD_DEVICE);			
		if(dversion>0){ //des devices sont portés par cet spot, et on doit analyser sa définition
			if(dhoster == null || dhoster.getDeviceListVersion()!=dversion){
				if(dhoster!=null) dhoster.setDeviceListVersion(dversion);
				else {				
					dhoster = new DeviceHoster(racontext, dspot, dversion);
					hosters.add(dhoster);
				}
//System.out.println("-> find device on spot port = "+dspot.getPort());				
				UdaUdp.exposedDevices(dspot, this); 
			}
			
			//on regarde si le noeud porte un nom
			dhoster.setNodeName(dspot.getString(Spot.HOST_NAME));
			//on regarde si l'hote porte un nom
			dhoster.setSpotName(dspot.getString(Spot.SPOT_NAME));			
		} else if(dhoster!=null) hosters.remove(dhoster); //ce noeud ne porte plus de device
	}
	public void down(Spot dspot){ //on doit supprimer tous les devices portés par ce noeud
		DeviceHoster dhoster = findHoster(dspot.getIdGlobal());
		if(dhoster !=null ) {
			hosters.remove(dhoster);
			dhoster.removeAllDevice();
		}
	}
	public void idContext(Spot dspot, String ch_context, String id_context){}
	//]
	
	//[Interface ExposedDeviceResponse
	public void exposedDevice(Spot source, int[][] exposed_device){ //exposed_device[i][0]=device_id , exposed_device[i][1]=device_type
//System.out.println("<- receive device on spot port = "+source.getPort());		
		DeviceHoster dhoster = findHoster(source.getIdGlobal());
		if(dhoster==null) return; //host non connu
		Vector<Integer> existing_devices = dhoster.getDeviceIds();
		
		int device_id;
		for(int i=0; i< exposed_device.length; i++){
//System.out.println("exposedDevice["+i+"][0] = "+exposed_device[i][0]+"  exposedDevice["+i+"][1] = "+exposed_device[i][1]);			
			device_id = exposed_device[i][0];
			if(!existing_devices.removeElement(device_id)) { //on doit créer l'instance SpotedDevice + capturer le nom du Device ++> déclencher une analyse des règles sur ce device
				dhoster.addDevice(device_id, exposed_device[i][1], null);
				//Capture du nom du device : nb cette capture va déclencher l'analyse des règles de plug sur ce device
				UdaUdp.requestDevice(dhoster.getSpot(), device_id, query_name_paquet, this, false);	
			}
		}	
		if(existing_devices.size()>0){ //des devices ont été supprimé sur ce spot
			for(Integer vi : existing_devices) dhoster.removeDevice(vi.intValue());
		}
	}
	
	//[interface ResponseDevice
	public void deviceResponse(Spot target, int device_id, int code_perform, DataReader paquet_response){ //pour déterminer le nom du device
		if(code_perform==UDA.OK && paquet_response!=null){ //il y a bien une information à exploiter
			int cmd_uda = paquet_response.readByte();
			if(cmd_uda==UDA.PARAMS){
				String new_name = paquet_response.readString();
				if(new_name!=null){
					DeviceHoster dhoster = findHoster(target);
					if(dhoster!=null) {
						HostedDevice hdevice = dhoster.getDevice(device_id);		
						if(hdevice!=null)  hdevice.setName(new_name);
					}
				}
			}
		}
	}
	//]
	protected DeviceHoster findHoster(Spot target){
		for(DeviceHoster dhoster : hosters) if(dhoster.getSpot()==target) return(dhoster);
		return(null);
	}
	protected DeviceHoster findHoster(String id){
		if(id!=null) for(DeviceHoster dhoster : hosters) if(id.equals(dhoster.getSpot().getIdGlobal())) return(dhoster);
		return(null);
	}
	
	protected HostedDevice findDevice(PhysicalPlugRule prule){
		HostedDevice hdevice=null;
		for(DeviceHoster dh : hosters){
			hdevice = dh.findDevice(prule);
			if(hdevice!=null) break;
		}
		return(hdevice);
	}	
}